Type: Package
Package: git2r
Title: Provides Access to Git Repositories
Authors@R: c(person("Stefan", "Widgren", role = c("aut", "cre"),
                    email = "stefan.widgren@gmail.com",
                    comment = c(ORCID = "0000-0001-5745-2284")),
             person("Gabor", "Csardi", role = "ctb"),
             person("Gregory", "Jefferis", role = "ctb"),
             person("Jennifer", "Bryan", role = "ctb"),
             person("Jeroen", "Ooms", role = "ctb"),
             person("Jim", "Hester", role = "ctb"),
             person("John", "Blischak", role = "ctb"),
             person("Karthik", "Ram", role = "ctb"),
             person("Peter", "Carbonetto", role = "ctb"),
             person("Scott", "Chamberlain", role = "ctb"),
             person("Thomas", "Rosendal", role = "ctb"))
Description: Interface to the 'libgit2' library, which is a pure C
    implementation of the 'Git' core methods. Provides access to 'Git'
    repositories to extract data and running some basic 'Git'
    commands.
Version: 0.36.2
License: GPL-2
URL: https://docs.ropensci.org/git2r/,
        https://github.com/ropensci/git2r
BugReports: https://github.com/ropensci/git2r/issues
Imports: graphics, utils
Depends: R (>= 4.0)
Suggests: getPass
NeedsCompilation: yes
SystemRequirements: libgit2 (>= 1.0) and its dependencies libssh2 and
        openssl header and libraries.
Collate: 'blame.R' 'blob.R' 'branch.R' 'bundle_r_package.R'
        'checkout.R' 'commit.R' 'config.R' 'contributions.R'
        'credential.R' 'diff.R' 'fetch.R' 'git2r.R' 'index.R'
        'libgit2.R' 'merge.R' 'note.R' 'odb.R' 'plot.R' 'pull.R'
        'punch_card.R' 'push.R' 'reference.R' 'reflog.R' 'refspec.R'
        'remote.R' 'repository.R' 'reset.R' 'revparse.R' 'sha.R'
        'signature.R' 'stash.R' 'status.R' 'tag.R' 'time.R' 'tree.R'
        'when.R'
Encoding: UTF-8
RoxygenNote: 7.3.2
Packaged: 2025-03-29 09:17:55 UTC; stefan
Author: Stefan Widgren [aut, cre] (<https://orcid.org/0000-0001-5745-2284>),
  Gabor Csardi [ctb],
  Gregory Jefferis [ctb],
  Jennifer Bryan [ctb],
  Jeroen Ooms [ctb],
  Jim Hester [ctb],
  John Blischak [ctb],
  Karthik Ram [ctb],
  Peter Carbonetto [ctb],
  Scott Chamberlain [ctb],
  Thomas Rosendal [ctb]
Maintainer: Stefan Widgren <stefan.widgren@gmail.com>
Repository: CRAN
Date/Publication: 2025-03-29 10:00:02 UTC
Built: R 4.6.0; aarch64-apple-darwin23; 2026-04-25 16:39:10 UTC; unix
Archs: git2r.so.dSYM

Package: git2r
Title: Provides Access to Git Repositories
Description: Interface to the 'libgit2' library, which is a pure C
    implementation of the 'Git' core methods. Provides access to 'Git'
    repositories to extract data and running some basic 'Git'
    commands.
Version: 0.33.0
License: GPL-2
Copyright: See COPYRIGHTS file.
URL: https://docs.ropensci.org/git2r/ (website),
        https://github.com/ropensci/git2r
BugReports: https://github.com/ropensci/git2r/issues
Maintainer: Stefan Widgren <stefan.widgren@gmail.com>
Author: See AUTHORS file.
Imports: graphics, utils
Depends: R (>= 4.0)
Suggests: getPass
Type: Package
Biarch: true
NeedsCompilation: yes
SystemRequirements: By default, git2r uses a system installation of the
        libgit2 headers and library. However, if a system installation
        is not available, builds and uses a bundled version of the
        libgit2 source. zlib headers and library. OpenSSL headers and
        library (non-macOS). LibSSH2 (optional on non-Windows) to
        enable the SSH transport.
Collate: 'blame.R' 'blob.R' 'branch.R' 'bundle_r_package.R'
        'checkout.R' 'commit.R' 'config.R' 'contributions.R'
        'credential.R' 'diff.R' 'fetch.R' 'git2r.R' 'index.R'
        'libgit2.R' 'merge.R' 'note.R' 'odb.R' 'plot.R' 'pull.R'
        'punch_card.R' 'push.R' 'reference.R' 'reflog.R' 'refspec.R'
        'remote.R' 'repository.R' 'reset.R' 'revparse.R' 'sha.R'
        'signature.R' 'stash.R' 'status.R' 'tag.R' 'time.R' 'tree.R'
        'when.R'
Encoding: UTF-8
RoxygenNote: 7.2.3
Packaged: 2023-11-26 16:10:10 UTC; stefan
Repository: CRAN
Date/Publication: 2023-11-26 16:50:05 UTC
Built: R 4.4.0; i386-portbld-midnightbsd4.0; 2026-05-12 21:45:20 UTC; unix

Package: git2r
Title: Provides Access to Git Repositories
Description: Interface to the 'libgit2' library, which is a pure C
    implementation of the 'Git' core methods. Provides access to 'Git'
    repositories to extract data and running some basic 'Git'
    commands.
Version: 0.33.0
License: GPL-2
Copyright: See COPYRIGHTS file.
URL: https://docs.ropensci.org/git2r/ (website),
        https://github.com/ropensci/git2r
BugReports: https://github.com/ropensci/git2r/issues
Maintainer: Stefan Widgren <stefan.widgren@gmail.com>
Author: See AUTHORS file.
Imports: graphics, utils
Depends: R (>= 4.0)
Suggests: getPass
Type: Package
Biarch: true
NeedsCompilation: yes
SystemRequirements: By default, git2r uses a system installation of the
        libgit2 headers and library. However, if a system installation
        is not available, builds and uses a bundled version of the
        libgit2 source. zlib headers and library. OpenSSL headers and
        library (non-macOS). LibSSH2 (optional on non-Windows) to
        enable the SSH transport.
Collate: 'blame.R' 'blob.R' 'branch.R' 'bundle_r_package.R'
        'checkout.R' 'commit.R' 'config.R' 'contributions.R'
        'credential.R' 'diff.R' 'fetch.R' 'git2r.R' 'index.R'
        'libgit2.R' 'merge.R' 'note.R' 'odb.R' 'plot.R' 'pull.R'
        'punch_card.R' 'push.R' 'reference.R' 'reflog.R' 'refspec.R'
        'remote.R' 'repository.R' 'reset.R' 'revparse.R' 'sha.R'
        'signature.R' 'stash.R' 'status.R' 'tag.R' 'time.R' 'tree.R'
        'when.R'
Encoding: UTF-8
RoxygenNote: 7.2.3
Packaged: 2023-11-26 16:10:10 UTC; stefan
Repository: CRAN
Date/Publication: 2023-11-26 16:50:05 UTC
Built: R 4.4.0; amd64-portbld-midnightbsd4.0.2; 2026-02-17 20:11:34 UTC; unix

Type: Package
Package: git2r
Title: Provides Access to Git Repositories
Authors@R: c(person("Stefan", "Widgren", role = c("aut", "cre"),
                    email = "stefan.widgren@gmail.com",
                    comment = c(ORCID = "0000-0001-5745-2284")),
             person("Gabor", "Csardi", role = "ctb"),
             person("Gregory", "Jefferis", role = "ctb"),
             person("Jennifer", "Bryan", role = "ctb"),
             person("Jeroen", "Ooms", role = "ctb"),
             person("Jim", "Hester", role = "ctb"),
             person("John", "Blischak", role = "ctb"),
             person("Karthik", "Ram", role = "ctb"),
             person("Peter", "Carbonetto", role = "ctb"),
             person("Scott", "Chamberlain", role = "ctb"),
             person("Thomas", "Rosendal", role = "ctb"))
Description: Interface to the 'libgit2' library, which is a pure C
    implementation of the 'Git' core methods. Provides access to 'Git'
    repositories to extract data and running some basic 'Git'
    commands.
Version: 0.36.2
License: GPL-2
URL: https://docs.ropensci.org/git2r/,
        https://github.com/ropensci/git2r
BugReports: https://github.com/ropensci/git2r/issues
Imports: graphics, utils
Depends: R (>= 4.0)
Suggests: getPass
NeedsCompilation: yes
SystemRequirements: libgit2 (>= 1.0) and its dependencies libssh2 and
        openssl header and libraries.
Collate: 'blame.R' 'blob.R' 'branch.R' 'bundle_r_package.R'
        'checkout.R' 'commit.R' 'config.R' 'contributions.R'
        'credential.R' 'diff.R' 'fetch.R' 'git2r.R' 'index.R'
        'libgit2.R' 'merge.R' 'note.R' 'odb.R' 'plot.R' 'pull.R'
        'punch_card.R' 'push.R' 'reference.R' 'reflog.R' 'refspec.R'
        'remote.R' 'repository.R' 'reset.R' 'revparse.R' 'sha.R'
        'signature.R' 'stash.R' 'status.R' 'tag.R' 'time.R' 'tree.R'
        'when.R'
Encoding: UTF-8
RoxygenNote: 7.3.2
Packaged: 2025-03-29 09:17:55 UTC; stefan
Author: Stefan Widgren [aut, cre] (<https://orcid.org/0000-0001-5745-2284>),
  Gabor Csardi [ctb],
  Gregory Jefferis [ctb],
  Jennifer Bryan [ctb],
  Jeroen Ooms [ctb],
  Jim Hester [ctb],
  John Blischak [ctb],
  Karthik Ram [ctb],
  Peter Carbonetto [ctb],
  Scott Chamberlain [ctb],
  Thomas Rosendal [ctb]
Maintainer: Stefan Widgren <stefan.widgren@gmail.com>
Repository: CRAN
Date/Publication: 2025-03-29 10:00:02 UTC

# Generated by roxygen2: do not edit by hand

S3method("[",git_tree)
S3method(as.POSIXct,git_time)
S3method(as.character,git_time)
S3method(as.data.frame,git_commit)
S3method(as.data.frame,git_repository)
S3method(as.data.frame,git_tree)
S3method(as.list,git_tree)
S3method(diff,git_repository)
S3method(diff,git_tree)
S3method(format,git_blob)
S3method(format,git_commit)
S3method(format,git_merge_result)
S3method(format,git_note)
S3method(format,git_reference)
S3method(format,git_signature)
S3method(format,git_stash)
S3method(format,git_tag)
S3method(format,git_time)
S3method(head,git_repository)
S3method(length,git_blob)
S3method(length,git_diff)
S3method(length,git_tree)
S3method(lookup_commit,git_branch)
S3method(lookup_commit,git_commit)
S3method(lookup_commit,git_reference)
S3method(lookup_commit,git_tag)
S3method(merge,character)
S3method(merge,git_branch)
S3method(merge,git_repository)
S3method(plot,git_repository)
S3method(print,git_blob)
S3method(print,git_branch)
S3method(print,git_commit)
S3method(print,git_config)
S3method(print,git_diff)
S3method(print,git_merge_result)
S3method(print,git_note)
S3method(print,git_reference)
S3method(print,git_reflog)
S3method(print,git_reflog_entry)
S3method(print,git_repository)
S3method(print,git_signature)
S3method(print,git_stash)
S3method(print,git_status)
S3method(print,git_tag)
S3method(print,git_time)
S3method(print,git_tree)
S3method(sha,git_blob)
S3method(sha,git_branch)
S3method(sha,git_commit)
S3method(sha,git_fetch_head)
S3method(sha,git_merge_result)
S3method(sha,git_note)
S3method(sha,git_reference)
S3method(sha,git_reflog_entry)
S3method(sha,git_tag)
S3method(sha,git_tree)
S3method(summary,git_commit)
S3method(summary,git_diff)
S3method(summary,git_repository)
S3method(summary,git_stash)
S3method(summary,git_tag)
S3method(summary,git_tree)
export(add)
export(ahead_behind)
export(as.data.frame)
export(blame)
export(blob_create)
export(branch_create)
export(branch_delete)
export(branch_get_upstream)
export(branch_remote_name)
export(branch_remote_url)
export(branch_rename)
export(branch_set_upstream)
export(branch_target)
export(branches)
export(bundle_r_package)
export(checkout)
export(clone)
export(commit)
export(commits)
export(config)
export(content)
export(contributions)
export(cred_env)
export(cred_ssh_key)
export(cred_token)
export(cred_user_pass)
export(default_signature)
export(descendant_of)
export(diff)
export(discover_repository)
export(fetch)
export(fetch_heads)
export(git_config_files)
export(hash)
export(hashfile)
export(head)
export(in_repository)
export(index_remove_bypath)
export(init)
export(is_bare)
export(is_binary)
export(is_blob)
export(is_branch)
export(is_commit)
export(is_detached)
export(is_empty)
export(is_head)
export(is_local)
export(is_merge)
export(is_shallow)
export(is_tag)
export(is_tree)
export(last_commit)
export(libgit2_features)
export(libgit2_version)
export(lookup)
export(lookup_commit)
export(ls_tree)
export(merge)
export(merge_base)
export(note_create)
export(note_default_ref)
export(note_remove)
export(notes)
export(odb_blobs)
export(odb_objects)
export(parents)
export(pull)
export(punch_card)
export(push)
export(references)
export(reflog)
export(remote_add)
export(remote_ls)
export(remote_remove)
export(remote_rename)
export(remote_set_url)
export(remote_url)
export(remotes)
export(repository)
export(repository_head)
export(reset)
export(revparse_single)
export(rm_file)
export(sha)
export(ssh_path)
export(ssl_cert_locations)
export(stash)
export(stash_apply)
export(stash_drop)
export(stash_list)
export(stash_pop)
export(status)
export(tag)
export(tag_delete)
export(tags)
export(tree)
export(when)
export(workdir)
importFrom(graphics,axis)
importFrom(graphics,barplot)
importFrom(graphics,par)
importFrom(graphics,plot.new)
importFrom(graphics,plot.window)
importFrom(graphics,symbols)
importFrom(graphics,title)
importFrom(utils,capture.output)
importFrom(utils,head)
importFrom(utils,sessionInfo)
useDynLib(git2r, .registration=TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test.txt"))

## Add and commit
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message")

## Add a remote
remote_add(repo, "playground",
           "https://github.com/gaborcsardi/playground")

stopifnot(identical(remotes(repo), "playground"))
stopifnot(identical(remote_url(repo, "playground"),
                    "https://github.com/gaborcsardi/playground"))
stopifnot(identical(remote_url(repo),
                    "https://github.com/gaborcsardi/playground"))

## Rename a remote
remote_rename(repo, "playground", "foobar")

stopifnot(identical(remotes(repo), "foobar"))
stopifnot(identical(remote_url(repo, "foobar"),
                    "https://github.com/gaborcsardi/playground"))

## Set remote url
remote_set_url(repo, "foobar", "https://github.com/stewid/playground")
stopifnot(identical(remote_url(repo, "foobar"),
                    "https://github.com/stewid/playground"))

## Remove a remote
remote_remove(repo, "foobar")

stopifnot(identical(remotes(repo), character(0)))

if (identical(Sys.getenv("NOT_CRAN"), "true")) {
    if (isTRUE(libgit2_features()$https)) {
        refs <- remote_ls("https://github.com/ropensci/git2r")
        stopifnot(length(refs) > 0)
        stopifnot(names(refs) > 0)
        stopifnot(any(names(refs) == "HEAD"))
    }
}

# an invalid URL should throw an error
tools::assertError(remote_ls("bad"))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
f <- file(file.path(path, "test.txt"), "wb")
writeChar("Hello world!\n", f, eos = NULL)
close(f)

## add and commit
add(repo, "test.txt")
commit(repo, "Commit message")

## Check tree
stopifnot(is_tree(lookup(repo, "a0b0b9e615e9e433eb5f11859e9feac4564c58c5")))
stopifnot(identical(
    sha(lookup(repo, "a0b0b9e615e9e433eb5f11859e9feac4564c58c5")),
    "a0b0b9e615e9e433eb5f11859e9feac4564c58c5"))
stopifnot(is_tree(tree(commits(repo)[[1]])))
stopifnot(identical(lookup(repo, "a0b0b9e615e9e433eb5f11859e9feac4564c58c5"),
                    tree(commits(repo)[[1]])))
stopifnot(identical(length(tree(commits(repo)[[1]])), 1L))

## Coerce to a data.frame and check column names
stopifnot(identical(names(as.data.frame(tree(commits(repo)[[1]]))),
                    c("mode", "type", "sha", "name")))

## Coerce to list and check length
stopifnot(identical(length(as.list(tree(last_commit(repo)))), 1L))

## Print and summary
stopifnot(identical(print(tree(last_commit(repo))), tree(last_commit(repo))))
summary(tree(last_commit(repo)))

## Check indexing
stopifnot(is_blob(tree(last_commit(repo))[TRUE]))
stopifnot(is_blob(tree(last_commit(repo))["test.txt"]))
res <- tools::assertError(tree(last_commit(repo))[data.frame()])
stopifnot(length(grep("Invalid index", res[[1]]$message)) > 0)

## Check ls_tree
stopifnot(identical(ls_tree(repo = repo), ls_tree(repo = path)))
stopifnot(identical(ls_tree(tree = sha(tree(last_commit(repo))), repo = repo),
                    ls_tree(repo = repo)))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create directories for repositories in tempdir
path_bare <- tempfile(pattern = "git2r-")
path_repo <- tempfile(pattern = "git2r-")

dir.create(path_bare)
dir.create(path_repo)

## Create bare repository
bare_repo <- init(path_bare, bare = TRUE)

## Clone to repo
repo <- clone(path_bare, path_repo)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Add changes to repo
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path_repo, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "First commit message")

## Add more changes to repo
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path_repo, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Second commit message")

## Check remote
stopifnot(identical(
    git2r:::get_refspec(repo, spec = "master")$remote,
    "origin"))

## Detach
checkout(commit_1)
tools::assertError(git2r:::get_refspec(repo))

## Cleanup
unlink(path_bare, recursive = TRUE)
unlink(path_repo, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create directory for repository in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Create repository
repo <- init(path, branch = "main")
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Add changes to repo
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path, "test-1.txt"))
add(repo, "test-1.txt")
commit_1 <- commit(repo, "First commit message")

## Create branch and checkout
checkout(branch_create(commit_1, name = "test"))

## Add changes to test branch
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test-1.txt"))
add(repo, "test-1.txt")
commit_2 <- commit(repo, "Second commit message")

# Checkout main and merge
b <- branches(repo)
checkout(b[sapply(b, "[", "name") == "main"][[1]], force = TRUE)
m <- merge(b[sapply(b, "[", "name") == "test"][[1]])

# Check merge
stopifnot(inherits(m, "git_merge_result"))
stopifnot(identical(m$up_to_date, FALSE))
stopifnot(identical(m$fast_forward, TRUE))
stopifnot(identical(m$conflicts, FALSE))
stopifnot(identical(sha(m), NA_character_))
stopifnot(identical(length(commits(repo)), 2L))

# Check reflog
r <- reflog(repo)
stopifnot(identical(r[[1]]$message, "merge test: Fast-forward"))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Status case 1
status_exp_1 <- structure(list(staged = empty_named_list(),
                               unstaged = empty_named_list(),
                               untracked = empty_named_list()),
                          class = "git_status")
status_obs_1 <- status(repo)
stopifnot(identical(print(status_obs_1), status_obs_1))
str(status_exp_1)
str(status_obs_1)
stopifnot(identical(status_obs_1, status_exp_1))
stopifnot(identical(capture.output(status(repo)),
                    "working directory clean"))

## Status case 2, include ignored files
status_exp_2 <- structure(list(staged = empty_named_list(),
                               unstaged = empty_named_list(),
                               untracked = empty_named_list(),
                               ignored = empty_named_list()),
                          class = "git_status")
status_obs_2 <- status(repo, ignored = TRUE)
status_obs_2
str(status_exp_2)
str(status_obs_2)
stopifnot(identical(status_obs_2, status_exp_2))
stopifnot(identical(capture.output(status(repo, ignored = TRUE)),
                    "working directory clean"))

## Create 4 files
writeLines("File-1", file.path(path, "test-1.txt"))
writeLines("File-2", file.path(path, "test-2.txt"))
writeLines("File-3", file.path(path, "test-3.txt"))
writeLines("File-4", file.path(path, "test-4.txt"))

## Status case 3: 4 untracked files
status_exp_3 <- structure(list(staged = empty_named_list(),
                               unstaged = empty_named_list(),
                               untracked = list(untracked = "test-1.txt",
                                                untracked = "test-2.txt",
                                                untracked = "test-3.txt",
                                                untracked = "test-4.txt")),
                          class = "git_status")
status_obs_3 <- status(repo)
status_obs_3
str(status_exp_3)
str(status_obs_3)
stopifnot(identical(status_obs_3, status_exp_3))

## Add file 1 and 2 to the repository and commit
add(repo, c("test-1.txt", "test-2.txt"))
commit(repo, "Commit message")

## Status case 4: 2 untracked files
status_exp_4 <- structure(list(staged = empty_named_list(),
                               unstaged = empty_named_list(),
                               untracked = list(untracked = "test-3.txt",
                                                untracked = "test-4.txt")),
                          class = "git_status")
status_obs_4 <- status(repo)
status_obs_4
str(status_exp_4)
str(status_obs_4)
stopifnot(identical(status_obs_4, status_exp_4))

## Update file 1 & 2
writeLines(c("File-1", "Hello world"), file.path(path, "test-1.txt"))
writeLines(c("File-2", "Hello world"), file.path(path, "test-2.txt"))

## Add file 1
add(repo, "test-1.txt")

## Status case 5: 1 staged file, 1 unstaged file and 2 untracked files
status_exp_5 <- structure(list(staged = list(modified = "test-1.txt"),
                               unstaged = list(modified = "test-2.txt"),
                               untracked = list(untracked = "test-3.txt",
                                                untracked = "test-4.txt")),
                          class = "git_status")
status_obs_5 <- status(repo)
status_obs_5
str(status_exp_5)
str(status_obs_5)
stopifnot(identical(status_obs_5, status_exp_5))

## Add .gitignore file with file test-4.txt
writeLines("test-4.txt", file.path(path, ".gitignore"))

## Status case 6: 1 staged file, 1 unstaged file, 2 untracked files
## and 1 ignored file
status_exp_6 <- structure(list(staged = list(modified = "test-1.txt"),
                               unstaged = list(modified = "test-2.txt"),
                               untracked = list(untracked = ".gitignore",
                                                untracked = "test-3.txt"),
                               ignored = list(ignored = "test-4.txt")),
                          class = "git_status")
status_obs_6 <- status(repo, ignore = TRUE)
status_obs_6
str(status_exp_6)
str(status_obs_6)
stopifnot(identical(status_obs_6, status_exp_6))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test.txt"))

## add and commit
add(repo, "test.txt")
commit(repo, "Commit message")

## Check tags, no tag added
stopifnot(identical(tags(repo), empty_named_list()))

## Create tag
new_tag <- tag(repo, "Tagname", "Tag message")
stopifnot(identical(print(new_tag), new_tag))
summary(new_tag)

## Check tag
stopifnot(identical(lookup(repo, sha(new_tag)), new_tag))
stopifnot(identical(new_tag$name, "Tagname"))
stopifnot(identical(new_tag$message, "Tag message"))
stopifnot(identical(new_tag$tagger$name, "Alice"))
stopifnot(identical(new_tag$tagger$email, "alice@example.org"))
stopifnot(identical(length(tags(repo)), 1L))
stopifnot(identical(tags(repo)[[1]]$name, "Tagname"))
stopifnot(identical(tags(repo)[[1]]$message, "Tag message"))
stopifnot(identical(tags(repo)[[1]]$tagger$name, "Alice"))
stopifnot(identical(tags(repo)[[1]]$tagger$email, "alice@example.org"))

## Check objects in object database
stopifnot(identical(table(odb_objects(repo)$type),
                    structure(c(1L, 1L, 1L, 1L),
                              .Dim = 4L,
                              .Dimnames = structure(list(
                                  c("blob", "commit", "tag", "tree")),
                                  .Names = ""),
                              class = "table")))

## Delete tag
tag_delete(new_tag)
stopifnot(identical(length(tags(repo)), 0L))

## Create tag with session info
tag(repo, "Tagname", "Tag message", session = TRUE)
stopifnot(grep("git2r", tags(repo)[[1]]$message) > 0)

## Check tags method with default repo argument
wd <- setwd(path)
stopifnot(identical(length(tags()), 1L))
tag_delete(name = "Tagname")
stopifnot(identical(length(tags()), 0L))
if (!is.null(wd))
    setwd(wd)

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test-1.txt"))

## Add and reset an empty repository using a path
add(repo, "test-1.txt")
stopifnot(identical(
    status(repo),
    structure(list(staged = list(new = "test-1.txt"),
                   unstaged = empty_named_list(),
                   untracked = empty_named_list()),
              class = "git_status")))
reset(repo, path = "test-1.txt")
stopifnot(identical(
    status(repo),
    structure(list(staged = empty_named_list(),
                   unstaged = empty_named_list(),
                   untracked = list(untracked = "test-1.txt")),
              class = "git_status")))

## Add and reset a non-empty repository using a path
add(repo, "test-1.txt")
commit(repo, "First commit")
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-1.txt"))
add(repo, "test-1.txt")
stopifnot(identical(
    status(repo),
    structure(list(staged = list(modified = "test-1.txt"),
                   unstaged = empty_named_list(),
                   untracked = empty_named_list()),
              class = "git_status")))
reset(repo, path = "test-1.txt")
stopifnot(identical(
    status(repo),
    structure(list(staged = empty_named_list(),
                   unstaged = list(modified = "test-1.txt"),
                   untracked = empty_named_list()),
              class = "git_status")))

## add and commit
add(repo, "test-1.txt")
commit_1 <- commit(repo, "Commit message")

## Make one more commit
writeLines(c("Hello world!", "HELLO WORLD!", "hello world!"),
           file.path(path, "test-1.txt"))
add(repo, "test-1.txt")
commit(repo, "Next commit message")

## Create one more file
writeLines("Hello world!", file.path(path, "test-2.txt"))

## 'soft' reset to first commit
reset(commit_1)
soft_exp <- structure(list(staged = list(modified = "test-1.txt"),
                           unstaged = empty_named_list(),
                           untracked = list(untracked = "test-2.txt")),
                      class = "git_status")
soft_obs <- status(repo)
stopifnot(identical(soft_obs, soft_exp))
stopifnot(identical(length(commits(repo)), 2L))
stopifnot(identical(commits(repo)[[1]], commit_1))

## 'mixed' reset to first commit
commit(repo, "Next commit message")
reset(commit_1, "mixed")
mixed_exp <- structure(list(staged = empty_named_list(),
                            unstaged = list(modified = "test-1.txt"),
                            untracked = list(untracked = "test-2.txt")),
                       class = "git_status")
mixed_obs <- status(repo)
stopifnot(identical(mixed_obs, mixed_exp))
stopifnot(identical(length(commits(repo)), 2L))
stopifnot(identical(commits(repo)[[1]], commit_1))

## 'hard' reset to first commit
add(repo, "test-1.txt")
commit(repo, "Next commit message")
reset(commit_1, "hard")
hard_exp <- structure(list(staged = empty_named_list(),
                           unstaged = empty_named_list(),
                           untracked = list(untracked = "test-2.txt")),
                      class = "git_status")
hard_obs <- status(repo)
stopifnot(identical(hard_obs, hard_exp))
stopifnot(identical(length(commits(repo)), 2L))
stopifnot(identical(commits(repo)[[1]], commit_1))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create 2 directories in tempdir
path_bare <- tempfile(pattern = "git2r-")
path_repo <- tempfile(pattern = "git2r-")
dir.create(path_bare)
dir.create(path_repo)

## Initialize a repository
repo <- init(path_repo)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Add commit to repo
writeLines("Hello world", con = file.path(path_repo, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message")

## Check bare argument
tools::assertError(clone(path_repo, path_bare, bare = c(TRUE, TRUE)))
tools::assertError(clone(path_repo, path_bare, bare = 1))
tools::assertError(clone(path_repo, path_bare, bare = 1L))
tools::assertError(clone(path_repo, path_bare, bare = "test"))

## Clone repo to bare repository
bare_repo <- clone(path_repo, path_bare, bare = TRUE)

## Check the repositores
stopifnot(identical(is_bare(bare_repo), TRUE))
stopifnot(identical(is_bare(repo), FALSE))

## Check result in bare repository
stopifnot(identical(length(commits(bare_repo)), 1L))
bare_commit_1 <- commits(bare_repo)[[1]]
stopifnot(identical(sha(commit_1), sha(bare_commit_1)))
stopifnot(identical(commit_1$author, bare_commit_1$author))
stopifnot(identical(commit_1$committer, bare_commit_1$committer))
stopifnot(identical(commit_1$summary, bare_commit_1$summary))
stopifnot(identical(commit_1$message, bare_commit_1$message))
stopifnot(!identical(commit_1$repo, bare_commit_1$repo))

## Cleanup
unlink(path_bare, recursive = TRUE)
unlink(path_repo, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path, branch = "main")
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test.txt"))

## add and commit
add(repo, "test.txt")
commit(repo, "Commit message")

## Check dwim of reference shorthand
stopifnot(identical(.Call(git2r:::git2r_reference_dwim, repo, "")$name,
                    "refs/heads/main"))
stopifnot(identical(.Call(git2r:::git2r_reference_dwim, repo, "main")$name,
                    "refs/heads/main"))
stopifnot(identical(
    .Call(git2r:::git2r_reference_dwim, repo, "refs/heads/main")$name,
    "refs/heads/main"))

## print reference
r <- .Call(git2r:::git2r_reference_dwim, repo, "refs/heads/main")
stopifnot(identical(print(r), r))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path, branch = "main")
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create first commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "First commit message")

## Edit file and checkout
writeLines(c("Hello world!", "Hello world!"), file.path(path, "test.txt"))
status_exp_1 <- structure(list(staged = structure(list(),
                                   .Names = character(0)),
                               unstaged = structure(list(modified = "test.txt"),
                                   .Names = "modified"),
                               untracked = structure(list(),
                                   .Names = character(0))),
                          .Names = c("staged", "unstaged", "untracked"),
                          class = "git_status")
status_obs_1 <- status(repo)
str(status_exp_1)
str(status_obs_1)
stopifnot(identical(status_obs_1, status_exp_1))
checkout(repo, path = "test.txt")
status_exp_2 <- structure(list(staged = structure(list(),
                                   .Names = character(0)),
                               unstaged = structure(list(),
                                   .Names = character(0)),
                               untracked = structure(list(),
                                   .Names = character(0))),
                          .Names = c("staged", "unstaged", "untracked"),
                          class = "git_status")
status_obs_2 <- status(repo)
str(status_exp_2)
str(status_obs_2)
stopifnot(identical(status_obs_2, status_exp_2))

## Create second commit
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Second commit message")
tag(repo, "commit_2", "Tag message")

## Create third commit
writeLines(c("Hello world!", "HELLO WORLD!", "HeLlO wOrLd!"),
           file.path(path, "test.txt"))
add(repo, "test.txt")
commit_3 <- commit(repo, "Third commit message")

## Check HEAD
stopifnot(identical(is_detached(repo), FALSE))
stopifnot(identical(repository_head(repo)$name, "main"))

## Check show and summary
repo
summary(repo)

## Checkout first commit
checkout(commit_1, TRUE)
stopifnot(identical(is_detached(repo), TRUE))
stopifnot(identical(repository_head(repo), commit_1))
stopifnot(identical(readLines(file.path(path, "test.txt")), "Hello world!"))

## Check show and summary
repo
summary(repo)

## Checkout tag
checkout(tags(repo)[[1]], TRUE)
stopifnot(identical(is_detached(repo), TRUE))
stopifnot(identical(readLines(file.path(path, "test.txt")),
                    c("Hello world!", "HELLO WORLD!")))

## Check is_detached with missing repo argument
wd <- setwd(path)
stopifnot(identical(is_detached(), TRUE))
if (!is.null(wd))
    setwd(wd)

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a '.gitignore' file
writeLines("test.txt", file.path(path, ".gitignore"))
add(repo, ".gitignore")
commit(repo, "First commit message")

## Create a file
writeLines("Hello world!", file.path(path, "test.txt"))

## Check status
s_1 <- structure(list(staged = empty_named_list(),
                      unstaged = empty_named_list(),
                      untracked = empty_named_list(),
                      ignored = list(ignored = "test.txt")),
                 class = "git_status")
stopifnot(identical(status(repo, ignored = TRUE), s_1))

## The file is ignored and should not be added
add(repo, "test.txt")
stopifnot(identical(status(repo, ignored = TRUE), s_1))

## The file is ignored but should be added with force
s_2 <- structure(list(staged = list(new = "test.txt"),
                      unstaged = empty_named_list(),
                      untracked = empty_named_list(),
                      ignored = empty_named_list()),
                 class = "git_status")

add(repo, "test.txt", force = TRUE)
stopifnot(identical(status(repo, ignored = TRUE), s_2))

## Commit and check status
s_3 <- structure(list(staged = empty_named_list(),
                      unstaged = empty_named_list(),
                      untracked = empty_named_list(),
                      ignored = empty_named_list()),
                 class = "git_status")

commit(repo, "Second commit message")
stopifnot(identical(status(repo, ignored = TRUE), s_3))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create directories for repositories in tempdir
path_bare <- tempfile(pattern = "git2r-")
path_repo_1 <- tempfile(pattern = "git2r-")
path_repo_2 <- tempfile(pattern = "git2r-")

dir.create(path_bare)
dir.create(path_repo_1)
dir.create(path_repo_2)

## Create bare repository
bare_repo <- init(path_bare, bare = TRUE)

## Clone to repo 1
repo_1 <- clone(path_bare, path_repo_1)
config(repo_1, user.name = "Alice", user.email = "alice@example.org")

## Add changes to repo 1 and push to bare
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path_repo_1, "test-1.txt"))
add(repo_1, "test-1.txt")
commit_1 <- commit(repo_1, "First commit message")
branch_name <- branches(repo_1)[[1]]$name
push(repo_1, "origin", paste0("refs/heads/", branch_name))

## Clone to repo 2
repo_2 <- clone(path_bare, path_repo_2)
config(repo_2, user.name = "Bob", user.email = "bob@example.org")

## Add more changes to repo 1 and push to bare
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path_repo_1, "test-1.txt"))
add(repo_1, "test-1.txt")
commit_2 <- commit(repo_1, "Second commit message")
push(repo_1, "origin", paste0("refs/heads/", branch_name))

## Pull changes to repo_2
pull(repo_2)
stopifnot(identical(length(commits(repo_2)), 2L))

## Check remote url of repo_2
stopifnot(identical(
    branch_remote_url(branch_get_upstream(repository_head(repo_2))),
    path_bare))

## Unset remote remote tracking branch
branch_set_upstream(repository_head(repo_2), NULL)
stopifnot(is.null(branch_get_upstream(repository_head(repo_2))))
tools::assertError(pull(repo_2))
tools::assertError(branch_set_upstream(repository_head(repo_2), NULL))

## Add more changes to repo 1 and push to bare
writeLines(
    c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
      "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
      "minim veniam, quis nostrud exercitation ullamco laboris nisi ut"),
    con = file.path(path_repo_1, "test-1.txt"))
add(repo_1, "test-1.txt")
commit_3 <- commit(repo_1, "Third commit message")
push(repo_1)

## Set remote tracking branch
branch_set_upstream(repository_head(repo_2),
                    paste0("origin/", branch_name))
stopifnot(identical(
    branch_remote_url(branch_get_upstream(repository_head(repo_2))),
    path_bare))

## Pull changes to repo_2
pull(repo_2)
stopifnot(identical(length(commits(repo_2)), 3L))

## Check references in repo_1 and repo_2. Must clear the repo item
## since the repositories have different paths.
stopifnot(identical(length(references(repo_1)), 2L))

ref_1 <- references(repo_1)
lapply(seq_len(length(ref_1)), function(i) {
    ref_1[[i]]$repo <<- NULL
})

ref_2 <- references(repo_2)
lapply(seq_len(length(ref_2)), function(i) {
    ref_2[[i]]$repo <<- NULL
})

name <- paste0("refs/heads/", branch_name)
stopifnot(identical(ref_1[[name]], ref_2[[name]]))

name <- paste0("refs/remotes/", branch_name)
stopifnot(identical(ref_1[[name]], ref_2[[name]]))

ref_1 <- references(repo_1)[[paste0("refs/heads/", branch_name)]]
stopifnot(identical(ref_1$name, paste0("refs/heads/", branch_name)))
stopifnot(identical(ref_1$type, 1L))
stopifnot(identical(sha(ref_1), sha(commit_3)))
stopifnot(identical(ref_1$target, NA_character_))
stopifnot(identical(ref_1$shorthand, branch_name))

ref_2 <- references(repo_1)[[paste0("refs/remotes/origin/", branch_name)]]
stopifnot(identical(ref_2$name, paste0("refs/remotes/origin/", branch_name)))
stopifnot(identical(ref_2$type, 1L))
stopifnot(identical(sha(ref_2), sha(commit_3)))
stopifnot(identical(ref_2$target, NA_character_))
stopifnot(identical(ref_2$shorthand, paste0("origin/", branch_name)))

## Check references with missing repo argument
wd <- setwd(path_repo_1)
stopifnot(identical(length(references()), 2L))
if (!is.null(wd))
    setwd(wd)

## Cleanup
unlink(path_bare, recursive = TRUE)
unlink(path_repo_1, recursive = TRUE)
unlink(path_repo_2, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Initialize a temporary repository
path <- tempfile(pattern = "git2r-")
dir.create(path)
dir.create(file.path(path, "subfolder"))
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create three files and commit
writeLines("First file",  file.path(path, "example-1.txt"))
writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
writeLines("Third file",  file.path(path, "example-3.txt"))
add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
commit(repo, "Commit message")

## Traverse tree entries and its subtrees.
## Various approaches that give identical result.
stopifnot(identical(ls_tree(tree = tree(last_commit(path))),
                    ls_tree(tree = tree(last_commit(repo)))))
stopifnot(identical(ls_tree(repo = path), ls_tree(repo = repo)))

## ls_tree(repo = repo) should match `git ls-tree -lr HEAD`
ls_tree_result <- ls_tree(repo = repo)
stopifnot(identical(ls_tree_result$name,
                    c("example-1.txt", "example-3.txt", "example-2.txt")))

# Argument `tree` can be a  'character that identifies a tree in the repository'
ls_tree(tree = tree(last_commit(path))$sha, repo = repo)

## Skip content in subfolder
ls_tree_toplevel <- ls_tree(repo = repo, recursive = FALSE)
stopifnot(nrow(ls_tree_toplevel) == 3)
stopifnot(identical(ls_tree_toplevel$name,
                    c("example-1.txt", "example-3.txt", "subfolder")))

## Start in subfolder
ls_tree_subfolder <- ls_tree(tree = "HEAD:subfolder", repo = repo)
stopifnot(nrow(ls_tree_subfolder) == 1)
stopifnot(identical(ls_tree_subfolder$name, "example-2.txt"))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Check when method
w1 <- structure(list(time = 1395567947, offset = 60),
               class = "git_time")
stopifnot(identical(when(w1), "2014-03-23 09:45:47 GMT"))
stopifnot(identical(when(w1, usetz = FALSE), "2014-03-23 09:45:47"))
stopifnot(identical(when(w1, tz = "Europe/Stockholm", origin = "1980-02-02"),
                    "2024-04-23 11:45:47 CEST"))

s1 <- structure(list(name = "Alice", email = "alice@example.org", when = w1),
                class = "git_signature")
stopifnot(identical(when(s1), "2014-03-23 09:45:47 GMT"))
stopifnot(identical(when(s1, usetz = FALSE), "2014-03-23 09:45:47"))
stopifnot(identical(when(s1, tz = "Europe/Stockholm", origin = "1980-02-02"),
                    "2024-04-23 11:45:47 CEST"))

w2 <- structure(list(time = 1395567950, offset = 60),
               class = "git_time")
s2 <- structure(list(name = "Alice", email = "alice@example.org", when = w2),
                class = "git_signature")
c1 <- structure(list(sha = "166f3f779fd7e4165aaa43f2828050ce040052b0",
                     author = s1,
                     committer = s2,
                     summary = "A commit summary",
                     message = "A commit message"),
                class = "git_commit")
stopifnot(identical(when(c1), "2014-03-23 09:45:47 GMT"))
stopifnot(identical(when(c1, usetz = FALSE), "2014-03-23 09:45:47"))
stopifnot(identical(when(c1, tz = "Europe/Stockholm", origin = "1980-02-02"),
                    "2024-04-23 11:45:47 CEST"))

t1 <- structure(list(sha = "166f3f779fd7e4165aaa43f2828050ce040052b0",
                     message = "A tag message",
                     name = "A tage name",
                     tagger = s1,
                     target = "166f3f779fd7e4165aaa43f2828050ce040052b0"),
                class = "git_tag")
stopifnot(identical(when(t1), "2014-03-23 09:45:47 GMT"))
stopifnot(identical(when(t1, usetz = FALSE), "2014-03-23 09:45:47"))
stopifnot(identical(when(t1, tz = "Europe/Stockholm", origin = "1980-02-02"),
                    "2024-04-23 11:45:47 CEST"))

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Test to coerce
git_t <- structure(list(time = 1395567947, offset = 60),
                   class = "git_time")
stopifnot(identical(as.character(git_t), "2014-03-23 09:45:47 GMT"))
stopifnot(identical(as.character(git_t, usetz = FALSE), "2014-03-23 09:45:47"))
stopifnot(identical(as.POSIXct(git_t),
                    as.POSIXct(1395567947, tz = "GMT", origin = "1970-01-01")))
stopifnot(identical(print(git_t), git_t))

as.POSIXct(1395567947, origin = "1970-01-01", tz = "-03")

## Test that origin/tz can be passed to as.POSIXct
stopifnot(identical(as.POSIXct(git_t, tz = "Europe/Stockholm",
                               origin = "1980-02-02"),
                    as.POSIXct(1395567947, tz = "Europe/Stockholm",
                               origin = "1980-02-02")))

## Test that origin/tz can be passed to as.character
stopifnot(identical(as.character(git_t, tz = "Europe/Stockholm",
                                 origin = "1980-02-02"),
                    "2024-04-23 11:45:47 CEST"))
stopifnot(identical(as.character(git_t, tz = "Europe/Stockholm",
                                 origin = "1980-02-02", usetz = FALSE),
                    "2024-04-23 11:45:47"))

## Test that origin/tz can be passed to print
stopifnot(identical(
    utils::capture.output(print(git_t, tz = "Europe/Stockholm",
                                origin = "1980-02-02")),
    "2024-04-23 11:45:47 CEST"
))
stopifnot(identical(
    utils::capture.output(print(git_t, tz = "Europe/Stockholm",
                                origin = "1980-02-02", usetz = FALSE)),
    "2024-04-23 11:45:47"
))

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path, branch = "main")
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create two files and alternate commits
writeLines("1", file.path(path, "odd.txt"))
add(repo, "odd.txt")
c1 <- commit(repo, "commit 1")

writeLines("2", file.path(path, "even.txt"))
add(repo, "even.txt")
c2 <- commit(repo, "commit 2")

writeLines("3", file.path(path, "odd.txt"))
add(repo, "odd.txt")
c3 <- commit(repo, "commit 3")

writeLines("4", file.path(path, "even.txt"))
add(repo, "even.txt")
c4 <- commit(repo, "commit 4")

writeLines("5", file.path(path, "odd.txt"))
add(repo, "odd.txt")
c5 <- commit(repo, "commit 5")

writeLines("6", file.path(path, "even.txt"))
add(repo, "even.txt")
c6 <- commit(repo, "commit 6")

commits_all <- commits(repo)
stopifnot(length(commits_all) == 6)

## Test path
commits_odd <- commits(repo, path = "odd.txt")
stopifnot(length(commits_odd) == 3)
stopifnot(commits_odd[[1]]$sha == c5$sha)
stopifnot(commits_odd[[2]]$sha == c3$sha)
stopifnot(commits_odd[[3]]$sha == c1$sha)

commits_even <- commits(repo, path = "even.txt")
stopifnot(length(commits_even) == 3)
stopifnot(commits_even[[1]]$sha == c6$sha)
stopifnot(commits_even[[2]]$sha == c4$sha)
stopifnot(commits_even[[3]]$sha == c2$sha)

## Test reverse
commits_odd_rev <- commits(repo, reverse = TRUE, path = "odd.txt")
stopifnot(length(commits_odd_rev) == 3)
stopifnot(commits_odd_rev[[1]]$sha == c1$sha)
stopifnot(commits_odd_rev[[2]]$sha == c3$sha)
stopifnot(commits_odd_rev[[3]]$sha == c5$sha)

commits_even_rev <- commits(repo, reverse = TRUE, path = "even.txt")
stopifnot(length(commits_even_rev) == 3)
stopifnot(commits_even_rev[[1]]$sha == c2$sha)
stopifnot(commits_even_rev[[2]]$sha == c4$sha)
stopifnot(commits_even_rev[[3]]$sha == c6$sha)

## Test n
commits_odd_n <- commits(repo, n = 2, path = "odd.txt")
stopifnot(length(commits_odd_n) == 2)
stopifnot(commits_odd_n[[1]]$sha == c5$sha)
stopifnot(commits_odd_n[[2]]$sha == c3$sha)

commits_even_n <- commits(repo, n = 2, path = "even.txt")
stopifnot(length(commits_even_n) == 2)
stopifnot(commits_even_n[[1]]$sha == c6$sha)
stopifnot(commits_even_n[[2]]$sha == c4$sha)

commits_odd_0 <- commits(repo, n = 0, path = "odd.txt")
stopifnot(length(commits_odd_0) == 0)
stopifnot(identical(commits_odd_0, list()))

commits_even_0 <- commits(repo, n = 0, path = "even.txt")
stopifnot(length(commits_even_0) == 0)
stopifnot(identical(commits_even_0, list()))

## Test ref
checkout(repo, branch = "test-ref", create = TRUE)

writeLines("7", file.path(path, "odd.txt"))
add(repo, "odd.txt")
c7 <- commit(repo, "commit 7")

writeLines("8", file.path(path, "even.txt"))
add(repo, "even.txt")
c8 <- commit(repo, "commit 8")

commits_odd_ref <- commits(repo, ref = "main", path = "odd.txt")
stopifnot(length(commits_odd_ref) == 3)
stopifnot(commits_odd_ref[[1]]$sha == c5$sha)
stopifnot(commits_odd_ref[[2]]$sha == c3$sha)
stopifnot(commits_odd_ref[[3]]$sha == c1$sha)

commits_even_ref <- commits(repo, ref = "main", path = "even.txt")
stopifnot(length(commits_even_ref) == 3)
stopifnot(commits_even_ref[[1]]$sha == c6$sha)
stopifnot(commits_even_ref[[2]]$sha == c4$sha)
stopifnot(commits_even_ref[[3]]$sha == c2$sha)

checkout(repo, branch = "main")

## Test renaming a file (path does not support --follow)
writeLines("a file to be renamed", file.path(path, "original.txt"))
add(repo, "original.txt")
c_original <- commit(repo, "commit original")

commits_original <- commits(repo, path = "original.txt")
stopifnot(length(commits_original) == 1)
stopifnot(commits_original[[1]]$sha == c_original$sha)

file.rename(file.path(path, "original.txt"), file.path(path, "new.txt"))
add(repo, c("original.txt", "new.txt"))
c_new <- commit(repo, "commit new")

commits_new <- commits(repo, path = "new.txt")
stopifnot(length(commits_new) == 1)
stopifnot(commits_new[[1]]$sha == c_new$sha)

## Test merge commits
writeLines(letters[1:5], file.path(path, "merge.txt"))
add(repo, "merge.txt")
c_merge_1 <- commit(repo, "commit merge 1")

checkout(repo, branch = "test-merge", create = TRUE)
cat("z", file = file.path(path, "merge.txt"), append = TRUE)
add(repo, "merge.txt")
c_merge_2 <- commit(repo, "commit merge 2")

checkout(repo, branch = "main")
writeLines(c("A", letters[2:5]), file.path(path, "merge.txt"))
add(repo, "merge.txt")
c_merge_3 <- commit(repo, "commit merge 3")

c_merge_4 <- merge(repo, "test-merge")
stopifnot(class(c_merge_4) == "git_merge_result")

commits_merge <- commits(repo, path = "merge.txt")
stopifnot(length(commits_merge) == 4)
stopifnot(commits_merge[[1]]$sha == c_merge_4$sha)
stopifnot(commits_merge[[2]]$sha == c_merge_3$sha)
stopifnot(commits_merge[[3]]$sha == c_merge_2$sha)
stopifnot(commits_merge[[4]]$sha == c_merge_1$sha)

## Test absolute path
writeLines("absolute", file.path(path, "abs.txt"))
add(repo, "abs.txt")
c_abs <- commit(repo, "commit absolute")

commits_abs <- commits(repo, path = file.path(path, "abs.txt"))
stopifnot(length(commits_abs) == 1)
stopifnot(commits_abs[[1]]$sha == c_abs$sha)

## Test topological and time
## Strategy:
##   - Commit a new file test-time.txt
##   - Commit a change on branch test-time-1 (a)
##   - Commit a change on branch test-time-2 (c)
##   - Commit a change on branch test-time-1 (b)
##   - Commit a change on branch test-time-2 (d)
##   - Merge branch test-time-2 into main (fast-forward)
##   - Merge branch test-time-1 into main (merge commit)
##
## $ git log --all --decorate --oneline --graph -n 6
## *   79e6880 (HEAD -> main) merge test-time-1
## |\
## | * e2f18f1 (test-time-1) commit b
## | * 5f34820 commit a
## * | b954ec9 (test-time-2) commit d
## * | 7ae2fd5 commit c
## |/
## * 923f3ea commit base
Sys.sleep(1)
writeLines(as.character(1:100), file.path(path, "test-time.txt"))
add(repo, "test-time.txt")
c_base <- commit(repo, "commit base")
Sys.sleep(1)
branch_create(commit = c_base, name = "test-time-1")
branch_create(commit = c_base, name = "test-time-2")

checkout(repo, branch = "test-time-1")
writeLines(c("edit", 2:100), file.path(path, "test-time.txt"))
add(repo, "test-time.txt")
c_a <- commit(repo, "commit a")
Sys.sleep(1)

checkout(repo, branch = "test-time-2")
writeLines(c(1:25, "edit", 27:100), file.path(path, "test-time.txt"))
add(repo, "test-time.txt")
c_c <- commit(repo, "commit c")
Sys.sleep(1)

checkout(repo, branch = "test-time-1")
writeLines(c(1:50, "edit", 52:100), file.path(path, "test-time.txt"))
add(repo, "test-time.txt")
c_b <- commit(repo, "commit b")
Sys.sleep(1)

checkout(repo, branch = "test-time-2")
writeLines(c(1:75, "edit", 77:100), file.path(path, "test-time.txt"))
add(repo, "test-time.txt")
c_d <- commit(repo, "commit d")
Sys.sleep(1)

checkout(repo, branch = "main")
merge(repo, "test-time-2") # Fast-forward
merge(repo, "test-time-1") # Merge commit
c_merge_time <- commits(repo, n = 1)[[1]]

## topological - commits in test-time-2 come first because it was
## merged first
stopifnot(identical(
    commits(repo, topological = TRUE, time = FALSE, path = "test-time.txt"),
    list(c_merge_time, c_b, c_a, c_d, c_c, c_base)
))
stopifnot(identical(
    commits(repo, topological = TRUE, time = FALSE, path = "test-time.txt"),
    commits(repo, topological = TRUE, time = FALSE)[1:6]
))

## time - commits ordered by time they were created, not merged into
## main
stopifnot(identical(
    commits(repo, topological = FALSE, time = TRUE, path = "test-time.txt"),
    list(c_merge_time, c_d, c_b, c_c, c_a, c_base)
))
stopifnot(identical(
    commits(repo, topological = FALSE, time = TRUE, path = "test-time.txt"),
    commits(repo, topological = FALSE, time = TRUE)[1:6]
))

## topological and time - dominated by time
stopifnot(identical(
    commits(repo, topological = TRUE, time = TRUE, path = "test-time.txt"),
    list(c_merge_time, c_d, c_b, c_c, c_a, c_base)
))
stopifnot(identical(
    commits(repo, topological = TRUE, time = TRUE, path = "test-time.txt"),
    commits(repo, topological = TRUE, time = TRUE)[1:6]
))

## reverse with topological and/or time
stopifnot(identical(
    commits(repo, topological = TRUE, time = FALSE, reverse = TRUE,
            path = "test-time.txt"),
    rev(list(c_merge_time, c_b, c_a, c_d, c_c, c_base))
))
stopifnot(identical(
    commits(repo, topological = FALSE, time = TRUE, reverse = TRUE,
            path = "test-time.txt"),
    rev(list(c_merge_time, c_d, c_b, c_c, c_a, c_base))
))
stopifnot(identical(
    commits(repo, topological = TRUE, time = TRUE, reverse = TRUE,
            path = "test-time.txt"),
    rev(list(c_merge_time, c_d, c_b, c_c, c_a, c_base))
))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path, branch = "main")
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message 1")

## Create first branch, checkout, add file and commit
b_1 <- branch_create(commit_1, "branch1")
checkout(b_1)
writeLines("Branch 1", file.path(path, "branch-1.txt"))
add(repo, "branch-1.txt")
commit_2 <- commit(repo, "Commit message branch 1")

## Create second branch, checkout, add file and commit
b_2 <- branch_create(commit_1, "branch2")
checkout(b_2)
writeLines("Branch 2", file.path(path, "branch-2.txt"))
add(repo, "branch-2.txt")
commit_3 <- commit(repo, "Commit message branch 2")
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit_4 <- commit(repo, "Second commit message branch 2")

## Check that merge base equals commit_1
stopifnot(identical(merge_base(commit_2, commit_3), commit_1))

## Checkout main
b <- branches(repo)
checkout(b[sapply(b, "[", "name") == "main"][[1]], force = TRUE)

## Merge branch 1
m_1 <- merge(b[sapply(b, "[", "name") == "branch1"][[1]])
stopifnot(identical(m_1$fast_forward, TRUE))
stopifnot(identical(m_1$conflicts, FALSE))
stopifnot(identical(sha(m_1), NA_character_))
stopifnot(identical(print(m_1), m_1))

## Merge branch 1 again
m_1_again <- merge(b[sapply(b, "[", "name") == "branch1"][[1]])
stopifnot(identical(m_1_again$up_to_date, TRUE))
stopifnot(identical(m_1_again$fast_forward, FALSE))
stopifnot(identical(m_1_again$conflicts, FALSE))
stopifnot(identical(sha(m_1_again), NA_character_))

## Merge branch 2
m_2 <- merge(b[sapply(b, "[", "name") == "branch2"][[1]])
stopifnot(identical(m_2$fast_forward, FALSE))
stopifnot(identical(m_2$conflicts, FALSE))
stopifnot(identical(sha(m_2), sha(commits(repo)[[1]])))

## Create third branch, checkout, change file and commit
b_3 <- branch_create(lookup(repo, sha(m_2)), "branch3")
checkout(b_3)
writeLines(c("Lorem ipsum dolor amet sit, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message branch 3")

## Checkout main and create a change that creates a conflict on
## merge
b <- branches(repo)
checkout(b[sapply(b, "[", "name") == "main"][[1]], force = TRUE)
writeLines(c("Lorem ipsum dolor sit amet, adipisicing consectetur elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Some commit message branch 1")

## Merge branch 3 with fail = TRUE
m_3 <- merge(b[sapply(b, "[", "name") == "branch3"][[1]], fail = TRUE)
stopifnot(identical(m_3$up_to_date, FALSE))
stopifnot(identical(m_3$fast_forward, FALSE))
stopifnot(identical(m_3$conflicts, TRUE))
stopifnot(identical(sha(m_3), NA_character_))
m_3

## Check status; Expect to have a clean working directory
wd <- structure(list(staged = empty_named_list(),
                     unstaged = empty_named_list(),
                     untracked = empty_named_list()),
                class = "git_status")
stopifnot(identical(status(repo), wd))

## Merge branch 3
m_3 <- merge(b[sapply(b, "[", "name") == "branch3"][[1]])
stopifnot(identical(m_3$up_to_date, FALSE))
stopifnot(identical(m_3$fast_forward, FALSE))
stopifnot(identical(m_3$conflicts, TRUE))
stopifnot(identical(sha(m_3), NA_character_))
m_3

## Check status; Expect to have one unstaged unmerged conflict.
stopifnot(identical(status(repo),
                    structure(list(staged = empty_named_list(),
                                   unstaged = list(conflicted = "test.txt"),
                                   untracked = empty_named_list()),
                              class = "git_status")))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


stopifnot(identical(names(libgit2_features()),
                    c("threads", "https", "ssh")))

stopifnot(identical(names(libgit2_version()),
                    c("major", "minor", "rev")))

tools::assertError(ssl_cert_locations())

if (identical(Sys.getenv("R_COVR"), "true")) {
    if (isTRUE(libgit2_features()$https)) {
        ## Create a directory in tempdir
        path <- tempfile(pattern = "git2r-")
        dir.create(path)
        stopifnot(is.null(ssl_cert_locations(path = path)))
        unlink(path)
    }
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create directories for repositories in tempdir
path_bare <- tempfile(pattern = "git2r-")
path_repo_1 <- tempfile(pattern = "git2r-")
path_repo_2 <- tempfile(pattern = "git2r-")

dir.create(path_bare)
dir.create(path_repo_1)
dir.create(path_repo_2)

## Create bare repository
bare_repo <- init(path_bare, bare = TRUE)

## Clone to repo 1
repo_1 <- clone(path_bare, path_repo_1)
config(repo_1, user.name = "Alice", user.email = "alice@example.org")

## Add changes to repo 1
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path_repo_1, "test-1.txt"))
add(repo_1, "test-1.txt")
commit_1 <- commit(repo_1, "First commit message")
branch_name <- branches(repo_1)[[1]]$name

## Create 'dev' branch
checkout(branch_create(commit_1, name = "dev"))

## Add more changes to repo 1
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path_repo_1, "test-1.txt"))
add(repo_1, "test-1.txt")
commit(repo_1, "Second commit message")

## Add more changes to repo 1
writeLines(
    c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
      "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
      "minim veniam, quis nostrud exercitation ullamco laboris nisi ut"),
    con = file.path(path_repo_1, "test-1.txt"))
add(repo_1, "test-1.txt")
commit(repo_1, "Third commit message")

## Push to bare
push(repo_1, "origin", paste0("refs/heads/", branch_name))
push(repo_1, "origin", "refs/heads/dev")

## Print branch
branches(repo_1)[[paste0("origin/", branch_name)]]

## Clone to repo 2
repo_2 <- clone(url = path_bare, local_path = path_repo_2, branch = "dev")
config(repo_2, user.name = "Bob", user.email = "bob@example.org")

## Check branch and commits
stopifnot(identical(length(commits(repo_2)), 3L))
stopifnot(identical(repository_head(repo_2)$name, "dev"))

## Cleanup
unlink(path_bare, recursive = TRUE)
unlink(path_repo_1, recursive = TRUE)
unlink(path_repo_2, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Commit without adding changes should produce an error
tools::assertError(commit(repo, "Test to commit"))

## Create a file
writeLines("Hello world!", file.path(path, "test.txt"))

## Commit without adding changes should produce an error
tools::assertError(commit(repo, "Test to commit"))

## Add
add(repo, "test.txt")

## Commit with empty message should produce an error
tools::assertError(commit(repo, ""))

## Commit
commit_1 <- commit(repo, "Commit message", session = TRUE)
summary(commit_1)
tag_1 <- tag(repo, "Tagname1", "Tag message 1")

## Check commit
stopifnot(identical(commit_1$author$name, "Alice"))
stopifnot(identical(commit_1$author$email, "alice@example.org"))
stopifnot(identical(lookup(repo, sha(commit_1)), commit_1))
stopifnot(identical(length(commits(repo)), 1L))
stopifnot(identical(commits(repo)[[1]]$author$name, "Alice"))
stopifnot(identical(commits(repo)[[1]]$author$email, "alice@example.org"))
stopifnot(identical(parents(commit_1), list()))
stopifnot(identical(print(commit_1), commit_1))

## Check is_commit
stopifnot(identical(is_commit(commit_1), TRUE))
stopifnot(identical(is_commit(5), FALSE))

## Commit without adding changes should produce an error
tools::assertError(commit(repo, "Test to commit"))

## Add another commit
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Commit message 2")
summary(commit_2)
tag_2 <- tag(repo, "Tagname2", "Tag message 2")

## Check relationship
stopifnot(identical(descendant_of(commit_2, commit_1), TRUE))
stopifnot(identical(descendant_of(commit_1, commit_2), FALSE))
stopifnot(identical(descendant_of(tag_2, tag_1), TRUE))
stopifnot(identical(descendant_of(tag_1, tag_2), FALSE))
stopifnot(identical(descendant_of(branches(repo)[[1]], commit_1), TRUE))
stopifnot(identical(descendant_of(commit_1, branches(repo)[[1]]), FALSE))
stopifnot(identical(length(parents(commit_2)), 1L))
stopifnot(identical(parents(commit_2)[[1]], commit_1))

## Check contributions
stopifnot(identical(
    colnames(contributions(repo, by = "author", breaks = "day")),
    c("when", "author", "n")))
stopifnot(identical(colnames(contributions(repo)),
                    c("when", "n")))
stopifnot(identical(nrow(contributions(repo)), 1L))
stopifnot(identical(contributions(repo)$n, 2L))
stopifnot(identical(contributions(repo, by = "author", breaks = "day")$n, 2L))

## Add another commit with 'all' argument
writeLines(c("Hello world!", "HELLO WORLD!", "HeLlO wOrLd!"),
           file.path(path, "test.txt"))
commit(repo, "Commit message 3", all = TRUE)

status_clean <- structure(list(staged = empty_named_list(),
                               unstaged = empty_named_list(),
                               untracked = empty_named_list()),
                          class = "git_status")
stopifnot(identical(status(repo), status_clean))

## Delete file and commit with 'all' argument
file.remove(file.path(path, "test.txt"))
commit(repo, "Commit message 4", all = TRUE)

stopifnot(identical(status(repo), status_clean))

## Add and commit multiple tracked files with 'all' argument
writeLines(sample(letters, 3), file.path(path, "test2.txt"))
add(repo, "test2.txt")
writeLines(sample(letters, 3), file.path(path, "test3.txt"))
add(repo, "test3.txt")
writeLines(sample(letters, 3), file.path(path, "test4.txt"))
add(repo, "test4.txt")
commit(repo, "Commit message 5")

stopifnot(identical(status(repo), status_clean))

writeLines(sample(letters, 3), file.path(path, "test2.txt"))
writeLines(sample(letters, 3), file.path(path, "test3.txt"))
writeLines(sample(letters, 3), file.path(path, "test4.txt"))
commit(repo, "Commit message 6", all = TRUE)

stopifnot(identical(status(repo), status_clean))

## Add one tracked file and delete another with 'all' argument
writeLines(sample(letters, 3), file.path(path, "test2.txt"))
file.remove(file.path(path, "test4.txt"))
commit(repo, "Commit message 7", all = TRUE)

stopifnot(identical(status(repo), status_clean))

## Delete multiple tracked files with 'all' argument
file.remove(file.path(path, "test2.txt"))
file.remove(file.path(path, "test3.txt"))
commit(repo, "Commit message 8", all = TRUE)

stopifnot(identical(status(repo), status_clean))

## Check max number of commits in output
stopifnot(identical(length(commits(repo)), 8L))
stopifnot(identical(length(commits(repo, n = -1)), 8L))
stopifnot(identical(length(commits(repo, n = 2)), 2L))
tools::assertError(commits(repo, n = 2.2))
tools::assertError(commits(repo, n = "2"))
tools::assertError(commits(repo, n = 1:2))

## Check to coerce repository to data.frame
df <- as.data.frame(repo)
stopifnot(identical(dim(df), c(8L, 6L)))
stopifnot(identical(names(df), c("sha", "summary", "message",
                                 "author", "email", "when")))

## Set working directory to path and check commits
setwd(path)
stopifnot(identical(sha(last_commit()), sha(commits(repo, n = 1)[[1]])))
stopifnot(identical(length(commits()), 8L))
stopifnot(identical(length(commits(n = -1)), 8L))
stopifnot(identical(length(commits(n = 2)), 2L))
tools::assertError(commits(n = 2.2))
tools::assertError(commits(n = "2"))

## Check plot method
plot_file <- tempfile(fileext = ".pdf")
pdf(plot_file)
plot(repo)
dev.off()
stopifnot(file.exists(plot_file))
unlink(plot_file)

## Check punch card plot method
punch_card_plot_file <- tempfile(fileext = ".pdf")
pdf(punch_card_plot_file)
punch_card(repo)
dev.off()
stopifnot(file.exists(punch_card_plot_file))
unlink(punch_card_plot_file)

## Check that 'git2r_arg_check_commit' raise error
res <- tools::assertError(.Call(git2r:::git2r_commit_tree, NULL))
stopifnot(length(grep("'commit' must be an S3 class git_commit",
                      res[[1]]$message)) > 0)
res <- tools::assertError(.Call(git2r:::git2r_commit_tree, 3))
stopifnot(length(grep("'commit' must be an S3 class git_commit",
                      res[[1]]$message)) > 0)
res <- tools::assertError(.Call(git2r:::git2r_commit_tree, repo))
stopifnot(length(grep("'commit' must be an S3 class git_commit",
                      res[[1]]$message)) > 0)
commit_1$sha <- NA_character_
res <- tools::assertError(.Call(git2r:::git2r_commit_tree, commit_1))
stopifnot(length(grep("'commit' must be an S3 class git_commit",
                      res[[1]]$message)) > 0)

## Cleanup
unlink(path, recursive = TRUE)

if (identical(Sys.getenv("NOT_CRAN"), "true") ||
    identical(Sys.getenv("R_COVR"), "true")) {
    path <- tempfile(pattern = "git2r-")
    dir.create(path)
    setwd(path)
    system("git clone --depth 2 https://github.com/ropensci/git2r.git")

    ## Check the number of commits in the shallow clone.
    stopifnot(identical(length(commits(repository("git2r"))), 2L))
    stopifnot(identical(length(commits(repository("git2r"), n = 1)), 1L))

    ## Cleanup
    unlink(path, recursive = TRUE)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message 1")

## Create another commit
writeLines(c("Hello world!",
             "HELLO WORLD!"),
           file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Commit message 2")

## Check default ref
stopifnot(identical(note_default_ref(repo),
                    "refs/notes/commits"))

## Check that an invalid object argument in note_create produce an
## error.
tools::assertError(note_create(object = NULL, message = "test"))
tools::assertError(note_create(object = 1, message = "test"))

## Check that notes is an empty list
stopifnot(identical(notes(repo), list()))

## Create note in default namespace
note_1 <- note_create(commit_1, "Note-1")
stopifnot(identical(print(note_1), note_1))
stopifnot(identical(length(notes(repo)), 1L))
stopifnot(identical(sha(note_1), note_1$sha))
tools::assertError(note_create(commit_1, "Note-2"))
note_2 <- note_create(commit_1, "Note-2", force = TRUE)
stopifnot(identical(length(notes(repo)), 1L))

## Check that an invalid note argument in note_remove produce an
## error.
tools::assertError(note_remove(note = 1))

## Create note in named (review) namespace
note_3 <- note_create(commit_1, "Note-3", ref = "refs/notes/review")
note_4 <- note_create(commit_2, "Note-4", ref = "refs/notes/review")
stopifnot(identical(length(notes(repo, ref = "refs/notes/review")), 2L))
note_remove(note_3)
note_remove(note_4)
stopifnot(identical(notes(repo, ref = "refs/notes/review"), list()))
note_5 <- note_create(commit_1, "Note-5", ref = "review")
note_6 <- note_create(commit_2, "Note-6", ref = "review")
stopifnot(identical(length(notes(repo, ref = "review")), 2L))
note_remove(note_5)
note_remove(note_6)
stopifnot(identical(length(notes(repo, ref = "review")), 0L))

## Create note on blob and tree
tree_1 <- tree(commit_1)
note_7 <- note_create(tree_1, "Note-7")
stopifnot(is(object = lookup(repo, note_7$annotated), class2 = "git_tree"))
stopifnot(identical(length(notes(repo)), 2L))
blob_1 <- lookup(repo, tree_1$id[1])
note_8 <- note_create(blob_1, "Note-8")
stopifnot(is(object = lookup(repo, note_8$annotated), class2 = "git_blob"))
stopifnot(identical(length(notes(repo)), 3L))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create first commit
writeLines("Hello world!", file.path(path, "test-1.txt"))
add(repo, "test-1.txt")
commit_1 <- commit(repo, "First commit message")
tag_1 <- tag(repo, "Tag1", "First tag message")

## Create and checkout dev branch in repo
checkout(repo, "dev", create = TRUE)

## Create second commit
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-2.txt"))
add(repo, "test-2.txt")
commit_2 <- commit(repo, "Second commit message")
tag_2 <- tag(repo, "Tag2", "Second tag message")

## Check files
stopifnot(identical(list.files(path), c("test-1.txt", "test-2.txt")))

## Checkout tag_1 and check files
checkout(tag_1)
stopifnot(identical(list.files(path), "test-1.txt"))

## Checkout tag_2 and check files
checkout(tag_2)
stopifnot(identical(list.files(path), c("test-1.txt", "test-2.txt")))

## Checkout "Tag1" and check files
checkout(repo, "Tag1")
stopifnot(identical(list.files(path), "test-1.txt"))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create 2 directories in tempdir
path_bare <- tempfile(pattern = "git2r-")
path_repo_1 <- tempfile(pattern = "git2r-")
path_repo_2 <- tempfile(pattern = "git2r-")

dir.create(path_bare)
dir.create(path_repo_1)
dir.create(path_repo_2)

## Create repositories
bare_repo <- init(path_bare, bare = TRUE)
repo_1 <- clone(path_bare, path_repo_1)
repo_2 <- clone(path_bare, path_repo_2)

## Config repositories
config(repo_1, user.name = "Alice", user.email = "alice@example.org")
config(repo_2, user.name = "Bob", user.email = "bob@example.org")

## Add changes to repo 1
writeLines("Hello world", con = file.path(path_repo_1, "test.txt"))
add(repo_1, "test.txt")
commit_1 <- commit(repo_1, "Commit message")
branch_name <- branches(repo_1)[[1]]$name

## Push changes from repo 1 to origin
push(repo_1, "origin", paste0("refs/heads/", branch_name))

## Check result in bare repository
stopifnot(identical(length(commits(bare_repo)), 1L))
bare_commit_1 <- commits(bare_repo)[[1]]
stopifnot(identical(sha(commit_1), sha(bare_commit_1)))
stopifnot(identical(commit_1$author, bare_commit_1$author))
stopifnot(identical(commit_1$committer, bare_commit_1$committer))
stopifnot(identical(commit_1$summary, bare_commit_1$summary))
stopifnot(identical(commit_1$message, bare_commit_1$message))
stopifnot(!identical(commit_1$repo, bare_commit_1$repo))

## Fetch
fetch(repo_2, "origin")
fh <- fetch_heads(repo_2)[[1]]
stopifnot(identical(sha(fh), fh$sha))

## Test show method of non-empty repository where head is null
show(repo_2)

## Check that 'git2r_arg_check_credentials' raise error
res <- tools::assertError(
                  .Call(git2r:::git2r_remote_fetch, repo_1, "origin",
                        3, "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", repo_1,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- cred_env(c("username", "username"), "password")
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- cred_env("username", c("password", "passowrd"))
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- cred_user_pass(c("username", "username"), "password")
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- cred_user_pass("username", c("password", "passowrd"))
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- cred_token(c("GITHUB_PAT", "GITHUB_PAT"))
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- structure(list(publickey  = c("id_rsa.pub", "id_rsa.pub"),
                              privatekey = "id_rsa",
                              passphrase = character(0)),
                         class = "cred_ssh_key")
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- structure(list(publickey  = "id_rsa.pub",
                              privatekey = c("id_rsa", "id_rsa"),
                              passphrase = character(0)),
                         class = "cred_ssh_key")
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- structure(list(publickey  = "id_rsa.pub",
                              privatekey = "id_rsa",
                              passphrase = NA_character_),
                         class = "cred_ssh_key")
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- structure(list(publickey  = "id_rsa.pub",
                              privatekey = "id_rsa",
                              passphrase = c("passphrase", "passphrase")),
                         class = "cred_ssh_key")
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

## Cleanup
unlink(path_bare, recursive = TRUE)
unlink(path_repo_1, recursive = TRUE)
unlink(path_repo_2, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a bare repository
repo <- init(path, bare = TRUE)

## Check that the state of the repository
stopifnot(identical(is_bare(repo), TRUE))
stopifnot(identical(is_empty(repo), TRUE))

## Check that workdir is NULL for a bare repository
stopifnot(is.null(workdir(repo)))

## Check with missing repo argument
setwd(path)
stopifnot(identical(is_bare(), TRUE))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create 2 directories in tempdir
path_bare <- tempfile(pattern = "git2r-")
path_repo <- tempfile(pattern = "git2r-")

dir.create(path_bare)
dir.create(path_repo)

## Create repositories
bare_repo <- init(path_bare, bare = TRUE)
repo <- clone(path_bare, path_repo)

## Check the repositores
stopifnot(identical(is_bare(bare_repo), TRUE))
stopifnot(identical(is_bare(repo), FALSE))

## Config repositories
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Add changes to repo
writeLines("Hello world", con = file.path(path_repo, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message")
branch_name <- branches(repo)[[1]]$name

## Check commit
stopifnot(identical(commit_1$author$name, "Alice"))
stopifnot(identical(commit_1$author$email, "alice@example.org"))
stopifnot(identical(length(commits(repo)), 1L))
stopifnot(identical(commits(repo)[[1]]$author$name, "Alice"))
stopifnot(identical(commits(repo)[[1]]$author$email, "alice@example.org"))

## Check push arguments
tools::assertError(push(repo, character(0), paste0("refs/heads/", branch_name)))
tools::assertError(push(repo, NA_character_, paste0("refs/heads/", branch_name)))
tools::assertError(push(repo, c("origin", "origin"), paste0("refs/heads/", branch_name)))
tools::assertError(push(repo, "origin"))
tools::assertError(push(repo, name = "origin"))
push(repo, "origin", character(0))
push(repo, "origin", NA_character_)
push(repo, "origin", c(NA_character_, NA_character_))
stopifnot(identical(reflog(repo, paste0("refs/remotes/origin/", branch_name)),
                    structure(list(), class = "git_reflog")))

## No tracking branch assigned to master
tools::assertError(push(branches(repo)[[1]]))

## Push changes from repo to origin
push(repo, "origin", paste0("refs/heads/", branch_name))
r <- reflog(repo, paste0("refs/remotes/origin/", branch_name))
stopifnot(identical(length(r), 1L))
r <- r[[1]]
stopifnot(identical(sha(r), sha(commit_1)))
stopifnot(identical(r$message, "update by push"))
stopifnot(identical(r$index, 0L))
stopifnot(identical(r$committer$name, "Alice"))
stopifnot(identical(r$committer$email, "alice@example.org"))
stopifnot(identical(r$refname, paste0("refs/remotes/origin/", branch_name)))
stopifnot(identical(r$repo$path, repo$path))
push(branches(repo)[[1]])

## Check result in bare repository
stopifnot(identical(length(commits(bare_repo)), 1L))
bare_commit_1 <- commits(bare_repo)[[1]]
stopifnot(identical(sha(commit_1), sha(bare_commit_1)))
stopifnot(identical(commit_1$author, bare_commit_1$author))
stopifnot(identical(commit_1$committer, bare_commit_1$committer))
stopifnot(identical(commit_1$summary, bare_commit_1$summary))
stopifnot(identical(commit_1$message, bare_commit_1$message))
stopifnot(!identical(commit_1$repo, bare_commit_1$repo))

## Add changes to repo and push head
writeLines(c("Hello world", "HELLO WORLD"),
           con = file.path(path_repo, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Commit message 2")
push(repo)
bare_commit_2 <- lookup(bare_repo, sha(commit_2))
stopifnot(identical(sha(commit_2), sha(bare_commit_2)))
stopifnot(identical(commit_2$author, bare_commit_2$author))
stopifnot(identical(commit_2$committer, bare_commit_2$committer))
stopifnot(identical(commit_2$summary, bare_commit_2$summary))
stopifnot(identical(commit_2$message, bare_commit_2$message))
stopifnot(!identical(commit_2$repo, bare_commit_2$repo))

## Check 'set_upstream'
branch_set_upstream(repository_head(repo), NULL)
push(repo, "origin", paste0("refs/heads/", branch_name))
stopifnot(is.null(branch_get_upstream(repository_head(repo))))
push(repo, "origin", paste0("refs/heads/", branch_name), set_upstream = TRUE)
stopifnot(!is.null(branch_get_upstream(repository_head(repo))))

## Cleanup
unlink(path_bare, recursive = TRUE)
unlink(path_repo, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Initialize a temporary repository
path <- tempfile(pattern = "git2r-")
dir.create(path)
repo <- init(path, branch = "main")

## Create a user and commit a file
config(repo, user.name = "Author", user.email = "author@example.org")
writeLines(c("First line in file 1.", "Second line in file 1."),
           file.path(path, "example-1.txt"))
add(repo, "example-1.txt")
commit(repo, "First commit message")

## Create and add one more file
writeLines(c("First line in file 2.", "Second line in file 2."),
           file.path(path, "example-2.txt"))
add(repo, "example-2.txt")
commit(repo, "Second commit message")

## Create a new branch 'fix'
checkout(repo, "fix", create = TRUE)

## Update 'example-1.txt' (swap words in first line) and commit
writeLines(c("line First in file 1.", "Second line in file 1."),
           file.path(path, "example-1.txt"))
add(repo, "example-1.txt")
commit(repo, "Third commit message")

checkout(repo, "main")

## Update 'example-2.txt' (swap words in second line) and commit
writeLines(c("First line in file 2.", "line Second in file 2."),
           file.path(path, "example-2.txt"))
add(repo, "example-2.txt")
commit(repo, "Fourth commit message")

# Missing branch to merge with should throw an error
tools::assertError(merge(repo))

## Merge 'fix'
m <- merge(repo, "fix", TRUE, default_signature(repo))
stopifnot(identical(format(m), "Merge"))

## Merge 'fix' again
m <- merge(repo, "fix", TRUE, default_signature(repo))
stopifnot(identical(format(m), "Already up-to-date"))

## Check number of parents of each commit
stopifnot(identical(sapply(commits(repo), function(x) length(parents(x))),
                    c(2L, 1L, 1L, 1L, 0L)))

## Check that last commit is a merge
stopifnot(is_merge(last_commit(repo)))
summary(last_commit(repo))

## Check that metadata associated with merge is removed
stopifnot(!file.exists(file.path(path, ".git", "MERGE_HEAD")))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)

## Config repository
cfg <- config(repo, user.name = "Alice", user.email = "alice@example.org")

## Check configuration
stopifnot(identical(print(cfg), cfg))
stopifnot("local" %in% names(cfg))
stopifnot("user.name" %in% names(cfg$local))
stopifnot(identical(cfg$local$user.name, "Alice"))
stopifnot(identical(cfg$local$user.email, "alice@example.org"))

## Check that config fails for non-character entry.
tools::assertError(config(repo, test = 5))

## Check config method with missing repo argument
wd <- setwd(path)
cfg <- config(user.name = "Alice", user.email = "alice@example.org")
stopifnot("local" %in% names(cfg))
stopifnot("user.name" %in% names(cfg$local))
stopifnot(identical(cfg$local$user.name, "Alice"))
stopifnot(identical(cfg$local$user.email, "alice@example.org"))
stopifnot(identical(git_config_files(repo = repo)$local,
                    git_config_files(repo = NULL)$local))
stopifnot(identical(git_config_files(repo = repo)$local,
                    git_config_files(repo = repo$path)$local))
if (!is.null(wd))
    setwd(wd)

## Delete entries
cfg <- config(repo, user.name = NULL, user.email = NULL)

## Check configuration
stopifnot(is.null(cfg$local$user.name))
stopifnot(is.null(cfg$local$user.email))

## Supply values as objects
user_name <- "Alice"
user_email <- "alice@example.org"
cfg <- config(repo, user.name = user_name, user.email = "alice@example.org")
stopifnot(identical(cfg$local$user.name, user_name))
stopifnot(identical(cfg$local$user.email, "alice@example.org"))
cfg <- config(repo, user.name = "Alice", user.email = user_email)
stopifnot(identical(cfg$local$user.name, "Alice"))
stopifnot(identical(cfg$local$user.email, user_email))

## Check git config files
cfg <- git_config_files(repo)
stopifnot(identical(nrow(cfg), 4L))
stopifnot(identical(names(cfg), c("file", "path")))
stopifnot(identical(cfg$file, c("system", "xdg", "global", "local")))
stopifnot(!is.na(cfg$path[4]))

## Check that the local config file is NA for an invalid repo
## argument.
stopifnot(is.na(git_config_files(5)$local))

## Check location of .gitconfig on Windows
if (identical(Sys.getenv("APPVEYOR"), "True")) {

  ## AppVeyor diagnostics
  str(Sys.getenv("USERPROFILE"))
  str(Sys.getenv("HOMEDRIVE"))
  str(normalizePath("~"))
  str(git_config_files())

  ## Temporarily move AppVeyor .gitconfig
  gitconfig_appveyor <- "C:/Users/appveyor/.gitconfig"
  gitconfig_tmp <- file.path(tempdir(), ".gitconfig")
  file.rename(gitconfig_appveyor, gitconfig_tmp)

  ## Test config() on Windows
  gitconfig_expected <- file.path(Sys.getenv("USERPROFILE"), ".gitconfig")
  ## .gitconfig should not be created if no configuration options specified
  config(global = TRUE)
  stopifnot(!file.exists(gitconfig_expected))
  ## .gitconfig should be created in the user's home directory
  config(global = TRUE, user.name = "name", user.email = "email")
  stopifnot(file.exists(gitconfig_expected))
  unlink(gitconfig_expected)
  ## .gitconfig should be created if user specifies option other than user.name
  ## and user.email
  config(global = TRUE, core.editor = "nano")
  stopifnot(file.exists(gitconfig_expected))
  unlink(gitconfig_expected)
  ## .gitconfig should not create a new .gitconfig if the user already has one
  ## in Documents/
  gitconfig_documents <- "~/.gitconfig"
  file.create(gitconfig_documents)
  config(global = TRUE, core.editor = "nano")
  stopifnot(!file.exists(gitconfig_expected))
  unlink(gitconfig_documents)

  ## Return AppVeyor .gitconfig
  file.rename(gitconfig_tmp, gitconfig_appveyor)
}

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
library(tools)
source("util/check.R")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
f <- file(file.path(path, "test.txt"), "wb")
writeChar("Hello world!\n", f, eos = NULL)
close(f)

## add and commit
add(repo, "test.txt")
new_commit <- commit(repo, "Commit message")

## Lookup blob
blob <- lookup(repo, "cd0875583aabe89ee197ea133980a9085d08e497")
stopifnot(isTRUE(is_blob(blob)))
stopifnot(identical(sha(blob), "cd0875583aabe89ee197ea133980a9085d08e497"))
stopifnot(identical(is_binary(blob), FALSE))
stopifnot(identical(blob, lookup(repo, "cd0875")))
stopifnot(identical(length(blob), 13L))
stopifnot(identical(content(blob), "Hello world!"))
stopifnot(identical(print(blob), blob))

## Add one more commit
f <- file(file.path(path, "test.txt"), "wb")
writeChar("Hello world!\nHELLO WORLD!\nHeLlO wOrLd!\n", f, eos = NULL)
close(f)
add(repo, "test.txt")
blob <- lookup(repo, tree(commit(repo, "New commit message"))$id[1])
stopifnot(identical(content(blob),
                    c("Hello world!", "HELLO WORLD!", "HeLlO wOrLd!")))
stopifnot(identical(rawToChar(content(blob, raw = TRUE)),
                    content(blob, split = FALSE)))

## Check content of binary file
set.seed(42)
x <- as.raw((sample(0:255, 1000, replace = TRUE)))
writeBin(x, con = file.path(path, "test.bin"))
add(repo, "test.bin")
commit(repo, "Add binary file")
blob <- tree(last_commit(repo))["test.bin"]
stopifnot(identical(content(blob), NA_character_))
stopifnot(identical(x, content(blob, raw = TRUE)))

## Hash
stopifnot(identical(hash("Hello, world!\n"),
                    "af5626b4a114abcb82d63db7c8082c3c4756e51b"))
stopifnot(identical(hash("test content\n"),
                    "d670460b4b4aece5915caf5c68d12f560a9fe3e4"))
stopifnot(identical(hash(c("Hello, world!\n",
                           "test content\n")),
                    c("af5626b4a114abcb82d63db7c8082c3c4756e51b",
                      "d670460b4b4aece5915caf5c68d12f560a9fe3e4")))
stopifnot(identical(hash(c("Hello, world!\n",
                           NA_character_,
                           "test content\n")),
                    c("af5626b4a114abcb82d63db7c8082c3c4756e51b",
                      NA_character_,
                      "d670460b4b4aece5915caf5c68d12f560a9fe3e4")))
stopifnot(identical(hash(character(0)), character(0)))

## Hash file
test_1_txt <- file(file.path(path, "test-1.txt"), "wb")
writeChar("Hello, world!\n", test_1_txt, eos = NULL)
close(test_1_txt)
test_2_txt <- file(file.path(path, "test-2.txt"), "wb")
writeChar("test content\n", test_2_txt, eos = NULL)
close(test_2_txt)
stopifnot(identical(hash("Hello, world!\n"),
                    hashfile(file.path(path, "test-1.txt"))))
stopifnot(identical(hash("test content\n"),
                    hashfile(file.path(path, "test-2.txt"))))
stopifnot(identical(hash(c("Hello, world!\n",
                           "test content\n")),
                    hashfile(c(file.path(path, "test-1.txt"),
                               file.path(path, "test-2.txt")))))
assertError(hashfile(c(file.path(path, "test-1.txt"),
                       NA_character_,
                       file.path(path, "test-2.txt"))))
stopifnot(identical(hashfile(character(0)), character(0)))

## Create blob from disk
tmp_file_1 <- tempfile()
tmp_file_2 <- tempfile()
f1 <- file(tmp_file_1, "wb")
writeChar("Hello, world!\n", f1, eos = NULL)
close(f1)
f2 <- file(tmp_file_2, "wb")
writeChar("test content\n", f2, eos = NULL)
close(f2)
blob_list_1 <- blob_create(repo, c(tmp_file_1, tmp_file_2), relative = FALSE)
unlink(tmp_file_1)
unlink(tmp_file_2)
stopifnot(identical(sapply(blob_list_1, "[[", "sha"),
                    c("af5626b4a114abcb82d63db7c8082c3c4756e51b",
                      "d670460b4b4aece5915caf5c68d12f560a9fe3e4")))

## Create blob from workdir
tmp_file_3 <- file.path(path, "test-workdir-1.txt")
tmp_file_4 <- file.path(path, "test-workdir-2.txt")
f3 <- file(tmp_file_3, "wb")
writeChar("Hello, world!\n", f3, eos = NULL)
close(f3)
f4 <- file(tmp_file_4, "wb")
writeChar("test content\n", f4, eos = NULL)
close(f4)
blob_list_2 <- blob_create(repo, c("test-workdir-1.txt",
                                   "test-workdir-2.txt"))
stopifnot(identical(sapply(blob_list_2, "[[", "sha"),
                    c("af5626b4a114abcb82d63db7c8082c3c4756e51b",
                      "d670460b4b4aece5915caf5c68d12f560a9fe3e4")))

## Test arguments
check_error(assertError(.Call(git2r:::git2r_blob_content, NULL, FALSE)),
            "'blob' must be an S3 class git_blob")
check_error(assertError(.Call(git2r:::git2r_blob_content, 3, FALSE)),
            "'blob' must be an S3 class git_blob")
check_error(assertError(.Call(git2r:::git2r_blob_content, repo, FALSE)),
            "'blob' must be an S3 class git_blob")

b <- blob_list_1[[1]]
b$sha <- NA_character_
check_error(assertError(.Call(git2r:::git2r_blob_content, b, FALSE)),
            "'blob' must be an S3 class git_blob")

check_error(assertError(hashfile(NA)), "invalid 'path' argument")

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create 2 directories in tempdir
path_src <- tempfile(pattern = "git2r-")
path_tgt <- tempfile(pattern = "git2r-")
dir.create(path_tgt)
dir.create(path_src)

## Initialize a repository
repo_src <- init(path_src)
config(repo_src, user.name = "Alice", user.email = "alice@example.org")

## Add commit to repo
filename <- "test.txt"
writeLines("Hello world", con = file.path(path_src, filename))
add(repo_src, "test.txt")
commit_src <- commit(repo_src, "Commit message")

## Check checkout argument
tools::assertError(clone(path_src, path_tgt, checkout = c(FALSE, TRUE)))
tools::assertError(clone(path_src, path_tgt, checkout = 1))
tools::assertError(clone(path_src, path_tgt, checkout = 1L))
tools::assertError(clone(path_src, path_tgt, checkout = "test"))

## Clone source to target repository without checking out any files
repo_tgt <- clone(path_src, path_tgt, checkout = FALSE)

## List files in the repositores
stopifnot(identical(list.files(path_src), filename))
stopifnot(identical(list.files(path_tgt), character(0)))

## Compare commits
stopifnot(identical(length(commits(repo_tgt)), 1L))
commit_tgt <- last_commit(repo_tgt)
stopifnot(identical(sha(last_commit(path_tgt)), sha(commit_tgt)))
stopifnot(identical(sha(commit_src), sha(commit_tgt)))
stopifnot(identical(commit_src$author, commit_tgt$author))
stopifnot(identical(commit_src$committer, commit_tgt$committer))
stopifnot(identical(commit_src$summary, commit_tgt$summary))
stopifnot(identical(commit_src$message, commit_tgt$message))
stopifnot(!identical(commit_src$repo, commit_tgt$repo))

## Cleanup
unlink(path_tgt, recursive = TRUE)
unlink(path_src, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file and commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "First commit message")
tag_1 <- tag(repo, "Tagname1", "Tag message 1")

## Change file and commit
writeLines(c("Hello world!", "HELLO WORLD!"),
           file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Second commit message")
tag_2 <- tag(repo, "Tagname2", "Tag message 2")

## Check ahead behind
stopifnot(identical(ahead_behind(commit_1, commit_2), c(0L, 1L)))
stopifnot(identical(ahead_behind(tag_1, tag_2), c(0L, 1L)))
stopifnot(identical(ahead_behind(tag_2, tag_1), c(1L, 0L)))
stopifnot(identical(ahead_behind(commit_1, branches(repo)[[1]]), c(0L, 1L)))
stopifnot(identical(ahead_behind(branches(repo)[[1]], commit_1), c(1L, 0L)))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(file.path(path, "bundle", "R"), recursive = TRUE)

## Initialize a repository
repo <- init(file.path(path, "bundle"))
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a DESCRIPTION file
writeLines(c(
    "package: bundle",
    "Title: Bundle Git Repository",
    "Description: Bundle a bare repository of the code in the 'inst' folder.",
    "Version: 0.1",
    "License: GPL-2",
    "Authors@R: person('Alice', role = c('aut', 'cre'),",
    "                  email = 'alice@example.org')"),
    con = file.path(path, "bundle", "DESCRIPTION"))
add(repo, file.path(path, "bundle", "DESCRIPTION"))
commit(repo, "Add DESCRIPTION file")

## Create R file
writeLines("f <- function(x, y) x+y",
           con = file.path(path, "bundle", "R", "bundle.R"))
add(repo, file.path(path, "bundle", "R", "bundle.R"))
commit(repo, "Add R file")

## Bundle package
bundle_r_package(repo)

## Fails if bundled package exists
tools::assertError(bundle_r_package(repo))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

empty_named_list <- function() {
    structure(list(), .Names = character(0))
}

## Raise an error if the error message doesn't match.
check_error <- function(current, target, exact = FALSE) {
    if (isTRUE(exact)) {
        stopifnot(identical(current[[1]]$message, target))
    } else {
        stopifnot(length(grep(target, current[[1]]$message)) > 0)
    }

    invisible(NULL)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)

## Config repository
config(repo, user.name = "Alice")

## Let's set one valid and one with variable with invalid format
res <- tools::assertWarning(config(repo,
                                   user.email = "alice@example.org",
                                   lol = "wut"))
stopifnot(length(grep("Variable was not in a valid format: 'lol'",
                      res[[1]]$message)) > 0)

cfg_exp <- structure(list(user.name = "Alice",
                          user.email = "alice@example.org",
                          "NA" = NULL),
                     .Names = c("user.name", "user.email", NA))

cfg_obs <- config(repo)$local
cfg_obs <- cfg_obs[c("user.name", "user.email", "lol")]
stopifnot(identical(cfg_obs, cfg_exp))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path, branch = "main")
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test.txt"))

## add and commit
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message")

## Check branch
stopifnot(identical(length(branches(repo)), 1L))
stopifnot(identical(is_head(branches(repo)[[1]]), TRUE))
stopifnot(identical(is_local(branches(repo)[[1]]), TRUE))
stopifnot(identical(branches(repo)[[1]]$name, "main"))
stopifnot(identical(branches(repo)[[1]], repository_head(repo)))
stopifnot(identical(branches(repo)$main, repository_head(repo)))

## Check branch argument
res <- tools::assertError(is_local(1))
stopifnot(length(grep("argument 'branch' must be a 'git_branch' object",
                      res[[1]]$message)) > 0)

## Check branch name argument
res <- tools::assertError(branch_set_upstream())
stopifnot(length(grep("Missing argument name",
                      res[[1]]$message)) > 0)

## Print branch
b <- repository_head(repo)
stopifnot(identical(print(b), b))

## Create a branch
b <- branch_create(commit_1, name = "test")
stopifnot(identical(b$name, "test"))
stopifnot(identical(b$type, 1L))
stopifnot(identical(sha(b), branch_target(b)))
stopifnot(identical(length(branches(repo)), 2L))
stopifnot(identical(branch_target(branches(repo)[[1]]),
                    branch_target(branches(repo)[[2]])))

## Check is_branch
stopifnot(identical(is_branch(b), TRUE))
stopifnot(identical(is_branch(5), FALSE))

## Add one more commit
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Another commit message")

## Now the first branch should have moved on
stopifnot(!identical(branch_target(branches(repo)[[1]]),
                     branch_target(branches(repo)[[2]])))

## Create a branch with the same name should fail
tools::assertError(branch_create(commit_2, name = "test"))

## Force it and check the branches are identical again
b <- branch_create(commit_2, name = "test", force = TRUE)
stopifnot(identical(branch_target(branches(repo)[[1]]),
                    branch_target(branches(repo)[[2]])))

## Test arguments
res <- tools::assertError(.Call(git2r:::git2r_branch_delete, NULL))
stopifnot(length(grep("'branch' must be an S3 class git_branch",
                      res[[1]]$message)) > 0)
res <- tools::assertError(.Call(git2r:::git2r_branch_delete, 3))
stopifnot(length(grep("'branch' must be an S3 class git_branch",
                      res[[1]]$message)) > 0)
res <- tools::assertError(.Call(git2r:::git2r_branch_delete, repo))
stopifnot(length(grep("'branch' must be an S3 class git_branch",
                      res[[1]]$message)) > 0)
b_tmp <- b
b_tmp$name <- NA_character_
res <- tools::assertError(.Call(git2r:::git2r_branch_delete, b_tmp))
stopifnot(length(grep("'branch' must be an S3 class git_branch",
                      res[[1]]$message)) > 0)
b_tmp <- b
b_tmp$type <- NA_integer_
res <- tools::assertError(.Call(git2r:::git2r_branch_delete, b_tmp))
stopifnot(length(grep("'branch' must be an S3 class git_branch",
                      res[[1]]$message)) > 0)
b_tmp$type <- 3L
res <- tools::assertError(.Call(git2r:::git2r_branch_delete, b_tmp))
stopifnot(length(grep("'branch' must be an S3 class git_branch",
                      res[[1]]$message)) > 0)

## Delete branch
branch_delete(b)
stopifnot(identical(length(branches(repo)), 1L))

## Add one more commit
writeLines(c("Hello world!", "HELLO WORLD!", "hello world"),
           file.path(path, "test.txt"))
add(repo, "test.txt")
commit_3 <- commit(repo, "Another third commit message")

## Create and test renaming of branches
b_1 <- branch_create(commit_1, name = "test-1")
b_2 <- branch_create(commit_2, name = "test-2")
b_3 <- branch_create(commit_3, name = "test-3")
stopifnot(identical(length(branches(repo)), 4L))
b_1 <- branch_rename(b_1, name = "test-1-new-name")
stopifnot(identical(length(branches(repo)), 4L))
stopifnot(identical(b_1$name, "test-1-new-name"))
tools::assertError(branch_rename(b_1, name = "test-2"))
branch_rename(b_1, name = "test-2", force = TRUE)
stopifnot(identical(length(branches(repo)), 3L))

## Check branches method with missing repo argument
wd <- setwd(path)
stopifnot(identical(length(branches()), 3L))
if (!is.null(wd))
    setwd(wd)

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file and commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "First commit message")

## Create new user and change file
config(repo, user.name = "Bob", user.email = "bob@example.org")
writeLines(c("Hello world!", "HELLO WORLD!", "HOLA"),
           file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Second commit message")

## Check blame
b <- blame(repo, "test.txt")
stopifnot(identical(length(b$hunks), 2L))

## Hunk: 1
stopifnot(identical(b$hunks[[1]]$lines_in_hunk, 1L))
stopifnot(identical(b$hunks[[1]]$final_commit_id, sha(commit_1)))
stopifnot(identical(b$hunks[[1]]$final_start_line_number, 1L))
stopifnot(identical(b$hunks[[1]]$final_signature$name, "Alice"))
stopifnot(identical(b$hunks[[1]]$final_signature$email, "alice@example.org"))
stopifnot(identical(b$hunks[[1]]$orig_commit_id, sha(commit_1)))
stopifnot(identical(b$hunks[[1]]$orig_start_line_number, 1L))
stopifnot(identical(b$hunks[[1]]$orig_signature$name, "Alice"))
stopifnot(identical(b$hunks[[1]]$orig_signature$email, "alice@example.org"))
stopifnot(identical(b$hunks[[1]]$orig_path, "test.txt"))
stopifnot(identical(b$hunks[[1]]$boundary, TRUE))

## Hunk: 2
stopifnot(identical(b$hunks[[2]]$lines_in_hunk, 2L))
stopifnot(identical(b$hunks[[2]]$final_commit_id, sha(commit_2)))
stopifnot(identical(b$hunks[[2]]$final_start_line_number, 2L))
stopifnot(identical(b$hunks[[2]]$final_signature$name, "Bob"))
stopifnot(identical(b$hunks[[2]]$final_signature$email, "bob@example.org"))
stopifnot(identical(b$hunks[[2]]$orig_commit_id, sha(commit_2)))
stopifnot(identical(b$hunks[[2]]$orig_start_line_number, 2L))
stopifnot(identical(b$hunks[[2]]$orig_signature$name, "Bob"))
stopifnot(identical(b$hunks[[2]]$orig_signature$email, "bob@example.org"))
stopifnot(identical(b$hunks[[2]]$orig_path, "test.txt"))
stopifnot(identical(b$hunks[[2]]$boundary, FALSE))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test-1.txt"))

## add and commit
add(repo, "test-1.txt")
commit(repo, "Commit message")

## Pop stash
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-1.txt"))
stash(repo)
stopifnot(identical("Hello world!",
                    readLines(file.path(path, "test-1.txt"))))
stash_pop(repo)
stopifnot(identical(c("Hello world!", "HELLO WORLD!"),
                    readLines(file.path(path, "test-1.txt"))))

## Make one more commit
add(repo, "test-1.txt")
commit(repo, "Next commit message")

## Check that there are no stashes
stopifnot(identical(stash_list(repo), list()))

## Apply stash
writeLines(c("Hello world!", "HELLO WORLD!", "hello world!"),
           file.path(path, "test-1.txt"))
stash(repo)
stopifnot(identical(c("Hello world!", "HELLO WORLD!"),
                    readLines(file.path(path, "test-1.txt"))))
stash_apply(repo)
stopifnot(identical(c("Hello world!", "HELLO WORLD!", "hello world!"),
                    readLines(file.path(path, "test-1.txt"))))
stopifnot(identical(length(stash_list(repo)), 1L))
stash_drop(repo, 1)
stopifnot(identical(stash_list(repo), list()))

## Make one more commit
add(repo, "test-1.txt")
commit(repo, "Apply stash commit message")

## Create one more file
writeLines("Hello world!", file.path(path, "test-2.txt"))

## Check that there are no stashes
stopifnot(identical(stash_list(repo), list()))

## Stash
stash(repo)
stopifnot(identical(stash_list(repo), list()))
s <- stash(repo, untracked = TRUE)
stopifnot(identical(print(s), s))
summary(s)
stopifnot(identical(length(stash_list(repo)), 1L))
tree(stash_list(repo)[[1]])

## Drop stash
stash_drop(repo, 1)
stopifnot(identical(stash_list(repo), list()))

## Check stash_drop argument
tools::assertError(stash_drop(repo))
tools::assertError(stash_drop(repo, -1))
tools::assertError(stash_drop(repo, 0.5))

## Create one more file
writeLines("Hello world!", file.path(path, "test-3.txt"))

## Create stash in repository
stash(repo, untracked = TRUE)
stopifnot(identical(length(stash_list(repo)), 1L))

## Check stash_list method with missing repo argument
wd <- setwd(path)
stopifnot(identical(length(stash_list()), 1L))
if (!is.null(wd))
    setwd(wd)

## Drop git_stash object in repository
stash_drop(stash_list(repo)[[1]])

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@@example.org")

## Create a file, add and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message 1")

## Change file and commit
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
             con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message 2")

## Commit same content under different name in a sub-directory
dir.create(file.path(path, "sub-directory"))
file.copy(file.path(path, "test.txt"),
          file.path(path, "sub-directory", "copy.txt"))
add(repo, "sub-directory/copy.txt")
commit(repo, "Commit message 3")

## List blobs
b <- odb_blobs(repo)

## Order the data.frame before checking
b <- b[order(b$name), ]

## Check blobs
stopifnot(identical(nrow(b), 3L))
stopifnot(identical(
    colnames(b),
    c("sha", "path", "name", "len", "commit", "author", "when")))
stopifnot(identical(b$path, c("sub-directory", "", "")))
stopifnot(identical(b$name, c("copy.txt", "test.txt", "test.txt")))
stopifnot(identical(b$author, c("Alice", "Alice", "Alice")))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Check printing of a class git_signature
when <- structure(list(time = 1395567947, offset = 60),
                  class = "git_time")

signature <- structure(list(name = "Alice",
                            email = "alice@example.org",
                            when = when),
                       class = "git_signature")
stopifnot(identical(print(signature), signature))

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path, branch = "main")
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create first commit
writeLines("Hello world!", file.path(path, "test-1.txt"))
add(repo, "test-1.txt")
commit(repo, "First commit message")

## Create and checkout dev branch in repo
checkout(repo, "dev", create = TRUE)

## Create second commit
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-2.txt"))
add(repo, "test-2.txt")
commit(repo, "Second commit message")

## Check files
stopifnot(identical(list.files(path), c("test-1.txt", "test-2.txt")))

## Checkout master branch and check files
checkout(repo, "main")
stopifnot(identical(list.files(path), "test-1.txt"))

## Cleanup
unlink(path, recursive = TRUE)

## Checkout branch in empty repository

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create and checkout dev branch in repo
checkout(repo, "dev", create = TRUE)

## Create first commit
writeLines("Hello world!", file.path(path, "test-1.txt"))
add(repo, "test-1.txt")
commit(repo, "First commit message")

stopifnot(identical(length(branches(repo)), 1L))
stopifnot(identical(branches(repo)[[1]]$name, "dev"))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create directories
dir.create(file.path(path, "sub-folder"));
dir.create(file.path(path, "sub-folder", "sub-sub-folder"));

## Create files
writeLines("Hello world!",
           file.path(path, "file-1.txt"))
writeLines("Hello world!",
           file.path(path, "sub-folder", "file-2.txt"))
writeLines("Hello world!",
           file.path(path, "sub-folder", "file-3.txt"))
writeLines("Hello world!",
           file.path(path, "sub-folder", "sub-sub-folder", "file-4.txt"))
writeLines("Hello world!",
           file.path(path, "sub-folder", "sub-sub-folder", "file-5.txt"))

## Add
add(repo, "file-1.txt")
status_exp <- structure(list(staged = list(new = "file-1.txt"),
                             unstaged = empty_named_list(),
                             untracked = list(untracked = "sub-folder/")),
                        class = "git_status")
status_obs <- status(repo)
stopifnot(identical(status_obs, status_exp))

## Index remove by path
index_remove_bypath(repo, "file-1.txt")
status_exp <- structure(list(staged = empty_named_list(),
                             unstaged = empty_named_list(),
                             untracked = list(untracked = "file-1.txt",
                                              untracked = "sub-folder/")),
                        class = "git_status")
status_obs <- status(repo)
stopifnot(identical(status_obs, status_exp))

## Add
add(repo, "sub-folder")
status_exp <- structure(list(staged = list(
                                 new = "sub-folder/file-2.txt",
                                 new = "sub-folder/file-3.txt",
                                 new = "sub-folder/sub-sub-folder/file-4.txt",
                                 new = "sub-folder/sub-sub-folder/file-5.txt"),
                             unstaged = empty_named_list(),
                             untracked = list(untracked = "file-1.txt")),
                        class = "git_status")
status_obs <- status(repo)
stopifnot(identical(status_obs, status_exp))

## Commit
commit(repo, "First commit message")

## It should fail to remove non-existing, untracked and ignored files
tools::assertError(rm_file(repo, c("file-1.txt", "file-2.txt")))
tools::assertError(rm_file(repo, c("file-1.txt", "")))
tools::assertError(rm_file(repo, c("file-1.txt")))
writeLines("/file-1.txt", file.path(path, ".gitignore"))
tools::assertError(rm_file(repo, "file-1.txt"))

## It should fail to remove files with staged changes
file.remove(file.path(path, ".gitignore"))
add(repo, "file-1.txt")
tools::assertError(rm_file(repo, "file-1.txt"))

## It should fail to remove files with unstaged changes
commit(repo, "Second commit message")
writeLines(c("Hello world!", "Hello world!"),
           file.path(path, "file-1.txt"))
tools::assertError(rm_file(repo, "file-1.txt"))

## Remove file
add(repo, "file-1.txt")
commit(repo, "Third commit message")
rm_file(repo, "file-1.txt")
status_exp <- structure(list(staged = list(deleted = "file-1.txt"),
                             unstaged = empty_named_list(),
                             untracked = empty_named_list()),
                        class = "git_status")
status_obs <- status(repo)
stopifnot(identical(status_obs, status_exp))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add, commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message")

## Change the file, diff between index and workdir
writeLines("Hello again!\nHere is a second line\nAnd a third",
           file.path(path, "test.txt"))
diff_1 <- diff(repo)
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(identical(diff_1$old, "index"))
stopifnot(identical(diff_1$new, "workdir"))
stopifnot(identical(length(diff_1$files), 1L))
stopifnot(identical(diff_1$files[[1]]$old_file, "test.txt"))
stopifnot(identical(diff_1$files[[1]]$new_file, "test.txt"))
stopifnot(identical(length(diff_1$files[[1]]$hunks), 1L))
stopifnot(identical(length(diff_1$files[[1]]$hunks[[1]]$lines), 4L))
## TODO: check actual diff

## Diff between index and HEAD is empty
diff_2 <- diff(repo, index = TRUE)
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(identical(diff_2$old, "HEAD"))
stopifnot(identical(diff_2$new, "index"))
stopifnot(identical(diff_2$files, list()))

## Diff between tree and working dir, same as diff_1
diff_3 <- diff(tree(commits(repo)[[1]]))
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(identical(diff_3$old, tree(commits(repo)[[1]])))
stopifnot(identical(diff_3$new, "workdir"))
stopifnot(identical(diff_3$files, diff_1$files))
stopifnot(identical(print(diff_3), diff_3))

## Add changes, diff between index and HEAD is the same as diff_1
add(repo, "test.txt")
diff_4 <- diff(repo, index = TRUE)
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(identical(diff_4$old, "HEAD"))
stopifnot(identical(diff_4$new, "index"))
stopifnot(identical(diff_4$files, diff_1$files))

## Diff between tree and index
diff_5 <- diff(tree(commits(repo)[[1]]), index = TRUE)
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(identical(diff_5$old, tree(commits(repo)[[1]])))
stopifnot(identical(diff_5$new, "index"))
stopifnot(identical(diff_5$files, diff_1$files))

## Diff between two trees
commit(repo, "Second commit")
tree_1 <- tree(commits(repo)[[2]])
tree_2 <- tree(commits(repo)[[1]])
diff_6 <- diff(tree_1, tree_2)
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(identical(diff_6$old, tree_1))
stopifnot(identical(diff_6$new, tree_2))
stopifnot(identical(diff_6$files, diff_1$files))

## Length of a diff
stopifnot(identical(length(diff_1), 1L))
stopifnot(identical(length(diff_2), 0L))
stopifnot(identical(length(diff_3), 1L))
stopifnot(identical(length(diff_4), 1L))
stopifnot(identical(length(diff_5), 1L))
stopifnot(identical(length(diff_6), 1L))

## Binary files
set.seed(42)
writeBin(as.raw((sample(0:255, 1000, replace = TRUE))),
         con = file.path(path, "test.bin"))
add(repo, "test.bin")
diff_7 <- diff(repo, index = TRUE)
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(any(grepl("binary file", capture.output(summary(diff_7)))))

## TODO: errors
## Check non-logical index argument
res <- tools::assertError(
                  .Call(git2r:::git2r_diff, NULL, NULL, NULL, "FALSE",
                        NULL, 3L, 0L, "a", "b", NULL, NULL, NULL))
stopifnot(length(grep(paste0("Error in 'git2r_diff': 'index' must be logical ",
                             "vector of length one with non NA value\n"),
                      res[[1]]$message)) > 0)

## Check various combinations of diff arguments
res <- tools::assertError(
                  .Call(git2r:::git2r_diff, NULL, NULL,
                        tree(commits(repo)[[1]]),
                        FALSE, NULL, 3L, 0L, "a", "b", NULL, NULL, NULL))
stopifnot(length(grep("Error in 'git2r_diff': Invalid diff parameters",
                      res[[1]]$message)) > 0)

res <- tools::assertError(
                  .Call(git2r:::git2r_diff, NULL, NULL,
                        tree(commits(repo)[[1]]),
                        TRUE, NULL, 3L, 0L, "a", "b", NULL, NULL, NULL))
stopifnot(length(grep("Error in 'git2r_diff': Invalid diff parameters",
                      res[[1]]$message)) > 0)

res <- tools::assertError(
                  .Call(git2r:::git2r_diff, repo, tree(commits(repo)[[1]]),
                        NULL, FALSE, NULL, 3L, 0L, "a", "b", NULL, NULL, NULL))
stopifnot(length(grep("Error in 'git2r_diff': Invalid diff parameters",
                      res[[1]]$message)) > 0)

res <- tools::assertError(
                  .Call(git2r:::git2r_diff, repo, tree(commits(repo)[[1]]),
                        NULL, TRUE, NULL, 3L, 0L, "a", "b", NULL, NULL, NULL))
stopifnot(length(grep("Error in 'git2r_diff': Invalid diff parameters",
                      res[[1]]$message)) > 0)

res <- tools::assertError(
                  .Call(git2r:::git2r_diff, repo, tree(commits(repo)[[1]]),
                        tree(commits(repo)[[2]]), FALSE, NULL, 3L, 0L, "a",
                        "b", NULL, NULL, NULL))
stopifnot(length(grep("Error in 'git2r_diff': Invalid diff parameters",
                      res[[1]]$message)) > 0)

res <- tools::assertError(
                  .Call(git2r:::git2r_diff, repo, tree(commits(repo)[[1]]),
                        tree(commits(repo)[[2]]), TRUE, NULL, 3L, 0L, "a",
                        "b", NULL, NULL, NULL))
stopifnot(length(grep("Error in 'git2r_diff': Invalid diff parameters",
                      res[[1]]$message)) > 0)

## TODO: printing

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create directories for repositories in tempdir
path_bare <- tempfile(pattern = "git2r-")
path_repo_1 <- tempfile(pattern = "git2r-")
path_repo_2 <- tempfile(pattern = "git2r-")

dir.create(path_bare)
dir.create(path_repo_1)
dir.create(path_repo_2)

## Create bare repository
bare_repo <- init(path_bare, bare = TRUE)

## Clone to repo 1
repo_1 <- clone(path_bare, path_repo_1)
config(repo_1, user.name = "Alice", user.email = "alice@example.org")

## Add changes to repo 1 and push to bare
writeLines(
    "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
    con = file.path(path_repo_1, "test.txt"))
add(repo_1, "test.txt")
commit(repo_1, "First commit message")
branch_name <- branches(repo_1)[[1]]$name
push(repo_1, "origin", paste0("refs/heads/", branch_name))

## Test checkout branch argument
tools::assertError(checkout(repo_1))
tools::assertError(checkout(repo_1, c("master", "master")))
tools::assertError(checkout(repo_1, "dev"))

## Create and checkout dev branch in repo 1
checkout(repo_1, "dev", create = TRUE)

## Add changes to dev branch in repo 1 and push to bare
writeLines(
    c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
      "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
    con = file.path(path_repo_1, "test.txt"))
add(repo_1, "test.txt")
commit(repo_1, "Second commit message")
push(repo_1, "origin", "refs/heads/dev")

## Clone to repo 2
repo_2 <- clone(path_bare, path_repo_2)
config(repo_2, user.name = "Bob", user.email = "bob@example.org")

stopifnot(identical(
    "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
    readLines(file.path(path_repo_2, "test.txt"))))

## Checkout dev branch
checkout(repo_2, "dev")

## Check content of file
stopifnot(identical(
    c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
      "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
    readLines(file.path(path_repo_2, "test.txt"))))

## Checkout previous branch
checkout(repo_2, "-")
stopifnot(identical(repository_head(repo_2)$name, branch_name))

## Cleanup
unlink(path_bare, recursive = TRUE)
unlink(path_repo_1, recursive = TRUE)
unlink(path_repo_2, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)
setwd(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Test to add file with a leading './'
writeLines("foo-1", file.path(path, "foo-1"))
add(repo, "./foo-1")
status_exp <- structure(list(staged = list(new = "foo-1"),
                             unstaged = empty_named_list(),
                             untracked = empty_named_list()),
                        class = "git_status")
status_obs <- status(repo)
str(status_exp)
str(status_obs)
stopifnot(identical(status_obs, status_exp))

## Test to add file in sub-folder with sub-folder as working directory
writeLines("foo-2", file.path(path, "foo-2"))
dir.create(file.path(path, "foo_dir"))
writeLines("foo-2", file.path(path, "foo_dir/foo-2"))
setwd("./foo_dir")
add(repo, "foo-2")
status_exp <- structure(list(staged = list(new = "foo-1",
                                           new = "foo_dir/foo-2"),
                             unstaged = empty_named_list(),
                             untracked = list(untracked = "foo-2")),
                        class = "git_status")
status_obs <- status(repo)
str(status_exp)
str(status_obs)
stopifnot(identical(status_obs, status_exp))

## Test glob expansion
setwd(tempdir())
dir.create(file.path(path, "glob_dir"))
writeLines("a", file.path(path, "glob_dir/a.txt"))
writeLines("b", file.path(path, "glob_dir/b.txt"))
writeLines("c", file.path(path, "glob_dir/c.txt"))
writeLines("d", file.path(path, "glob_dir/d.md"))
add(repo, "glob_dir/*txt")
status_exp <- structure(list(staged = list(new = "foo-1",
                                           new = "foo_dir/foo-2",
                                           new = "glob_dir/a.txt",
                                           new = "glob_dir/b.txt",
                                           new = "glob_dir/c.txt"),
                             unstaged = empty_named_list(),
                             untracked = list(untracked = "foo-2",
                                              untracked = "glob_dir/d.md")),
                        class = "git_status")
status_obs <- status(repo)
str(status_exp)
str(status_obs)
stopifnot(identical(status_obs, status_exp))

## Test glob expansion with relative path
setwd(path)
add(repo, "./glob_dir/*md")
status_exp <- structure(list(staged = list(new = "foo-1",
                                           new = "foo_dir/foo-2",
                                           new = "glob_dir/a.txt",
                                           new = "glob_dir/b.txt",
                                           new = "glob_dir/c.txt",
                                           new = "glob_dir/d.md"),
                             unstaged = empty_named_list(),
                             untracked = list(untracked = "foo-2")),
                        class = "git_status")
status_obs <- status(repo)
str(status_exp)
str(status_obs)
stopifnot(identical(status_obs, status_exp))

## Test to add file in root of workdir when the file also exists in
## current workdir.
setwd(tempdir())
writeLines("e", file.path(path, "e.txt"))
writeLines("e", file.path(tempdir(), "e.txt"))
add(repo, "e.txt")
status_exp <- structure(list(staged = list(new = "e.txt",
                                           new = "foo-1",
                                           new = "foo_dir/foo-2",
                                           new = "glob_dir/a.txt",
                                           new = "glob_dir/b.txt",
                                           new = "glob_dir/c.txt",
                                           new = "glob_dir/d.md"),
                             unstaged = empty_named_list(),
                             untracked = list(untracked = "foo-2")),
                        class = "git_status")
status_obs <- status(repo)
str(status_exp)
str(status_obs)
stopifnot(identical(status_obs, status_exp))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## is_bare: "Invalid repository"
tools::assertError(is_bare(new("git_repository")))

## is_empty: "Invalid repository"
tools::assertError(is_empty(new("git_repository")))

## Check that open an invalid repository fails
tools::assertError(repository(path))
tools::assertError(repository(path, discover = FALSE))

## Check that it fails to open/init a repository with a path to a
## file.
writeLines("test", file.path(path, "test.txt"))
tools::assertError(repository(file.path(path, "test.txt"),
                              discover = FALSE))
tools::assertError(init(file.path(path, "test.txt")))
unlink(file.path(path, "test.txt"))

## Initialize a repository
repo <- init(path)
stopifnot(identical(print(repo), repo))

## Check the state of the repository
stopifnot(identical(is_bare(repo), FALSE))
stopifnot(identical(is_empty(repo), TRUE))
stopifnot(identical(is_shallow(repo), FALSE))
stopifnot(identical(branches(repo), empty_named_list()))
stopifnot(identical(references(repo), empty_named_list()))
stopifnot(identical(commits(repo), list()))
stopifnot(identical(repository_head(repo), NULL))

# check that we can find repository from a path
wd <- sub(paste0("[", .Platform$file.sep, "]$"), "",  workdir(repo))
writeLines("test file", con = file.path(wd, "myfile.txt"))
stopifnot(identical(discover_repository(file.path(wd, "myfile.txt")),
                    file.path(wd, ".git")))
stopifnot(identical(discover_repository(file.path(wd, "doesntexist.txt")),
                    NULL))

# Check that we can use ceiling in discover repostiory
dir.create(file.path(wd, "temp"))
stopifnot(identical(discover_repository(file.path(wd, "temp"), 0), NULL))
stopifnot(identical(discover_repository(file.path(wd, "temp"), 1),
                    file.path(wd, ".git")))
tools::assertError(discover_repository(file.path(wd, "temp"), 2))

## Check that lookup with a sha of less than 4 characters or more than
## 40 characters fail.
tools::assertError(lookup(repo, paste0(rep("a", 3), collapse = "")))
tools::assertError(lookup(repo, paste0(rep("a", 41), collapse = "")))

## Check in_repository
stopifnot(identical(in_repository(path), TRUE))

## Check:
## - in_repository method with missing path argument
## - repository method with missing path argument
## - workdir method with missing path argument
## - is_empty method with missing repo argument
## - is_shallow method with missing repo argument
wd <- setwd(path)
stopifnot(identical(in_repository(), TRUE))
stopifnot(identical(workdir(repository(path)), workdir(repository())))
stopifnot(identical(workdir(repository(path)), workdir()))
stopifnot(identical(is_empty(), TRUE))
stopifnot(identical(is_shallow(), FALSE))
if (!is.null(wd))
    setwd(wd)

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Add files
invisible(lapply(file.path(path, letters[1:4]), writeLines, text = ""))
add(repo, letters)
commit(repo, "init")

## Remove one file
rm_file(repo, letters[1])
commit(repo, "remove")

## Remove two files. Don't raise warnings
withCallingHandlers(rm_file(repo, letters[2:3]), warning = function(w) stop(w))

## Remove one file using the absolute path to the file.
rm_file(repo, file.path(path, letters[4]))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013 - 2022 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Create a new environmental credential object
##'
##' Environmental variables can be written to the file
##' \code{.Renviron}. This file is read by \emph{R} during startup,
##' see \code{\link[base]{Startup}}.
##' @family git credential functions
##' @param username The name of the environmental variable that holds
##'     the username for the authentication.
##' @param password The name of the environmental variable that holds
##'     the password for the authentication.
##' @return A list of class \code{cred_env} with entries:
##' \describe{
##'   \item{username}{
##'     The name of the environmental variable that holds
##'     the username for the authentication.
##'   }
##'   \item{password}{
##'     The name of the environmental variable that holds
##'     the password for the authentication.
##'   }
##' }
##' @export
##' @examples
##' \dontrun{
##' ## Create an environmental credential object for the username and
##' ## password.
##' cred <- cred_env("NAME_OF_ENV_VARIABLE_WITH_USERNAME",
##'                  "NAME_OF_ENV_VARIABLE_WITH_PASSWORD")
##' repo <- repository("git2r")
##' push(repo, credentials = cred)
##' }
cred_env <- function(username = NULL, password = NULL) {
  structure(list(username = username, password = password),
            class = "cred_env")
}

##' Create a new personal access token credential object
##'
##' The personal access token is stored in an envrionmental variable.
##' Environmental variables can be written to the file
##' \code{.Renviron}. This file is read by \emph{R} during startup,
##' see \code{\link[base]{Startup}}. On GitHub, personal access tokens
##' function like ordinary OAuth access tokens. They can be used
##' instead of a password for Git over HTTPS, see the \dQuote{Creating
##' a personal access token} article on GitHub Docs.
##' @family git credential functions
##' @param token The name of the environmental variable that holds the
##'     personal access token for the authentication. Default is
##'     \code{GITHUB_PAT}.
##' @return A list of class \code{cred_token} with entry:
##' \describe{
##'   \item{token}{
##'     The name of the environmental variable that holds
##'     the personal access token for the authentication.
##'   }
##' }
##' @export
##' @examples
##' \dontrun{
##' ## Create a personal access token credential object.
##' ## This example assumes that the token is stored in
##' ## the 'GITHUB_PAT' environmental variable.
##' repo <- repository("git2r")
##' cred <- cred_token()
##' push(repo, credentials = cred)
##' }
cred_token <- function(token = "GITHUB_PAT") {
  structure(list(token = token), class = "cred_token")
}

##' Create a new plain-text username and password credential object
##'
##' @family git credential functions
##' @param username The username of the credential
##' @param password The password of the credential. If getPass is installed
##'     and the only input is username, \code{getPass::getPass()} will be
##'     called to allow for interactive and obfuscated interactive
##'     input of the password.
##' @return A list of class \code{cred_user_pass} with entries:
##' \describe{
##'   \item{username}{
##'     The username of the credential
##'   }
##'   \item{password}{
##'     The password of the credential
##'   }
##' }
##' @export
##' @examples
##' \dontrun{
##' ## Create a plain-text username and password credential object
##' cred_user_pass("Random Developer", "SecretPassword")
##' }
cred_user_pass <- function(username = NULL, password = NULL) {

  if (!is.null(username)) {
    if (is.null(password)) {
      if (requireNamespace("getPass", quietly = TRUE)) {
        password <- getPass::getPass()
      }
    }
  }

  structure(list(username = username, password = password),
            class = "cred_user_pass")
}

##' Create a new passphrase-protected ssh key credential object
##'
##' @family git credential functions
##' @param publickey The path to the public key of the
##'     credential. Default is \code{ssh_path("id_rsa.pub")}
##' @param privatekey The path to the private key of the
##'     credential. Default is \code{ssh_path("id_rsa")}
##' @param passphrase The passphrase of the credential. Default is
##'     \code{character(0)}. If getPass is installed and private key
##'     is passphrase protected \code{getPass::getPass()} will be
##'     called to allow for interactive and obfuscated interactive
##'     input of the passphrase.
##' @return A list of class \code{cred_ssh_key} with entries:
##' \describe{
##'   \item{publickey}{
##'     The path to the public key of the credential
##'   }
##'   \item{privatekey}{
##'     The path to the private key of the credential
##'   }
##'   \item{passphrase}{
##'     The passphrase of the credential
##'   }
##' }
##' @export
##' @examples
##' \dontrun{
##' ## Create a ssh key credential object. It can optionally be
##' ## passphrase-protected
##' cred <- cred_ssh_key(ssh_path("id_rsa.pub"), ssh_path("id_rsa"))
##' repo <- repository("git2r")
##' push(repo, credentials = cred)
##' }
cred_ssh_key <- function(publickey = ssh_path("id_rsa.pub"),
                         privatekey = ssh_path("id_rsa"),
                         passphrase = character(0)) {
  publickey <- normalizePath(publickey, mustWork = TRUE)
  privatekey <- normalizePath(privatekey, mustWork = TRUE)

  if (length(passphrase) == 0) {
    if (ssh_key_needs_passphrase(privatekey)) {
      if (requireNamespace("getPass", quietly = TRUE)) {
        passphrase <- getPass::getPass()
      }
    }
  }

  structure(list(publickey  = publickey,
                 privatekey = privatekey,
                 passphrase = passphrase),
            class = "cred_ssh_key")
}

##' Check if private key is passphrase protected
##' @param privatekey The path to the private key of the
##' credential. Default is \code{ssh_path("id_rsa")}
##' @noRd
ssh_key_needs_passphrase <- function(privatekey = ssh_path("id_rsa")) {
  private_content <- readLines(privatekey, n = 3)
  contains_encrypted <- grepl("encrypted", private_content, ignore.case = TRUE)
  any(contains_encrypted)
}

##' Compose usual path to ssh keys
##'
##' This function provides a consistent means across OS-types to access the
##' \code{.ssh} directory.
##'
##' On Windows-based systems,
##' \code{path.expand("~")} returns \code{"C:/Users/username/Documents"},
##' whereas the usual path to the \code{.ssh} directory is
##' \code{"C:/Users/username"}.
##'
##' On other operating systems, \code{path.expand("~")} returns the usual path
##' to the \code{.ssh} directory.
##'
##' Calling \code{ssh_path()} with no arguments will return the usual path to
##' the \code{.ssh} directory.
##'
##' @param file basename of file for which path is requested
##' @return Full path to the file
##' @export
##' @examples
##' ssh_path()
##' ssh_path("is_rsa.pub")
ssh_path <- function(file = "") {
  file.path(home_dir(), ".ssh", file)
}

# Return the user's home directory regardless of operating system
home_dir <- function() {
  if (.Platform$OS.type == "windows") {
    home <- Sys.getenv("USERPROFILE")
  } else {
    home <- path.expand("~")
  }
  return(home)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2018 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Punch card
##'
##' @template repo-param
##' @param main Default title for the plot is "Punch card on repo:"
##'     and repository workdir basename. Supply a new title if you
##'     desire one.
##' @param ... Additional arguments affecting the plot
##' @return invisible NULL
##' @importFrom graphics axis
##' @importFrom graphics par
##' @importFrom graphics plot.new
##' @importFrom graphics plot.window
##' @importFrom graphics symbols
##' @importFrom graphics title
##' @export
##' @examples
##' \dontrun{
##' ## Initialize repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- clone("https://github.com/ropensci/git2r.git", path)
##'
##' ## Plot
##' punch_card(repo)
##' }
punch_card <- function(repo = ".", main = NULL, ...) {
    savepar <- graphics::par(las = 1, mar = c(2.2, 6, 2, 0))
    on.exit(par(savepar))

    wd <- c("Saturday", "Friday", "Thursday", "Wednesday", "Tuesday",
            "Monday", "Sunday")

    ## Extract information from repository
    repo <- lookup_repository(repo)
    df <- as.data.frame(repo)
    df$when <- as.POSIXlt(df$when)
    df$hour <- df$when$hour
    df$weekday <- df$when$wday

    ## Create a key and tabulate
    df$key <- paste0(df$weekday, "-", df$hour)
    df <- as.data.frame(table(df$key), stringsAsFactors = FALSE)
    names(df) <- c("key", "Commits")

    ## Convert key to Weekday and Hour
    df$Weekday <- sapply(strsplit(df$key, "-"), "[", 1)
    df$Weekday <- factor(df$Weekday, levels = c(6, 5, 4, 3, 2, 1, 0),
                         labels = wd)
    df$Hour <- as.integer(sapply(strsplit(df$key, "-"), "[", 2))
    df$key <- paste0(df$Weekday, "-", df$Hour)

    ## Scale
    df$Commits <- sqrt((df$Commits / max(df$Commits)) / pi)

    plot.new()
    plot.window(xlim = c(0, 23), ylim = c(0.8, 7.2))
    symbols(df$ Hour, df$Weekday, circles = df$Commits,
            xaxt = "n", yaxt = "n", inches = FALSE,
            fg = "white", bg = "black", add = TRUE, ...)
    h <- 0:23
    h <- paste0(ifelse(h > 9, as.character(h), paste0("0", as.character(h))),
                ":00")
    axis(1, at = 0:23, labels = h)
    axis(2, at = 1:7, labels = wd)

    if (is.null(main)) {
        if (is_bare(repo)) {
            main <- "Punch card"
        } else {
            main <- sprintf("Punch card on repository: %s",
                            basename(workdir(repo)))
        }
    }

    title(main)

    invisible(NULL)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2024 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Create blob from file on disk
##'
##' Read a file from the filesystem and write its content to the
##' Object Database as a loose blob. The method is vectorized and
##' accepts a vector of files to create blobs from.
##' @param repo The repository where the blob(s) will be written. Can
##'     be a bare repository. A \code{git_repository} object, or a
##'     path to a repository, or \code{NULL}.  If the \code{repo}
##'     argument is \code{NULL}, the repository is searched for with
##'     \code{\link{discover_repository}} in the current working
##'     directory.
##' @param path The file(s) from which the blob will be created.
##' @param relative TRUE if the file(s) from which the blob will be
##'     created is relative to the repository's working dir. Default
##'     is TRUE.
##' @return list of S3 class git_blob \code{objects}
##' @export
##' @useDynLib git2r git2r_blob_create_fromdisk
##' @useDynLib git2r git2r_blob_create_fromworkdir
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create blobs from files relative to workdir
##' writeLines("Hello, world!", file.path(path, "example-1.txt"))
##' writeLines("test content", file.path(path, "example-2.txt"))
##' blob_list_1 <- blob_create(repo, c("example-1.txt",
##'                                    "example-2.txt"))
##'
##' ## Create blobs from files not relative to workdir
##' temp_file_1 <- tempfile()
##' temp_file_2 <- tempfile()
##' writeLines("Hello, world!", temp_file_1)
##' writeLines("test content", temp_file_2)
##' blob_list_2 <- blob_create(repo, c(temp_file_1, temp_file_2),
##'                            relative = FALSE)
##' }
blob_create <- function(repo = ".", path = NULL, relative = TRUE) {
    repo <- lookup_repository(repo)
    if (isTRUE(relative))
        return(.Call(git2r_blob_create_fromworkdir, repo, path))
    path <- normalizePath(path, mustWork = TRUE)
    .Call(git2r_blob_create_fromdisk, repo, path)
}

##' Content of blob
##'
##' @param blob The blob object.
##' @param split Split blob content to text lines. Default TRUE.
##' @param raw When \code{TRUE}, get the content of the blob as a raw
##'     vector, else as a character vector. Default is \code{FALSE}.
##' @return The content of the blob. NA_character_ if the blob is
##'     binary and \code{raw} is \code{FALSE}.
##' @export
##' @useDynLib git2r git2r_blob_content
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Display content of blob.
##' content(tree(commits(repo)[[1]])["example.txt"])
##' }
content <- function(blob = NULL, split = TRUE, raw = FALSE) {
    result <- .Call(git2r_blob_content, blob, raw)
    if (isTRUE(raw))
        return(result)
    if (isTRUE(split))
        result <- strsplit(result, "\n")[[1]]
    result
}

##' Determine the sha from a blob string
##'
##' The blob is not written to the object database.
##' @param data The string vector to hash.
##' @return A string vector with the sha for each string in data.
##' @export
##' @useDynLib git2r git2r_odb_hash
##' @examples
##' \dontrun{
##' identical(hash(c("Hello, world!\n",
##'                  "test content\n")),
##'                c("af5626b4a114abcb82d63db7c8082c3c4756e51b",
##'                  "d670460b4b4aece5915caf5c68d12f560a9fe3e4"))
##' }
hash <- function(data = NULL) {
    .Call(git2r_odb_hash, data)
}

##' Determine the sha from a blob in a file
##'
##' The blob is not written to the object database.
##' @param path The path vector with files to hash.
##' @return A vector with the sha for each file in path.
##' @export
##' @useDynLib git2r git2r_odb_hashfile
##' @examples
##' \dontrun{
##' ## Create a file. NOTE: The line endings from writeLines gives
##' ## LF (line feed) on Unix/Linux and CRLF (carriage return, line feed)
##' ## on Windows. The example use writeChar to have more control.
##' path <- tempfile()
##' f <- file(path, "wb")
##' writeChar("Hello, world!\n", f, eos = NULL)
##' close(f)
##'
##' ## Generate hash
##' hashfile(path)
##' identical(hashfile(path), hash("Hello, world!\n"))
##' }
hashfile <- function(path = NULL) {
    path <- normalizePath(path, mustWork = TRUE)
    if (any(is.na(path)))
        stop("Invalid 'path' argument")
    .Call(git2r_odb_hashfile, path)
}

##' Is blob binary
##'
##' @param blob The blob \code{object}.
##' @return TRUE if binary data, FALSE if not.
##' @export
##' @useDynLib git2r git2r_blob_is_binary
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##'
##' ## Check if binary
##' b_text <- tree(commit_1)["example.txt"]
##' is_binary(b_text)
##'
##' ## Commit plot file (binary)
##' x <- 1:100
##' y <- x^2
##' png(file.path(path, "plot.png"))
##' plot(y ~ x, type = "l")
##' dev.off()
##' add(repo, "plot.png")
##' commit_2 <- commit(repo, "Second commit message")
##'
##' ## Check if binary
##' b_png <- tree(commit_2)["plot.png"]
##' is_binary(b_png)
##' }
is_binary <- function(blob = NULL) {
    .Call(git2r_blob_is_binary, blob)
}

##' Check if object is S3 class git_blob
##'
##' @param object Check if object is S3 class git_blob
##' @return TRUE if object is S3 class git_blob, else FALSE
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##' blob_1 <- tree(commit_1)["example.txt"]
##'
##' ## Check if blob
##' is_blob(commit_1)
##' is_blob(blob_1)
##' }
is_blob <- function(object) {
    inherits(object, "git_blob")
}

##' Size in bytes of the contents of a blob
##'
##' @param x The blob \code{object}
##' @return a non-negative integer
##' @export
##' @useDynLib git2r git2r_blob_rawsize
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##' blob_1 <- tree(commit_1)["example.txt"]
##'
##' ## Get length in size of bytes of the content of the blob
##' length(blob_1)
##' }
length.git_blob <- function(x) {
    .Call(git2r_blob_rawsize, x)
}

##' @export
format.git_blob <- function(x, ...) {
    sprintf("blob:  %s\nsize:  %i bytes", x$sha, length(x))
}

##' @export
print.git_blob <- function(x, ...) {
    cat(format(x, ...), "\n", sep = "")
    invisible(x)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2024 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Coerce entries in a git_tree to a \code{data.frame}
##'
##' The entries in a tree are coerced to a \code{data.frame}
##'
##'
##' The \code{data.frame} have the following columns:
##' \describe{
##'
##'   \item{filemode}{
##'     The UNIX file attributes of a tree entry
##'   }
##'
##'   \item{type}{
##'     String representation of the tree entry type
##'   }
##'
##'   \item{sha}{
##'     The sha of a tree entry
##'   }
##'
##'   \item{name}{
##'     The filename of a tree entry
##'   }
##'
##' }
##' @param x The tree \code{object}
##' @param ... Additional arguments. Not used.
##' @return \code{data.frame}
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' dir.create(file.path(path, "subfolder"))
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create three files and commit
##' writeLines("First file",  file.path(path, "example-1.txt"))
##' writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
##' writeLines("Third file",  file.path(path, "example-3.txt"))
##' add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
##' commit(repo, "Commit message")
##'
##' ## Display tree
##' tree(last_commit(repo))
##'
##' ## Coerce tree to a data.frame
##' df <- as.data.frame(tree(last_commit(repo)))
##' df
##' }
as.data.frame.git_tree <- function(x, ...) {
    data.frame(mode = sprintf("%06o", x$filemode),
               type = x$type,
               sha  = x$id,
               name = x$name,
               stringsAsFactors = FALSE)
}

##' @export
base::as.data.frame

##' Coerce entries in a git_tree to a list of entry objects
##'
##' @param x The tree \code{object}
##' @param ... Unused
##' @return list of entry objects
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' dir.create(file.path(path, "subfolder"))
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create three files and commit
##' writeLines("First file",  file.path(path, "example-1.txt"))
##' writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
##' writeLines("Third file",  file.path(path, "example-3.txt"))
##' add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
##' commit(repo, "Commit message")
##'
##' ## Inspect size of each blob in tree
##' invisible(lapply(as(tree(last_commit(repo)), "list"),
##'   function(obj) {
##'     if (is_blob(obj))
##'       summary(obj)
##'     NULL
##'   }))
##' }
as.list.git_tree <- function(x, ...) {
    lapply(x$id, function(sha) lookup(x$repo, sha))
}

##' Tree
##'
##' Get the tree pointed to by a commit or stash.
##' @param object the \code{commit} or \code{stash} object
##' @return A S3 class git_tree object
##' @export
##' @useDynLib git2r git2r_commit_tree
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a first user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' tree(last_commit(repo))
##' }
tree <- function(object = NULL) {
    .Call(git2r_commit_tree, object)
}

##' @export
print.git_tree <- function(x, ...) {
    cat(sprintf("tree: %s\n\n", x$sha))
    print(as.data.frame(x))
    invisible(x)
}

##' Summary of tree
##'
##' @param object The tree \code{object}
##' @param ... Additional arguments affecting the summary produced.
##' @return None (invisible 'NULL').
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' summary(tree(last_commit(repo)))
##' }
summary.git_tree <- function(object, ...) {
    print(as.data.frame(object))
}

##' Extract object from tree
##'
##' Lookup a tree entry by its position in the tree
##' @param x The tree \code{object}
##' @param i The index (integer or logical) of the tree object to
##' extract. If negative values, all elements except those indicated
##' are selected. A character vector to match against the names of
##' objects to extract.
##' @return Git object
##' @export
##' @examples
##' \dontrun{
##' ##' Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' dir.create(file.path(path, "subfolder"))
##' repo <- init(path)
##'
##' ##' Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ##' Create three files and commit
##' writeLines("First file",  file.path(path, "example-1.txt"))
##' writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
##' writeLines("Third file",  file.path(path, "example-3.txt"))
##' add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
##' new_commit <- commit(repo, "Commit message")
##'
##' ##' Pick a tree in the repository
##' tree_object <- tree(new_commit)
##'
##' ##' Display tree
##' tree_object
##'
##' ##' Select item by name
##' tree_object["example-1.txt"]
##'
##' ##' Select first item in tree
##' tree_object[1]
##'
##' ##' Select first three items in tree
##' tree_object[1:3]
##'
##' ##' Select all blobs in tree
##' tree_object[vapply(as(tree_object, 'list'), is_blob, logical(1))]
##' }
"[.git_tree" <- function(x, i) {
    if (is.logical(i))
        return(x[seq_along(x)[i]])

    if (is.character(i))
        return(x[which(x$name %in% i)])

    if (!is.numeric(i))
        stop("Invalid index")

    i <- seq_len(length(x))[as.integer(i)]
    ret <- lapply(i, function(j) lookup(x$repo, x$id[j]))
    if (identical(length(ret), 1L))
        ret <- ret[[1]]
    ret
}

##' Number of entries in tree
##'
##' @param x The tree \code{object}
##' @return a non-negative integer or double (which will be rounded
##' down)
##' @export
length.git_tree <- function(x) {
    length(x$id)
}

##' Check if object is S3 class git_tree
##'
##' @param object Check if object is S3 class git_tree
##' @return TRUE if object is S3 class git_tree, else FALSE
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##' tree_1 <- tree(commit_1)
##'
##' ## Check if tree
##' is_tree(commit_1)
##' is_tree(tree_1)
##' }
is_tree <- function(object) {
    inherits(object, "git_tree")
}

##' List the contents of a tree object
##'
##' Traverse the entries in a tree and its subtrees.  Akin to the 'git
##' ls-tree' command.
##' @param tree default (\code{NULL}) is the tree of the last commit
##'     in \code{repo}. Can also be a \code{git_tree} object or a
##'     character that identifies a tree in the repository (see
##'     \sQuote{Examples}).
##' @param repo never used if \code{tree} is a \code{git_tree}
##'     object. A \code{git_repository} object, or a path (default =
##'     '.') to a repository.
##' @param recursive default is to recurse into sub-trees.
##' @return A data.frame with the following columns: \describe{
##'     \item{mode}{UNIX file attribute of the tree entry}
##'     \item{type}{type of object} \item{sha}{sha of the object}
##'     \item{path}{path relative to the root tree}
##'     \item{name}{filename of the tree entry} \item{len}{object size
##'     of blob (file) entries. NA for other objects.}  }
##' @export
##' @useDynLib git2r git2r_tree_walk
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' dir.create(file.path(path, "subfolder"))
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create three files and commit
##' writeLines("First file",  file.path(path, "example-1.txt"))
##' writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
##' writeLines("Third file",  file.path(path, "example-3.txt"))
##' add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
##' commit(repo, "Commit message")
##'
##' ## Traverse tree entries and its subtrees.
##' ## Various approaches that give identical result.
##' ls_tree(tree = tree(last_commit(path)))
##' ls_tree(tree = tree(last_commit(repo)))
##' ls_tree(repo = path)
##' ls_tree(repo = repo)
##'
##' ## Skip content in subfolder
##' ls_tree(repo = repo, recursive = FALSE)
##'
##' ## Start in subfolder
##' ls_tree(tree = "HEAD:subfolder", repo = repo)
##' }
ls_tree <- function(tree = NULL, repo = ".", recursive = TRUE) {
    if (is.null(tree)) {
        tree <- tree(last_commit(lookup_repository(repo)))
    } else if (is.character(tree)) {
        tree <- revparse_single(repo = lookup_repository(repo), revision = tree)
        if (!is_tree(tree))
            tree <- tree(tree)
    }

    data.frame(.Call(git2r_tree_walk, tree, recursive),
               stringsAsFactors = FALSE)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file and commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "First commit message")

## Create new user and change file
config(repo, user.name = "Bob", user.email = "bob@example.org")
writeLines(c("Hello world!", "HELLO WORLD!", "HOLA"),
           file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Second commit message")

## Check blame
b <- blame(repo, "test.txt")
stopifnot(identical(length(b$hunks), 2L))

## Hunk: 1
stopifnot(identical(b$hunks[[1]]$lines_in_hunk, 1L))
stopifnot(identical(b$hunks[[1]]$final_commit_id, sha(commit_1)))
stopifnot(identical(b$hunks[[1]]$final_start_line_number, 1L))
stopifnot(identical(b$hunks[[1]]$final_signature$name, "Alice"))
stopifnot(identical(b$hunks[[1]]$final_signature$email, "alice@example.org"))
stopifnot(identical(b$hunks[[1]]$orig_commit_id, sha(commit_1)))
stopifnot(identical(b$hunks[[1]]$orig_start_line_number, 1L))
stopifnot(identical(b$hunks[[1]]$orig_signature$name, "Alice"))
stopifnot(identical(b$hunks[[1]]$orig_signature$email, "alice@example.org"))
stopifnot(identical(b$hunks[[1]]$orig_path, "test.txt"))
stopifnot(identical(b$hunks[[1]]$boundary, TRUE))

## Hunk: 2
stopifnot(identical(b$hunks[[2]]$lines_in_hunk, 2L))
stopifnot(identical(b$hunks[[2]]$final_commit_id, sha(commit_2)))
stopifnot(identical(b$hunks[[2]]$final_start_line_number, 2L))
stopifnot(identical(b$hunks[[2]]$final_signature$name, "Bob"))
stopifnot(identical(b$hunks[[2]]$final_signature$email, "bob@example.org"))
stopifnot(identical(b$hunks[[2]]$orig_commit_id, sha(commit_2)))
stopifnot(identical(b$hunks[[2]]$orig_start_line_number, 2L))
stopifnot(identical(b$hunks[[2]]$orig_signature$name, "Bob"))
stopifnot(identical(b$hunks[[2]]$orig_signature$email, "bob@example.org"))
stopifnot(identical(b$hunks[[2]]$orig_path, "test.txt"))
stopifnot(identical(b$hunks[[2]]$boundary, FALSE))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Status
##'
##' Display state of the repository working directory and the staging
##' area.
##' @template repo-param
##' @param staged Include staged files. Default TRUE.
##' @param unstaged Include unstaged files. Default TRUE.
##' @param untracked Include untracked files and directories. Default
##'     TRUE.
##' @param ignored Include ignored files. Default FALSE.
##' @param all_untracked Shows individual files in untracked
##'     directories if \code{untracked} is \code{TRUE}.
##' @return \code{git_status} with repository status
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file
##' writeLines("Hello world!", file.path(path, "test.txt"))
##'
##' ## Check status; untracked file
##' status(repo)
##'
##' ## Add file
##' add(repo, "test.txt")
##'
##' ## Check status; staged file
##' status(repo)
##'
##' ## Commit
##' commit(repo, "First commit message")
##'
##' ## Check status; clean
##' status(repo)
##'
##' ## Change the file
##' writeLines(c("Hello again!", "Here is a second line", "And a third"),
##'            file.path(path, "test.txt"))
##'
##' ## Check status; unstaged file
##' status(repo)
##'
##' ## Add file and commit
##' add(repo, "test.txt")
##' commit(repo, "Second commit message")
##'
##' ## Check status; clean
##' status(repo)
##'}
status <- function(repo      = ".",
                   staged    = TRUE,
                   unstaged  = TRUE,
                   untracked = TRUE,
                   ignored   = FALSE,
                   all_untracked = FALSE) {
    structure(.Call(git2r_status_list, lookup_repository(repo), staged,
                    unstaged, untracked, all_untracked, ignored),
              class = "git_status")
}

##' @export
print.git_status <- function(x, ...) {
    display_status <- function(title, section) {
        cat(sprintf("%s:\n", title))

        for (i in seq_len(length(section))) {
            label <- names(section)[i]
            label <- paste0(toupper(substr(label, 1, 1)),
                            substr(label, 2, nchar(label)))
            cat(sprintf("\t%-12s%s\n", paste0(label, ":"), section[[i]]))
        }

        invisible(NULL)
    }

    if (max(sapply(x, length)) == 0L)
      cat("working directory clean\n")

    if (length(x$ignored)) {
        display_status("Ignored files", x$ignored)
        cat("\n")
    }

    if (length(x$untracked)) {
        display_status("Untracked files", x$untracked)
        cat("\n")
    }

    if (length(x$unstaged)) {
        display_status("Unstaged changes", x$unstaged)
        cat("\n")
    }

    if (length(x$staged)) {
        display_status("Staged changes", x$staged)
        cat("\n")
    }

    invisible(x)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2024 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

get_upstream_name <- function(object) {
    upstream <- branch_get_upstream(object)

    if (is.null(upstream)) {
        stop("The branch '", object$name, "' that you are ",
             "trying to push does not track an upstream branch.")
    }

    branch_remote_name(upstream)
}

##' Push
##'
##' @param object path to repository, or a \code{git_repository} or
##'     \code{git_branch}.
##' @param name The remote's name. Default is NULL.
##' @param refspec The refspec to be pushed. Default is NULL.
##' @param force Force your local revision to the remote repo. Use it
##'     with care. Default is FALSE.
##' @param credentials The credentials for remote repository
##'     access. Default is NULL. To use and query an ssh-agent for the
##'     ssh key credentials, let this parameter be NULL (the default).
##' @param set_upstream Set the current local branch to track the
##'     remote branch. Default is FALSE.
##' @param proxy Either \code{NULL} (the default) to disable proxy usage,
##'     \code{TRUE} for automatic proxy detection, or a character string
##'     with a proxy URL (for example, \code{"http://proxy.example.org:3128"}).
##' @return invisible(NULL)
##' @seealso \code{\link{cred_user_pass}}, \code{\link{cred_ssh_key}}
##' @export
##' @useDynLib git2r git2r_branch_canonical_name
##' @useDynLib git2r git2r_branch_upstream_canonical_name
##' @useDynLib git2r git2r_push
##' @examples
##' \dontrun{
##' ## Initialize two temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##'
##' ## Clone the bare repository. This creates remote-tracking
##' ## branches for each branch in the cloned repository.
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' push(repo, "origin", "refs/heads/master")
##'
##' ## Now, unset the remote-tracking branch to NULL to demonstrate
##' ## the 'set_upstream' argument. Then push with 'set_upstream = TRUE'
##' ## to add the upstream tracking branch to branch 'master' again.
##' branch_get_upstream(repository_head(repo))
##' branch_set_upstream(repository_head(repo), NULL)
##' branch_get_upstream(repository_head(repo))
##' push(repo, "origin", "refs/heads/master", set_upstream = TRUE)
##' branch_get_upstream(repository_head(repo))
##'
##' ## Change file and commit
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Second commit message")
##'
##' ## Push commits from repository to bare repository
##' push(repo)
##'
##' ## List commits in repository and bare repository
##' commits(repo)
##' commits(repo_bare)
##' }
push <- function(object       = ".",
                 name         = NULL,
                 refspec      = NULL,
                 force        = FALSE,
                 credentials  = NULL,
                 set_upstream = FALSE,
                 proxy        = NULL) {
    if (is_branch(object)) {
        name <- get_upstream_name(object)

        src <- .Call(git2r_branch_canonical_name, object)
        dst <- .Call(git2r_branch_upstream_canonical_name, object)
        refspec <- paste0(src, ":", dst)
        object <- object$repo
    } else {
        object <- lookup_repository(object)
    }

    if (all(is.null(name), is.null(refspec))) {
        b <- repository_head(object)
        name <- get_upstream_name(b)
        src <- .Call(git2r_branch_canonical_name, b)
        dst <- .Call(git2r_branch_upstream_canonical_name, b)
        refspec <- paste0(src, ":", dst)

        if (isTRUE(force))
            refspec <- paste0("+", refspec)
    } else {
        opts <- list(force = force)
        tmp <- get_refspec(object, name, refspec, opts)
        name <- tmp$remote
        refspec <- tmp$refspec
    }

    .Call(git2r_push, object, name, refspec, credentials, proxy)

    if (isTRUE(set_upstream)) {
        b <- repository_head(object)
        if (is_local(b))
            branch_set_upstream(b, paste0(name, "/", b$name))
    }

    invisible(NULL)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test-1.txt"))

## add and commit
add(repo, "test-1.txt")
commit(repo, "Commit message")

## Pop stash
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-1.txt"))
stash(repo)
stopifnot(identical("Hello world!",
                    readLines(file.path(path, "test-1.txt"))))
stash_pop(repo)
stopifnot(identical(c("Hello world!", "HELLO WORLD!"),
                    readLines(file.path(path, "test-1.txt"))))

## Make one more commit
add(repo, "test-1.txt")
commit(repo, "Next commit message")

## Check that there are no stashes
stopifnot(identical(stash_list(repo), list()))

## Apply stash
writeLines(c("Hello world!", "HELLO WORLD!", "hello world!"),
           file.path(path, "test-1.txt"))
stash(repo)
stopifnot(identical(c("Hello world!", "HELLO WORLD!"),
                    readLines(file.path(path, "test-1.txt"))))
stash_apply(repo)
stopifnot(identical(c("Hello world!", "HELLO WORLD!", "hello world!"),
                    readLines(file.path(path, "test-1.txt"))))
stopifnot(identical(length(stash_list(repo)), 1L))
stash_drop(repo, 1)
stopifnot(identical(stash_list(repo), list()))

## Make one more commit
add(repo, "test-1.txt")
commit(repo, "Apply stash commit message")

## Create one more file
writeLines("Hello world!", file.path(path, "test-2.txt"))

## Check that there are no stashes
stopifnot(identical(stash_list(repo), list()))

## Stash
stash(repo)
stopifnot(identical(stash_list(repo), list()))
s <- stash(repo, untracked = TRUE)
stopifnot(identical(print(s), s))
summary(s)
stopifnot(identical(length(stash_list(repo)), 1L))
tree(stash_list(repo)[[1]])

## Drop stash
stash_drop(repo, 1)
stopifnot(identical(stash_list(repo), list()))

## Check stash_drop argument
tools::assertError(stash_drop(repo))
tools::assertError(stash_drop(repo, -1))
tools::assertError(stash_drop(repo, 0.5))

## Create one more file
writeLines("Hello world!", file.path(path, "test-3.txt"))

## Create stash in repository
stash(repo, untracked = TRUE)
stopifnot(identical(length(stash_list(repo)), 1L))

## Check stash_list method with missing repo argument
wd <- setwd(path)
stopifnot(identical(length(stash_list()), 1L))
if (!is.null(wd))
    setwd(wd)

## Drop git_stash object in repository
stash_drop(stash_list(repo)[[1]])

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a bare repository
repo <- init(path, bare = TRUE)

## Check that the state of the repository
stopifnot(identical(is_bare(repo), TRUE))
stopifnot(identical(is_empty(repo), TRUE))

## Check that workdir is NULL for a bare repository
stopifnot(is.null(workdir(repo)))

## Check with missing repo argument
setwd(path)
stopifnot(identical(is_bare(), TRUE))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2018 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Blobs in the object database
##'
##' List all blobs reachable from the commits in the object
##' database. For each commit, list blob's in the commit tree and
##' sub-trees.
##' @template repo-param
##' @return A data.frame with the following columns:
##' \describe{
##'   \item{sha}{The sha of the blob}
##'   \item{path}{The path to the blob from the tree and sub-trees}
##'   \item{name}{The name of the blob from the tree that contains the blob}
##'   \item{len}{The length of the blob}
##'   \item{commit}{The sha of the commit}
##'   \item{author}{The author of the commit}
##'   \item{when}{The timestamp of the author signature in the commit}
##' }
##' @note A blob sha can have several entries
##' @export
##' @examples \dontrun{
##' ## Create a directory in tempdir
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##'
##' ## Initialize a repository
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit(repo, "Commit message 1")
##'
##' ## Change file and commit
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit(repo, "Commit message 2")
##'
##' ## Commit same content under different name in a sub-directory
##' dir.create(file.path(path, "sub-directory"))
##' file.copy(file.path(path, "test.txt"),
##'           file.path(path, "sub-directory", "copy.txt"))
##' add(repo, "sub-directory/copy.txt")
##' commit(repo, "Commit message 3")
##'
##' ## List blobs
##' odb_blobs(repo)
##' }
odb_blobs <- function(repo = ".") {
    blobs <- .Call(git2r_odb_blobs, lookup_repository(repo))
    blobs <- data.frame(blobs, stringsAsFactors = FALSE)
    blobs <- blobs[order(blobs$when), ]
    index <- paste0(blobs$sha, ":", blobs$path, "/", blobs$name)
    blobs <- blobs[!duplicated(index), ]
    rownames(blobs) <- NULL
    blobs$when <- as.POSIXct(blobs$when, origin = "1970-01-01", tz = "GMT")
    blobs
}

##' List all objects available in the database
##'
##' @template repo-param
##' @return A data.frame with the following columns:
##' \describe{
##'   \item{sha}{The sha of the object}
##'   \item{type}{The type of the object}
##'   \item{len}{The length of the object}
##' }
##' @export
##' @examples \dontrun{
##' ## Create a directory in tempdir
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##'
##' ## Initialize a repository
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit(repo, "Commit message 1")
##'
##' ## Create tag
##' tag(repo, "Tagname", "Tag message")
##'
##' ## List objects in repository
##' odb_objects(repo)
##' }
odb_objects <- function(repo = ".") {
    data.frame(.Call(git2r_odb_objects, lookup_repository(repo)),
               stringsAsFactors = FALSE)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Time
##'
##' The class \code{git_time} stores the time a Git object was created.
##'
##' The default is to use \code{tz = "GMT"} and \code{origin =
##' "1970-01-01"}. To use your local timezone, set \code{tz =
##' Sys.timezone()}.
##'
##' @inheritParams base::as.POSIXct
##' @inheritParams base::strptime
##' @seealso \code{\link{when}}
##' @name git_time
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a first user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create tag
##' tag(repo, "Tagname", "Tag message")
##'
##' as.POSIXct(commits(repo)[[1]]$author$when)
##' as.POSIXct(tags(repo)[[1]]$tagger$when)
##' as.POSIXct(tags(repo)[[1]]$tagger$when, tz = Sys.timezone())
##' }
NULL

##' @rdname git_time
##' @export
as.character.git_time <- function(x,  tz = "GMT", origin = "1970-01-01",
                                  usetz = TRUE, ...) {
    as.character(format(as.POSIXct(x, tz = tz, origin = origin),
                        usetz = usetz), ...)
}

##' @rdname git_time
##' @export
format.git_time <- function(x,  tz = "GMT", origin = "1970-01-01",
                            usetz = TRUE, ...) {
    format(as.POSIXct(x, tz = tz, origin = origin), usetz = usetz, ...)
}

##' @rdname git_time
##' @export
as.POSIXct.git_time <- function(x, tz = "GMT", origin = "1970-01-01", ...) {
    as.POSIXct(x$time, tz = tz, origin = origin, ...)
}

##' @rdname git_time
##' @export
print.git_time <- function(x,  tz = "GMT", origin = "1970-01-01",
                           usetz = TRUE, ...) {
    cat(sprintf("%s\n", as.character(x, tz = tz, origin = origin,
                                     usetz = usetz, ...)))
    invisible(x)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2024 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Find a merge base between two commits
##'
##' @param one One of the commits
##' @param two The other commit
##' @return git_commit
##' @export
##' @useDynLib git2r git2r_merge_base
##' @examples \dontrun{
##' ## Create a directory in tempdir
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##'
##' ## Initialize a repository
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Master branch", file.path(path, "master_branch.txt"))
##' add(repo, "master_branch.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##'
##' ## Create first branch, checkout, add file and commit
##' branch_1 <- branch_create(commit_1, "branch_1")
##' checkout(branch_1)
##' writeLines("Branch 1", file.path(path, "branch_1.txt"))
##' add(repo, "branch_1.txt")
##' commit_2 <- commit(repo, "Commit message branch_1")
##'
##' ## Create second branch, checkout, add file and commit
##' branch_2 <- branch_create(commit_1, "branch_2")
##' checkout(branch_2)
##' writeLines("Branch 2", file.path(path, "branch_2.txt"))
##' add(repo, "branch_2.txt")
##' commit_3 <- commit(repo, "Commit message branch_2")
##'
##' ## Check that merge base equals commit_1
##' stopifnot(identical(merge_base(commit_2, commit_3), commit_1))
##' }
merge_base <- function(one = NULL, two = NULL) {
    .Call(git2r_merge_base, one, two)
}

##' Merge a branch into HEAD
##'
##' @rdname merge
##' @param x A path (default '.') to a repository, or a
##'     \code{git_repository} object, or a \code{git_branch}.
##' @param y If \code{x} is a \code{git_repository}, the name of the
##'     branch to merge into HEAD. Not used if \code{x} is a
##'     \code{git_branch}.
##' @param commit_on_success If there are no conflicts written to the
##'     index, the merge commit will be committed. Default is TRUE.
##' @param merger Who made the merge. The default (\code{NULL}) is to
##'     use \code{default_signature} for the repository.
##' @param fail If a conflict occurs, exit immediately instead of
##'     attempting to continue resolving conflicts. Default is
##'     \code{FALSE}.
##' @param ... Additional arguments (unused).
##' @template return-git_merge_result
##' @export
##' @useDynLib git2r git2r_merge_branch
##' @template merge-example
merge.git_branch <- function(x, y = NULL, commit_on_success = TRUE,
                             merger = NULL, fail = FALSE, ...) {
    if (is.null(merger))
        merger <- default_signature(x$repo)
    .Call(git2r_merge_branch, x, merger, commit_on_success, fail)
}

##' @export
##' @rdname merge
merge.git_repository <- function(x, y = NULL, commit_on_success = TRUE,
                                 merger = NULL, fail = FALSE, ...) {
    ## Check branch argument
    if (!is.character(y) || !identical(length(y), 1L))
        stop("'branch' must be a character vector of length one")
    b <- branches(x)
    b <- b[vapply(b, "[[", character(1), "name") == y][[1]]
    merge.git_branch(b, commit_on_success = commit_on_success,
                     merger = merger, fail = fail)
}

##' @export
##' @rdname merge
merge.character <- function(x = ".", y = NULL, commit_on_success = TRUE,
                            merger = NULL, fail = FALSE, ...) {
    x <- lookup_repository(x)
    merge.git_repository(x, y, commit_on_success, merger, fail)
}

##' @export
base::merge

##' @export
format.git_merge_result <- function(x, ...) {
    if (isTRUE(x$up_to_date))
        return("Already up-to-date")
    if (isTRUE(x$conflicts))
        return("Merge: Conflicts")
    if (isTRUE(x$fast_forward))
        return("Merge: Fast-forward")
    return("Merge")
}

##' @export
print.git_merge_result <- function(x, ...) {
    cat(format(x, ...), "\n", sep = "")
    invisible(x)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Punch card
##'
##' @rdname punch_card-methods
##' @docType methods
##' @param repo The repository
##' @param main Default title for the plot is "Punch card on repo:"
##' and repository workdir basename. Supply a new title if you desire
##' one.
##' @param ... Additional arguments affecting the plot
##' @return invisible NULL
##' @keywords methods
##' @include S4_classes.r
##' @importFrom graphics axis
##' @importFrom graphics par
##' @importFrom graphics plot.new
##' @importFrom graphics plot.window
##' @importFrom graphics symbols
##' @export
##' @examples
##' \dontrun{
##' ## Initialize repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- clone("https://github.com/ropensci/git2r.git", path)
##'
##' ## Plot
##' punch_card(repo)
##' }
setGeneric("punch_card",
           signature = "repo",
           function(repo, main = NULL, ...)
           standardGeneric("punch_card"))

##' @rdname punch_card-methods
##' @export
setMethod("punch_card",
          signature(repo = "git_repository"),
          function(repo, main, ...)
          {
              savepar <- graphics::par(las = 1, mar = c(2.2, 6, 2, 0))
              on.exit(par(savepar))

              wd <- c("Saturday", "Friday", "Thursday", "Wednesday",
                      "Tuesday", "Monday", "Sunday")

              ## Extract information from repository
              df <- as(repo, "data.frame")
              df$when <- as.POSIXlt(df$when)
              df$hour <- df$when$hour
              df$weekday <- df$when$wday

              ## Create a key and tabulate
              df$key <- paste0(df$weekday, "-", df$hour)
              df <- as.data.frame(table(df$key), stringsAsFactors = FALSE)
              names(df) <- c("key", "Commits")

              ## Convert key to Weekday and Hour
              df$Weekday <- sapply(strsplit(df$key, "-"), "[", 1)
              df$Weekday <- factor(df$Weekday,
                                   levels = c(6, 5, 4, 3, 2, 1, 0),
                                   labels = wd)
              df$Hour <- as.integer(sapply(strsplit(df$key, "-"), "[", 2))
              df$key <- paste0(df$Weekday, "-", df$Hour)

              ## Scale
              df$Commits <- sqrt((df$Commits / max(df$Commits)) / pi)

              graphics::plot.new()
              graphics::plot.window(xlim = c(0, 23), ylim = c(0.8, 7.2))
              graphics::symbols(df$Hour, df$Weekday, circles = df$Commits,
                                xaxt = "n", yaxt = "n", inches = FALSE,
                                fg = "white", bg = "black", add = TRUE, ...)
              h <- 0:23
              h <- paste0(ifelse(h > 9, as.character(h),
                                 paste0("0", as.character(h))), ":00")
              graphics::axis(1, at = 0:23, labels = h)
              graphics::axis(2, at = 1:7, labels = wd)

              if (is.null(main)) {
                  if (is_bare(repo)) {
                      main <- "Punch card"
                  } else {
                      main <- sprintf("Punch card on repository: %s",
                                      basename(workdir(repo)))
                  }
              }

              graphics::title(main)

              invisible(NULL)
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2024 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Coerce Git repository to a \code{data.frame}
##'
##' The commits in the repository are coerced to a \code{data.frame}
##'
##'
##' The \code{data.frame} have the following columns:
##' \describe{
##'
##'   \item{sha}{
##'     The 40 character hexadecimal string of the SHA-1
##'   }
##'
##'   \item{summary}{
##'     the short "summary" of the git commit message.
##'   }
##'
##'   \item{message}{
##'     the full message of a commit
##'   }
##'
##'   \item{author}{
##'     full name of the author
##'   }
##'
##'   \item{email}{
##'     email of the author
##'   }
##'
##'   \item{when}{
##'     time when the commit happened
##'   }
##'
##' }
##' @param x The repository \code{object}
##' @param ... Additional arguments. Not used.
##' @return \code{data.frame}
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create three files and commit
##' writeLines("First file",  file.path(path, "example-1.txt"))
##' writeLines("Second file", file.path(path, "example-2.txt"))
##' writeLines("Third file",  file.path(path, "example-3.txt"))
##' add(repo, "example-1.txt")
##' commit(repo, "Commit first file")
##' add(repo, "example-2.txt")
##' commit(repo, "Commit second file")
##' add(repo, "example-3.txt")
##' commit(repo, "Commit third file")
##'
##' ## Coerce commits to a data.frame
##' df <- as.data.frame(repo)
##' df
##' }
as.data.frame.git_repository <- function(x, ...) {
    do.call("rbind", lapply(commits(x), as.data.frame))
}

##' Open a repository
##'
##' @param path A path to an existing local git repository.
##' @param discover Discover repository from path. Default is TRUE.
##' @return A \code{git_repository} object with entries:
##' \describe{
##'   \item{path}{
##'     Path to a git repository
##'   }
##' }
##' @export
##' @useDynLib git2r git2r_repository_can_open
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test-1.txt"))
##' add(repo, 'test-1.txt')
##' commit_1 <- commit(repo, "Commit message")
##'
##' ## Make one more commit
##' writeLines(c("Hello world!", "HELLO WORLD!"),
##'            file.path(path, "test-1.txt"))
##' add(repo, 'test-1.txt')
##' commit(repo, "Next commit message")
##'
##' ## Create one more file
##' writeLines("Hello world!",
##'            file.path(path, "test-2.txt"))
##'
##' ## Brief summary of repository
##' repo
##'
##' ## Summary of repository
##' summary(repo)
##'
##' ## Workdir of repository
##' workdir(repo)
##'
##' ## Check if repository is bare
##' is_bare(repo)
##'
##' ## Check if repository is empty
##' is_empty(repo)
##'
##' ## Check if repository is a shallow clone
##' is_shallow(repo)
##'
##' ## List all references in repository
##' references(repo)
##'
##' ## List all branches in repository
##' branches(repo)
##'
##' ## Get HEAD of repository
##' repository_head(repo)
##'
##' ## Check if HEAD is head
##' is_head(repository_head(repo))
##'
##' ## Check if HEAD is local
##' is_local(repository_head(repo))
##'
##' ## List all tags in repository
##' tags(repo)
##' }
repository <- function(path = ".", discover = TRUE) {
    if (isTRUE(discover)) {
        path <- discover_repository(path)
        if (is.null(path))
            stop("The 'path' is not in a git repository")
    } else {
        path <- normalizePath(path, winslash = "/", mustWork = TRUE)
        if (!file.info(path)$isdir)
            stop("'path' is not a directory")
    }

    if (!isTRUE(.Call(git2r_repository_can_open, path)))
        stop("Unable to open repository at 'path'")

    structure(list(path = path), class = "git_repository")
}

##' Init a repository
##'
##' @param path A path to where to init a git repository
##' @param bare If TRUE, a Git repository without a working directory
##'     is created at the pointed path. If FALSE, provided path will
##'     be considered as the working directory into which the .git
##'     directory will be created.
##' @param branch Use the specified name for the initial branch in the
##'     newly created repository. If \code{branch=NULL}, fall back to
##'     the default name.
##' @return A \code{git_repository} object
##' @export
##' @useDynLib git2r git2r_repository_init
##' @seealso \link{repository}
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' is_bare(repo)
##'
##' ## Initialize a bare repository
##' path_bare <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' repo_bare <- init(path_bare, bare = TRUE)
##' is_bare(repo_bare)
##' }
init <- function(path = ".", bare = FALSE, branch = NULL) {
    path <- normalizePath(path, winslash = "/", mustWork = TRUE)
    if (!file.info(path)$isdir)
        stop("'path' is not a directory")
    .Call(git2r_repository_init, path, bare, branch)
    repository(path)
}

##' Clone a remote repository
##'
##' @param url The remote repository to clone
##' @param local_path Local directory to clone to.
##' @param bare Create a bare repository. Default is FALSE.
##' @param branch The name of the branch to checkout. Default is NULL
##'     which means to use the remote's default branch.
##' @param checkout Checkout HEAD after the clone is complete. Default
##'     is TRUE.
##' @param credentials The credentials for remote repository
##'     access. Default is NULL. To use and query an ssh-agent for the
##'     ssh key credentials, let this parameter be NULL (the default).
##' @param progress Show progress. Default is TRUE.
##' @return A \code{git_repository} object.
##' @seealso \link{repository}, \code{\link{cred_user_pass}},
##'     \code{\link{cred_ssh_key}}
##' @export
##' @useDynLib git2r git2r_clone
##' @examples
##' \dontrun{
##' ## Initialize repository
##' path_repo_1 <- tempfile(pattern="git2r-")
##' path_repo_2 <- tempfile(pattern="git2r-")
##' dir.create(path_repo_1)
##' dir.create(path_repo_2)
##' repo_1 <- init(path_repo_1)
##'
##' ## Config user and commit a file
##' config(repo_1, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines(
##'     "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'     file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "First commit message")
##'
##' ## Change file and commit
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "Second commit message")
##'
##' ## Change file again and commit.
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
##'   "minim veniam, quis nostrud exercitation ullamco laboris nisi ut")
##' writeLines(lines, file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "Third commit message")
##'
##' ## Clone to second repository
##' repo_2 <- clone(path_repo_1, path_repo_2)
##'
##' ## List commits in repositories
##' commits(repo_1)
##' commits(repo_2)
##' }
clone <- function(url         = NULL,
                  local_path  = NULL,
                  bare        = FALSE,
                  branch      = NULL,
                  checkout    = TRUE,
                  credentials = NULL,
                  progress    = TRUE) {
    .Call(git2r_clone, url, local_path, bare,
          branch, checkout, credentials, progress)
    repository(local_path)
}

##' Get HEAD for a repository
##'
##' @param x The repository \code{x} to check head
##' @param ... Additional arguments. Unused.
##' @return NULL if unborn branch or not found. A git_branch if not a
##'     detached head. A git_commit if detached head
##' @importFrom utils head
##' @export
##' @useDynLib git2r git2r_repository_head
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Commit message")
##'
##' ## Get HEAD of repository
##' repository_head(repo)
##' }
head.git_repository <- function(x, ...) {
    .Deprecated("repository_head")
    .Call(git2r_repository_head, x)
}

##' @export
utils::head

##' Get HEAD for a repository
##'
##' @template repo-param
##' @return NULL if unborn branch or not found. A git_branch if not a
##'     detached head. A git_commit if detached head
##' @export
##' @useDynLib git2r git2r_repository_head
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Commit message")
##'
##' ## Get HEAD of repository
##' repository_head(repo)
##' }
repository_head <- function(repo = ".") {
    .Call(git2r_repository_head, lookup_repository(repo))
}

##' Check if repository is bare
##'
##' @template repo-param
##' @return \code{TRUE} if bare repository, else \code{FALSE}
##' @seealso \link{init}
##' @export
##' @useDynLib git2r git2r_repository_is_bare
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' is_bare(repo)
##'
##' ## Initialize a bare repository
##' path_bare <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' repo_bare <- init(path_bare, bare = TRUE)
##' is_bare(repo_bare)
##' }
is_bare <- function(repo = ".") {
    .Call(git2r_repository_is_bare, lookup_repository(repo))
}

##' Check if HEAD of repository is detached
##'
##' @template repo-param
##' @return \code{TRUE} if repository HEAD is detached, else
##'     \code{FALSE}.
##' @export
##' @useDynLib git2r git2r_repository_head_detached
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##'
##' ## Change file, add and commit
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Commit message 2")
##'
##' ## HEAD of repository is not detached
##' is_detached(repo)
##'
##' ## Checkout first commit
##' checkout(commit_1)
##'
##' ## HEAD of repository is detached
##' is_detached(repo)
##' }
is_detached <- function(repo = ".") {
    .Call(git2r_repository_head_detached, lookup_repository(repo))
}

##' Check if repository is empty
##'
##' @template repo-param
##' @return \code{TRUE} if repository is empty else \code{FALSE}.
##' @export
##' @useDynLib git2r git2r_repository_is_empty
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Check if it's an empty repository
##' is_empty(repo)
##'
##' ## Commit a file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Check if it's an empty repository
##' is_empty(repo)
##' }
is_empty <- function(repo = ".") {
    .Call(git2r_repository_is_empty, lookup_repository(repo))
}

##' Determine if a directory is in a git repository
##'
##' The lookup start from path and walk across parent directories if
##' nothing has been found.
##' @param path The path to the directory.
##' @return TRUE if directory is in a git repository else FALSE
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Check if path is in a git repository
##' in_repository(path)
##'
##' ## Check if working directory is in a git repository
##' setwd(path)
##' in_repository()
##' }
in_repository <- function(path = ".") {
    !is.null(discover_repository(path))
}

##' Determine if the repository is a shallow clone
##'
##' @template repo-param
##' @return \code{TRUE} if shallow clone, else \code{FALSE}
##' @export
##' @useDynLib git2r git2r_repository_is_shallow
##' @examples
##' \dontrun{
##' ## Initialize repository
##' path_repo_1 <- tempfile(pattern="git2r-")
##' path_repo_2 <- tempfile(pattern="git2r-")
##' dir.create(path_repo_1)
##' dir.create(path_repo_2)
##' repo_1 <- init(path_repo_1)
##'
##' ## Config user and commit a file
##' config(repo_1, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "First commit message")
##'
##' ## Change file and commit
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "Second commit message")
##'
##' ## Change file again and commit.
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
##'   "minim veniam, quis nostrud exercitation ullamco laboris nisi ut")
##' writeLines(lines, file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "Third commit message")
##'
##' ## Clone to second repository
##' repo_2 <- clone(path_repo_1, path_repo_2)
##'
##' ## Check if it's a shallow clone
##' is_shallow(repo_2)
##' }
is_shallow <- function(repo = ".") {
    .Call(git2r_repository_is_shallow, lookup_repository(repo))
}

##' Lookup
##'
##' Lookup one object in a repository.
##' @template repo-param
##' @param sha The identity of the object to lookup. Must be 4 to 40
##' characters long.
##' @return a \code{git_blob} or \code{git_commit} or \code{git_tag}
##' or \code{git_tree} object
##' @export
##' @useDynLib git2r git2r_object_lookup
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##'
##' ## Create tag
##' tag(repo, "Tagname", "Tag message")
##'
##' ## First, get SHAs to lookup in the repository
##' sha_commit <- sha(commit_1)
##' sha_tree <- sha(tree(commit_1))
##' sha_blob <- sha(tree(commit_1)["example.txt"])
##' sha_tag <- sha(tags(repo)[[1]])
##'
##' ## SHAs
##' sha_commit
##' sha_tree
##' sha_blob
##' sha_tag
##'
##' ## Lookup objects
##' lookup(repo, sha_commit)
##' lookup(repo, sha_tree)
##' lookup(repo, sha_blob)
##' lookup(repo, sha_tag)
##'
##' ## Lookup objects, using only the first seven characters
##' lookup(repo, substr(sha_commit, 1, 7))
##' lookup(repo, substr(sha_tree, 1, 7))
##' lookup(repo, substr(sha_blob, 1, 7))
##' lookup(repo, substr(sha_tag, 1, 7))
##' }
lookup <- function(repo = ".", sha = NULL) {
    .Call(git2r_object_lookup, lookup_repository(repo), sha)
}

##' Lookup the commit related to a git object
##'
##' Lookup the commit related to a git_reference, git_tag or
##' git_branch object.
##' @param object a git object to get the related commit from.
##' @return A git commit object.
##' @export
##' @examples \dontrun{
##' ## Create a directory in tempdir
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##'
##' ## Initialize a repository
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, con = file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit(repo, "Commit message 1")
##'
##' ## Get the commit pointed to by the 'master' branch
##' lookup_commit(repository_head(repo))
##'
##' ## Create a tag
##' a_tag <- tag(repo, "Tagname", "Tag message")
##'
##' ## Get the commit pointed to by 'a_tag'
##' lookup_commit(a_tag)
##' }
lookup_commit <- function(object) {
    UseMethod("lookup_commit", object)
}

##' @rdname lookup_commit
##' @export
lookup_commit.git_branch <- function(object) {
    lookup(object$repo, branch_target(object))
}

##' @rdname lookup_commit
##' @export
lookup_commit.git_commit <- function(object) {
    object
}

##' @rdname lookup_commit
##' @export
lookup_commit.git_tag <- function(object) {
    lookup(object$repo, object$target)
}

##' @rdname lookup_commit
##' @export
lookup_commit.git_reference <- function(object) {
    lookup_commit(lookup(object$repo, object$sha))
}

##' @export
print.git_repository <- function(x, ...) {
    if (any(is_empty(x), is.null(repository_head(x)))) {
        cat(sprintf("Local:    %s\n", workdir(x)))
        cat("Head:     nothing commited (yet)\n")
    } else {
        if (is_detached(x)) {
            cat(sprintf("Local:    (detached) %s\n", workdir(x)))

            h <- repository_head(x)
        } else {
            cat(sprintf("Local:    %s %s\n",
                        repository_head(x)$name,
                        workdir(x)))

            h <- repository_head(x)
            u <- branch_get_upstream(h)
            if (!is.null(u)) {
                rn <- branch_remote_name(u)
                cat(sprintf("Remote:   %s @ %s (%s)\n",
                            substr(u$name, nchar(rn) + 2, nchar(u$name)),
                            rn,
                            branch_remote_url(u)))
            }

            h <- lookup(x, branch_target(repository_head(x)))
        }

        cat(sprintf("Head:     [%s] %s: %s\n",
                    substring(h$sha, 1, 7),
                    substring(as.character(h$author$when), 1, 10),
                    h$summary))
    }

    invisible(x)
}

##' Summary of repository
##'
##' @param object The repository \code{object}
##' @param ... Additional arguments affecting the summary produced.
##' @return None (invisible 'NULL').
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' summary(repo)
##'
##' ## Add file
##' add(repo, "test.txt")
##' summary(repo)
##'
##' ## Commit
##' commit(repo, "First commit message")
##' summary(repo)
##'
##' ## Change the file
##' writeLines(c("Hello again!", "Here is a second line", "And a third"),
##'            file.path(path, "test.txt"))
##' summary(repo)
##'
##' ## Add file and commit
##' add(repo, "test.txt")
##' commit(repo, "Second commit message")
##' summary(repo)
##'}
summary.git_repository <- function(object, ...) {
    print(object)
    cat("\n")

    n_branches <- sum(!is.na(unique(sapply(branches(object),
                                           branch_target))))
    n_tags <- sum(!is.na(unique(vapply(tags(object), "[[",
                                       character(1), "sha"))))

    work <- commits(object)
    n_commits <- length(work)
    n_authors <- length(unique(vapply(lapply(work, "[[", "author"),
                                      "[[", character(1), "name")))

    s <- .Call(git2r_status_list, object, TRUE, TRUE, TRUE, FALSE, TRUE)
    n_ignored <- length(s$ignored)
    n_untracked <- length(s$untracked)
    n_unstaged <- length(s$unstaged)
    n_staged <- length(s$staged)

    n_stashes <- length(stash_list(object))

    ## Determine max characters needed to display numbers
    n <- max(vapply(c(n_branches, n_tags, n_commits, n_authors,
                      n_stashes, n_ignored, n_untracked,
                      n_unstaged, n_staged),
                    nchar,
                    numeric(1)))

    fmt <- paste0("Branches:        %", n, "i\n",
                  "Tags:            %", n, "i\n",
                  "Commits:         %", n, "i\n",
                  "Contributors:    %", n, "i\n",
                  "Stashes:         %", n, "i\n",
                  "Ignored files:   %", n, "i\n",
                  "Untracked files: %", n, "i\n",
                  "Unstaged files:  %", n, "i\n",
                  "Staged files:    %", n, "i\n")
    cat(sprintf(fmt, n_branches, n_tags, n_commits, n_authors,
                n_stashes, n_ignored, n_untracked, n_unstaged,
                n_staged))

    cat("\nLatest commits:\n")
    lapply(commits(object, n = 5), print)

    invisible(NULL)
}

## Strip trailing slash or backslash, unless it's the current drive
## root (/) or a Windows drive, for example, 'c:\'.
strip_trailing_slash <- function(path) {
    if (!is.null(path) && grep("^(/|[a-zA-Z]:[/\\\\]?)$", path, invert = TRUE))
        path <- sub("/?$", "", path)
    path
}

##' Workdir of repository
##'
##' @template repo-param
##' @return Character vector with the path of the workdir. If the
##' repository is bare, \code{NULL} will be returned.
##' @export
##' @useDynLib git2r git2r_repository_workdir
##' @examples
##' \dontrun{
##' ## Create a directory in tempdir
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##'
##' ## Initialize a repository
##' repo <- init(path)
##'
##' ## Get the path of the workdir for repository
##' workdir(repo)
##' }
workdir <- function(repo = ".") {
    path <- .Call(git2r_repository_workdir, lookup_repository(repo))
    strip_trailing_slash(path)
}

##' Find path to repository for any file
##'
##' @param path A character vector specifying the path to a file or
##'     folder
##' @param ceiling The default is to not use the ceiling argument and
##'     start the lookup from path and walk across parent
##'     directories. When ceiling is 0, the lookup is only in
##'     path. When ceiling is 1, the lookup is in both the path and
##'     the parent to path.
##' @return Character vector with path (terminated by a file
##'     separator) to repository or NULL if this cannot be
##'     established.
##' @export
##' @useDynLib git2r git2r_repository_discover
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "example-1.txt"))
##' add(repo, "example-1.txt")
##' commit(repo, "First commit message")
##'
##' ## Create a second file. The file is not added for version control
##' ## in the repository.
##' dir.create(file.path(path, "example"))
##' file_2 <- file.path(path, "example/example-2.txt")
##' writeLines("Not under version control", file_2)
##'
##' ## Find the path to the repository using the path to the second file
##' discover_repository(file_2)
##'
##' ## Demonstrate the 'ceiling' argument
##' wd <- workdir(repo)
##' dir.create(file.path(wd, "temp"))
##'
##' ## Lookup repository in 'file.path(wd, "temp")'. Should return NULL
##' discover_repository(file.path(wd, "temp"), ceiling = 0)
##'
##' ## Lookup repository in parent to 'file.path(wd, "temp")'.
##' ## Should not return NULL
##' discover_repository(file.path(wd, "temp"), ceiling = 1)
##' }
discover_repository <- function(path = ".", ceiling = NULL) {
    if (identical(path, "."))
        path <- getwd()
    path <- normalizePath(path)

    if (!is.null(ceiling)) {
        ceiling <- as.integer(ceiling)
        if (identical(ceiling, 0L)) {
            ceiling <- dirname(path)
        } else if (identical(ceiling, 1L)) {
            ceiling <- dirname(dirname(path))
        } else {
            stop("'ceiling' must be either 0 or 1")
        }
    }

    path <- .Call(git2r_repository_discover, path, ceiling)
    strip_trailing_slash(path)
}

##' Internal utility function to lookup repository for methods
##'
##' @param repo repository \code{object} \code{git_repository}, or a
##'     path to a repository, or \code{NULL}.  If the \code{repo}
##'     argument is \code{NULL}, the repository is searched for with
##'     \code{\link{discover_repository}} in the current working
##'     directory.
##' @return git_repository
##' @noRd
lookup_repository <- function(repo = NULL) {
    if (is.null(repo) || identical(repo, ".")) {
        ## Try current working directory
        repo <- discover_repository(getwd())
        if (is.null(repo))
            stop("The working directory is not in a git repository")
    } else if (inherits(repo, "git_repository")) {
        return(repo)
    }

    repository(repo)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2018 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Get the SHA-1 of a git object
##'
##' Get the 40 character hexadecimal string of the SHA-1.
##' @param object a git object to get the SHA-1 from.
##' @return The 40 character hexadecimal string of the SHA-1.
##' @export
##' @examples \dontrun{
##' ## Create a directory in tempdir
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##'
##' ## Initialize a repository
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit(repo, "Commit message 1")
##'
##' ## Get the SHA-1 of the last commit
##' sha(last_commit(repo))
##' }
sha <- function(object) {
    UseMethod("sha", object)
}

##' @rdname sha
##' @export
sha.git_blob <- function(object) {
    object$sha
}

##' @rdname sha
##' @export
sha.git_branch <- function(object) {
    branch_target(object)
}

##' @rdname sha
##' @export
sha.git_commit <- function(object) {
    object$sha
}

##' @rdname sha
##' @export
sha.git_note <- function(object) {
    object$sha
}

##' @rdname sha
##' @export
sha.git_reference <- function(object) {
    object$sha
}

##' @rdname sha
##' @export
sha.git_reflog_entry <- function(object) {
    object$sha
}

##' @rdname sha
##' @export
sha.git_tag <- function(object) {
    object$sha
}

##' @rdname sha
##' @export
sha.git_tree <- function(object) {
    object$sha
}

##' @rdname sha
##' @export
sha.git_fetch_head <- function(object) {
    object$sha
}

##' @rdname sha
##' @export
sha.git_merge_result <- function(object) {
    object$sha
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2024 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Contributions
##'
##' See contributions to a Git repo
##' @template repo-param
##' @param breaks Default is \code{month}. Change to year, quarter,
##' week or day as necessary.
##' @param by Contributions by "commits" or "author". Default is "commits".
##' @return A \code{data.frame} with contributions.
##' @export
##' @useDynLib git2r git2r_revwalk_contributions
##' @examples
##' \dontrun{
##' ## Create directories and initialize repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo_1 <- tempfile(pattern="git2r-")
##' path_repo_2 <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo_1)
##' dir.create(path_repo_2)
##' repo_bare <- init(path_bare, bare = TRUE)
##'
##' ## Clone to repo 1 and config user
##' repo_1 <- clone(path_bare, path_repo_1)
##' config(repo_1, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Add changes to repo 1 and push to bare
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path_repo_1, "test.txt"))
##' add(repo_1, "test.txt")
##' commit(repo_1, "First commit message")
##'
##' ## Add more changes to repo 1
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path_repo_1, "test.txt"))
##' add(repo_1, "test.txt")
##' commit(repo_1, "Second commit message")
##'
##' ## Push to bare
##' push(repo_1, "origin", "refs/heads/master")
##'
##' ## Clone to repo 2
##' repo_2 <- clone(path_bare, path_repo_2)
##' config(repo_2, user.name = "Bob", user.email = "bob@@example.org")
##'
##' ## Add changes to repo 2
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
##'   "minim veniam, quis nostrud exercitation ullamco laboris nisi ut")
##' writeLines(lines, file.path(path_repo_2, "test.txt"))
##' add(repo_2, "test.txt")
##' commit(repo_2, "Third commit message")
##'
##' ## Push to bare
##' push(repo_2, "origin", "refs/heads/master")
##'
##' ## Pull changes to repo 1
##' pull(repo_1)
##'
##' ## View contributions by day
##' contributions(repo_1)
##'
##' ## View contributions by author and day
##' contributions(repo_1, by = "author")
##' }
contributions <- function(repo = ".",
                          breaks = c("month", "year", "quarter", "week", "day"),
                          by = c("commits", "author")) {
    breaks <- match.arg(breaks)
    by <- match.arg(by)

    ctbs <- .Call(git2r_revwalk_contributions, lookup_repository(repo),
                  TRUE, TRUE, FALSE)
    ctbs$when <- as.POSIXct(ctbs$when, origin = "1970-01-01", tz = "GMT")
    ctbs$when <- as.POSIXct(cut(ctbs$when, breaks = breaks))

    if (identical(by, "commits")) {
        ctbs <- as.data.frame(table(ctbs$when))
        names(ctbs) <- c("when", "n")
        ctbs$when <- as.Date(ctbs$when)
    } else {
        ## Create an index and tabulate
        ctbs$index <- paste0(ctbs$when, ctbs$author, ctbs$email)
        count <- as.data.frame(table(ctbs$index),
                               stringsAsFactors = FALSE)
        names(count) <- c("index", "n")

        ## Match counts and clean result
        ctbs <- as.data.frame(ctbs)
        ctbs$n <- count$n[match(ctbs$index, count$index)]
        ctbs <- unique(ctbs[, c("when", "author", "n")])
        ctbs$when <- as.Date(substr(as.character(ctbs$when), 1, 10))
        ctbs <- ctbs[order(ctbs$when, ctbs$author), ]
        row.names(ctbs) <- NULL
    }

    ctbs
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2017 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Status
##'
##' Display state of the repository working directory and the staging
##' area.
##' @rdname status-methods
##' @docType methods
##' @param repo The repository \code{object}
##'     \code{\linkS4class{git_repository}} to get status from. If the
##'     \code{repo} argument is missing, the repository is searched
##'     for with \code{\link{discover_repository}} in the current
##'     working directory.
##' @param staged Include staged files. Default TRUE.
##' @param unstaged Include unstaged files. Default TRUE.
##' @param untracked Include untracked files and directories. Default
##'     TRUE.
##' @param ignored Include ignored files. Default FALSE.
##' @param all_untracked Shows individual files in untracked
##'     directories if \code{untracked} is \code{TRUE}.
##' @param ... Additional arguments to status.
##' @return S3 class \code{git_status} with repository status
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file
##' writeLines("Hello world!", file.path(path, "test.txt"))
##'
##' ## Check status; untracked file
##' status(repo)
##'
##' ## Add file
##' add(repo, "test.txt")
##'
##' ## Check status; staged file
##' status(repo)
##'
##' ## Commit
##' commit(repo, "First commit message")
##'
##' ## Check status; clean
##' status(repo)
##'
##' ## Change the file
##' writeLines(c("Hello again!", "Here is a second line", "And a third"),
##'            file.path(path, "test.txt"))
##'
##' ## Check status; unstaged file
##' status(repo)
##'
##' ## Add file and commit
##' add(repo, "test.txt")
##' commit(repo, "Second commit message")
##'
##' ## Check status; clean
##' status(repo)
##'}
setGeneric("status",
           signature = "repo",
           function(repo,
                    staged    = TRUE,
                    unstaged  = TRUE,
                    untracked = TRUE,
                    ignored   = FALSE,
                    all_untracked = FALSE,
                    ...)
           standardGeneric("status"))

##' @rdname status-methods
##' @export
setMethod("status",
          signature(repo = "missing"),
          function(staged, unstaged, untracked, ignored, all_untracked, ...)
          {
              callGeneric(repo          = lookup_repository(),
                          staged        = staged,
                          unstaged      = unstaged,
                          untracked     = untracked,
                          ignored       = ignored,
                          all_untracked = all_untracked,
                          ...)
          }
)

##' @rdname status-methods
##' @export
setMethod("status",
          signature(repo = "git_repository"),
          function(repo, staged, unstaged, untracked, ignored, all_untracked, ...)
          {
              structure(.Call(git2r_status_list, repo, staged,
                              unstaged, untracked, all_untracked,
                              ignored),
                        class = "git_status")
          }
)

##' @export
print.git_status <- function(x, ...)
{
    display_status <- function(title, section) {
        cat(sprintf("%s:\n", title))

        for(i in seq_len(length(section))) {
            label <- names(section)[i]
            label <- paste0(toupper(substr(label, 1, 1)),
                            substr(label, 2, nchar(label)))
            cat(sprintf("\t%-12s%s\n", paste0(label, ":"), section[[i]]))
        }

        invisible(NULL)
    }

    if (max(sapply(x, length)) == 0L)
      cat("working directory clean\n")

    if (length(x$ignored)) {
        display_status("Ignored files", x$ignored)
        cat("\n")
    }

    if (length(x$untracked)) {
        display_status("Untracked files", x$untracked)
        cat("\n")
    }

    if (length(x$unstaged)) {
        display_status("Unstaged changes", x$unstaged)
        cat("\n")
    }

    if (length(x$staged)) {
        display_status("Staged changes", x$staged)
        cat("\n")
    }

    invisible(NULL)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create 2 directories in tempdir
path_bare <- tempfile(pattern = "git2r-")
path_repo <- tempfile(pattern = "git2r-")

dir.create(path_bare)
dir.create(path_repo)

## Create repositories
bare_repo <- init(path_bare, bare = TRUE)
repo <- clone(path_bare, path_repo)

## Check the repositores
stopifnot(identical(is_bare(bare_repo), TRUE))
stopifnot(identical(is_bare(repo), FALSE))

## Config repositories
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Add changes to repo
writeLines("Hello world", con = file.path(path_repo, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message")
branch_name <- branches(repo)[[1]]$name

## Check commit
stopifnot(identical(commit_1$author$name, "Alice"))
stopifnot(identical(commit_1$author$email, "alice@example.org"))
stopifnot(identical(length(commits(repo)), 1L))
stopifnot(identical(commits(repo)[[1]]$author$name, "Alice"))
stopifnot(identical(commits(repo)[[1]]$author$email, "alice@example.org"))

## Check push arguments
tools::assertError(push(repo, character(0), paste0("refs/heads/", branch_name)))
tools::assertError(push(repo, NA_character_, paste0("refs/heads/", branch_name)))
tools::assertError(push(repo, c("origin", "origin"), paste0("refs/heads/", branch_name)))
tools::assertError(push(repo, "origin"))
tools::assertError(push(repo, name = "origin"))
push(repo, "origin", character(0))
push(repo, "origin", NA_character_)
push(repo, "origin", c(NA_character_, NA_character_))
stopifnot(identical(reflog(repo, paste0("refs/remotes/origin/", branch_name)),
                    structure(list(), class = "git_reflog")))

## No tracking branch assigned to master
tools::assertError(push(branches(repo)[[1]]))

## Push changes from repo to origin
push(repo, "origin", paste0("refs/heads/", branch_name))
r <- reflog(repo, paste0("refs/remotes/origin/", branch_name))
stopifnot(identical(length(r), 1L))
r <- r[[1]]
stopifnot(identical(sha(r), sha(commit_1)))
stopifnot(identical(r$message, "update by push"))
stopifnot(identical(r$index, 0L))
stopifnot(identical(r$committer$name, "Alice"))
stopifnot(identical(r$committer$email, "alice@example.org"))
stopifnot(identical(r$refname, paste0("refs/remotes/origin/", branch_name)))
stopifnot(identical(r$repo$path, repo$path))
push(branches(repo)[[1]])

## Check result in bare repository
stopifnot(identical(length(commits(bare_repo)), 1L))
bare_commit_1 <- commits(bare_repo)[[1]]
stopifnot(identical(sha(commit_1), sha(bare_commit_1)))
stopifnot(identical(commit_1$author, bare_commit_1$author))
stopifnot(identical(commit_1$committer, bare_commit_1$committer))
stopifnot(identical(commit_1$summary, bare_commit_1$summary))
stopifnot(identical(commit_1$message, bare_commit_1$message))
stopifnot(!identical(commit_1$repo, bare_commit_1$repo))

## Add changes to repo and push head
writeLines(c("Hello world", "HELLO WORLD"),
           con = file.path(path_repo, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Commit message 2")
push(repo)
bare_commit_2 <- lookup(bare_repo, sha(commit_2))
stopifnot(identical(sha(commit_2), sha(bare_commit_2)))
stopifnot(identical(commit_2$author, bare_commit_2$author))
stopifnot(identical(commit_2$committer, bare_commit_2$committer))
stopifnot(identical(commit_2$summary, bare_commit_2$summary))
stopifnot(identical(commit_2$message, bare_commit_2$message))
stopifnot(!identical(commit_2$repo, bare_commit_2$repo))

## Check 'set_upstream'
branch_set_upstream(repository_head(repo), NULL)
push(repo, "origin", paste0("refs/heads/", branch_name))
stopifnot(is.null(branch_get_upstream(repository_head(repo))))
push(repo, "origin", paste0("refs/heads/", branch_name), set_upstream = TRUE)
stopifnot(!is.null(branch_get_upstream(repository_head(repo))))

if (identical(Sys.getenv("NOT_CRAN"), "true")) {
  ## Attempt to push with proxy=TRUE. If no actual proxy is configured, it may
  ## still succeed for local operations, or fail if there's a real network call.
  ## We'll wrap it in a tryCatch to avoid a hard failure.
  tryCatch({
    push(repo, "origin", paste0("refs/heads/", branch_name), proxy = TRUE)
    message("push() with proxy=TRUE succeeded (or no-op if local).")
  }, error = function(e) {
    message("push() with proxy=TRUE gave an error (likely no real proxy set): ",
            e$message)
  })
}

## Cleanup
unlink(path_bare, recursive = TRUE)
unlink(path_repo, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()

## Test to coerce
git_t <- structure(list(time = 1395567947, offset = 60),
                   class = "git_time")
stopifnot(identical(as.character(git_t), "2014-03-23 09:45:47 GMT"))
stopifnot(identical(as.character(git_t, usetz = FALSE), "2014-03-23 09:45:47"))
stopifnot(identical(as.POSIXct(git_t),
                    as.POSIXct(1395567947, tz = "GMT", origin = "1970-01-01")))
stopifnot(identical(print(git_t), git_t))

as.POSIXct(1395567947, origin = "1970-01-01", tz = "-03")

## Test that origin/tz can be passed to as.POSIXct
stopifnot(identical(as.POSIXct(git_t, tz = "Europe/Stockholm",
                               origin = "1980-02-02"),
                    as.POSIXct(1395567947, tz = "Europe/Stockholm",
                               origin = "1980-02-02")))

## Test that origin/tz can be passed to as.character
stopifnot(identical(as.character(git_t, tz = "Europe/Stockholm",
                                 origin = "1980-02-02"),
                    "2024-04-23 11:45:47 CEST"))
stopifnot(identical(as.character(git_t, tz = "Europe/Stockholm",
                                 origin = "1980-02-02", usetz = FALSE),
                    "2024-04-23 11:45:47"))

## Test that origin/tz can be passed to print
stopifnot(identical(
    utils::capture.output(print(git_t, tz = "Europe/Stockholm",
                                origin = "1980-02-02")),
    "2024-04-23 11:45:47 CEST"
))
stopifnot(identical(
    utils::capture.output(print(git_t, tz = "Europe/Stockholm",
                                origin = "1980-02-02", usetz = FALSE)),
    "2024-04-23 11:45:47"
))

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message 1")

## Create another commit
writeLines(c("Hello world!",
             "HELLO WORLD!"),
           file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Commit message 2")

## Check default ref
stopifnot(identical(note_default_ref(repo),
                    "refs/notes/commits"))

## Check that an invalid object argument in note_create produce an
## error.
tools::assertError(note_create(object = NULL, message = "test"))
tools::assertError(note_create(object = 1, message = "test"))

## Check that notes is an empty list
stopifnot(identical(notes(repo), list()))

## Create note in default namespace
note_1 <- note_create(commit_1, "Note-1")
stopifnot(identical(print(note_1), note_1))
stopifnot(identical(length(notes(repo)), 1L))
stopifnot(identical(sha(note_1), note_1$sha))
tools::assertError(note_create(commit_1, "Note-2"))
note_2 <- note_create(commit_1, "Note-2", force = TRUE)
stopifnot(identical(length(notes(repo)), 1L))

## Check that an invalid note argument in note_remove produce an
## error.
tools::assertError(note_remove(note = 1))

## Create note in named (review) namespace
note_3 <- note_create(commit_1, "Note-3", ref = "refs/notes/review")
note_4 <- note_create(commit_2, "Note-4", ref = "refs/notes/review")
stopifnot(identical(length(notes(repo, ref = "refs/notes/review")), 2L))
note_remove(note_3)
note_remove(note_4)
stopifnot(identical(notes(repo, ref = "refs/notes/review"), list()))
note_5 <- note_create(commit_1, "Note-5", ref = "review")
note_6 <- note_create(commit_2, "Note-6", ref = "review")
stopifnot(identical(length(notes(repo, ref = "review")), 2L))
note_remove(note_5)
note_remove(note_6)
stopifnot(identical(length(notes(repo, ref = "review")), 0L))

## Create note on blob and tree
tree_1 <- tree(commit_1)
note_7 <- note_create(tree_1, "Note-7")
stopifnot(is(object = lookup(repo, note_7$annotated), class2 = "git_tree"))
stopifnot(identical(length(notes(repo)), 2L))
blob_1 <- lookup(repo, tree_1$id[1])
note_8 <- note_create(blob_1, "Note-8")
stopifnot(is(object = lookup(repo, note_8$annotated), class2 = "git_blob"))
stopifnot(identical(length(notes(repo)), 3L))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2024 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Get blame for file
##'
##' @template repo-param
##' @param path Path to the file to consider
##' @return git_blame object with the following entries:
##' \describe{
##'   \item{path}{
##'     The path to the file of the blame
##'   }
##'   \item{hunks}{
##'     List of blame hunks
##'   }
##'   \item{repo}{
##'     The git_repository that contains the file
##'   }
##' }
##' \describe{
##'   \item{lines_in_hunk}{
##'     The number of lines in this hunk
##'   }
##'   \item{final_commit_id}{
##'     The sha of the commit where this line was last changed
##'   }
##'   \item{final_start_line_number}{
##'     The 1-based line number where this hunk begins, in the final
##'     version of the file
##'   }
##'   \item{final_signature}{
##'     Final committer
##'   }
##'   \item{orig_commit_id}{
##'     The sha of the commit where this hunk was found. This will usually
##'     be the same as 'final_commit_id'.
##'   }
##'   \item{orig_start_line_number}{
##'      The 1-based line number where this hunk begins in the file
##'      named by 'orig_path' in the commit specified by 'orig_commit_id'.
##'   }
##'   \item{orig_signature}{
##'     Origin committer
##'   }
##'   \item{orig_path}{
##'     The path to the file where this hunk originated, as of the commit
##'     specified by 'orig_commit_id'
##'   }
##'   \item{boundary}{
##'     TRUE iff the hunk has been tracked to a boundary commit.
##'   }
##'   \item{repo}{
##'     The \code{git_repository} object that contains the blame hunk
##'   }
##' }
##' @export
##' @useDynLib git2r git2r_blame_file
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a first user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create a second user and change the file
##' config(repo, user.name = "Bob", user.email = "bob@@example.org")
##' writeLines(c("Hello world!", "HELLO WORLD!", "HOLA"),
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Second commit message")
##'
##' ## Check blame
##' blame(repo, "example.txt")
##' }
blame <- function(repo = ".", path = NULL) {
    .Call(git2r_blame_file, lookup_repository(repo), path)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2024 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Number of files in git_diff object
##'
##' @param x The git_diff \code{object}
##' @return a non-negative integer
##' @export
length.git_diff <- function(x) {
    length(x$files)
}

##' @export
print.git_diff <- function(x, ...) {
    cat("Old:  ")
    if (is.character(x$old)) {
        cat(x$old, "\n")
    } else if (inherits(x$old, "git_tree")) {
        print(x$old)
    } else {
        cat("\n")
        print(x$old)
    }

    cat("New:  ")
    if (is.character(x$new)) {
        cat(x$new, "\n")
    } else if (inherits(x$new, "git_tree")) {
        print(x$new)
    } else {
        cat("\n")
        print(x$new)
    }

    invisible(x)
}

lines_per_file <- function(diff) {
    lapply(diff$files, function(x) {
        del <- add <- 0
        for (h in x$hunks) {
            for (l in h$lines) {
                if (l$origin == 45) {
                    del <- del + l$num_lines
                } else if (l$origin == 43) {
                    add <- add + l$num_lines
                }
            }
        }
        list(file = x$new_file, del = del, add = add)
    })
}

print_lines_per_file <- function(diff) {
  lpf <- lines_per_file(diff)
  files <- vapply(lpf, function(x) x$file, character(1))
  del <- vapply(lpf, function(x) x$del, numeric(1))
  add <- vapply(lpf, function(x) x$add, numeric(1))
  paste0(format(files), " | ", "-", format(del), " +", format(add))
}

hunks_per_file <- function(diff) {
    vapply(diff$files, function(x) length(x$hunks), numeric(1))
}

##' @export
summary.git_diff <- function(object, ...) {
    print(object)
    if (length(object) > 0) {
        plpf <- print_lines_per_file(object)
        hpf <- hunks_per_file(object)
        hunk_txt <- ifelse(hpf > 1, " hunks",
                    ifelse(hpf > 0, " hunk",
                           " hunk (binary file)"))
        phpf <- paste0("  in ", format(hpf), hunk_txt)
        cat("Summary:", paste0(plpf, phpf), sep = "\n")
    } else {
        cat("No changes.\n")
    }
}

##' Changes between commits, trees, working tree, etc.
##'
##' @rdname diff-methods
##' @export
##' @useDynLib git2r git2r_diff
##' @param x A \code{git_repository} object or the old \code{git_tree}
##'     object to compare to.
##' @param index \describe{
##'   \item{\emph{When object equals a git_repository}}{
##'     Whether to compare the index to HEAD. If FALSE (the default),
##'     then the working tree is compared to the index.
##'   }
##'   \item{\emph{When object equals a git_tree}}{
##'     Whether to use the working directory (by default), or the index
##'     (if set to TRUE) in the comparison to \code{object}.
##'   }
##' }
##' @param as_char logical: should the result be converted to a
##'     character string?. Default is FALSE.
##' @param filename If as_char is TRUE, then the diff can be written
##'     to a file with name filename (the file is overwritten if it
##'     exists). Default is NULL.
##' @param context_lines The number of unchanged lines that define the
##'     boundary of a hunk (and to display before and after). Defaults
##'     to 3.
##' @param interhunk_lines The maximum number of unchanged lines
##'     between hunk boundaries before the hunks will be merged into
##'     one. Defaults to 0.
##' @param old_prefix The virtual "directory" prefix for old file
##'     names in hunk headers. Default is "a".
##' @param new_prefix The virtual "directory" prefix for new file
##'     names in hunk headers. Defaults to "b".
##' @param id_abbrev The abbreviation length to use when formatting
##'     object ids. Defaults to the value of 'core.abbrev' from the
##'     config, or 7 if NULL.
##' @param path A character vector of paths / fnmatch patterns to
##'     constrain diff. Default is NULL which include all paths.
##' @param max_size A size (in bytes) above which a blob will be
##'     marked as binary automatically; pass a negative value to
##'     disable. Defaults to 512MB when max_size is NULL.
##' @return A \code{git_diff} object if as_char is FALSE. If as_char
##'     is TRUE and filename is NULL, a character string, else NULL.
##' @section Line endings:
##'
##' Different operating systems handle line endings
##' differently. Windows uses both a carriage-return character and a
##' linefeed character to represent a newline in a file. While Linux
##' and macOS use only the linefeed character for a newline in a
##' file. To avoid problems in your diffs, you can configure Git to
##' properly handle line endings using the \verb{core.autocrlf}
##' setting in the Git config file, see the Git documentation
##' (\url{https://git-scm.com/}).
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add, commit
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit(repo, "Commit message")
##'
##' ## Change the file
##' writeLines(c("Hello again!", "Here is a second line", "And a third"),
##'            file.path(path, "test.txt"))
##'
##' ## diff between index and workdir
##' diff_1 <- diff(repo)
##' summary(diff_1)
##' cat(diff(repo, as_char=TRUE))
##'
##' ## Diff between index and HEAD is empty
##' diff_2 <- diff(repo, index=TRUE)
##' summary(diff_2)
##' cat(diff(repo, index=TRUE, as_char=TRUE))
##'
##' ## Diff between tree and working dir, same as diff_1
##' diff_3 <- diff(tree(commits(repo)[[1]]))
##' summary(diff_3)
##' cat(diff(tree(commits(repo)[[1]]), as_char=TRUE))
##'
##' ## Add changes, diff between index and HEAD is the same as diff_1
##' add(repo, "test.txt")
##' diff_4 <- diff(repo, index=TRUE)
##' summary(diff_4)
##' cat(diff(repo, index=TRUE, as_char=TRUE))
##'
##' ## Diff between tree and index
##' diff_5 <- diff(tree(commits(repo)[[1]]), index=TRUE)
##' summary(diff_5)
##' cat(diff(tree(commits(repo)[[1]]), index=TRUE, as_char=TRUE))
##'
##' ## Diff between two trees
##' commit(repo, "Second commit")
##' tree_1 <- tree(commits(repo)[[2]])
##' tree_2 <- tree(commits(repo)[[1]])
##' diff_6 <- diff(tree_1, tree_2)
##' summary(diff_6)
##' cat(diff(tree_1, tree_2, as_char=TRUE))
##'
##' ## Binary files
##' set.seed(42)
##' writeBin(as.raw((sample(0:255, 1000, replace=TRUE))),
##'          con=file.path(path, "test.bin"))
##' add(repo, "test.bin")
##' diff_7 <- diff(repo, index=TRUE)
##' summary(diff_7)
##' cat(diff(repo, index=TRUE, as_char=TRUE))
##' }
diff.git_repository <- function(x,
                                index    = FALSE,
                                as_char  = FALSE,
                                filename = NULL,
                                context_lines = 3,
                                interhunk_lines = 0,
                                old_prefix = "a",
                                new_prefix = "b",
                                id_abbrev = NULL,
                                path = NULL,
                                max_size = NULL,
                                ...) {
    if (isTRUE(as_char)) {
        ## Make sure filename is character(0) to write to a
        ## character vector or a character vector with path in
        ## order to write to a file.
        filename <- as.character(filename)
        if (any(identical(filename, NA_character_),
                identical(nchar(filename), 0L))) {
            filename <- character(0)
        } else if (length(filename)) {
            filename <- normalizePath(filename, mustWork = FALSE)
        }
    } else {
        ## Make sure filename is NULL
        filename <- NULL
    }

    if (!is.null(id_abbrev))
        id_abbrev <- as.integer(id_abbrev)

    if (!is.null(max_size))
        max_size <- as.integer(max_size)

    .Call(git2r_diff, x, NULL, NULL, index, filename,
          as.integer(context_lines), as.integer(interhunk_lines),
          old_prefix, new_prefix, id_abbrev, path, max_size)
}

##' @rdname diff-methods
##' @param new_tree The new git_tree object to compare, or NULL.  If
##'     NULL, then we use the working directory or the index (see the
##'     \code{index} argument).
##' @param ... Not used.
##' @export
##' @useDynLib git2r git2r_diff
diff.git_tree <- function(x,
                          new_tree = NULL,
                          index    = FALSE,
                          as_char  = FALSE,
                          filename = NULL,
                          context_lines = 3,
                          interhunk_lines = 0,
                          old_prefix = "a",
                          new_prefix = "b",
                          id_abbrev = NULL,
                          path = NULL,
                          max_size = NULL,
                          ...) {
    if (isTRUE(as_char)) {
        ## Make sure filename is character(0) to write to a character
        ## vector or a character vector with path in order to write to
        ## a file.
        filename <- as.character(filename)
        if (any(identical(filename, NA_character_),
                identical(nchar(filename), 0L))) {
            filename <- character(0)
        } else if (length(filename)) {
            filename <- normalizePath(filename, mustWork = FALSE)
        }
    } else {
        ## Make sure filename is NULL
        filename <- NULL
    }

    if (!is.null(new_tree)) {
        if (!inherits(new_tree, "git_tree")) {
            stop("Not a git tree")
        }
        if (x$repo$path != new_tree$repo$path) {
            stop("Cannot compare trees in different repositories")
        }
    }

    if (!is.null(id_abbrev))
        id_abbrev <- as.integer(id_abbrev)

    if (!is.null(max_size))
        max_size <- as.integer(max_size)

    .Call(git2r_diff, NULL, x, new_tree, index, filename,
          as.integer(context_lines), as.integer(interhunk_lines),
          old_prefix, new_prefix, id_abbrev, path, max_size)
}

##' @export
base::diff

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2018 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Check printing of a class git_signature
when <- structure(list(time = 1395567947, offset = 60),
                  class = "git_time")

signature <- structure(list(name = "Alice",
                            email = "alice@example.org",
                            when = when),
                       class = "git_signature")
stopifnot(identical(print(signature), signature))

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013 - 2016 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Create a new environmental credential object
##'
##' Environmental variables can be written to the file
##' \code{.Renviron}. This file is read by \emph{R} during startup,
##' see \code{\link[base]{Startup}}.
##' @rdname cred_env-methods
##' @family git credential functions
##' @param username The name of the environmental variable that holds
##' the username for the authentication.
##' @param password The name of the environmental variable that holds
##' the password for the authentication.
##' @return A S4 \code{cred_env} object
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Create an environmental credential object for the username and
##' ## password.
##' cred <- cred_env("NAME_OF_ENV_VARIABLE_WITH_USERNAME",
##'                  "NAME_OF_ENV_VARIABLE_WITH_PASSWORD")
##' repo <- repository("git2r")
##' push(repo, credentials = cred)
##' }
setGeneric("cred_env",
           signature = c("username", "password"),
           function(username, password)
           standardGeneric("cred_env"))

##' @rdname cred_env-methods
##' @export
setMethod("cred_env",
          signature(username = "character",
                    password = "character"),
          function(username, password)
          {
              new("cred_env",
                  username = username,
                  password = password)
          }
)

##' Create a new personal access token credential object
##'
##' The personal access token is stored in an envrionmental variable.
##' Environmental variables can be written to the file
##' \code{.Renviron}. This file is read by \emph{R} during startup,
##' see \code{\link[base]{Startup}}. On GitHub, personal access tokens
##' function like ordinary OAuth access tokens. They can be used
##' instead of a password for Git over HTTPS, see
##' \url{https://help.github.com/articles/creating-an-access-token-for-command-line-use/}
##' @family git credential functions
##' @param token The name of the environmental variable that holds the
##' personal access token for the authentication. Defualt is
##' \code{GITHUB_PAT}.
##' @return A S4 \code{cred_token} object
##' @export
##' @examples
##' \dontrun{
##' ## Create a personal access token credential object.
##' ## This example assumes that the token is stored in
##' ## the 'GITHUB_PAT' environmental variable.
##' repo <- repository("git2r")
##' cred <- cred_token()
##' push(repo, credentials = cred)
##' }
cred_token <- function(token = "GITHUB_PAT")
{
    new("cred_token", token = token)
}

##' Create a new plain-text username and password credential object
##'
##' @rdname cred_user_pass-methods
##' @family git credential functions
##' @param username The username of the credential
##' @param password The password of the credential
##' @return A S4 \code{cred_user_pass} object
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Create a plain-text username and password credential object
##' cred_user_pass("Random Developer", "SecretPassword")
##' }
setGeneric("cred_user_pass",
           signature = c("username", "password"),
           function(username, password)
           standardGeneric("cred_user_pass"))

##' @rdname cred_user_pass-methods
##' @export
setMethod("cred_user_pass",
          signature(username = "character",
                    password = "character"),
          function(username, password)
          {
              new("cred_user_pass",
                  username = username,
                  password = password)
          }
)

##' Create a new passphrase-protected ssh key credential object
##'
##' @family git credential functions
##' @param publickey The path to the public key of the
##' credential. Default is \code{'~/.ssh/id_rsa.pub'}
##' @param privatekey The path to the private key of the
##' credential. Default is \code{'~/.ssh/id_rsa'}
##' @param passphrase The passphrase of the credential. Default is
##' \code{character(0)}. If getPass is installed and private key is
##' passphrase protected \code{getPass::getPass()} will be called
##' to allow for interactive and obfuscated interactive input of
##' the passphrase.
##' @return A S4 \code{cred_ssh_key} object
##' @export
##' @examples
##' \dontrun{
##' ## Create a ssh key credential object. It can optionally be
##' ## passphrase-protected
##' cred <- cred_ssh_key("~/.ssh/id_rsa.pub", "~/.ssh/id_rsa")
##' repo <- repository("git2r")
##' push(repo, credentials = cred)
##' }
cred_ssh_key <-  function (publickey = "~/.ssh/id_rsa.pub",
                           privatekey = "~/.ssh/id_rsa",
                           passphrase = character(0))
{
    publickey = normalizePath(publickey, mustWork = TRUE)
    privatekey = normalizePath(privatekey, mustWork = TRUE)

    if (length(passphrase) == 0) {
        if (ssh_key_needs_passphrase(privatekey)) {
            if (requireNamespace("getPass", quietly = TRUE)) {
                passphrase <- getPass::getPass()
            }
        }
    }

    new("cred_ssh_key",
        publickey  = publickey,
        privatekey = privatekey,
        passphrase = passphrase)
}

##' Check if private key is passphrase protected
##' @param privatekey The path to the private key of the
##' credential. Default is \code{'~/.ssh/id_rsa'}
##' @keywords internal
ssh_key_needs_passphrase <- function(privatekey = "~/.ssh/id_rsa"){
  private_content    <- readLines(privatekey, n=3)
  contains_encrypted <- grepl("encrypted",  private_content , ignore.case = TRUE)
  return(
    any(contains_encrypted)
  )
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' When
##'
##' Help method to extract the time as a character string from a
##' git_commit, git_signature, git_tag and git_time object.
##' @param object the \code{object} to extract the time slot from.
##' @inheritParams base::as.POSIXct
##' @inheritParams base::strptime
##' @return A \code{character} vector of length one.
##' @seealso \code{\link{git_time}}
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a first user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create tag
##' tag(repo, "Tagname", "Tag message")
##'
##' when(commits(repo)[[1]])
##' when(tags(repo)[[1]])
##' when(tags(repo)[[1]], tz = Sys.timezone())
##' }
when <- function(object, tz = "GMT", origin = "1970-01-01", usetz = TRUE) {
    if (inherits(object, "git_commit"))
        return(as.character(object$author$when, tz = tz, origin = origin,
                            usetz = usetz))
    if (inherits(object, "git_signature"))
        return(as.character(object$when, tz = tz, origin = origin,
                            usetz = usetz))
    if (inherits(object, "git_stash"))
        return(as.character(object$stasher$when, tz = tz, origin = origin,
                            usetz = usetz))
    if (inherits(object, "git_tag"))
        return(as.character(object$tagger$when, tz = tz, origin = origin,
                            usetz = usetz))
    if (inherits(object, "git_time"))
        return(as.character(object, tz = tz, origin = origin,
                            usetz = usetz))
    stop("Invalid 'object'")
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Create blob from file on disk
##'
##' Read a file from the filesystem and write its content to the
##' Object Database as a loose blob. The method is vectorized and
##' accepts a vector of files to create blobs from.
##' @rdname blob_create-methods
##' @docType methods
##' @param repo The repository where the blob(s) will be written. Can
##' be a bare repository.
##' @param path The file(s) from which the blob will be created.
##' @param relative TRUE if the file(s) from which the blob will be
##' created is relative to the repository's working dir. Default is
##' TRUE.
##' @return list of S4 class git_blob \code{objects}
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create blobs from files relative to workdir
##' writeLines("Hello, world!", file.path(path, "example-1.txt"))
##' writeLines("test content", file.path(path, "example-2.txt"))
##' blob_list_1 <- blob_create(repo, c("example-1.txt",
##'                                    "example-2.txt"))
##'
##' ## Create blobs from files not relative to workdir
##' temp_file_1 <- tempfile()
##' temp_file_2 <- tempfile()
##' writeLines("Hello, world!", temp_file_1)
##' writeLines("test content", temp_file_2)
##' blob_list_2 <- blob_create(repo, c(temp_file_1, temp_file_2), relative = FALSE)
##' }
setGeneric("blob_create",
           signature = c("repo", "path"),
           function(repo, path, relative = TRUE)
           standardGeneric("blob_create"))

##' @rdname blob_create-methods
##' @export
setMethod("blob_create",
          signature(repo = "git_repository",
                    path = "character"),
          function(repo, path, relative)
          {
              ## Argument checking
              stopifnot(is.logical(relative),
                        identical(length(relative), 1L))

              if (relative) {
                  result <- .Call(git2r_blob_create_fromworkdir,
                                  repo,
                                  path)
              } else {
                  path <- normalizePath(path, mustWork = TRUE)
                  result <- .Call(git2r_blob_create_fromdisk,
                                  repo,
                                  path)
              }

              result
          }
)

##' Content of blob
##'
##' @rdname content-methods
##' @docType methods
##' @param blob The blob \code{object}.
##' @param split Split blob content to text lines. Default TRUE.
##' @return The content of the blob. NA_character_ if the blob is binary.
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Display content of blob.
##' content(tree(commits(repo)[[1]])["example.txt"])
##' }
setGeneric("content",
           signature = "blob",
           function(blob,
                    split = TRUE)
           standardGeneric("content"))

##' @rdname content-methods
##' @export
setMethod("content",
          signature(blob = "git_blob"),
          function(blob, split)
          {
              if (is_binary(blob))
                  return(NA_character_)

              ret <- .Call(git2r_blob_content, blob)
              if (isTRUE(split))
                  ret <- strsplit(ret, "\n")[[1]]
              ret
          }
)

##' Determine the sha from a blob string
##'
##' The blob is not written to the object database.
##' @rdname hash-methods
##' @docType methods
##' @param data The string vector to hash.
##' @return A string vector with the sha for each string in data.
##' @keywords methods
##' @examples
##' \dontrun{
##' identical(hash(c("Hello, world!\n",
##'                  "test content\n")),
##'                c("af5626b4a114abcb82d63db7c8082c3c4756e51b",
##'                  "d670460b4b4aece5915caf5c68d12f560a9fe3e4"))
##' }
setGeneric("hash",
           signature = "data",
           function(data)
           standardGeneric("hash"))

##' @rdname hash-methods
##' @export
setMethod("hash",
          signature(data = "character"),
          function(data)
          {
              .Call(git2r_odb_hash, data)
          }
)

##' Determine the sha from a blob in a file
##'
##' The blob is not written to the object database.
##' @rdname hashfile-methods
##' @docType methods
##' @param path The path vector with files to hash.
##' @return A vector with the sha for each file in path.
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Create a file. NOTE: The line endings from writeLines gives
##' ## LF (line feed) on Unix/Linux and CRLF (carriage return, line feed)
##' ## on Windows. The example use writeChar to have more control.
##' path <- tempfile()
##' f <- file(path, "wb")
##' writeChar("Hello, world!\n", f, eos = NULL)
##' close(f)
##'
##' ## Generate hash
##' hashfile(path)
##' identical(hashfile(path), hash("Hello, world!\n"))
##' }
setGeneric("hashfile",
           signature = "path",
           function(path)
           standardGeneric("hashfile"))

##' @rdname hashfile-methods
##' @export
setMethod("hashfile",
          signature(path = "character"),
          function(path)
          {
              path <- normalizePath(path, mustWork = TRUE)
              .Call(git2r_odb_hashfile, path)
          }
)

##' Is blob binary
##'
##' @rdname is_binary-methods
##' @docType methods
##' @param blob The blob \code{object}.
##' @return TRUE if binary data, FALSE if not.
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##'
##' ## Check if binary
##' b_text <- tree(commit_1)["example.txt"]
##' is_binary(b_text)
##'
##' ## Commit plot file (binary)
##' x <- 1:100
##' y <- x^2
##' png(file.path(path, "plot.png"))
##' plot(y ~ x, type = "l")
##' dev.off()
##' add(repo, "plot.png")
##' commit_2 <- commit(repo, "Second commit message")
##'
##' ## Check if binary
##' b_png <- tree(commit_2)["plot.png"]
##' is_binary(b_png)
##' }
setGeneric("is_binary",
           signature = "blob",
           function(blob)
           standardGeneric("is_binary"))

##' @rdname is_binary-methods
##' @export
setMethod("is_binary",
          signature(blob = "git_blob"),
          function(blob)
          {
              .Call(git2r_blob_is_binary, blob)
          }
)

##' Check if object is S4 class git_blob
##'
##' @param object Check if object is S4 class git_blob
##' @return TRUE if object is S4 class git_blob, else FALSE
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##' blob_1 <- tree(commit_1)["example.txt"]
##'
##' ## Check if blob
##' is_blob(commit_1)
##' is_blob(blob_1)
##' }
is_blob <- function(object) {
    is(object = object, class2 = "git_blob")
}

##' Size in bytes of the contents of a blob
##'
##' @docType methods
##' @param x The blob \code{object}
##' @return a non-negative integer
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##' blob_1 <- tree(commit_1)["example.txt"]
##'
##' ## Get length in size of bytes of the content of the blob
##' length(blob_1)
##' }
setMethod("length",
          signature(x = "git_blob"),
          function(x)
          {
              .Call(git2r_blob_rawsize, x)
          }
)

##' Brief summary of blob
##'
##' @aliases show,git_blob-methods
##' @docType methods
##' @param object The blob \code{object}
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Brief summary of the blob in the repository
##' tree(commits(repo)[[1]])["example.txt"]
##' }
setMethod("show",
          signature(object = "git_blob"),
          function(object)
          {
              cat(sprintf("blob:  %s\n", object@sha))
          }
)

##' Summary of blob
##'
##' @docType methods
##' @param object The blob \code{object}
##' @param ... Additional arguments affecting the summary produced.
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##' blob_1 <- tree(commit_1)["example.txt"]
##'
##' ## Get summary of the blob
##' summary(blob_1)
##' }
setMethod("summary",
          signature(object = "git_blob"),
          function(object, ...)
          {
              cat(sprintf("blob:  %s\nsize:  %i bytes\n",
                          object@sha,
                          length(object)))
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2024 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Pull
##'
##' @template repo-param
##' @param credentials The credentials for remote repository
##'     access. Default is NULL. To use and query an ssh-agent for the
##'     ssh key credentials, let this parameter be NULL (the default).
##' @param merger Who made the merge, if the merge is non-fast forward
##'     merge that creates a merge commit. The
##'     \code{default_signature} for \code{repo} is used if this
##'     parameter is \code{NULL}.
##' @template return-git_merge_result
##' @export
##' @useDynLib git2r git2r_merge_fetch_heads
##' @examples
##' \dontrun{
##' ## Initialize repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo_1 <- tempfile(pattern="git2r-")
##' path_repo_2 <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo_1)
##' dir.create(path_repo_2)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo_1 <- clone(path_bare, path_repo_1)
##'
##' ## Config first user and commit a file
##' config(repo_1, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "First commit message")
##'
##' ## Push commits from first repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo_1, "origin", "refs/heads/master")
##'
##' ## Clone to second repository
##' repo_2 <- clone(path_bare, path_repo_2)
##' config(repo_2, user.name = "Bob", user.email = "bob@@example.org")
##'
##' ## Change file and commit
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "Second commit message")
##'
##' ## Push commits from first repository to bare repository
##' push(repo_1)
##'
##' ## Pull changes to repo_2
##' pull(repo_2)
##'
##' ## Change file again and commit. This time in repository 2
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
##'   "minim veniam, quis nostrud exercitation ullamco laboris nisi ut")
##' writeLines(lines, file.path(path_repo_2, "example.txt"))
##' add(repo_2, "example.txt")
##' commit(repo_2, "Third commit message")
##'
##' ## Push commits from second repository to bare repository
##' push(repo_2)
##'
##' ## Pull changes to repo_1
##' pull(repo_1)
##'
##' ## List commits in repositories
##' commits(repo_1)
##' commits(repo_2)
##' commits(repo_bare)
##' }
pull <- function(repo = ".", credentials = NULL, merger = NULL) {
    repo <- lookup_repository(repo)
    if (is.null(merger))
        merger <- default_signature(repo)
    current_branch <- repository_head(repo)

    if (is.null(current_branch))
        stop("'branch' is NULL")
    if (!is_local(current_branch))
        stop("'branch' is not local")
    upstream_branch <- branch_get_upstream(current_branch)
    if (is.null(upstream_branch))
        stop("'branch' is not tracking a remote branch")

    fetch(repo        = repo,
          name        = branch_remote_name(upstream_branch),
          credentials = credentials)

    ## fetch heads marked for merge
    fh <- fetch_heads(repo)
    fh <- fh[vapply(fh, "[[", logical(1), "is_merge")]

    if (identical(length(fh), 0L))
        stop("Remote ref was not fetched")

    .Call(git2r_merge_fetch_heads, fh, merger)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Coerce entries in a git_tree to a \code{data.frame}
##'
##' The entries in a tree are coerced to a \code{data.frame}
##'
##'
##' The \code{data.frame} have the following columns:
##' \describe{
##'
##'   \item{filemode}{
##'     The UNIX file attributes of a tree entry
##'   }
##'
##'   \item{type}{
##'     String representation of the tree entry type
##'   }
##'
##'   \item{sha}{
##'     The sha of a tree entry
##'   }
##'
##'   \item{name}{
##'     The filename of a tree entry
##'   }
##'
##' }
##' @name coerce-git_tree-data.frame-method
##' @aliases coerce,git_tree,data.frame-method
##' @docType methods
##' @param from The tree \code{object}
##' @return \code{data.frame}
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' dir.create(file.path(path, "subfolder"))
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create three files and commit
##' writeLines("First file",  file.path(path, "example-1.txt"))
##' writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
##' writeLines("Third file",  file.path(path, "example-3.txt"))
##' add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
##' new_commit <- commit(repo, "Commit message")
##'
##' ## Coerce tree to a data.frame
##' df <- as(tree(new_commit), "data.frame")
##' df
##' }
setAs(from = "git_tree",
      to = "data.frame",
      def = function(from)
      {
          data.frame(mode = sprintf("%06o", from@filemode),
                     type = from@type,
                     sha  = from@id,
                     name = from@name,
                     stringsAsFactors = FALSE)
      }
)

##' Coerce entries in a git_tree to a list of entry objects
##'
##' @name coerce-git_tree-list-method
##' @docType methods
##' @param from The tree \code{object}
##' @return list of entry objects
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' dir.create(file.path(path, "subfolder"))
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create three files and commit
##' writeLines("First file",  file.path(path, "example-1.txt"))
##' writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
##' writeLines("Third file",  file.path(path, "example-3.txt"))
##' add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
##' new_commit <- commit(repo, "Commit message")
##'
##' ## Inspect size of each blob in tree
##' invisible(lapply(as(tree(new_commit), "list"),
##'   function(obj) {
##'     if (is_blob(obj))
##'       summary(obj)
##'     NULL
##'   }))
##' }
setAs(from = "git_tree",
      to = "list",
      def = function(from)
      {
          lapply(from@id, function(sha) lookup(from@repo, sha))
      }
)

##' Tree
##'
##' Get the tree pointed to by a commit or stash.
##' @rdname tree-methods
##' @docType methods
##' @param object the \code{commit} or \code{stash} object
##' @return A S4 class git_tree object
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a first user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' tree(commits(repo)[[1]])
##' summary(tree(commits(repo)[[1]]))
##' }
setGeneric("tree",
           signature = "object",
           function(object) standardGeneric("tree"))

##' @rdname tree-methods
##' @export
setMethod("tree",
          signature(object = "git_commit"),
          function(object)
          {
              .Call(git2r_commit_tree, object)
          }
)

##' @rdname tree-methods
##' @export
setMethod("tree",
          signature(object = "git_stash"),
          function(object)
          {
              .Call(git2r_commit_tree, object)
          }
)

##' Brief summary of tree
##'
##' @aliases show,git_tree-methods
##' @docType methods
##' @param object The tree \code{object}
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Brief summary of the tree in the repository
##' tree(commits(repo)[[1]])
##' }
setMethod("show",
          signature(object = "git_tree"),
          function(object)
          {
              cat(sprintf("tree:  %s\n", object@sha))
          }
)

##' Summary of tree
##'
##' @aliases summary,git_tree-methods
##' @docType methods
##' @param object The tree \code{object}
##' @param ... Additional arguments affecting the summary produced.
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' summary(tree(commits(repo)[[1]]))
##' }
setMethod("summary",
          signature(object = "git_tree"),
          function(object, ...)
          {
              show(as(object, "data.frame"))
          }
)

##' Extract object from tree
##'
##' Lookup a tree entry by its position in the tree
##' @rdname tree-index-methods
##' @docType methods
##' @param x The tree \code{object}
##' @param i The index (integer or logical) of the tree object to
##' extract. If negative values, all elements except those indicated
##' are selected. A character vector to match against the names of
##' objects to extract.
##' @return Git object
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' dir.create(file.path(path, "subfolder"))
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create three files and commit
##' writeLines("First file",  file.path(path, "example-1.txt"))
##' writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
##' writeLines("Third file",  file.path(path, "example-3.txt"))
##' add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
##' new_commit <- commit(repo, "Commit message")
##'
##' ## Pick a tree in the repository
##' tree_object <- tree(new_commit)
##'
##' ## Summarize tree
##' summary(tree_object)
##'
##' ## Select item by name
##' tree_object["example-1.txt"]
##'
##' ## Select first item in tree
##' tree_object[1]
##'
##' ## Select first three items in tree
##' tree_object[1:3]
##'
##' ## Select all blobs in tree
##' tree_object[vapply(as(tree_object, 'list'), is_blob, logical(1))]
##' }
setMethod("[",
          signature(x = "git_tree", i = "integer", j = "missing"),
          function(x, i)
          {
              i <- seq_len(length(x))[i]
              ret <- lapply(i, function(j) lookup(x@repo, x@id[j]))
              if (identical(length(ret), 1L))
                  ret <- ret[[1]]
              ret
          }
)

##' @rdname tree-index-methods
##' @export
setMethod("[",
          signature(x = "git_tree", i = "numeric", j = "missing"),
          function(x, i)
          {
              x[as.integer(i)]
          }
)

##' @rdname tree-index-methods
##' @export
setMethod("[",
          signature(x = "git_tree", i = "logical", j = "missing"),
          function(x, i)
          {
              x[seq_along(x)[i]]
          }
)

##' @rdname tree-index-methods
##' @export
setMethod("[",
          signature(x = "git_tree", i = "character", j = "missing"),
          function(x, i)
          {
              x[which(x@name %in% i)]
          }
)

##' Number of entries in tree
##'
##' @docType methods
##' @param x The tree \code{object}
##' @return a non-negative integer or double (which will be rounded
##' down)
##' @keywords methods
##' @export
setMethod("length",
          signature(x = "git_tree"),
          function(x)
          {
              length(x@id)
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Plot commits over time
##'
##' @param x The repository to plot
##' @param breaks Default is \code{month}. Change to year, quarter,
##' week or day as necessary.
##' @param main Default title for the plot is "Commits on repo:" and
##' repository workdir basename. Supply a new title if you desire one.
##' @param ... Additional arguments affecting the plot
##' @importFrom graphics axis
##' @importFrom graphics barplot
##' @export
##' @examples
##' \dontrun{
##' ## Initialize repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- clone("https://github.com/ropensci/git2r.git", path)
##'
##' ## Plot commits
##' plot(repo)
##' }
plot.git_repository <- function(x,
                                breaks = c("month",
                                           "year",
                                           "quarter",
                                           "week",
                                           "day"),
                                main = NULL,
                                ...) {
    breaks <- match.arg(breaks)

    df <- contributions(x, breaks = breaks, by = "commits")
    tmp <- data.frame(when = seq(from = min(df$when),
                                 to   = max(df$when),
                                 by   = breaks),
                      n = 0)
    i <- match(df$when, tmp$when)
    tmp$n[i] <- df$n
    df <- tmp

    xlab <- sprintf("Time [%s]", breaks)
    ylab <- "Number of commits"
    if (is.null(main)) {
        if (is_bare(x)) {
            main <- "Commits"
        } else {
            main <- sprintf("Commits on repository: %s",
                            basename(workdir(x)))
        }
    }

    mp <- barplot(df$n, xlab = xlab, ylab = ylab, main = main, ...)
    axis(1, at = mp, labels = seq(min(df$when), max(df$when), breaks))
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2024 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Determine previous branch name
##'
##' @param repo The repository.
##' @noRd
previous_branch_name <- function(repo) {
    branch <- revparse_single(repo, "@{-1}")$sha
    branch <- sapply(references(repo), function(x) {
        if (is.null(x$sha))
            return(NA_character_)
        if (x$sha == branch)
            return(x$shorthand)
        NA_character_
    })
    branch <- branch[vapply(branch, Negate(is.na), logical(1))]

    branch <- sapply(branches(repo, "local"), function(x) {
        ifelse(x$name %in% branch, x$name, NA_character_)
    })
    branch <- branch[vapply(branch, Negate(is.na), logical(1))]

    if (any(!is.character(branch), !identical(length(branch), 1L))) {
        stop("'branch' must be a character vector of length one")
    }

    branch
}

##' @useDynLib git2r git2r_checkout_tree
##' @useDynLib git2r git2r_repository_set_head
##' @noRd
checkout_branch <- function(object, force) {
    ref_name <- paste0("refs/heads/", object$name)
    .Call(git2r_checkout_tree, object$repo, ref_name, force)
    .Call(git2r_repository_set_head, object$repo, ref_name)
}

##' @useDynLib git2r git2r_checkout_tree
##' @useDynLib git2r git2r_repository_set_head_detached
##' @noRd
checkout_commit <- function(object, force) {
    .Call(git2r_checkout_tree, object$repo, object$sha, force)
    .Call(git2r_repository_set_head_detached, object)
}

##' @useDynLib git2r git2r_checkout_tree
##' @useDynLib git2r git2r_repository_set_head_detached
##' @noRd
checkout_tag <- function(object, force) {
    .Call(git2r_checkout_tree, object$repo, object$target, force)
    .Call(git2r_repository_set_head_detached,
          lookup(object$repo, object$target))
}

checkout_git_object <- function(object, force) {
    if (is_branch(object)) {
        checkout_branch(object, force)
        return(TRUE)
    }

    if (is_commit(object)) {
        checkout_commit(object, force)
        return(TRUE)
    }

    if (is_tag(object)) {
        checkout_tag(object, force)
        return(TRUE)
    }

    FALSE
}

##' Checkout
##'
##' Update files in the index and working tree to match the content of
##' the tree pointed at by the treeish object (commit, tag or tree).
##' The default checkout strategy (\code{force = FALSE}) will only
##' make modifications that will not lose changes. Use \code{force =
##' TRUE} to force working directory to look like index.
##' @param object A path to a repository, or a \code{git_repository}
##'     object, or a \code{git_commit} object, or a \code{git_tag}
##'     object, or a \code{git_tree} object.
##' @param branch name of the branch to check out. Only used if object
##'     is a path to a repository or a \code{git_repository} object.
##' @param create create branch if it doesn't exist. Only used if
##'     object is a path to a repository or a \code{git_repository}
##'     object.
##' @param force If \code{TRUE}, then make working directory match
##'     target. This will throw away local changes. Default is
##'     \code{FALSE}.
##' @param path Limit the checkout operation to only certain
##'     paths. This argument is only used if branch is NULL. Default
##'     is \code{NULL}.
##' @param ...  Additional arguments. Not used.
##' @return invisible NULL
##' @export
##' @useDynLib git2r git2r_checkout_path
##' @useDynLib git2r git2r_repository_set_head
##' @examples
##' \dontrun{
##' ## Create directories and initialize repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo_1 <- tempfile(pattern="git2r-")
##' path_repo_2 <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo_1)
##' dir.create(path_repo_2)
##' repo_bare <- init(path_bare, bare = TRUE)
##'
##' ## Clone to repo 1 and config user
##' repo_1 <- clone(path_bare, path_repo_1)
##' config(repo_1, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Add changes to repo 1 and push to bare
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path_repo_1, "test.txt"))
##' add(repo_1, "test.txt")
##' commit(repo_1, "First commit message")
##' push(repo_1, "origin", "refs/heads/master")
##'
##' ## Create and checkout 'dev' branch in repo 1
##' checkout(repo_1, "dev", create = TRUE)
##'
##' ## Add changes to 'dev' branch in repo 1 and push to bare
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path_repo_1, "test.txt"))
##' add(repo_1, "test.txt")
##' commit(repo_1, "Second commit message")
##' push(repo_1, "origin", "refs/heads/dev")
##'
##' ## Clone to repo 2
##' repo_2 <- clone(path_bare, path_repo_2)
##' config(repo_2, user.name = "Bob", user.email = "bob@@example.org")
##'
##' ## Read content of 'test.txt'
##' readLines(file.path(path_repo_2, "test.txt"))
##'
##' ## Checkout dev branch
##' checkout(repo_2, "dev")
##'
##' ## Read content of 'test.txt'
##' readLines(file.path(path_repo_2, "test.txt"))
##'
##' ## Edit "test.txt" in repo_2
##' writeLines("Hello world!", con = file.path(path_repo_2, "test.txt"))
##'
##' ## Check status
##' status(repo_2)
##'
##' ## Checkout "test.txt"
##' checkout(repo_2, path = "test.txt")
##'
##' ## Check status
##' status(repo_2)
##' }
checkout <- function(object = NULL,
                     branch = NULL,
                     create = FALSE,
                     force  = FALSE,
                     path   = NULL,
                     ...) {
    if (checkout_git_object(object, force))
        return(invisible(NULL))

    object <- lookup_repository(object)
    if (is.null(branch)) {
        if (is.null(path))
            stop("missing 'branch' or 'path' argument")
        .Call(git2r_checkout_path, object, path)
        return(invisible(NULL))
    }

    if (!is.character(branch) || !identical(length(branch), 1L))
        stop("'branch' must be a character vector of length one")

    if (is_empty(object)) {
        if (!isTRUE(create))
            stop(sprintf("'%s' did not match any branch", branch))
        ref_name <- paste0("refs/heads/", branch)
        .Call(git2r_repository_set_head, object, ref_name)
        return(invisible(NULL))
    }

    if (identical(branch, "-"))
        branch <- previous_branch_name(object)

    ## Check if branch exists in a local branch
    lb <- branches(object, "local")
    lb <- lb[vapply(lb, "[[", character(1), "name") == branch]
    if (length(lb)) {
        checkout_branch(lb[[1]], force)
        return(invisible(NULL))
    }

    ## Check if there exists exactly one remote branch with a matching
    ## name.
    rb <- branches(object, "remote")

    ## Split remote/name to check for a unique name
    name <- vapply(rb, function(x) {
        remote <- strsplit(x$name, "/")[[1]][1]
        sub(paste0("^", remote, "/"), "", x$name)
    }, character(1))
    i <- which(name == branch)
    if (identical(length(i), 1L)) {
        ## Create branch and track remote
        commit <- lookup(object, branch_target(rb[[i]]))
        branch <- branch_create(commit, branch)
        branch_set_upstream(branch, rb[[i]]$name)
        checkout_branch(branch, force)
        return(invisible(NULL))
    }

    if (isTRUE(create)) {
        ## Create branch
        commit <- lookup(object, branch_target(repository_head(object)))
        checkout_branch(branch_create(commit, branch), force)
        return(invisible(NULL))
    }

    ## Check if branch object is specified by revision.
    if (checkout_git_object(revparse_single(object, branch), force))
        return(invisible(NULL))

    stop(sprintf("'%s' did not match any branch", branch))
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2024 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Ahead Behind
##'
##' Count the number of unique commits between two commit objects.
##' @param local a git_commit object. Can also be a tag or a branch,
##'     and in that case the commit will be the target of the tag or
##'     branch.
##' @param upstream a git_commit object. Can also be a tag or a
##'     branch, and in that case the commit will be the target of the
##'     tag or branch.
##' @return An integer vector of length 2 with number of commits that
##'     the upstream commit is ahead and behind the local commit
##' @export
##' @useDynLib git2r git2r_graph_ahead_behind
##' @examples \dontrun{
##' ## Create a directory in tempdir
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##'
##' ## Initialize a repository
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##' tag_1 <- tag(repo, "Tagname1", "Tag message 1")
##'
##' # Change file and commit
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit_2 <- commit(repo, "Commit message 2")
##' tag_2 <- tag(repo, "Tagname2", "Tag message 2")
##'
##' ahead_behind(commit_1, commit_2)
##' ahead_behind(tag_1, tag_2)
##' }
ahead_behind <- function(local = NULL, upstream = NULL) {
    .Call(git2r_graph_ahead_behind,
          lookup_commit(local),
          lookup_commit(upstream))
}

##' Add sessionInfo to message
##'
##' @param message The message.
##' @return message with appended sessionInfo
##' @importFrom utils capture.output
##' @importFrom utils sessionInfo
##' @noRd
add_session_info <- function(message) {
    paste0(message, "\n\nsessionInfo:\n",
           paste0(utils::capture.output(utils::sessionInfo()),
                  collapse = "\n"))
}

##' Commit
##'
##' @template repo-param
##' @param message The commit message.
##' @param all Stage modified and deleted files. Files not added to
##'     Git are not affected.
##' @param session Add sessionInfo to commit message. Default is
##'     FALSE.
##' @param author Signature with author and author time of commit.
##' @param committer Signature with committer and commit time of
##'     commit.
##' @return A list of class \code{git_commit} with entries:
##' \describe{
##'   \item{sha}{
##'     The 40 character hexadecimal string of the SHA-1
##'   }
##'   \item{author}{
##'     An author signature
##'   }
##'   \item{committer}{
##'     The committer signature
##'   }
##'   \item{summary}{
##'     The short "summary" of a git commit message, comprising the first
##'     paragraph of the message with whitespace trimmed and squashed.
##'   }
##'   \item{message}{
##'     The message of a commit
##'   }
##'   \item{repo}{
##'     The \code{git_repository} object that contains the commit
##'   }
##' }
##' @export
##' @useDynLib git2r git2r_commit
##' @useDynLib git2r git2r_index_remove_bypath
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##' }
commit <- function(repo      = ".",
                   message   = NULL,
                   all       = FALSE,
                   session   = FALSE,
                   author    = NULL,
                   committer = NULL) {
    repo <- lookup_repository(repo)
    if (is.null(author))
        author <- default_signature(repo)
    if (is.null(committer))
        committer <- default_signature(repo)

    stopifnot(is.character(message), identical(length(message), 1L))
    if (!nchar(message[1]))
        stop("Aborting commit due to empty commit message.")

    if (isTRUE(all)) {
        s <- status(repo,
                    unstaged  = TRUE,
                    staged    = FALSE,
                    untracked = FALSE,
                    ignored   = FALSE)

        ## Convert list of lists to character vector
        unstaged <- unlist(s$unstaged)
        for (i in seq_along(unstaged)) {
            if (names(unstaged)[i] == "modified") {
                ## Stage modified files
                add(repo, unstaged[i])
            } else if (names(unstaged)[i] == "deleted") {
                ## Stage deleted files
                .Call(git2r_index_remove_bypath, repo, unstaged[i])
            }
        }

    }

    if (isTRUE(session))
        message <- add_session_info(message)

    .Call(git2r_commit, repo, message, author, committer)
}

##' Check limit in number of commits
##' @noRd
get_upper_limit_of_commits <- function(n) {
    if (is.null(n)) {
        n <- -1L
    } else if (is.numeric(n)) {
        if (!identical(length(n), 1L))
            stop("'n' must be integer")
        if (abs(n - round(n)) >= .Machine$double.eps^0.5)
            stop("'n' must be integer")
        n <- as.integer(n)
    } else {
        stop("'n' must be integer")
    }

    n
}

shallow_commits <- function(repo, sha, n) {
    ## List to hold result
    result <- list()

    ## Get latest commit
    x <- lookup(repo, sha)

    ## Repeat until no more parent commits
    repeat {
        if (n == 0) {
            break
        } else if (n > 0) {
            n <- n - 1
        }

        if (is.null(x))
            break
        result[[length(result) + 1]] <- x

        ## Get parent to commit
        x <- tryCatch(parents(x)[[1]], error = function(e) NULL)
    }

    result
}

##' Commits
##'
##' @template repo-param
##' @param topological Sort the commits in topological order (parents
##'     before children); can be combined with time sorting. Default
##'     is TRUE.
##' @param time Sort the commits by commit time; Can be combined with
##'     topological sorting. Default is TRUE.
##' @param reverse Sort the commits in reverse order; can be combined
##'     with topological and/or time sorting. Default is FALSE.
##' @param n The upper limit of the number of commits to output. The
##'     default is NULL for unlimited number of commits.
##' @param ref The name of a reference to list commits from e.g. a tag
##'     or a branch. The default is NULL for the current branch.
##' @param path The path to a file. If not NULL, only commits modifying
##'     this file will be returned. Note that modifying commits that
##'     occurred before the file was given its present name are not
##'     returned; that is, the output of \code{git log} with
##'     \code{--no-follow} is reproduced.
##' @return list of commits in repository
##' @export
##' @useDynLib git2r git2r_reference_dwim
##' @useDynLib git2r git2r_revwalk_list2
##' @useDynLib git2r git2r_revwalk_list
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Change file and commit
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Second commit message")
##'
##' ## Create a tag
##' tag(repo, "Tagname", "Tag message")
##'
##' ## Change file again and commit
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
##'   "minim veniam, quis nostrud exercitation ullamco laboris nisi ut")
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Third commit message")
##'
##' ## Create a new file containing R code, and commit.
##' writeLines(c("x <- seq(1,100)",
##'              "print(mean(x))"),
##'            file.path(path, "mean.R"))
##' add(repo, "mean.R")
##' commit(repo, "Fourth commit message")
##'
##' ## List the commits in the repository
##' commits(repo)
##'
##' ## List the commits starting from the tag
##' commits(repo, ref = "Tagname")
##'
##' ## List the commits modifying example.txt and mean.R.
##' commits(repo, path = "example.txt")
##' commits(repo, path = "mean.R")
##'
##' ## Create and checkout 'dev' branch in the repo
##' checkout(repo, "dev", create = TRUE)
##'
##' ## Add changes to the 'dev' branch
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Commit message in dev branch")
##'
##' ## Checkout the 'master' branch again and list the commits
##' ## starting from the 'dev' branch.
##' checkout(repo, "master")
##' commits(repo, ref = "dev")
##' }
commits <- function(repo        = ".",
                    topological = TRUE,
                    time        = TRUE,
                    reverse     = FALSE,
                    n           = NULL,
                    ref         = NULL,
                    path        = NULL) {
    ## Check limit in number of commits
    n <- get_upper_limit_of_commits(n)

    if (!is.null(path)) {
        if (!(is.character(path) && length(path) == 1)) {
            stop("path must be a single file")
        }
    }

    repo <- lookup_repository(repo)
    if (is_empty(repo))
        return(list())

    if (is.null(ref)) {
        sha <- sha(repository_head(repo))
    } else {
        sha <- sha(lookup_commit(.Call(git2r_reference_dwim, repo, ref)))
    }

    if (is_shallow(repo)) {
        ## FIXME: Remove this if-statement when libgit2 supports
        ## shallow clones, see #219.  Note: This workaround does not
        ## use the 'topological', 'time' and 'reverse' flags.
        return(shallow_commits(repo, sha, n))
    }

    if (!is.null(path)) {
        repo_wd <- normalizePath(workdir(repo), winslash = "/")
        path <- sanitize_path(path, repo_wd)
        path_revwalk <- .Call(git2r_revwalk_list2, repo, sha, topological,
                              time, reverse, n, path)
        return(path_revwalk[!vapply(path_revwalk, is.null, logical(1))])
    }

    .Call(git2r_revwalk_list, repo, sha, topological, time, reverse, n)
}

##' Last commit
##'
##' Get last commit in the current branch.
##' @template repo-param
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Get last commit
##' last_commit(repo)
##' last_commit(path)
##'
##' ## Coerce the last commit to a data.frame
##' as.data.frame(last_commit(path), "data.frame")
##'
##' ## Summary of last commit in repository
##' summary(last_commit(repo))
##' }
last_commit <- function(repo = ".") {
    commits(lookup_repository(repo), n = 1)[[1]]
}

##' Descendant
##'
##' Determine if a commit is the descendant of another commit
##' @param commit a git_commit object. Can also be a tag or a branch,
##'     and in that case the commit will be the target of the tag or
##'     branch.
##' @param ancestor a git_commit object to check if ancestor to
##'     \code{commit}. Can also be a tag or a branch, and in that case
##'     the commit will be the target of the tag or branch.
##' @return TRUE if \code{commit} is descendant of \code{ancestor},
##'     else FALSE
##' @export
##' @useDynLib git2r git2r_graph_descendant_of
##' @examples
##' \dontrun{
##' ## Create a directory in tempdir
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##'
##' ## Initialize a repository
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##' tag_1 <- tag(repo, "Tagname1", "Tag message 1")
##'
##' # Change file and commit
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit_2 <- commit(repo, "Commit message 2")
##' tag_2 <- tag(repo, "Tagname2", "Tag message 2")
##'
##' descendant_of(commit_1, commit_2)
##' descendant_of(commit_2, commit_1)
##' descendant_of(tag_1, tag_2)
##' descendant_of(tag_2, tag_1)
##' }
descendant_of <- function(commit = NULL, ancestor = NULL) {
    .Call(git2r_graph_descendant_of,
          lookup_commit(commit),
          lookup_commit(ancestor))
}

##' Check if object is a git_commit object
##'
##' @param object Check if object is a git_commit object
##' @return TRUE if object is a git_commit, else FALSE
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##'
##' ## Check if commit
##' is_commit(commit_1)
##' }
is_commit <- function(object) {
    inherits(object, "git_commit")
}

##' Is merge
##'
##' Determine if a commit is a merge commit, i.e. has more than one
##' parent.
##' @param commit a git_commit object.
##' @return TRUE if commit has more than one parent, else FALSE
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' writeLines(c("First line in file 1.", "Second line in file 1."),
##'            file.path(path, "example-1.txt"))
##' add(repo, "example-1.txt")
##' commit(repo, "First commit message")
##'
##' ## Create and add one more file
##' writeLines(c("First line in file 2.", "Second line in file 2."),
##'            file.path(path, "example-2.txt"))
##' add(repo, "example-2.txt")
##' commit(repo, "Second commit message")
##'
##' ## Create a new branch 'fix'
##' checkout(repo, "fix", create = TRUE)
##'
##' ## Update 'example-1.txt' (swap words in first line) and commit
##' writeLines(c("line First in file 1.", "Second line in file 1."),
##'            file.path(path, "example-1.txt"))
##' add(repo, "example-1.txt")
##' commit(repo, "Third commit message")
##'
##' checkout(repo, "master")
##'
##' ## Update 'example-2.txt' (swap words in second line) and commit
##' writeLines(c("First line in file 2.", "line Second in file 2."),
##'            file.path(path, "example-2.txt"))
##' add(repo, "example-2.txt")
##' commit(repo, "Fourth commit message")
##'
##' ## Merge 'fix'
##' merge(repo, "fix")
##'
##' ## Display parents of last commit
##' parents(lookup(repo, branch_target(repository_head(repo))))
##'
##' ## Check that last commit is a merge
##' is_merge(lookup(repo, branch_target(repository_head(repo))))
##' }
is_merge <- function(commit = NULL) {
    length(parents(commit)) > 1
}

##' Parents
##'
##' Get parents of a commit.
##' @param object a git_commit object.
##' @return list of git_commit objects
##' @export
##' @useDynLib git2r git2r_commit_parent_list
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' writeLines("First line.",
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##'
##' ## commit_1 has no parents
##' parents(commit_1)
##'
##' ## Update 'example.txt' and commit
##' writeLines(c("First line.", "Second line."),
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_2 <- commit(repo, "Second commit message")
##'
##' ## commit_2 has commit_1 as parent
##' parents(commit_2)
##' }
parents <- function(object = NULL) {
    .Call(git2r_commit_parent_list, object)
}

##' @export
format.git_commit <- function(x, ...) {
    sprintf("[%s] %s: %s",
            substring(x$sha, 1, 7),
            substring(as.character(x$author$when), 1, 10),
            x$summary)
}

##' @export
print.git_commit <- function(x, ...) {
    cat(format(x, ...), "\n", sep = "")
    invisible(x)
}

##' @export
summary.git_commit <- function(object, ...) {
    is_merge_commit <- is_merge(object)
    po <- parents(object)

    cat(sprintf("Commit:  %s\n", object$sha))

    if (is_merge_commit) {
        sha <- vapply(po, "[[", character(1), "sha")
        cat(sprintf("Merge:   %s\n", sha[1]))
        cat(paste0("         ", sha[-1]), sep = "\n")
    }

    cat(sprintf(paste0("Author:  %s <%s>\n",
                       "When:    %s\n\n"),
                object$author$name,
                object$author$email,
                as.character(object$author$when)))

    msg <- paste0("    ", readLines(textConnection(object$message)))
    cat("", sprintf("%s\n", msg))

    if (is_merge_commit) {
        cat("\n")
        lapply(po, function(parent) {
            cat("Commit message: ", parent$sha, "\n")
            msg <- paste0("    ",
                          readLines(textConnection(parent$message)))
            cat("", sprintf("%s\n", msg), "\n")
        })
    }

    if (identical(length(po), 1L)) {
        df <- diff(tree(po[[1]]), tree(object))
        if (length(df) > 0) {
            if (length(df) > 1) {
                cat(sprintf("%i files changed, ", length(df)))
            } else {
                cat("1 file changed, ")
            }

            cat(sprintf(
                "%i insertions, %i deletions\n",
                sum(vapply(lines_per_file(df), "[[", numeric(1), "add")),
                sum(vapply(lines_per_file(df), "[[", numeric(1), "del"))))

            plpf <- print_lines_per_file(df)
            hpf <- hunks_per_file(df)
            hunk_txt <- ifelse(hpf > 1, " hunks",
                        ifelse(hpf > 0, " hunk",
                               " hunk (binary file)"))
            phpf <- paste0("  in ", format(hpf), hunk_txt)
            cat(paste0(plpf, phpf), sep = "\n")
        }

        cat("\n")
    }

    invisible(NULL)
}

##' @export
as.data.frame.git_commit <- function(x, ...) {
    data.frame(sha              = x$sha,
               summary          = x$summary,
               message          = x$message,
               author           = x$author$name,
               email            = x$author$email,
               when             = as.POSIXct(x$author$when),
               stringsAsFactors = FALSE)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Class \code{"git_repository"}
##'
##' @title  S4 class to handle a git repository
##' @section Slots:
##' \describe{
##'   \item{path}{
##'     Path to a git repository
##'   }
##' }
##' @rdname git_repository-class
##' @keywords classes
##' @section Methods:
##' \describe{
##'   \item{is_bare}{\code{signature(object = "git_repository")}}
##'   \item{is_empty}{\code{signature(object = "git_repository")}}
##' }
##' @keywords methods
##' @export
setClass("git_repository",
         slots = c(path = "character"),
         validity = function(object) {
             errors <- character()

             can_open <- .Call(git2r_repository_can_open, object@path)
             if (!isTRUE(can_open))
                 errors <- c(errors, "Unable to open repository at 'path'")

             if (length(errors) == 0) TRUE else errors
         }
)

##' Class \code{"git_time"}
##'
##' @title S4 class to handle a git time in a signature
##' @section Slots:
##' \describe{
##'   \item{time}{
##'     Time in seconds from epoch
##'   }
##'   \item{offset}{
##'     Timezone offset, in minutes
##'   }
##' }
##' @name git_time-class
##' @aliases coerce,git_time,character-method
##' @aliases coerce,git_time,POSIXct-method
##' @aliases show,git_time-method
##' @keywords classes
##' @section Methods:
##' \describe{
##'   \item{show}{\code{signature(object = "git_time")}}
##' }
##' @keywords methods
##' @export
setClass("git_time",
         slots = c(time   = "numeric",
                   offset = "numeric"),
         validity = function(object)
         {
             errors <- character()

             if (!identical(length(object@time), 1L))
                 errors <- c(errors, "time must have length equal to one")
             if (!identical(length(object@offset), 1L))
                 errors <- c(errors, "offset must have length equal to one")

             if (length(errors) == 0) TRUE else errors
         }
)

##' Class \code{"git_signature"}
##'
##' @title S4 class to handle a git signature
##' @section Slots:
##' \describe{
##'   \item{name}{
##'     The full name of the author.
##'   }
##'   \item{email}{
##'     Email of the author.
##'   }
##'   \item{when}{
##'     Time when the action happened.
##'   }
##' }
##' @name git_signature-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("git_signature",
         slots = c(name  = "character",
                   email = "character",
                   when  = "git_time"),
         validity = function(object)
         {
             errors <- validObject(object@when)

             if (isTRUE(errors))
               errors <- character()

             if (!identical(length(object@name), 1L))
                 errors <- c(errors, "name must have length equal to one")
             if (!identical(length(object@email), 1L))
                 errors <- c(errors, "email must have length equal to one")

             if (length(errors) == 0) TRUE else errors
         }
)

##' Class \code{"cred_env"}
##'
##' @title S4 class to handle environmental variables credential
##' object
##' @section Slots:
##' \describe{
##'   \item{username}{
##'     The name of the environmental variable that holds
##'     the username for the authentication.
##'   }
##'   \item{password}{
##'     The name of the environmental variable that holds
##'     the password for the authentication.
##'   }
##' }
##' @rdname cred_env-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("cred_env",
         slots = c(username = "character",
                   password = "character")
)

##' Class \code{"cred_token"}
##'
##' @title S4 class to handle a personal access token credential
##' object
##' @section Slots:
##' \describe{
##'   \item{token}{
##'     The name of the environmental variable that holds
##'     the personal access token for the authentication.
##'   }
##' }
##' @rdname cred_token-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("cred_token",
         slots = c(token = "character")
)

##' Class \code{"cred_user_pass"}
##'
##' @title S4 class to handle plain-text username and password
##' credential object
##' @section Slots:
##' \describe{
##'   \item{username}{
##'     The username of the credential
##'   }
##'   \item{password}{
##'     The password of the credential
##'   }
##' }
##' @rdname cred_user_pass-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("cred_user_pass",
         slots = c(username = "character",
                   password = "character")
)

##' Class \code{"cred_ssh_key"}
##'
##' @title S4 class to handle a passphrase-protected ssh key
##' credential object
##' @section Slots:
##' \describe{
##'   \item{publickey}{
##'     The path to the public key of the credential
##'   }
##'   \item{privatekey}{
##'     The path to the private key of the credential
##'   }
##'   \item{passphrase}{
##'     The passphrase of the credential
##'   }
##' }
##' @rdname cred_ssh_key-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("cred_ssh_key",
         slots = c(publickey  = "character",
                   privatekey = "character",
                   passphrase = "character")
)

##' Class \code{"git_blame"}
##'
##' @title  S4 class to handle a git blame for a single file
##' @section Slots:
##' \describe{
##'   \item{path}{
##'     The path to the file of the blame
##'   }
##'   \item{hunks}{
##'     List of blame hunks
##'   }
##'   \item{repo}{
##'     The S4 class git_repository that contains the file
##'   }
##' }
##' @rdname git_blame-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("git_blame",
         slots = c(path  = "character",
                   hunks = "list",
                   repo  = "git_repository"),
         validity = function(object) {
             errors <- character()

             if (length(errors) == 0) TRUE else errors
         }
)

##' Class \code{"git_blame_hunk"}
##'
##' @title  S4 class to represent a blame hunk
##' @section Slots:
##' \describe{
##'   \item{lines_in_hunk}{
##'     The number of lines in this hunk
##'   }
##'   \item{final_commit_id}{
##'     The sha of the commit where this line was last changed
##'   }
##'   \item{final_start_line_number}{
##'     The 1-based line number where this hunk begins, in the final
##'     version of the file
##'   }
##'   \item{final_signature}{
##'     Final committer
##'   }
##'   \item{orig_commit_id}{
##'     The sha of the commit where this hunk was found. This will usually
##'     be the same as 'final_commit_id'.
##'   }
##'   \item{orig_start_line_number}{
##'      The 1-based line number where this hunk begins in the file
##'      named by 'orig_path' in the commit specified by 'orig_commit_id'.
##'   }
##'   \item{orig_signature}{
##'     Origin committer
##'   }
##'   \item{orig_path}{
##'     The path to the file where this hunk originated, as of the commit
##'     specified by 'orig_commit_id'
##'   }
##'   \item{boundary}{
##'     TRUE iff the hunk has been tracked to a boundary commit.
##'   }
##'   \item{repo}{
##'     The S4 class git_repository that contains the blame hunk
##'   }
##' }
##' @rdname git_blame_hunk-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("git_blame_hunk",
         slots = c(lines_in_hunk           = "integer",
                   final_commit_id         = "character",
                   final_start_line_number = "integer",
                   final_signature         = "git_signature",
                   orig_commit_id          = "character",
                   orig_start_line_number  = "integer",
                   orig_signature          = "git_signature",
                   orig_path               = "character",
                   boundary                = "logical",
                   repo                    = "git_repository"),
         validity = function(object) {
             errors <- character()

             if (length(errors) == 0) TRUE else errors
         }
)

##' Class \code{"git_blob"}
##'
##' @title  S4 class to handle a git blob
##' @section Slots:
##' \describe{
##'   \item{sha}{
##'     The 40 character hexadecimal string of the SHA-1
##'   }
##'   \item{repo}{
##'     The S4 class git_repository that contains the blob
##'   }
##' }
##' @rdname git_blob-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("git_blob",
         slots = c(sha  = "character",
                   repo = "git_repository"),
         validity = function(object) {
             errors <- character()

             if (length(errors) == 0) TRUE else errors
         }
)

##' Class \code{git_branch}
##'
##' @title S4 class to handle a git branch
##' @section Slots:
##' \describe{
##'   \item{name}{
##'     Name of the branch.
##'   }
##'   \item{type}{
##'     Type of the branch, either 1 (local) or 2 (remote).
##'   }
##'   \item{repo}{
##'     The S4 class git_repository that contains the branch
##'   }
##' }
##' @name git_branch-class
##' @keywords classes
##' @section Methods:
##' \describe{
##'   \item{is_head}{\code{signature(object = "git_branch")}}
##'   \item{is_local}{\code{signature(object = "git_branch")}}
##'   \item{show}{\code{signature(object = "git_branch")}}
##' }
##' @keywords methods
##' @export
setClass("git_branch",
         slots = c(name = "character",
                   type = "integer",
                   repo = "git_repository"))

##' Class \code{"git_commit"}
##'
##' @title S4 class to handle a git commit.
##' @section Slots:
##' \describe{
##'   \item{sha}{
##'     The 40 character hexadecimal string of the SHA-1
##'   }
##'   \item{author}{
##'     An author signature
##'   }
##'   \item{committer}{
##'     The committer signature
##'   }
##'   \item{summary}{
##'     The short "summary" of a git commit message, comprising the first
##'     paragraph of the message with whitespace trimmed and squashed.
##'   }
##'   \item{message}{
##'     The message of a commit
##'   }
##'   \item{repo}{
##'     The S4 class git_repository that contains the commit
##'   }
##' }
##' @name git_commit-class
##' @keywords classes
##' @section Methods:
##' \describe{
##'   \item{show}{\code{signature(object = "git_commit")}}
##' }
##' @keywords methods
##' @export
setClass("git_commit",
         slots = c(sha       = "character",
                   author    = "git_signature",
                   committer = "git_signature",
                   summary   = "character",
                   message   = "character",
                   repo      = "git_repository"),
         prototype = list(summary = NA_character_,
                          message = NA_character_))

##' Git diff
##'
##' @section Slots:
##' \describe{
##'   \item{old}{
##'     Represents the 'from' side of the diff.
##'   }
##'   \item{new}{
##'     Represents the 'to' side of the diff.
##'   }
##'   \item{files}{
##'     List of \code{\linkS4class{git_diff_file}} objects.
##'   }
##' }
##' @name git_diff-class
##' @keywords classes
##' @export
setClass("git_diff",
         slots = c(old   = "ANY",
                   new   = "ANY",
                   files = "list"),
         prototype = list(old = NA_character_,
                          new = NA_character_))

##' Git diff file
##'
##' @section Slots:
##' \describe{
##'   \item{old_file}{
##'     Filename for old side of diff.
##'   }
##'   \item{new_file}{
##'     Filename for new side of diff.
##'   }
##'   \item{hunks}{
##'     List of \code{\linkS4class{git_diff_hunk}} objects.
##'   }
##' }
##' @name git_diff_file-class
##' @keywords classes
##' @export
setClass("git_diff_file",
         slots = c(old_file = "character",
                   new_file = "character",
                   hunks    = "list"))

##' Git diff hunk
##'
##' @section Slots:
##' \describe{
##'   \item{old_start}{
##'     Starting line number in old_file.
##'   }
##'   \item{old_lines}{
##'     Number of lines in old_file.
##'   }
##'   \item{new_start}{
##'     Starting line number in new_file.
##'   }
##'   \item{new_lines}{
##'     Number of lines in new_file.
##'   }
##'   \item{header}{
##'     Header text.
##'   }
##'   \item{lines}{
##'     List with \code{\linkS4class{git_diff_line}} objects.
##'   }
##' }
##' @name git_diff_hunk-class
##' @keywords classes
##' @export
setClass("git_diff_hunk",
         slots = c(old_start = "integer",
                   old_lines = "integer",
                   new_start = "integer",
                   new_lines = "integer",
                   header    = "character",
                   lines     = "list"))

##' Git diff line
##'
##' @section Slots:
##' \describe{
##'   \item{origin}{
##'     Line origin constants.
##'   }
##'   \item{old_lineno}{
##'     Line number in old file or -1 for added line.
##'   }
##'   \item{new_lineno}{
##'     Line number in new file or -1 for deleted line.
##'   }
##'   \item{num_lines}{
##'     Number of newline characters in content.
##'   }
##'   \item{content}{
##'     The diff text.
##'   }
##' }
##' @name git_diff_line-class
##' @keywords classes
##' @export
setClass("git_diff_line",
         slots = c(origin     = "integer",
                   old_lineno = "integer",
                   new_lineno = "integer",
                   num_lines  = "integer",
                   content    = "character"))

##' Class \code{git_note}
##'
##' @title S4 class to handle a git note
##' @section Slots:
##' \describe{
##'   \item{sha}{
##'     The 40 character hexadecimal string of the SHA-1 of the blob
##'     containing the message
##'   }
##'   \item{annotated}{
##'     The 40 character hexadecimal string of the SHA-1 of the git
##'     object being annotated
##'   }
##'   \item{message}{
##'     The note message
##'   }
##'   \item{refname}{
##'     Name of the reference
##'   }
##'   \item{repo}{
##'     The S4 class git_repository that contains the note
##'   }
##' }
##' @name git_note-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("git_note",
         slots = c(sha       = "character",
                   annotated = "character",
                   message   = "character",
                   refname   = "character",
                   repo      = "git_repository"))

##' Class \code{"git_reference"}
##'
##' @title S4 class to handle a git reference
##' @section Slots:
##' \describe{
##'   \item{name}{
##'     The full name of the reference.
##'   }
##'   \item{type}{
##'     Type of the reference, either direct (GIT_REF_OID == 1) or
##'     symbolic (GIT_REF_SYMBOLIC == 2)
##'   }
##'   \item{sha}{
##'     The 40 character hexadecimal string of the SHA-1
##'   }
##'   \item{target}{
##'     The full name to the reference pointed to by a
##'     symbolic reference.
##'   }
##'   \item{shorthand}{
##'     The reference's short name
##'   }
##' }
##' @rdname git_reference-class
##' @keywords classes
##' @section Methods:
##' \describe{
##'   \item{show}{\code{signature(object = "git_reference")}}
##' }
##' @keywords methods
##' @export
setClass("git_reference",
         slots = c(name      = "character",
                   type      = "integer",
                   sha       = "character",
                   target    = "character",
                   shorthand = "character"),
         prototype = list(sha    = NA_character_,
                          target = NA_character_))

##' Class \code{"git_reflog_entry"}
##'
##' @title S4 class to handle a git reflog entry.
##' @section Slots:
##' \describe{
##'   \item{sha}{
##'     The 40 character hexadecimal string of the SHA-1
##'   }
##'   \item{message}{
##'     The log message of the entry
##'   }
##'   \item{index}{
##'     The index (zero-based) of the entry in the reflog
##'   }
##'   \item{committer}{
##'     The committer signature
##'   }
##'   \item{refname}{
##'     Name of the reference
##'   }
##'   \item{repo}{
##'     The S4 class git_repository that contains the reflog entry
##'   }
##' }
##' @name git_reflog_entry-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("git_reflog_entry",
         slots = c(sha       = "character",
                   message   = "character",
                   index     = "integer",
                   committer = "git_signature",
                   refname   = "character",
                   repo      = "git_repository"))

##' Class \code{"git_stash"}
##'
##' @title S4 class to handle a git stash
##' @name git_stash-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("git_stash", contains = "git_commit")

##' Class \code{"git_tag"}
##'
##' @title S4 class to handle a git tag
##' @section Slots:
##' \describe{
##'   \item{sha}{
##'     The 40 character hexadecimal string of the SHA-1
##'   }
##'   \item{message}{
##'     The message of the tag
##'   }
##'   \item{name}{
##'     The name of the tag
##'   }
##'   \item{tagger}{
##'     The tagger (author) of the tag
##'   }
##'   \item{target}{
##'     The target of the tag
##'   }
##'   \item{repo}{
##'     The S4 class git_repository that contains the tag
##'   }
##' }
##' @name git_tag-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("git_tag",
         slots = c(sha     = "character",
                   message = "character",
                   name    = "character",
                   tagger  = "git_signature",
                   target  = "character",
                   repo    = "git_repository"),
         validity = function(object)
         {
             errors <- validObject(object@tagger)

             if (isTRUE(errors))
               errors <- character()

             if (!identical(length(object@sha), 1L))
                 errors <- c(errors, "sha must have length equal to one")
             if (!identical(length(object@message), 1L))
                 errors <- c(errors, "message must have length equal to one")
             if (!identical(length(object@name), 1L))
                 errors <- c(errors, "name must have length equal to one")
             if (!identical(length(object@target), 1L))
                 errors <- c(errors, "target must have length equal to one")

             if (length(errors) == 0) TRUE else errors
         }
)

##' Class \code{"git_tree"}
##'
##' @title S4 class to handle a git tree
##' @section Slots:
##' \describe{
##'   \item{sha}{
##'     The 40 character hexadecimal string of the SHA-1
##'   }
##'   \item{filemode}{
##'     The UNIX file attributes of a tree entry
##'   }
##'   \item{type}{
##'     String representation of the tree entry type
##'   }
##'   \item{id}{
##'     The sha id of a tree entry
##'   }
##'   \item{name}{
##'     The filename of a tree entry
##'   }
##'   \item{repo}{
##'     The S4 class git_repository that contains the commit
##'   }
##' }
##' @name git_tree-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("git_tree",
         slots = c(sha      = "character",
                   filemode = "integer",
                   type     = "character",
                   id       = "character",
                   name     = "character",
                   repo     = "git_repository"),
         validity = function(object)
         {
             errors <- character(0)

             if (!identical(length(object@sha), 1L))
                 errors <- c(errors, "sha must have length equal to one")

             if (length(errors) == 0) TRUE else errors
         }
)

##' Class \code{"git_transfer_progress"}
##'
##' Statistics from the fetch operation.
##' @section Slots:
##' \describe{
##'   \item{total_objects}{
##'     Number of objects in the packfile being downloaded
##'   }
##'   \item{indexed_objects}{
##'     Received objects that have been hashed
##'   }
##'   \item{received_objects}{
##'     Objects which have been downloaded
##'   }
##'   \item{total_deltas}{
##'     Total number of deltas in the pack
##'   }
##'   \item{indexed_deltas}{
##'     Deltas which have been indexed
##'   }
##'   \item{local_objects}{
##'     Locally-available objects that have been injected in order to
##'     fix a thin pack
##'   }
##'   \item{received_bytes}{
##'     Size of the packfile received up to now
##'   }
##' }
##' @name git_transfer_progress-class
##' @keywords classes
##' @keywords methods
##' @export
setClass("git_transfer_progress",
         slots = c(total_objects    = "integer",
                   indexed_objects  = "integer",
                   received_objects = "integer",
                   local_objects    = "integer",
                   total_deltas     = "integer",
                   indexed_deltas   = "integer",
                   received_bytes   = "integer"))

##' Class \code{"git_fetch_head"}
##'
##' @title S4 class to handle a fetch head
##' @section Slots:
##' \describe{
##'   \item{ref_name}{
##'     The name of the ref.
##'   }
##'   \item{remote_url}{
##'     The url of the remote.
##'   }
##'   \item{sha}{
##'     The SHA-1 of the remote head that were updated during the last
##'     fetch.
##'   }
##'   \item{is_merge}{
##'     Is head for merge.
##'   }
##'   \item{repo}{
##'     The S4 class git_repository that contains the fetch head.
##'   }
##' }
##' @rdname git_fetch_head
##' @keywords classes
##' @keywords methods
##' @export
setClass("git_fetch_head",
         slots = c(ref_name   = "character",
                   remote_url = "character",
                   sha        = "character",
                   is_merge   = "logical",
                   repo       = "git_repository")
)

##' Class \code{"git_merge_result"}
##'
##' @title S4 class to handle the merge result
##' @section Slots:
##' \describe{
##'   \item{up_to_date}{
##'     TRUE if the merge is already up-to-date, else FALSE.
##'   }
##'   \item{fast_forward}{
##'     TRUE if a fast-forward merge, else FALSE.
##'   }
##'   \item{conflicts}{
##'     TRUE if the index contain entries representing file conflicts,
##'     else FALSE.
##'   }
##'   \item{sha}{
##'     If the merge created a merge commit, the sha of the merge
##'     commit. NA if no merge commit created.
##'   }
##' }
##' @rdname git_merge_result
##' @keywords classes
##' @keywords methods
##' @export
setClass("git_merge_result",
         slots = c(up_to_date   = "logical",
                   fast_forward = "logical",
                   conflicts    = "logical",
                   sha          = "character")
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2024 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Revparse
##'
##' Find object specified by revision.
##' @template repo-param
##' @param revision The revision string, see
##' http://git-scm.com/docs/git-rev-parse.html#_specifying_revisions
##' @return a \code{git_commit} or \code{git_tag} or \code{git_tree}
##' object
##' @export
##' @useDynLib git2r git2r_revparse_single
##' @examples
##' \dontrun{
##' ## Create a directory in tempdir
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##'
##' ## Initialize a repository
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit(repo, "First commit message")
##'
##' # Change file and commit
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit(repo, "Second commit message")
##'
##' revparse_single(repo, "HEAD^")
##' revparse_single(repo, "HEAD:test.txt")
##' }
revparse_single <- function(repo = ".", revision = NULL) {
    .Call(git2r_revparse_single, lookup_repository(repo), revision)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Add file(s) to index
##'
##' @template repo-param
##' @param path Character vector with file names or shell glob
##'     patterns that will matched against files in the repository's
##'     working directory. Each file that matches will be added to the
##'     index (either updating an existing entry or adding a new
##'     entry).
##' @param force Add ignored files. Default is FALSE.
##' @return invisible(NULL)
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file
##' writeLines("a", file.path(path, "a.txt"))
##'
##' ## Add file to repository and view status
##' add(repo, "a.txt")
##' status(repo)
##'
##' ## Add file with a leading './' when the repository working
##' ## directory is the current working directory
##' setwd(path)
##' writeLines("b", file.path(path, "b.txt"))
##' add(repo, "./b.txt")
##' status(repo)
##'
##' ## Add a file in a sub-folder with sub-folder as the working
##' ## directory. Create a file in the root of the repository
##' ## working directory that will remain untracked.
##' dir.create(file.path(path, "sub_dir"))
##' setwd("./sub_dir")
##' writeLines("c", file.path(path, "c.txt"))
##' writeLines("c", file.path(path, "sub_dir/c.txt"))
##' add(repo, "c.txt")
##' status(repo)
##'
##' ## Add files with glob expansion when the current working
##' ## directory is outside the repository's working directory.
##' setwd(tempdir())
##' dir.create(file.path(path, "glob_dir"))
##' writeLines("d", file.path(path, "glob_dir/d.txt"))
##' writeLines("e", file.path(path, "glob_dir/e.txt"))
##' writeLines("f", file.path(path, "glob_dir/f.txt"))
##' writeLines("g", file.path(path, "glob_dir/g.md"))
##' add(repo, "glob_dir/*txt")
##' status(repo)
##'
##' ## Add file with glob expansion with a relative path when
##' ## the current working directory is inside the repository's
##' ## working directory.
##' setwd(path)
##' add(repo, "./glob_dir/*md")
##' status(repo)
##' }
add <- function(repo = ".", path = NULL, force = FALSE) {
    ## Documentation for the pathspec argument in the libgit2 function
    ## 'git_index_add_all' that git2r use internally:
    ##
    ## The pathspec is a list of file names or shell glob patterns
    ## that will matched against files in the repository's working
    ## directory. Each file that matches will be added to the index
    ## (either updating an existing entry or adding a new entry).

    if (is.null(path) || !is.character(path))
        stop("'path' must be a character vector")

    repo <- lookup_repository(repo)
    repo_wd <- normalizePath(workdir(repo), winslash = "/")
    path <- vapply(path, sanitize_path, character(1), repo_wd = repo_wd)

    .Call(git2r_index_add_all, repo, path, isTRUE(force))

    invisible(NULL)
}

sanitize_path <- function(p, repo_wd) {
    np <- suppressWarnings(normalizePath(p, winslash = "/"))

    if (!length(grep("/$", repo_wd)))
        repo_wd <- paste0(repo_wd, "/")

    ## Check if the normalized path is a non-file e.g. a glob.
    if (!file.exists(np)) {
        ## Check if the normalized path starts with a leading './'
        if (length(grep("^[.]/", np))) {
            nd <- suppressWarnings(normalizePath(dirname(p), winslash = "/"))
            if (!length(grep("/$", nd)))
                nd <- paste0(nd, "/")
            np <- paste0(nd, basename(np))
        }
    }

    ## Check if the file is in the repository's working directory,
    ## else let libgit2 handle this path unmodified.
    if (!length(grep(paste0("^", repo_wd), np)))
        return(p)

    ## Change the path to be relative to the repository's working
    ## directory. Substitute common prefix with ""
    sub(paste0("^", repo_wd), "", np)
}

##' Remove files from the working tree and from the index
##'
##' @template repo-param
##' @param path character vector with filenames to remove. Only files
##'     known to Git are removed.
##' @return invisible(NULL)
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file
##' writeLines("Hello world!", file.path(path, "file-to-remove.txt"))
##'
##' ## Add file to repository
##' add(repo, "file-to-remove.txt")
##' commit(repo, "First commit message")
##'
##' ## Remove file
##' rm_file(repo, "file-to-remove.txt")
##'
##' ## View status of repository
##' status(repo)
##' }
rm_file <- function(repo = ".", path = NULL) {
    if (is.null(path) || !is.character(path))
        stop("'path' must be a character vector")

    repo <- lookup_repository(repo)

    if (length(path)) {
        repo_wd <- workdir(repo)
        repo_wd <- normalizePath(workdir(repo), winslash = "/")
        path <- vapply(path, sanitize_path, character(1), repo_wd = repo_wd)

        ## Check that files exists and are known to Git
        if (!all(file.exists(file.path(repo_wd, path)))) {
            stop(sprintf("pathspec '%s' did not match any files. ",
                         path[!file.exists(file.path(repo_wd, path))]))
        }

        if (any(file.info(file.path(repo_wd, path))$isdir)) {
            stop(sprintf("pathspec '%s' did not match any files. ",
                         path[exists(file.path(repo_wd, path))]))
        }

        s <- status(repo, staged = TRUE, unstaged = TRUE,
                    untracked = TRUE, ignored = TRUE)
        if (any(path %in% c(s$ignored, s$untracked))) {
            stop(sprintf("pathspec '%s' did not match any files. ",
                         path[path %in% c(s$ignored, s$untracked)]))
        }

        if (any(path %in% s$staged)) {
            stop(sprintf("'%s' has changes staged in the index. ",
                         path[path %in% s$staged]))
        }

        if (any(path %in% s$unstaged)) {
            stop(sprintf("'%s' has local modifications. ",
                         path[path %in% s$unstaged]))
        }

        ## Remove and stage files
        lapply(path, function(x) {
            file.remove(file.path(repo_wd, x))
            .Call(git2r_index_remove_bypath, repo, x)
        })
    }

    invisible(NULL)
}

##' Remove an index entry corresponding to a file on disk
##'
##' @template repo-param
##' @param path character vector with filenames to remove. The path
##'     must be relative to the repository's working folder. It may
##'     exist. If this file currently is the result of a merge
##'     conflict, this file will no longer be marked as
##'     conflicting. The data about the conflict will be moved to the
##'     "resolve undo" (REUC) section.
##' @return invisible(NULL)
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file
##' writeLines("Hello world!", file.path(path, "file-to-remove.txt"))
##'
##' ## Add file to repository
##' add(repo, "file-to-remove.txt")
##'
##' ## View status of repository
##' status(repo)
##'
##' ## Remove file
##' index_remove_bypath(repo, "file-to-remove.txt")
##'
##' ## View status of repository
##' status(repo)
##' }
index_remove_bypath <- function(repo = ".", path = NULL) {
    .Call(git2r_index_remove_bypath, lookup_repository(repo), path)
    invisible(NULL)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Reset current HEAD to the specified state
##'
##' @param object Either a \code{git_commit}, a \code{git_repository}
##'     or a character vector. If \code{object} is a
##'     \code{git_commit}, HEAD is moved to the \code{git_commit}. If
##'     \code{object} is a \code{git_repository}, resets the index
##'     entries in the \code{path} argument to their state at HEAD. If
##'     \code{object} is a character vector with paths, resets the
##'     index entries in \code{object} to their state at HEAD if the
##'     current working directory is in a repository.
##' @param reset_type If object is a 'git_commit', the kind of reset
##'     operation to perform. 'soft' means the HEAD will be moved to
##'     the commit. 'mixed' reset will trigger a 'soft' reset, plus
##'     the index will be replaced with the content of the commit
##'     tree. 'hard' reset will trigger a 'mixed' reset and the
##'     working directory will be replaced with the content of the
##'     index.
##' @param path If object is a 'git_repository', resets the index
##'     entries for all paths to their state at HEAD.
##' @return invisible NULL
##' @export
##' @examples \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test-1.txt"))
##' add(repo, "test-1.txt")
##' commit_1 <- commit(repo, "Commit message")
##'
##' ## Change and stage the file
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-1.txt"))
##' add(repo, "test-1.txt")
##' status(repo)
##'
##' ## Unstage file
##' reset(repo, path = "test-1.txt")
##' status(repo)
##'
##' ## Make one more commit
##' add(repo, "test-1.txt")
##' commit(repo, "Next commit message")
##'
##' ## Create one more file
##' writeLines("Hello world!", file.path(path, "test-2.txt"))
##'
##' ## 'soft' reset to first commit and check status
##' reset(commit_1)
##' status(repo)
##'
##' ## 'mixed' reset to first commit and check status
##' commit(repo, "Next commit message")
##' reset(commit_1, "mixed")
##' status(repo)
##'
##' ## 'hard' reset to first commit and check status
##' add(repo, "test-1.txt")
##' commit(repo, "Next commit message")
##' reset(commit_1, "hard")
##' status(repo)
##' }
reset <- function(object,
                  reset_type = c("soft", "mixed", "hard"),
                  path = NULL) {
    if (is_commit(object)) {
        reset_type <- switch(match.arg(reset_type),
                             soft  = 1L,
                             mixed = 2L,
                             hard  = 3L)

        .Call(git2r_reset, object, reset_type)
    } else {
        object <- lookup_repository(object)
        if (is_empty(object)) {
            .Call(git2r_index_remove_bypath, object, path)
        } else {
            .Call(git2r_reset_default, object, path)
        }
    }

    invisible(NULL)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2023 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()
libgit2_version()
libgit2_features()


## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test.txt"))

## Add and commit
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message")

## Add a remote
remote_add(repo, "playground",
           "https://github.com/gaborcsardi/playground")

stopifnot(identical(remotes(repo), "playground"))
stopifnot(identical(remote_url(repo, "playground"),
                    "https://github.com/gaborcsardi/playground"))
stopifnot(identical(remote_url(repo),
                    "https://github.com/gaborcsardi/playground"))

## Rename a remote
remote_rename(repo, "playground", "foobar")

stopifnot(identical(remotes(repo), "foobar"))
stopifnot(identical(remote_url(repo, "foobar"),
                    "https://github.com/gaborcsardi/playground"))

## Set remote url
remote_set_url(repo, "foobar", "https://github.com/stewid/playground")
stopifnot(identical(remote_url(repo, "foobar"),
                    "https://github.com/stewid/playground"))

## Remove a remote
remote_remove(repo, "foobar")

stopifnot(identical(remotes(repo), character(0)))

if (identical(Sys.getenv("NOT_CRAN"), "true")) {
    if (isTRUE(libgit2_features()$https)) {
        ## Standard test for remote_ls
        refs <- remote_ls("https://github.com/ropensci/git2r")
        stopifnot(length(refs) > 0)
        stopifnot(names(refs) > 0)
        stopifnot(any(names(refs) == "HEAD"))

        ## -- PROXY TESTS --
        ## Proxy = TRUE => automatic detection. We expect no error.
        if (nzchar(Sys.getenv("http_proxy")) || nzchar(Sys.getenv("HTTP_PROXY")) ||
            nzchar(Sys.getenv("https_proxy")) || nzchar(Sys.getenv("HTTPS_PROXY"))) {
          ## If a proxy is actually set, we can test the code path that picks it up.
          refs_proxy_auto <- remote_ls("https://github.com/ropensci/git2r", proxy = TRUE)
          stopifnot(length(refs_proxy_auto) > 0)
          stopifnot(any(names(refs_proxy_auto) == "HEAD"))
        }

        ## Proxy = "string" => a manual proxy URL. This will fail with a fake URL.
        test_proxy_url <- Sys.getenv("TEST_PROXY_URL", "http://fake-proxy.example.com:8080")
        message("Testing remote_ls with proxy='", test_proxy_url, "'")
        tryCatch(
          {
            refs_proxy_str <- remote_ls("https://github.com/ropensci/git2r", proxy = test_proxy_url)
            stopifnot(length(refs_proxy_str) >= 0)  # minimal check
          },
          error = function(e) {
            message("As expected, connecting via a fake/invalid proxy might fail: ", e$message)
          }
        )
    }
}


# an invalid URL should throw an error
tools::assertError(remote_ls("bad"))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Push
##'
##' @rdname push-methods
##' @docType methods
##' @param object S4 class \code{git_repository} or \code{git_branch}.
##' @param ... Additional arguments affecting the push.
##' @param force Force your local revision to the remote repo. Use it
##' with care. Default is FALSE.
##' @param credentials The credentials for remote repository
##' access. Default is NULL. To use and query an ssh-agent for the ssh
##' key credentials, let this parameter be NULL (the default).
##' @return invisible(NULL)
##' @seealso \code{\linkS4class{cred_user_pass}},
##' \code{\linkS4class{cred_ssh_key}}
##' @keywords methods
##' @include refspec.r
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize two temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## Change file and commit
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
##'            file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Second commit message")
##'
##' ## Push commits from repository to bare repository
##' push(repo)
##'
##' ## List commits in repository and bare repository
##' commits(repo)
##' commits(repo_bare)
##' }
setGeneric("push",
           signature = "object",
           function(object, ...)
           standardGeneric("push"))

##' @rdname push-methods
##' @export
setMethod("push",
          signature(object = "git_branch"),
          function(object,
                   force       = FALSE,
                   credentials = NULL)
          {
              upstream <- branch_get_upstream(object)
              if (is.null(upstream)) {
                  stop("The branch '", object@name, "' that you are ",
                       "trying to push does not track an upstream branch.")
              }

              src <- .Call(git2r_branch_canonical_name, object)
              dst <- .Call(git2r_branch_upstream_canonical_name, object)

              push(object      = object@repo,
                   name        = branch_remote_name(upstream),
                   refspec     = paste0(src, ":", dst),
                   force       = force,
                   credentials = credentials)
          }
)

##' @rdname push-methods
##' @param name The remote's name. Default is NULL.
##' @param refspec The refspec to be pushed. Default is NULL.
##' @export
setMethod("push",
          signature(object = "git_repository"),
          function(object,
                   name        = NULL,
                   refspec     = NULL,
                   force       = FALSE,
                   credentials = NULL)
          {
              if (all(is.null(name), is.null(refspec))) {
                  b <- head(object)
                  upstream <- branch_get_upstream(b)
                  if (is.null(upstream)) {
                      stop("The branch '", b@name, "' that you are ",
                           "trying to push does not track an upstream branch.")
                  }

                  src <- .Call(git2r_branch_canonical_name, b)
                  dst <- .Call(git2r_branch_upstream_canonical_name, b)
                  name <- branch_remote_name(upstream)
                  refspec <- paste0(src, ":", dst)

                  if (isTRUE(force))
                      refspec <- paste0("+", refspec)
              } else {
                  opts <- list(force = force)
                  tmp <- get_refspec(object, name, refspec, opts)
                  name <- tmp$remote
                  refspec <- tmp$refspec
              }

              result <- .Call(git2r_push, object, name, refspec, credentials)

              invisible(result)
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2018 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Bundle bare repo of package
##'
##' Clone the package git repository as a bare repository to
##' \code{pkg/inst/pkg.git}
##' @template repo-param
##' @return Invisible bundled \code{git_repository} object
##' @export
##' @examples
##' \dontrun{
##' ## Initialize repository
##' path <- tempfile()
##' dir.create(path)
##' path <- file.path(path, "git2r")
##' repo <- clone("https://github.com/ropensci/git2r.git", path)
##'
##' ## Bundle bare repository in package
##' bundle_r_package(repo)
##'
##' ## Build and install bundled package
##' wd <- setwd(dirname(path))
##' system(sprintf("R CMD build %s", path))
##' pkg <- list.files(".", pattern = "[.]tar[.]gz$")
##' system(sprintf("R CMD INSTALL %s", pkg))
##' setwd(wd)
##'
##' ## Reload package
##' detach("package:git2r", unload = TRUE)
##' library(git2r)
##'
##' ## Summarize last five commits of bundled repo
##' repo <- repository(system.file("git2r.git", package = "git2r"))
##' invisible(lapply(commits(repo, n = 5), summary))
##'
##' ## Plot content of bundled repo
##' plot(repo)
##' }
bundle_r_package <- function(repo = ".") {
    repo <- lookup_repository(repo)

    ## Check for 'inst' folder
    inst <- file.path(workdir(repo), "inst")
    if (!isTRUE(file.info(inst)$isdir))
        dir.create(inst)

    ## Check for 'pkg.git' folder
    local_path <- paste0(basename(workdir(repo)), ".git", sep = "")
    local_path <- file.path(inst, local_path)
    if (file.exists(local_path))
        stop("Repo already exists:", local_path)
    invisible(clone(workdir(repo), local_path, bare = TRUE))
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2018 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create first commit
writeLines("Hello world!", file.path(path, "test-1.txt"))
add(repo, "test-1.txt")
commit(repo, "First commit message")

## Create and checkout dev branch in repo
checkout(repo, "dev", create = TRUE)

## Create second commit
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-2.txt"))
add(repo, "test-2.txt")
commit(repo, "Second commit message")

## Check files
stopifnot(identical(list.files(path), c("test-1.txt", "test-2.txt")))

## Checkout master branch and check files
checkout(repo, "master")
stopifnot(identical(list.files(path), "test-1.txt"))

## Cleanup
unlink(path, recursive = TRUE)

## Checkout branch in empty repository

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create and checkout dev branch in repo
checkout(repo, "dev", create = TRUE)

## Create first commit
writeLines("Hello world!", file.path(path, "test-1.txt"))
add(repo, "test-1.txt")
commit(repo, "First commit message")

stopifnot(identical(length(branches(repo)), 1L))
stopifnot(identical(branches(repo)[[1]]$name, "dev"))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2018 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create 2 directories in tempdir
path_bare <- tempfile(pattern = "git2r-")
path_repo <- tempfile(pattern = "git2r-")
dir.create(path_bare)
dir.create(path_repo)

## Initialize a repository
repo <- init(path_repo)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Add commit to repo
writeLines("Hello world", con = file.path(path_repo, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message")

## Check bare argument
tools::assertError(clone(path_repo, path_bare, bare = c(TRUE, TRUE)))
tools::assertError(clone(path_repo, path_bare, bare = 1))
tools::assertError(clone(path_repo, path_bare, bare = 1L))
tools::assertError(clone(path_repo, path_bare, bare = "test"))

## Clone repo to bare repository
bare_repo <- clone(path_repo, path_bare, bare = TRUE)

## Check the repositores
stopifnot(identical(is_bare(bare_repo), TRUE))
stopifnot(identical(is_bare(repo), FALSE))

## Check result in bare repository
stopifnot(identical(length(commits(bare_repo)), 1L))
bare_commit_1 <- commits(bare_repo)[[1]]
stopifnot(identical(sha(commit_1), sha(bare_commit_1)))
stopifnot(identical(commit_1$author, bare_commit_1$author))
stopifnot(identical(commit_1$committer, bare_commit_1$committer))
stopifnot(identical(commit_1$summary, bare_commit_1$summary))
stopifnot(identical(commit_1$message, bare_commit_1$message))
stopifnot(!identical(commit_1$repo, bare_commit_1$repo))

## Cleanup
unlink(path_bare, recursive = TRUE)
unlink(path_repo, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2016 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Add file(s) to index
##'
##' @rdname add-methods
##' @docType methods
##' @param repo The repository \code{object}.
##' @param path Character vector with file names or shell glob
##' patterns that will matched against files in the repository's
##' working directory. Each file that matches will be added to the
##' index (either updating an existing entry or adding a new entry).
##' @param ... Additional arguments to the method
##' @param force Add ignored files. Default is FALSE.
##' @return invisible(NULL)
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file
##' writeLines("a", file.path(path, "a.txt"))
##'
##' ## Add file to repository and view status
##' add(repo, "a.txt")
##' status(repo)
##'
##' ## Add file with a leading './' when the repository working
##' ## directory is the current working directory
##' setwd(path)
##' writeLines("b", file.path(path, "b.txt"))
##' add(repo, "./b.txt")
##' status(repo)
##'
##' ## Add a file in a sub-folder with sub-folder as the working
##' ## directory. Create a file in the root of the repository
##' ## working directory that will remain untracked.
##' dir.create(file.path(path, "sub_dir"))
##' setwd("./sub_dir")
##' writeLines("c", file.path(path, "c.txt"))
##' writeLines("c", file.path(path, "sub_dir/c.txt"))
##' add(repo, "c.txt")
##' status(repo)
##'
##' ## Add files with glob expansion when the current working
##' ## directory is outside the repository's working directory.
##' setwd(tempdir())
##' dir.create(file.path(path, "glob_dir"))
##' writeLines("d", file.path(path, "glob_dir/d.txt"))
##' writeLines("e", file.path(path, "glob_dir/e.txt"))
##' writeLines("f", file.path(path, "glob_dir/f.txt"))
##' writeLines("g", file.path(path, "glob_dir/g.md"))
##' add(repo, "glob_dir/*txt")
##' status(repo)
##'
##' ## Add file with glob expansion with a relative path when
##' ## the current working directory is inside the repository's
##' ## working directory.
##' setwd(path)
##' add(repo, "./glob_dir/*md")
##' status(repo)
##' }
setGeneric("add",
           signature = c("repo", "path"),
           function(repo, path, ...)
           standardGeneric("add"))

##' @rdname add-methods
##' @export
setMethod("add",
          signature(repo = "git_repository",
                    path = "character"),
          function(repo, path, force = FALSE)
          {
              ## Documentation for the pathspec argument in the
              ## libgit2 function 'git_index_add_all' that git2r use
              ## internally:
              ##
              ## The pathspec is a list of file names or shell glob
              ## patterns that will matched against files in the
              ## repository's working directory. Each file that
              ## matches will be added to the index (either updating
              ## an existing entry or adding a new entry).

              repo_wd <- normalizePath(workdir(repo), winslash = "/")
              if (!length(grep("/$", repo_wd)))
                  repo_wd <- paste0(repo_wd, "/")

              path <- vapply(path, function(p) {
                  np <- suppressWarnings(normalizePath(p, winslash = "/"))

                  ## Check if the normalized path is a non-file e.g. a
                  ## glob.
                  if (!file.exists(np)) {
                      ## Check if the normalized path starts with a
                      ## leading './'
                      if (length(grep("^[.]/", np))) {
                          nd <- suppressWarnings(normalizePath(dirname(p),
                                                               winslash = "/"))
                          if (!length(grep("/$", nd)))
                              nd <- paste0(nd, "/")
                          np <- paste0(nd, basename(np))
                      }
                  }

                  ## Check if the file is in the repository's working
                  ## directory, else let libgit2 handle this path
                  ## unmodified.
                  if (!length(grep(paste0("^", repo_wd), np)))
                      return(p)

                  ## Change the path to be relative to the repository's
                  ## working directory. Substitute common prefix with ""
                  sub(paste0("^", repo_wd), "", np)
              },
              character(1))

              .Call(git2r_index_add_all, repo, path, force)

              invisible(NULL)
          }
)

##' Remove files from the working tree and from the index
##'
##' @rdname rm_file-methods
##' @docType methods
##' @param repo The repository \code{object}.
##' @param path character vector with filenames to remove. The path
##' must be relative to the repository's working folder. Only files
##' known to Git are removed.
##' @return invisible(NULL)
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file
##' writeLines("Hello world!", file.path(path, "file-to-remove.txt"))
##'
##' ## Add file to repository
##' add(repo, "file-to-remove.txt")
##' commit(repo, "First commit message")
##'
##' ## Remove file
##' rm_file(repo, "file-to-remove.txt")
##'
##' ## View status of repository
##' status(repo)
##' }
##'
setGeneric("rm_file",
           signature = c("repo", "path"),
           function(repo, path)
           standardGeneric("rm_file"))

##' @rdname rm_file-methods
##' @export
setMethod("rm_file",
          signature(repo = "git_repository",
                    path = "character"),
          function(repo, path)
          {
              if (length(path)) {
                  ## Check that files exists and are known to Git
                  if (!all(file.exists(paste0(workdir(repo), path)))) {
                      stop(sprintf("pathspec '%s' did not match any files. ",
                                   path[!file.exists(paste0(workdir(repo), path))]))
                  }

                  if (any(file.info(paste0(workdir(repo), path))$isdir)) {
                      stop(sprintf("pathspec '%s' did not match any files. ",
                                   path[exists(paste0(workdir(repo), path))]))
                  }

                  s <- status(repo, staged = TRUE, unstaged = TRUE,
                              untracked = TRUE, ignored = TRUE)
                  if (any(path %in% c(s$ignored, s$untracked))) {
                      stop(sprintf("pathspec '%s' did not match any files. ",
                                   path[path %in% c(s$ignored, s$untracked)]))
                  }

                  if (any(path %in% s$staged)) {
                      stop(sprintf("'%s' has changes staged in the index. ",
                                   path[path %in% s$staged]))
                  }

                  if (any(path %in% s$unstaged)) {
                      stop(sprintf("'%s' has local modifications. ",
                                   path[path %in% s$unstaged]))
                  }

                  ## Remove and stage files
                  lapply(path, function(x) {
                      file.remove(paste0(workdir(repo), x))
                      .Call(git2r_index_remove_bypath, repo, x)
                  })
              }

              invisible(NULL)
          }
)

##' Remove an index entry corresponding to a file on disk
##'
##' @rdname index_remove_bypath-methods
##' @docType methods
##' @param repo The repository \code{object}.
##' @param path character vector with filenames to remove. The path
##'     must be relative to the repository's working folder. It may
##'     exist. If this file currently is the result of a merge
##'     conflict, this file will no longer be marked as
##'     conflicting. The data about the conflict will be moved to the
##'     "resolve undo" (REUC) section.
##' @return invisible(NULL)
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file
##' writeLines("Hello world!", file.path(path, "file-to-remove.txt"))
##'
##' ## Add file to repository
##' add(repo, "file-to-remove.txt")
##'
##' ## View status of repository
##' status(repo)
##'
##' ## Remove file
##' index_remove_bypath(repo, "file-to-remove.txt")
##'
##' ## View status of repository
##' status(repo)
##' }
##'
setGeneric("index_remove_bypath",
           signature = c("repo", "path"),
           function(repo, path)
           standardGeneric("index_remove_bypath"))

##' @rdname index_remove_bypath-methods
##' @export
setMethod("index_remove_bypath",
          signature(repo = "git_repository",
                    path = "character"),
          function(repo, path)
          {
              if (length(path)) {
                  .Call(git2r_index_remove_bypath, repo, path)
              }

              invisible(NULL)
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Initialize a temporary repository
path <- tempfile(pattern = "git2r-")
dir.create(path)
dir.create(file.path(path, "subfolder"))
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create three files and commit
writeLines("First file",  file.path(path, "example-1.txt"))
writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
writeLines("Third file",  file.path(path, "example-3.txt"))
add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
commit(repo, "Commit message")

## Traverse tree entries and its subtrees.
## Various approaches that give identical result.
stopifnot(identical(ls_tree(tree = tree(last_commit(path))),
                    ls_tree(tree = tree(last_commit(repo)))))
stopifnot(identical(ls_tree(repo = path), ls_tree(repo = repo)))

## ls_tree(repo = repo) should match `git ls-tree -lr HEAD`
ls_tree_result <- ls_tree(repo = repo)
stopifnot(identical(ls_tree_result$name,
                    c("example-1.txt", "example-3.txt", "example-2.txt")))

# Argument `tree` can be a  'character that identifies a tree in the repository'
ls_tree(tree = tree(last_commit(path))$sha, repo = repo)

## Skip content in subfolder
ls_tree_toplevel <- ls_tree(repo = repo, recursive = FALSE)
stopifnot(nrow(ls_tree_toplevel) == 3)
stopifnot(identical(ls_tree_toplevel$name,
                    c("example-1.txt", "example-3.txt", "subfolder")))

## Start in subfolder
ls_tree_subfolder <- ls_tree(tree = "HEAD:subfolder", repo = repo)
stopifnot(nrow(ls_tree_subfolder) == 1)
stopifnot(identical(ls_tree_subfolder$name, "example-2.txt"))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)
setwd(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Test to add file with a leading './'
writeLines("foo-1", file.path(path, "foo-1"))
add(repo, "./foo-1")
status_exp <- structure(list(staged = list(new = "foo-1"),
                             unstaged = empty_named_list(),
                             untracked = empty_named_list()),
                        class = "git_status")
status_obs <- status(repo)
str(status_exp)
str(status_obs)
stopifnot(identical(status_obs, status_exp))

## Test to add file in sub-folder with sub-folder as working directory
writeLines("foo-2", file.path(path, "foo-2"))
dir.create(file.path(path, "foo_dir"))
writeLines("foo-2", file.path(path, "foo_dir/foo-2"))
setwd("./foo_dir")
add(repo, "foo-2")
status_exp <- structure(list(staged = list(new = "foo-1",
                                           new = "foo_dir/foo-2"),
                             unstaged = empty_named_list(),
                             untracked = list(untracked = "foo-2")),
                        class = "git_status")
status_obs <- status(repo)
str(status_exp)
str(status_obs)
stopifnot(identical(status_obs, status_exp))

## Test glob expansion
setwd(tempdir())
dir.create(file.path(path, "glob_dir"))
writeLines("a", file.path(path, "glob_dir/a.txt"))
writeLines("b", file.path(path, "glob_dir/b.txt"))
writeLines("c", file.path(path, "glob_dir/c.txt"))
writeLines("d", file.path(path, "glob_dir/d.md"))
add(repo, "glob_dir/*txt")
status_exp <- structure(list(staged = list(new = "foo-1",
                                           new = "foo_dir/foo-2",
                                           new = "glob_dir/a.txt",
                                           new = "glob_dir/b.txt",
                                           new = "glob_dir/c.txt"),
                             unstaged = empty_named_list(),
                             untracked = list(untracked = "foo-2",
                                              untracked = "glob_dir/d.md")),
                        class = "git_status")
status_obs <- status(repo)
str(status_exp)
str(status_obs)
stopifnot(identical(status_obs, status_exp))

## Test glob expansion with relative path
setwd(path)
add(repo, "./glob_dir/*md")
status_exp <- structure(list(staged = list(new = "foo-1",
                                           new = "foo_dir/foo-2",
                                           new = "glob_dir/a.txt",
                                           new = "glob_dir/b.txt",
                                           new = "glob_dir/c.txt",
                                           new = "glob_dir/d.md"),
                             unstaged = empty_named_list(),
                             untracked = list(untracked = "foo-2")),
                        class = "git_status")
status_obs <- status(repo)
str(status_exp)
str(status_obs)
stopifnot(identical(status_obs, status_exp))

## Test to add file in root of workdir when the file also exists in
## current workdir.
setwd(tempdir())
writeLines("e", file.path(path, "e.txt"))
writeLines("e", file.path(tempdir(), "e.txt"))
add(repo, "e.txt")
status_exp <- structure(list(staged = list(new = "e.txt",
                                           new = "foo-1",
                                           new = "foo_dir/foo-2",
                                           new = "glob_dir/a.txt",
                                           new = "glob_dir/b.txt",
                                           new = "glob_dir/c.txt",
                                           new = "glob_dir/d.md"),
                             unstaged = empty_named_list(),
                             untracked = list(untracked = "foo-2")),
                        class = "git_status")
status_obs <- status(repo)
str(status_exp)
str(status_obs)
stopifnot(identical(status_obs, status_exp))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

setAs(from="git_time",
      to="character",
      def=function(from)
      {
          as.character(as(from, "POSIXct"))
      }
)

setAs(from="git_time",
      to="POSIXct",
      def=function(from)
      {
          as.POSIXct(from@time + from@offset*60,
                     origin="1970-01-01",
                     tz="GMT")
      }
)

setAs(from = "POSIXlt",
      to   = "git_time",
      def=function(from)
      {
          if (is.null(from$gmtoff) || is.na(from$gmtoff)) {
              offset <- 0
          } else {
              offset <- from$gmtoff / 60
          }

          new("git_time",
              time   = as.numeric(from),
              offset = offset)
      }
)

##' Brief summary of \code{git_time}
##'
##' @aliases show,git_time-methods
##' @docType methods
##' @param object The time \code{object}
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Brief summary of git_time from the default signature
##' default_signature(repo)@@when
##' }
setMethod("show",
          signature(object = "git_time"),
          function(object)
          {
              cat(sprintf("%s\n", as(object, "character")))
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Config
##'
##' Config file management. To display the configuration variables,
##' call method \code{config} without the \code{user.name},
##' \code{user.email} or \code{...} options.
##'
##' There are two ways git2r can find the local repository when
##' writing local options (1) Use the \code{repo} argument. (2) If the
##' \code{repo} argument is \code{NULL} but the current working
##' directory is inside the local repository, then \code{git2r} uses
##' that repository.
##' @param repo The \code{repository}. Default is NULL.
##' @param global Write option(s) to global configuration
##' file. Default is FALSE.
##' @param user.name The user name. Use NULL to delete the entry
##' @param user.email The e-mail address. Use NULL to delete the entry
##' @param ... Additional options to write or delete from the
##' configuration.
##' @return S3 class \code{git_config}. When writing options, the
##' configuration is returned invisible.
##' @export
##' @examples \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern = "git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Set user name and email.
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Display configuration
##' config(repo)
##'
##' ## Delete user email.
##' config(repo, user.email = NULL)
##'
##' ## Display configuration
##' config(repo)
##' }
config <- function(repo = NULL, global = FALSE, user.name, user.email, ...) {
    if (is.null(repo)) {
        repo <- discover_repository(getwd())
        if (!is.null(repo))
            repo <- repository(repo)
    }

    ## Check that 'global' is either TRUE or FALSE
    stopifnot(any(identical(global, TRUE), identical(global, FALSE)))

    variables <- list(...)
    if (!missing(user.name))
        variables <- c(variables, list(user.name = user.name))

    if (!missing(user.email))
        variables <- c(variables, list(user.email = user.email))

    if (length(variables)) {
        for (i in seq_len(length(variables))) {
            if (!is.null(variables[[i]])) {
                if (!is.character(variables[[i]])) {
                    stop("'", names(variables)[i],
                         "' must be a character vector")
                }
            }
        }

        if (isTRUE(global)) {
            repo <- NULL
            if (.Platform$OS.type == "windows") {
                ## Ensure that git2r writes the config file to the
                ## root of the user's home directory by first creating
                ## an empty file. Otherwise it may be written to the
                ## user's Documents/ directory. Only create the empty
                ## file if the user has specified configuration
                ## options to set and no global config file exists.
                if (is.na(git_config_files()[["path"]][3])) {
                    if (length(variables) > 0) {
                        file.create(file.path(home_dir(), ".gitconfig"))
                    }
                }
            }
        } else if (is.null(repo)) {
            stop("Unable to locate local repository")
        }

        .Call(git2r_config_set, repo, variables)
    }

    cfg <- .Call(git2r_config_get, repo)

    ## Sort the variables within levels by name
    cfg <- structure(lapply(cfg, function(x) x[order(names(x))]),
                     class = "git_config")

    if (length(variables)) {
        invisible(cfg)
    } else {
        return(cfg)
    }
}

##' @export
print.git_config <- function(x, ...) {
    lapply(names(x), function(level) {
        cat(sprintf("%s:\n", level))
        lapply(names(x[[level]]), function(entry) {
            cat(sprintf("        %s=%s\n", entry, x[[level]][[entry]][1]))
        })
    })

    invisible(x)
}

##' Locate the path to configuration files
##'
##' Potential configuration files:
##' \describe{
##'   \item{system}{
##'     Locate the path to the system configuration file. If
##'     '/etc/gitconfig' doesn't exist, it will look for
##'     '\%PROGRAMFILES\%'.
##'   }
##'   \item{xdg}{
##'     Locate the path to the global xdg compatible configuration
##'     file. The xdg compatible configuration file is usually located
##'     in '$HOME/.config/git/config'. This method will try to guess
##'     the full path to that file, if the file exists.
##'   }
##'   \item{global}{
##'     The user or global configuration file is usually located in
##'     '$HOME/.gitconfig'. This method will try to guess the full
##'     path to that file, if the file exists.
##'   }
##'   \item{local}{
##'     Locate the path to the repository specific configuration file,
##'     if the file exists.
##'   }
##' }
##' @template repo-param
##' @return a \code{data.frame} with one row per potential
##'     configuration file where \code{NA} means not found.
##' @export
git_config_files <- function(repo = ".") {
    ## Lookup repository
    if (inherits(repo, "git_repository")) {
        repo <- repo$path
    } else if (is.null(repo)) {
        repo <- discover_repository(getwd())
    } else if (is.character(repo) && (length(repo) == 1) &&
               !is.na(repo) && isTRUE(file.info(repo)$isdir)) {
        repo <- discover_repository(repo)
    } else {
        repo <- NULL
    }

    ## Find local configuration file
    if (is.null(repo)) {
        path <- NA_character_
    } else {
        path <- file.path(normalizePath(repo), "config")
        if (!isTRUE(!file.info(path)$isdir))
            path <- NA_character_
    }

    data.frame(file = c("system", "xdg", "global", "local"),
               path = c(.Call(git2r_config_find_file, "system"),
                        .Call(git2r_config_find_file, "xdg"),
                        .Call(git2r_config_find_file, "global"),
                        path),
               stringsAsFactors = FALSE)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Find a merge base between two commits
##'
##' @rdname merge_base-methods
##' @docType methods
##' @param one One of the commits
##' @param two The other commit
##' @return S4 class git_commit
##' @keywords methods
##' @include S4_classes.r
##' @examples \dontrun{
##' ## Create a directory in tempdir
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##'
##' ## Initialize a repository
##' repo <- init(path)
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Master branch", file.path(path, "master_branch.txt"))
##' add(repo, "master_branch.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##'
##' ## Create first branch, checkout, add file and commit
##' branch_1 <- branch_create(commit_1, "branch_1")
##' checkout(branch_1)
##' writeLines("Branch 1", file.path(path, "branch_1.txt"))
##' add(repo, "branch_1.txt")
##' commit_2 <- commit(repo, "Commit message branch_1")
##'
##' ## Create second branch, checkout, add file and commit
##' branch_2 <- branch_create(commit_1, "branch_2")
##' checkout(branch_2)
##' writeLines("Branch 2", file.path(path, "branch_2.txt"))
##' add(repo, "branch_2.txt")
##' commit_3 <- commit(repo, "Commit message branch_2")
##'
##' ## Check that merge base equals commit_1
##' stopifnot(identical(merge_base(commit_2, commit_3), commit_1))
##' }
setGeneric("merge_base",
           signature = c("one", "two"),
           function(one, two)
           standardGeneric("merge_base"))

##' @rdname merge_base-methods
##' @export
setMethod("merge_base",
          signature(one = "git_commit",
                    two = "git_commit"),
          function(one, two)
          {
              stopifnot(identical(one@repo, two@repo))
              .Call(git2r_merge_base, one, two)
          }
)

##' Merge branch into HEAD
##'
##' @param branch The branch
##' @param commit_on_success If there are no conflicts written to the
##' index, the merge commit will be committed.
##' @param merger Who made the merge.
##' @return A \code{\linkS4class{git_merge_result}} object.
##' @keywords internal
merge_branch <- function(branch, commit_on_success, merger)
{
    .Call(git2r_merge_branch, branch, merger, commit_on_success)
}

##' Merge named branch into HEAD
##'
##' @param repo The repository
##' @param branch Name of branch
##' @param commit_on_success If there are no conflicts written to the
##' index, the merge commit will be committed.
##' @param merger Who made the merge.
##' @return A \code{\linkS4class{git_merge_result}} object.
##' @keywords internal
merge_named_branch <- function(repo, branch, commit_on_success, merger)
{
    ## Check branch argument
    if (missing(branch))
        stop("missing 'branch' argument")
    if (any(!is.character(branch), !identical(length(branch), 1L)))
        stop("'branch' must be a character vector of length one")

    b <- branches(repo)
    b <- b[vapply(b, slot, character(1), "name") == branch][[1]]

    merge_branch(b, commit_on_success, merger)
}

##' Merge a branch into HEAD
##'
##' @rdname merge-methods
##' @export
##' @docType methods
##' @param x A \code{\linkS4class{git_branch}} or
##' \code{\linkS4class{git_repository}} object.
##' @param y If \code{x} is a \code{\linkS4class{git_repository}}, the
##' name of the branch to merge into HEAD. Not used if \code{x} is a
##' \code{\linkS4class{git_branch}}.
##' @param ... Additional arguments affecting the merge
##' @param commit_on_success If there are no conflicts written to the
##' index, the merge commit will be committed. Default is TRUE.
##' @param merger Who made the merge.
##' @return A \code{\linkS4class{git_merge_result}} object.
##' @keywords methods
##' @include S4_classes.r
##' @examples \dontrun{
##' ## Create a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name="Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            con = file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##'
##' ## Create first branch, checkout, add file and commit
##' checkout(repo, "branch1", create = TRUE)
##' writeLines("Branch 1", file.path(path, "branch-1.txt"))
##' add(repo, "branch-1.txt")
##' commit(repo, "Commit message branch 1")
##'
##' ## Create second branch, checkout, add file and commit
##' b_2 <- branch_create(commit_1, "branch2")
##' checkout(b_2)
##' writeLines("Branch 2", file.path(path, "branch-2.txt"))
##' add(repo, "branch-2.txt")
##' commit(repo, "Commit message branch 2")
##'
##' ## Make a change to 'test.txt'
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
##'            con = file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit(repo, "Second commit message branch 2")
##'
##' ## Checkout master
##' checkout(repo, "master", force = TRUE)
##'
##' ## Merge branch 1
##' merge(repo, "branch1")
##'
##' ## Merge branch 2
##' merge(repo, "branch2")
##'
##' ## Create third branch, checkout, change file and commit
##' checkout(repo, "branch3", create=TRUE)
##' writeLines(c("Lorem ipsum dolor amet sit, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
##'            con = file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit(repo, "Commit message branch 3")
##'
##' ## Checkout master and create a change that creates a merge conflict
##' checkout(repo, "master", force=TRUE)
##' writeLines(c("Lorem ipsum dolor sit amet, adipisicing consectetur elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
##'            con = file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit(repo, "Some commit message branch 1")
##'
##' ## Merge branch 3
##' merge(repo, "branch3")
##'
##' ## Check status; Expect to have one unstaged unmerged conflict.
##' status(repo)
##' }
setMethod("merge",
          signature(x = "git_repository", y = "character"),
          function(x,
                   y,
                   ...,
                   commit_on_success = TRUE,
                   merger = default_signature(x))
          {
              merge_named_branch(x, y, commit_on_success, merger)
          }
)

##' @rdname merge-methods
##' @export
setMethod("merge",
          signature(x = "git_branch", y = "missing"),
          function(x,
                   ...,
                   commit_on_success = TRUE,
                   merger = default_signature(x@repo))
          {
              merge_branch(x, commit_on_success, merger)
          }
)

##' Brief summary of merge result
##'
##' @aliases show,git_merge_result-methods
##' @docType methods
##' @param object The \code{\linkS4class{git_merge_result}} \code{object}
##' @return None (invisible 'NULL').
##' @keywords methods
##' @include S4_classes.r
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("First line.",
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create and checkout a new branch. Update 'example.txt' and commit
##' checkout(repo, "new_branch", create=TRUE)
##' writeLines(c("First line.", "Second line."),
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Second commit message")
##'
##' ## Checkout 'master' branch
##' checkout(repo, "master", force = TRUE)
##'
##' ## Display 'example.txt'
##' readLines(file.path(path, "example.txt"))
##'
##' ## Merge and display brief summary of the fast-forward merge
##' merge(repo, "new_branch")
##'
##' ## Display 'example.txt'
##' readLines(file.path(path, "example.txt"))
##' }
setMethod("show",
          signature(object = "git_merge_result"),
          function(object)
          {
              if (isTRUE(object@up_to_date)) {
                  cat("Already up-to-date")
              } else if (isTRUE(object@conflicts)) {
                  cat("Merge: Conflicts")
              } else if (isTRUE(object@fast_forward)) {
                  cat("Merge: Fast-forward")
              } else {
                  cat("Merge")
              }

              cat("\n")
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test-1.txt"))

## Add and reset an empty repository using a path
add(repo, "test-1.txt")
stopifnot(identical(
    status(repo),
    structure(list(staged = list(new = "test-1.txt"),
                   unstaged = empty_named_list(),
                   untracked = empty_named_list()),
              class = "git_status")))
reset(repo, path = "test-1.txt")
stopifnot(identical(
    status(repo),
    structure(list(staged = empty_named_list(),
                   unstaged = empty_named_list(),
                   untracked = list(untracked = "test-1.txt")),
              class = "git_status")))

## Add and reset a non-empty repository using a path
add(repo, "test-1.txt")
commit(repo, "First commit")
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-1.txt"))
add(repo, "test-1.txt")
stopifnot(identical(
    status(repo),
    structure(list(staged = list(modified = "test-1.txt"),
                   unstaged = empty_named_list(),
                   untracked = empty_named_list()),
              class = "git_status")))
reset(repo, path = "test-1.txt")
stopifnot(identical(
    status(repo),
    structure(list(staged = empty_named_list(),
                   unstaged = list(modified = "test-1.txt"),
                   untracked = empty_named_list()),
              class = "git_status")))

## add and commit
add(repo, "test-1.txt")
commit_1 <- commit(repo, "Commit message")

## Make one more commit
writeLines(c("Hello world!", "HELLO WORLD!", "hello world!"),
           file.path(path, "test-1.txt"))
add(repo, "test-1.txt")
commit(repo, "Next commit message")

## Create one more file
writeLines("Hello world!", file.path(path, "test-2.txt"))

## 'soft' reset to first commit
reset(commit_1)
soft_exp <- structure(list(staged = list(modified = "test-1.txt"),
                           unstaged = empty_named_list(),
                           untracked = list(untracked = "test-2.txt")),
                      class = "git_status")
soft_obs <- status(repo)
stopifnot(identical(soft_obs, soft_exp))
stopifnot(identical(length(commits(repo)), 2L))
stopifnot(identical(commits(repo)[[1]], commit_1))

## 'mixed' reset to first commit
commit(repo, "Next commit message")
reset(commit_1, "mixed")
mixed_exp <- structure(list(staged = empty_named_list(),
                            unstaged = list(modified = "test-1.txt"),
                            untracked = list(untracked = "test-2.txt")),
                       class = "git_status")
mixed_obs <- status(repo)
stopifnot(identical(mixed_obs, mixed_exp))
stopifnot(identical(length(commits(repo)), 2L))
stopifnot(identical(commits(repo)[[1]], commit_1))

## 'hard' reset to first commit
add(repo, "test-1.txt")
commit(repo, "Next commit message")
reset(commit_1, "hard")
hard_exp <- structure(list(staged = empty_named_list(),
                           unstaged = empty_named_list(),
                           untracked = list(untracked = "test-2.txt")),
                      class = "git_status")
hard_obs <- status(repo)
stopifnot(identical(hard_obs, hard_exp))
stopifnot(identical(length(commits(repo)), 2L))
stopifnot(identical(commits(repo)[[1]], commit_1))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2016 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Coerce Git repository to a \code{data.frame}
##'
##' The commits in the repository are coerced to a \code{data.frame}
##'
##'
##' The \code{data.frame} have the following columns:
##' \describe{
##'
##'   \item{sha}{
##'     The 40 character hexadecimal string of the SHA-1
##'   }
##'
##'   \item{summary}{
##'     the short "summary" of the git commit message.
##'   }
##'
##'   \item{message}{
##'     the full message of a commit
##'   }
##'
##'   \item{author}{
##'     full name of the author
##'   }
##'
##'   \item{email}{
##'     email of the author
##'   }
##'
##'   \item{when}{
##'     time when the commit happened
##'   }
##'
##' }
##' @name coerce-git_repository-method
##' @aliases coerce,git_repository,data.frame-method
##' @docType methods
##' @param from The repository \code{object}
##' @return \code{data.frame}
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create three files and commit
##' writeLines("First file",  file.path(path, "example-1.txt"))
##' writeLines("Second file", file.path(path, "example-2.txt"))
##' writeLines("Third file",  file.path(path, "example-3.txt"))
##' add(repo, "example-1.txt")
##' commit(repo, "Commit first file")
##' add(repo, "example-2.txt")
##' commit(repo, "Commit second file")
##' add(repo, "example-3.txt")
##' commit(repo, "Commit third file")
##'
##' ## Coerce commits to a data.frame
##' df <- as(repo, "data.frame")
##' df
##' }
setAs(from="git_repository",
      to="data.frame",
      def=function(from)
      {
          do.call("rbind", lapply(commits(from), as, "data.frame"))
      }
)

##' Open a repository
##'
##' @rdname repository-methods
##' @docType methods
##' @param path A path to an existing local git repository
##' @param ... Additional arguments to \code{repository} method.
##' @param discover Discover repository from path. Default is FALSE.
##' @return A S4 \code{\linkS4class{git_repository}} object
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test-1.txt"))
##' add(repo, 'test-1.txt')
##' commit_1 <- commit(repo, "Commit message")
##'
##' ## Make one more commit
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-1.txt"))
##' add(repo, 'test-1.txt')
##' commit(repo, "Next commit message")
##'
##' ## Create one more file
##' writeLines("Hello world!", file.path(path, "test-2.txt"))
##'
##' ## Brief summary of repository
##' repo
##'
##' ## Summary of repository
##' summary(repo)
##'
##' ## Workdir of repository
##' workdir(repo)
##'
##' ## Check if repository is bare
##' is_bare(repo)
##'
##' ## Check if repository is empty
##' is_empty(repo)
##'
##' ## Check if repository is a shallow clone
##' is_shallow(repo)
##'
##' ## List all references in repository
##' references(repo)
##'
##' ## List all branches in repository
##' branches(repo)
##'
##' ## Get HEAD of repository
##' head(repo)
##'
##' ## Check if HEAD is head
##' is_head(head(repo))
##'
##' ## Check if HEAD is local
##' is_local(head(repo))
##'
##' ## List all tags in repository
##' tags(repo)
##' }
setGeneric("repository",
           signature = "path",
           function(path, ...)
           standardGeneric("repository"))

##' @rdname repository-methods
##' @export
setMethod("repository",
          signature(path = "missing"),
          function()
          {
              callGeneric(path = getwd(), discover = TRUE)
          }
)

##' @rdname repository-methods
##' @export
setMethod("repository",
          signature(path = "character"),
          function(path, discover = FALSE, ...)
          {
              ## Argument checking
              stopifnot(identical(length(path), 1L),
                        nchar(path) > 0,
                        is.logical(discover),
                        identical(length(discover), 1L))

              if (discover) {
                  path <- discover_repository(path)
                  if (is.null(path))
                      stop("The 'path' is not in a git repository")
              } else {
                  path <- normalizePath(path, winslash = "/", mustWork = TRUE)
                  if (!file.info(path)$isdir)
                      stop("'path' is not a directory")
              }

              new("git_repository", path = path)
          }
)

##' Init a repository
##'
##' @rdname init-methods
##' @docType methods
##' @param path A path to where to init a git repository
##' @param bare If TRUE, a Git repository without a working directory
##' is created at the pointed path. If FALSE, provided path will be
##' considered as the working directory into which the .git directory
##' will be created.
##' @return A S4 \code{\linkS4class{git_repository}} object
##' @keywords methods
##' @seealso \link{repository}
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' is_bare(repo)
##'
##' ## Initialize a bare repository
##' path_bare <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' repo_bare <- init(path_bare, bare = TRUE)
##' is_bare(repo_bare)
##' }
setGeneric("init",
           signature = "path",
           function(path,
                    bare = FALSE)
           standardGeneric("init"))

##' @rdname init-methods
##' @export
setMethod("init",
          signature(path = "character"),
          function(path, bare)
          {
              ## Argument checking
              stopifnot(identical(length(path), 1L),
                        nchar(path) > 0,
                        is.logical(bare),
                        identical(length(bare), 1L))

              path <- normalizePath(path, winslash = "/", mustWork = TRUE)
              if (!file.info(path)$isdir)
                  stop("path is not a directory")

              .Call(git2r_repository_init, path, bare)

              new("git_repository", path=path)
          }
)

##' Clone a remote repository
##'
##' @rdname clone-methods
##' @docType methods
##' @param url The remote repository to clone
##' @param local_path Local directory to clone to.
##' @param bare Create a bare repository. Default is FALSE.
##' @param branch The name of the branch to checkout. Default is NULL
##' which means to use the remote's default branch.
##' @param checkout Checkout HEAD after the clone is complete. Default is TRUE.
##' @param credentials The credentials for remote repository
##' access. Default is NULL. To use and query an ssh-agent for the ssh
##' key credentials, let this parameter be NULL (the default).
##' @param progress Show progress. Default is TRUE.
##' @return A S4 \code{\linkS4class{git_repository}} object
##' @seealso \code{\linkS4class{cred_user_pass}},
##' \code{\linkS4class{cred_ssh_key}}
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize repository
##' path_repo_1 <- tempfile(pattern="git2r-")
##' path_repo_2 <- tempfile(pattern="git2r-")
##' dir.create(path_repo_1)
##' dir.create(path_repo_2)
##' repo_1 <- init(path_repo_1)
##'
##' ## Config user and commit a file
##' config(repo_1, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "First commit message")
##'
##' ## Change file and commit
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
##'            file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "Second commit message")
##'
##' ## Change file again and commit.
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
##'              "minim veniam, quis nostrud exercitation ullamco laboris nisi ut"),
##'            file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "Third commit message")
##'
##' ## Clone to second repository
##' repo_2 <- clone(path_repo_1, path_repo_2)
##'
##' ## List commits in repositories
##' commits(repo_1)
##' commits(repo_2)
##' }
setGeneric("clone",
           signature = c("url", "local_path"),
           function(url,
                    local_path,
                    bare        = FALSE,
                    branch      = NULL,
                    checkout    = TRUE,
                    credentials = NULL,
                    progress    = TRUE)
           standardGeneric("clone"))

##' @rdname clone-methods
##' @export
setMethod("clone",
          signature(url        = "character",
                    local_path = "character"),
          function(url,
                   local_path,
                   bare,
                   branch,
                   checkout,
                   credentials,
                   progress)
          {
              .Call(git2r_clone, url, local_path, bare,
                    branch, checkout, credentials, progress)

              repository(local_path)
          }
)

##' Get HEAD for a repository
##'
##' @rdname head-methods
##' @docType methods
##' @param x The repository \code{x} to check head
##' @return NULL if unborn branch or not found. S4 class git_branch if
##' not a detached head. S4 class git_commit if detached head
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Commit message")
##'
##' ## Get HEAD of repository
##' head(repo)
##' }
setMethod("head",
          signature(x = "git_repository"),
          function(x)
          {
              .Call(git2r_repository_head, x)
          }
)

##' Check if repository is bare
##'
##' @rdname is_bare-methods
##' @docType methods
##' @param repo The repository \code{object}
##' \code{\linkS4class{git_repository}} to check if it's bare. If the
##' \code{repo} argument is missing, the repository is searched for
##' with \code{\link{discover_repository}} in the current working
##' directory.
##' @return TRUE if bare repository, else FALSE
##' @keywords methods
##' @seealso \link{init}
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' is_bare(repo)
##'
##' ## Initialize a bare repository
##' path_bare <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' repo_bare <- init(path_bare, bare = TRUE)
##' is_bare(repo_bare)
##' }
setGeneric("is_bare",
           signature = "repo",
           function(repo)
           standardGeneric("is_bare"))

##' @rdname is_bare-methods
##' @export
setMethod("is_bare",
          signature(repo = "git_repository"),
          function(repo)
          {
              .Call(git2r_repository_is_bare, repo)
          }
)

##' @rdname is_bare-methods
##' @export
setMethod("is_bare",
          signature(repo = "missing"),
          function()
          {
              callGeneric(repo = lookup_repository())
          }
)

##' Check if HEAD of repository is detached
##'
##' @rdname is_detached-methods
##' @docType methods
##' @param repo The repository \code{object}
##' \code{\linkS4class{git_repository}}. If the \code{repo} argument
##' is missing, the repository is searched for with
##' \code{\link{discover_repository}} in the current working
##' directory.
##' @return TRUE if repository HEAD is detached, else FALSE
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##'
##' ## Change file, add and commit
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
##'              file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Commit message 2")
##'
##' ## HEAD of repository is not detached
##' is_detached(repo)
##'
##' ## Checkout first commit
##' checkout(commit_1)
##'
##' ## HEAD of repository is detached
##' is_detached(repo)
##' }
setGeneric("is_detached",
           signature = "repo",
           function(repo)
           standardGeneric("is_detached"))


##' @rdname is_detached-methods
##' @export
setMethod("is_detached",
          signature(repo = "missing"),
          function()
          {
              callGeneric(repo = lookup_repository())
          }
)

##' @rdname is_detached-methods
##' @export
setMethod("is_detached",
          signature(repo = "git_repository"),
          function(repo)
          {
              .Call(git2r_repository_head_detached, repo)
          }
)

##' Check if repository is empty
##'
##' @rdname is_empty-methods
##' @docType methods
##' @param repo The repository \code{object}
##' \code{\linkS4class{git_repository}} to check if it's empty. If the
##' \code{repo} argument is missing, the repository is searched for
##' with \code{\link{discover_repository}} in the current working
##' directory.
##' @return TRUE if empty else FALSE
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Check if it's an empty repository
##' is_empty(repo)
##'
##' ## Commit a file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Check if it's an empty repository
##' is_empty(repo)
##' }
setGeneric("is_empty",
           signature = "repo",
           function(repo)
           standardGeneric("is_empty"))


##' @rdname is_empty-methods
##' @export
setMethod("is_empty",
          signature(repo = "missing"),
          function()
          {
              callGeneric(repo = lookup_repository())
          }
)

##' @rdname is_empty-methods
##' @export
setMethod("is_empty",
          signature(repo = "git_repository"),
          function(repo)
          {
              .Call(git2r_repository_is_empty, repo)
          }
)

##' Determine if a directory is in a git repository
##'
##' The lookup start from path and walk across parent directories if
##' nothing has been found.
##' @rdname in_repository-methods
##' @docType methods
##' @param path The path to the directory. The working directory is
##' used if path is missing.
##' @return TRUE if directory is in a git repository else FALSE
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Check if path is in a git repository
##' in_repository(path)
##'
##' ## Check if working directory is in a git repository
##' setwd(path)
##' in_repository()
##' }
setGeneric("in_repository",
           signature = "path",
           function(path)
           standardGeneric("in_repository"))

##' @rdname in_repository-methods
##' @export
setMethod("in_repository",
          signature(path = "missing"),
          function()
          {
              callGeneric(path = getwd())
          }
)

##' @rdname in_repository-methods
##' @export
setMethod("in_repository",
          signature(path = "character"),
          function(path)
          {
              !is.null(discover_repository(path))
          }
)

##' Determine if the repository was a shallow clone
##'
##' @rdname is_shallow-methods
##' @docType methods
##' @param repo The repository \code{object}
##' \code{\linkS4class{git_repository}}. If the \code{repo} argument
##' is missing, the repository is searched for with
##' \code{\link{discover_repository}} in the current working
##' directory.
##' @return TRUE if shallow clone, else FALSE
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize repository
##' path_repo_1 <- tempfile(pattern="git2r-")
##' path_repo_2 <- tempfile(pattern="git2r-")
##' dir.create(path_repo_1)
##' dir.create(path_repo_2)
##' repo_1 <- init(path_repo_1)
##'
##' ## Config user and commit a file
##' config(repo_1, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "First commit message")
##'
##' ## Change file and commit
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
##'            file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "Second commit message")
##'
##' ## Change file again and commit.
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
##'              "minim veniam, quis nostrud exercitation ullamco laboris nisi ut"),
##'            file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "Third commit message")
##'
##' ## Clone to second repository
##' repo_2 <- clone(path_repo_1, path_repo_2)
##'
##' ## Check if it's a shallow clone
##' is_shallow(repo_2)
##' }
setGeneric("is_shallow",
           signature = "repo",
           function(repo)
           standardGeneric("is_shallow"))


##' @rdname is_shallow-methods
##' @export
setMethod("is_shallow",
          signature(repo = "missing"),
          function()
          {
              callGeneric(repo = lookup_repository())
          }
)

##' @rdname is_shallow-methods
##' @export
setMethod("is_shallow",
          signature(repo = "git_repository"),
          function(repo)
          {
              .Call(git2r_repository_is_shallow, repo)
          }
)

##' Lookup
##'
##' Lookup one object in a repository.
##' @rdname lookup-methods
##' @docType methods
##' @param repo The repository.
##' @param sha The identity of the object to lookup. Must be 4 to 40
##' characters long.
##' @return a \code{git_blob} or \code{git_commit} or \code{git_tag}
##' or \code{git_tree} object
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##'
##' ## Create tag
##' tag(repo, "Tagname", "Tag message")
##'
##' ## First, get SHAs to lookup in the repository
##' sha_commit <- commit_1@@sha
##' sha_tree <- tree(commit_1)@@sha
##' sha_blob <- tree(commit_1)["example.txt"]@@sha
##' sha_tag <- tags(repo)[[1]]@@sha
##'
##' ## SHAs
##' sha_commit
##' sha_tree
##' sha_blob
##' sha_tag
##'
##' ## Lookup objects
##' lookup(repo, sha_commit)
##' lookup(repo, sha_tree)
##' lookup(repo, sha_blob)
##' lookup(repo, sha_tag)
##'
##' ## Lookup objects, using only the first seven characters
##' lookup(repo, substr(sha_commit, 1, 7))
##' lookup(repo, substr(sha_tree, 1, 7))
##' lookup(repo, substr(sha_blob, 1, 7))
##' lookup(repo, substr(sha_tag, 1, 7))
##' }
setGeneric("lookup",
           signature = c("repo", "sha"),
           function(repo, sha)
           standardGeneric("lookup"))

##' @rdname lookup-methods
##' @export
setMethod("lookup",
          signature(repo = "git_repository",
                    sha  = "character"),
          function(repo, sha)
          {
              .Call(git2r_object_lookup, repo, sha)
          }
)

##' Get the signature
##'
##' Get the signature according to the repository's configuration
##' @rdname default_signature-methods
##' @docType methods
##' @param repo The repository \code{object} to check signature
##' @return S4 class git_signature
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Get the default signature
##' default_signature(repo)
##'
##' ## Change user
##' config(repo, user.name="Bob", user.email="bob@@example.org")
##'
##' ## Get the default signature
##' default_signature(repo)
##' }
setGeneric("default_signature",
           signature = "repo",
           function(repo)
           standardGeneric("default_signature"))

##' @rdname default_signature-methods
##' @export
setMethod("default_signature",
          signature(repo = "git_repository"),
          function(repo)
          {
              .Call(git2r_signature_default, repo)
          }
)

##' Brief summary of repository
##'
##' @aliases show,git_repository-methods
##' @docType methods
##' @param object The repository \code{object}
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Brief summary of the repository
##' repo
##'
##' ## Create and commit a file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Brief summary of the repository
##' repo
##' }
setMethod("show",
          signature(object = "git_repository"),
          function(object)
          {
              if (any(is_empty(object), is.null(head(object)))) {
                  cat(sprintf("Local:    %s\n", workdir(object)))
                  cat("Head:     nothing commited (yet)\n")
              } else {
                  if (is_detached(object)) {
                      cat(sprintf("Local:    (detached) %s\n", workdir(object)))

                      h <- git2r::head(object)
                  } else {
                      cat(sprintf("Local:    %s %s\n",
                                  head(object)@name,
                                  workdir(object)))

                      h <- head(object)
                      u <- branch_get_upstream(h)
                      if (!is.null(u)) {
                          rn <- branch_remote_name(u)
                          cat(sprintf("Remote:   %s @ %s (%s)\n",
                                      substr(u@name, nchar(rn) + 2, nchar(u@name)),
                                      rn,
                                      branch_remote_url(u)))
                      }

                      h <- lookup(object, branch_target(head(object)))
                  }

                  cat(sprintf("Head:     [%s] %s: %s\n",
                              substring(h@sha, 1, 7),
                              substring(as(h@author@when, "character"), 1, 10),
                              h@summary))
              }
          }
)

##' Summary of repository
##'
##' @aliases summary,git_repository-methods
##' @docType methods
##' @param object The repository \code{object}
##' @param ... Additional arguments affecting the summary produced.
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' summary(repo)
##'
##' ## Add file
##' add(repo, "test.txt")
##' summary(repo)
##'
##' ## Commit
##' commit(repo, "First commit message")
##' summary(repo)
##'
##' ## Change the file
##' writeLines(c("Hello again!", "Here is a second line", "And a third"),
##'            file.path(path, "test.txt"))
##' summary(repo)
##'
##' ## Add file and commit
##' add(repo, "test.txt")
##' commit(repo, "Second commit message")
##' summary(repo)
##'}
setMethod("summary",
          signature(object = "git_repository"),
          function(object, ...)
          {
              show(object)
              cat("\n")

              n_branches <- sum(!is.na(unique(sapply(branches(object),
                                                     branch_target))))
              n_tags <- sum(!is.na(unique(vapply(tags(object), slot, character(1), "sha"))))

              work <- commits(object)
              n_commits <- length(work)
              n_authors <- length(unique(vapply(lapply(work, slot, "author"),
                                                slot, character(1), "name")))

              s <- .Call(git2r_status_list, object, TRUE, TRUE, TRUE, FALSE, TRUE)
              n_ignored <- length(s$ignored)
              n_untracked <- length(s$untracked)
              n_unstaged <- length(s$unstaged)
              n_staged <- length(s$staged)

              n_stashes <- length(stash_list(object))

              ## Determine max characters needed to display numbers
              n <- max(vapply(c(n_branches, n_tags, n_commits, n_authors,
                                n_stashes, n_ignored, n_untracked,
                                n_unstaged, n_staged),
                              nchar,
                              numeric(1)))

              fmt <- paste0("Branches:        %", n, "i\n",
                            "Tags:            %", n, "i\n",
                            "Commits:         %", n, "i\n",
                            "Contributors:    %", n, "i\n",
                            "Stashes:         %", n, "i\n",
                            "Ignored files:   %", n, "i\n",
                            "Untracked files: %", n, "i\n",
                            "Unstaged files:  %", n, "i\n",
                            "Staged files:    %", n, "i\n")
              cat(sprintf(fmt, n_branches, n_tags, n_commits, n_authors,
                          n_stashes, n_ignored, n_untracked, n_unstaged,
                          n_staged))

              cat("\nLatest commits:\n")
              lapply(commits(object, n = 5), show)

              invisible(NULL)
          }
)

##' Workdir of repository
##'
##' @rdname workdir-methods
##' @docType methods
##' @param repo The repository \code{object}
##' \code{\linkS4class{git_repository}}. If the \code{repo} argument
##' is missing, the repository is searched for with
##' \code{\link{discover_repository}} in the current working
##' directory.
##' @return Character vector with the path of the workdir. If the
##' repository is bare, \code{NULL} will be returned.
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Create a directory in tempdir
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##'
##' ## Initialize a repository
##' repo <- init(path)
##'
##' ## Get the path of the workdir for repository
##' workdir(repo)
##' }
setGeneric("workdir",
           signature = "repo",
           function(repo)
           standardGeneric("workdir"))

##' @rdname workdir-methods
##' @export
setMethod("workdir",
          signature(repo = "missing"),
          function()
          {
              callGeneric(repo = lookup_repository())
          }
)

##' @rdname workdir-methods
##' @export
setMethod("workdir",
          signature(repo = "git_repository"),
          function(repo)
          {
              .Call(git2r_repository_workdir, repo)
          }
)

##' Find path to repository for any file
##'
##' libgit's git_discover_repository is used to identify the location
##' of the repository. The path will therefore be terminated by a file
##' separator.
##' @rdname discover_repository-methods
##' @docType methods
##' @param path A character vector specifying the path to a file or folder
##' @param ceiling The defult is to not use the ceiling argument and
##' start the lookup from path and walk across parent
##' directories. When ceiling is 0, the lookup is only in path. When
##' ceiling is 1, the lookup is in both the path and the parent to
##' path.
##' @return Character vector with path to repository or NULL if this
##' cannot be established.
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path, "example-1.txt"))
##' add(repo, "example-1.txt")
##' commit(repo, "First commit message")
##'
##' ## Create a second file. The file is not added for version control
##' ## in the repository.
##' dir.create(file.path(path, "example"))
##' file_2 <- file.path(path, "example/example-2.txt")
##' writeLines("Not under version control", file_2)
##'
##' ## Find the path to the repository using the path to the second file
##' discover_repository(file_2)
##'
##' ## Demonstrate the 'ceiling' argument
##' wd <- workdir(repo)
##' dir.create(file.path(wd, "temp"))
##'
##' ## Lookup repository in 'file.path(wd, "temp")'. Should return NULL
##' discover_repository(file.path(wd, "temp"), ceiling = 0)
##'
##' ## Lookup repository in parent to 'file.path(wd, "temp")'.
##' ## Should not return NULL
##' discover_repository(file.path(wd, "temp"), ceiling = 1)
##' }
setGeneric("discover_repository",
           signature = c("path", "ceiling"),
           function(path, ceiling)
           standardGeneric("discover_repository"))

##' @rdname discover_repository-methods
##' @export
setMethod("discover_repository",
          signature(path = "character", ceiling = "missing"),
          function(path)
          {
              callGeneric(path = path, ceiling = as.numeric(NA))
          }
)

##' @rdname discover_repository-methods
##' @export
setMethod("discover_repository",
          signature(path = "character", ceiling = "numeric"),
          function(path, ceiling)
          {
              path <- normalizePath(path)

              if (is.na(ceiling)) {
                  ceiling <- NULL
              } else {
                  ceiling <- as.integer(ceiling)
                  if (identical(ceiling, 0L)) {
                      ceiling <- dirname(path)
                  } else if (identical(ceiling, 1L)) {
                      ceiling <- dirname(dirname(path))
                  } else {
                      stop("'ceiling' must be either 0 or 1")
                  }
              }


              .Call(git2r_repository_discover, path, ceiling)
          }
)

##' Lookup repository for methods with missing repo argument
##'
##' @return S4 class git_repository
##' @keywords internal
lookup_repository <- function()
{
    ## Try current working directory
    repo <- discover_repository(getwd())
    if (is.null(repo))
        stop("The working directory is not in a git repository")

    repository(repo)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name="Alice", user.email="alice@example.org")

## Create a file, add and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message 1")

## Create first branch, checkout, add file and commit
checkout(repo, "branch1", create=TRUE)
writeLines("Branch 1", file.path(path, "branch-1.txt"))
add(repo, "branch-1.txt")
commit_2 <- commit(repo, "Commit message branch 1")

## Create second branch, checkout, add file and commit
b_2 <- branch_create(commit_1, "branch2")
checkout(b_2)
writeLines("Branch 2", file.path(path, "branch-2.txt"))
add(repo, "branch-2.txt")
commit_3 <- commit(repo, "Commit message branch 2")
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit_4 <- commit(repo, "Second commit message branch 2")

## Check that merge base equals commit_1
stopifnot(identical(merge_base(commit_2, commit_3), commit_1))

## Checkout master
b <- branches(repo)
checkout(b[sapply(b, slot, "name") == "master"][[1]], force=TRUE)

## Merge branch 1
m_1 <- git2r:::merge_named_branch(repo, "branch1", TRUE, default_signature(repo))
stopifnot(identical(m_1@fast_forward, TRUE))
stopifnot(identical(m_1@conflicts, FALSE))
stopifnot(identical(m_1@sha, character(0)))

## Merge branch 2
m_2 <- git2r:::merge_named_branch(repo, "branch2", TRUE, default_signature(repo))
stopifnot(identical(m_2@fast_forward, FALSE))
stopifnot(identical(m_2@conflicts, FALSE))
stopifnot(identical(m_2@sha, commits(repo)[[1]]@sha))

## Create third branch, checkout, change file and commit
checkout(repo, "branch3", create=TRUE)
writeLines(c("Lorem ipsum dolor amet sit, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message branch 3")

## Checkout master and create a change that creates a merge conflict
checkout(repo, "master", force=TRUE)
writeLines(c("Lorem ipsum dolor sit amet, adipisicing consectetur elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Some commit message branch 1")

## Merge branch 3
m_3 <- git2r:::merge_named_branch(repo, "branch3", TRUE, default_signature(repo))
stopifnot(identical(m_3@up_to_date, FALSE))
stopifnot(identical(m_3@fast_forward, FALSE))
stopifnot(identical(m_3@conflicts, TRUE))
stopifnot(identical(m_3@sha, character(0)))

## Check status; Expect to have one unstaged unmerged conflict.
stopifnot(identical(status(repo),
                    structure(list(staged = structure(list(),
                                       .Names = character(0)),
                                   unstaged = structure(list(conflicted = "test.txt"),
                                       .Names = "conflicted"),
                                   untracked = structure(list(),
                                       .Names = character(0))),
                              .Names = c("staged", "unstaged", "untracked"),
                              class = "git_status")))

## Cleanup
unlink(path, recursive=TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add, commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message")

## Change the file, diff between index and workdir
writeLines("Hello again!\nHere is a second line\nAnd a third",
           file.path(path, "test.txt"))
diff_1 <- diff(repo)
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(identical(diff_1$old, "index"))
stopifnot(identical(diff_1$new, "workdir"))
stopifnot(identical(length(diff_1$files), 1L))
stopifnot(identical(diff_1$files[[1]]$old_file, "test.txt"))
stopifnot(identical(diff_1$files[[1]]$new_file, "test.txt"))
stopifnot(identical(length(diff_1$files[[1]]$hunks), 1L))
stopifnot(identical(length(diff_1$files[[1]]$hunks[[1]]$lines), 4L))
## TODO: check actual diff

## Diff between index and HEAD is empty
diff_2 <- diff(repo, index = TRUE)
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(identical(diff_2$old, "HEAD"))
stopifnot(identical(diff_2$new, "index"))
stopifnot(identical(diff_2$files, list()))

## Diff between tree and working dir, same as diff_1
diff_3 <- diff(tree(commits(repo)[[1]]))
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(identical(diff_3$old, tree(commits(repo)[[1]])))
stopifnot(identical(diff_3$new, "workdir"))
stopifnot(identical(diff_3$files, diff_1$files))
stopifnot(identical(print(diff_3), diff_3))

## Add changes, diff between index and HEAD is the same as diff_1
add(repo, "test.txt")
diff_4 <- diff(repo, index = TRUE)
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(identical(diff_4$old, "HEAD"))
stopifnot(identical(diff_4$new, "index"))
stopifnot(identical(diff_4$files, diff_1$files))

## Diff between tree and index
diff_5 <- diff(tree(commits(repo)[[1]]), index = TRUE)
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(identical(diff_5$old, tree(commits(repo)[[1]])))
stopifnot(identical(diff_5$new, "index"))
stopifnot(identical(diff_5$files, diff_1$files))

## Diff between two trees
commit(repo, "Second commit")
tree_1 <- tree(commits(repo)[[2]])
tree_2 <- tree(commits(repo)[[1]])
diff_6 <- diff(tree_1, tree_2)
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(identical(diff_6$old, tree_1))
stopifnot(identical(diff_6$new, tree_2))
stopifnot(identical(diff_6$files, diff_1$files))

## Length of a diff
stopifnot(identical(length(diff_1), 1L))
stopifnot(identical(length(diff_2), 0L))
stopifnot(identical(length(diff_3), 1L))
stopifnot(identical(length(diff_4), 1L))
stopifnot(identical(length(diff_5), 1L))
stopifnot(identical(length(diff_6), 1L))

## Binary files
set.seed(42)
writeBin(as.raw((sample(0:255, 1000, replace = TRUE))),
         con = file.path(path, "test.bin"))
add(repo, "test.bin")
diff_7 <- diff(repo, index = TRUE)
diff(repo, as_char = TRUE)
diff(repo, as_char = TRUE, filename = file.path(path, "test.diff"))

stopifnot(any(grepl("binary file", capture.output(summary(diff_7)))))

## TODO: errors
## Check non-logical index argument
res <- tools::assertError(
                  .Call(git2r:::git2r_diff, NULL, NULL, NULL, "FALSE",
                        NULL, 3L, 0L, "a", "b", NULL, NULL, NULL))
stopifnot(length(grep(paste0("Error in 'git2r_diff': 'index' must be logical ",
                             "vector of length one with non NA value\n"),
                      res[[1]]$message)) > 0)

## Check various combinations of diff arguments
res <- tools::assertError(
                  .Call(git2r:::git2r_diff, NULL, NULL,
                        tree(commits(repo)[[1]]),
                        FALSE, NULL, 3L, 0L, "a", "b", NULL, NULL, NULL))
stopifnot(length(grep("Error in 'git2r_diff': Invalid diff parameters",
                      res[[1]]$message)) > 0)

res <- tools::assertError(
                  .Call(git2r:::git2r_diff, NULL, NULL,
                        tree(commits(repo)[[1]]),
                        TRUE, NULL, 3L, 0L, "a", "b", NULL, NULL, NULL))
stopifnot(length(grep("Error in 'git2r_diff': Invalid diff parameters",
                      res[[1]]$message)) > 0)

res <- tools::assertError(
                  .Call(git2r:::git2r_diff, repo, tree(commits(repo)[[1]]),
                        NULL, FALSE, NULL, 3L, 0L, "a", "b", NULL, NULL, NULL))
stopifnot(length(grep("Error in 'git2r_diff': Invalid diff parameters",
                      res[[1]]$message)) > 0)

res <- tools::assertError(
                  .Call(git2r:::git2r_diff, repo, tree(commits(repo)[[1]]),
                        NULL, TRUE, NULL, 3L, 0L, "a", "b", NULL, NULL, NULL))
stopifnot(length(grep("Error in 'git2r_diff': Invalid diff parameters",
                      res[[1]]$message)) > 0)

res <- tools::assertError(
                  .Call(git2r:::git2r_diff, repo, tree(commits(repo)[[1]]),
                        tree(commits(repo)[[2]]), FALSE, NULL, 3L, 0L, "a",
                        "b", NULL, NULL, NULL))
stopifnot(length(grep("Error in 'git2r_diff': Invalid diff parameters",
                      res[[1]]$message)) > 0)

res <- tools::assertError(
                  .Call(git2r:::git2r_diff, repo, tree(commits(repo)[[1]]),
                        tree(commits(repo)[[2]]), TRUE, NULL, 3L, 0L, "a",
                        "b", NULL, NULL, NULL))
stopifnot(length(grep("Error in 'git2r_diff': Invalid diff parameters",
                      res[[1]]$message)) > 0)

## TODO: printing

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)

## Config repository
config(repo, user.name = "Alice")

## Let's set one valid and one with variable with invalid format
res <- tools::assertWarning(config(repo,
                                   user.email = "alice@example.org",
                                   lol = "wut"))
stopifnot(length(grep("Variable was not in a valid format: 'lol'",
                      res[[1]]$message)) > 0)

cfg_exp <- structure(list(user.name = "Alice",
                          user.email = "alice@example.org",
                          "NA" = NULL),
                     .Names = c("user.name", "user.email", NA))

cfg_obs <- config(repo)$local
cfg_obs <- cfg_obs[c("user.name", "user.email", "lol")]
stopifnot(identical(cfg_obs, cfg_exp))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create 2 directories in tempdir
path_src <- tempfile(pattern = "git2r-")
path_tgt <- tempfile(pattern = "git2r-")
dir.create(path_tgt)
dir.create(path_src)

## Initialize a repository
repo_src <- init(path_src)
config(repo_src, user.name = "Alice", user.email = "alice@example.org")

## Add commit to repo
filename <- "test.txt"
writeLines("Hello world", con = file.path(path_src, filename))
add(repo_src, "test.txt")
commit_src <- commit(repo_src, "Commit message")

## Check checkout argument
tools::assertError(clone(path_src, path_tgt, checkout = c(FALSE, TRUE)))
tools::assertError(clone(path_src, path_tgt, checkout = 1))
tools::assertError(clone(path_src, path_tgt, checkout = 1L))
tools::assertError(clone(path_src, path_tgt, checkout = "test"))

## Clone source to target repository without checking out any files
repo_tgt <- clone(path_src, path_tgt, checkout = FALSE)

## List files in the repositores
stopifnot(identical(list.files(path_src), filename))
stopifnot(identical(list.files(path_tgt), character(0)))

## Compare commits
stopifnot(identical(length(commits(repo_tgt)), 1L))
commit_tgt <- last_commit(repo_tgt)
stopifnot(identical(sha(last_commit(path_tgt)), sha(commit_tgt)))
stopifnot(identical(sha(commit_src), sha(commit_tgt)))
stopifnot(identical(commit_src$author, commit_tgt$author))
stopifnot(identical(commit_src$committer, commit_tgt$committer))
stopifnot(identical(commit_src$summary, commit_tgt$summary))
stopifnot(identical(commit_src$message, commit_tgt$message))
stopifnot(!identical(commit_src$repo, commit_tgt$repo))

## Cleanup
unlink(path_tgt, recursive = TRUE)
unlink(path_src, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2018 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Get the configured remotes for a repo
##'
##' @template repo-param
##' @return Character vector with remotes
##' @export
##' @template remote-example
remotes <- function(repo = ".") {
    .Call(git2r_remote_list, lookup_repository(repo))
}

##' Add a remote to a repo
##'
##' @template repo-param
##' @param name Short name of the remote repository
##' @param url URL of the remote repository
##' @return NULL, invisibly
##' @export
##' @template remote-example
remote_add <- function(repo = ".", name = NULL, url = NULL) {
    .Call(git2r_remote_add, lookup_repository(repo), name, url)
    invisible(NULL)
}

##' Rename a remote
##'
##' @template repo-param
##' @param oldname Old name of the remote
##' @param newname New name of the remote
##' @return NULL, invisibly
##' @export
##' @template remote-example
remote_rename <- function(repo = ".", oldname = NULL, newname = NULL) {
    .Call(git2r_remote_rename, lookup_repository(repo), oldname, newname)
    invisible(NULL)
}

##' Remove a remote
##'
##' All remote-tracking branches and configuration settings for the
##' remote will be removed.
##' @template repo-param
##' @param name The name of the remote to remove
##' @return NULL, invisibly
##' @export
##' @template remote-example
remote_remove <- function(repo = ".", name = NULL) {
    .Call(git2r_remote_remove, lookup_repository(repo), name)
    invisible(NULL)
}

##' Set the remote's url in the configuration
##'
##' This assumes the common case of a single-url remote and will
##' otherwise raise an error.
##' @template repo-param
##' @param name The name of the remote
##' @param url The \code{url} to set
##' @return NULL, invisibly
##' @export
##' @template remote-example
remote_set_url <- function(repo = ".", name = NULL, url = NULL) {
    .Call(git2r_remote_set_url, lookup_repository(repo), name, url)
    invisible(NULL)
}

##' Get the remote url for remotes in a repo
##'
##' @template repo-param
##' @param remote Character vector with the remotes to get the url
##'     from. Default is the remotes of the repository.
##' @return Character vector with remote_url for each of the remote
##' @export
##' @template remote-example
remote_url <- function(repo = ".", remote = NULL) {
    repo <- lookup_repository(repo)
    if (is.null(remote))
        remote <- remotes(repo)
    .Call(git2r_remote_url, repo, remote)
}

##' List references in a remote repository
##'
##' Displays references available in a remote repository along with
##' the associated commit IDs.  Akin to the 'git ls-remote' command.
##' @param name Character vector with the "remote" repository URL to
##'     query or the name of the remote if a \code{repo} argument is
##'     given.
##' @param repo an optional repository object used if remotes are
##'     specified by name.
##' @param credentials The credentials for remote repository
##'     access. Default is NULL. To use and query an ssh-agent for the
##'     ssh key credentials, let this parameter be NULL (the default).
##' @return Character vector for each reference with the associated
##'     commit IDs.
##' @export
##' @examples
##' \dontrun{
##' remote_ls("https://github.com/ropensci/git2r")
##' }
remote_ls <- function(name = NULL, repo = NULL, credentials = NULL) {
    if (is.null(repo)) {
        ver <- libgit2_version()
        if (ver$major == 0 && ver$minor < 27) {
            path <- tempdir()
            repo <- init(path)
            on.exit(unlink(file.path(path, ".git"), recursive = TRUE))
        }
    } else {
        repo <- lookup_repository(repo)
    }

    .Call(git2r_remote_ls, name, repo, credentials)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2018 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Compile time options for libgit2.
##'
##' @return A list with threads, https and ssh set to TRUE/FALSE.
##' @keywords methods
##' @export
##' @examples
##' libgit2_features()
libgit2_features <- function() {
    .Call(git2r_libgit2_features)
}

##' Version of the libgit2 library
##'
##' Version of the libgit2 library that the bundled source code is
##' based on
##' @return A list with major, minor and rev
##' @keywords methods
##' @export
##' @examples
##' libgit2_version()
libgit2_version <- function() {
    .Call(git2r_libgit2_version)
}

##' Set the SSL certificate-authority locations
##'
##' @note Either parameter may be 'NULL', but not both.
##' @param filename Location of a file containing several certificates
##' concatenated together. Default NULL.
##' @param path Location of a directory holding several certificates,
##' one per file. Default NULL.
##' @return invisible(NULL)
##' @keywords methods
##' @export
ssl_cert_locations <- function(filename = NULL, path = NULL) {
    .Call(git2r_ssl_cert_locations, filename, path)
    invisible(NULL)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create directories for repositories in tempdir
path_bare <- tempfile(pattern = "git2r-")
path_repo_1 <- tempfile(pattern = "git2r-")
path_repo_2 <- tempfile(pattern = "git2r-")

dir.create(path_bare)
dir.create(path_repo_1)
dir.create(path_repo_2)

## Create bare repository
bare_repo <- init(path_bare, bare = TRUE)

## Clone to repo 1
repo_1 <- clone(path_bare, path_repo_1)
config(repo_1, user.name = "Alice", user.email = "alice@example.org")

## Add changes to repo 1
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path_repo_1, "test-1.txt"))
add(repo_1, "test-1.txt")
commit_1 <- commit(repo_1, "First commit message")

## Create 'dev' branch
checkout(branch_create(commit_1, name = "dev"))

## Add more changes to repo 1
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path_repo_1, "test-1.txt"))
add(repo_1, "test-1.txt")
commit(repo_1, "Second commit message")

## Add more changes to repo 1
writeLines(
    c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
      "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
      "minim veniam, quis nostrud exercitation ullamco laboris nisi ut"),
    con = file.path(path_repo_1, "test-1.txt"))
add(repo_1, "test-1.txt")
commit(repo_1, "Third commit message")

## Push to bare
push(repo_1, "origin", "refs/heads/master")
push(repo_1, "origin", "refs/heads/dev")

## Print branch
branches(repo_1)[["origin/master"]]

## Clone to repo 2
repo_2 <- clone(url = path_bare, local_path = path_repo_2, branch = "dev")
config(repo_2, user.name = "Bob", user.email = "bob@example.org")

## Check branch and commits
stopifnot(identical(length(commits(repo_2)), 3L))
stopifnot(identical(repository_head(repo_2)$name, "dev"))

## Cleanup
unlink(path_bare, recursive = TRUE)
unlink(path_repo_1, recursive = TRUE)
unlink(path_repo_2, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2020 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' git2r: R bindings to the libgit2 library
##'
##' git2r: R bindings to the libgit2 library.
##' @docType package
##' @name git2r
##' @useDynLib git2r, .registration=TRUE
NULL

##' Unload hook function
##'
##' @param libpath A character string giving the complete path to the
##' package.
##' @noRd
.onUnload <- function(libpath) {
    library.dynam.unload("git2r", libpath)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create 2 directories in tempdir
path_bare <- tempfile(pattern = "git2r-")
path_repo_1 <- tempfile(pattern = "git2r-")
path_repo_2 <- tempfile(pattern = "git2r-")

dir.create(path_bare)
dir.create(path_repo_1)
dir.create(path_repo_2)

## Create repositories
bare_repo <- init(path_bare, bare = TRUE)
repo_1 <- clone(path_bare, path_repo_1)
repo_2 <- clone(path_bare, path_repo_2)

## Config repositories
config(repo_1, user.name = "Alice", user.email = "alice@example.org")
config(repo_2, user.name = "Bob", user.email = "bob@example.org")

## Add changes to repo 1
writeLines("Hello world", con = file.path(path_repo_1, "test.txt"))
add(repo_1, "test.txt")
commit_1 <- commit(repo_1, "Commit message")

## Push changes from repo 1 to origin
push(repo_1, "origin", "refs/heads/master")

## Check result in bare repository
stopifnot(identical(length(commits(bare_repo)), 1L))
bare_commit_1 <- commits(bare_repo)[[1]]
stopifnot(identical(sha(commit_1), sha(bare_commit_1)))
stopifnot(identical(commit_1$author, bare_commit_1$author))
stopifnot(identical(commit_1$committer, bare_commit_1$committer))
stopifnot(identical(commit_1$summary, bare_commit_1$summary))
stopifnot(identical(commit_1$message, bare_commit_1$message))
stopifnot(!identical(commit_1$repo, bare_commit_1$repo))

## Fetch
fetch(repo_2, "origin")
fh <- fetch_heads(repo_2)[[1]]
stopifnot(identical(sha(fh), fh$sha))

## Test show method of non-empty repository where head is null
show(repo_2)

## Check that 'git2r_arg_check_credentials' raise error
res <- tools::assertError(
                  .Call(git2r:::git2r_remote_fetch, repo_1, "origin",
                        3, "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", repo_1,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- cred_env(c("username", "username"), "password")
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- cred_env("username", c("password", "passowrd"))
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- cred_user_pass(c("username", "username"), "password")
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- cred_user_pass("username", c("password", "passowrd"))
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- cred_token(c("GITHUB_PAT", "GITHUB_PAT"))
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- structure(list(publickey  = c("id_rsa.pub", "id_rsa.pub"),
                              privatekey = "id_rsa",
                              passphrase = character(0)),
                         class = "cred_ssh_key")
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- structure(list(publickey  = "id_rsa.pub",
                              privatekey = c("id_rsa", "id_rsa"),
                              passphrase = character(0)),
                         class = "cred_ssh_key")
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- structure(list(publickey  = "id_rsa.pub",
                              privatekey = "id_rsa",
                              passphrase = NA_character_),
                         class = "cred_ssh_key")
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

credentials <- structure(list(publickey  = "id_rsa.pub",
                              privatekey = "id_rsa",
                              passphrase = c("passphrase", "passphrase")),
                         class = "cred_ssh_key")
res <- tools::assertError(
    .Call(git2r:::git2r_remote_fetch, repo_1, "origin", credentials,
          "fetch", FALSE, NULL))
stopifnot(length(grep("'credentials' must be an S3 class with credentials",
                      res[[1]]$message)) > 0)

## Cleanup
unlink(path_bare, recursive = TRUE)
unlink(path_repo_1, recursive = TRUE)
unlink(path_repo_2, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Create tag targeting HEAD commit in repository
##'
##' @param object The repository \code{object}.
##' @param name Name for the tag.
##' @param message The tag message. Specify a tag message to create an
##'     annotated tag. A lightweight tag is created if the message
##'     parameter is \code{NULL}.
##' @param session Add sessionInfo to tag message. Default is FALSE.
##' @param tagger The tagger (author) of the tag
##' @param force Overwrite existing tag. Default = FALSE
##' @return invisible(\code{git_tag}) object
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Commit a text file
##' filename <- file.path(path, "example.txt")
##' writeLines("Hello world!", filename)
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create an annotated tag
##' tag(repo, "v1.0", "Tag message")
##'
##' ## List tags
##' tags(repo)
##'
##' ## Make a change to the text file and commit.
##' writeLines(c("Hello world!", "HELLO WORLD!"), filename)
##' add(repo, "example.txt")
##' commit(repo, "Second commit message")
##'
##' ## Create a lightweight tag
##' tag(repo, "v2.0")
##'
##' ## List tags
##' tags(repo)
##' }
tag <- function(object = ".",
                name    = NULL,
                message = NULL,
                session = FALSE,
                tagger  = NULL,
                force   = FALSE) {
    object <- lookup_repository(object)

    if (isTRUE(session))
        message <- add_session_info(message)

    if (is.null(tagger))
        tagger <- default_signature(object)

    invisible(.Call(git2r_tag_create, object, name, message, tagger, force))
}

##' Check if object is a git_tag object
##'
##' @param object Check if object is a git_tag object
##' @return TRUE if object is a git_tag, else FALSE
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create tag
##' tag(repo, "Tagname", "Tag message")
##'
##' is_tag(tags(repo)[[1]])
##' is_tag(last_commit(repo))
##' }
is_tag <- function(object) {
    inherits(object, "git_tag")
}

##' Delete an existing tag reference
##'
##' @param object Can be either the path (default is ".") to a
##'     repository, or a \code{git_repository} object, or a
##'     \code{git_tag} object. or the tag name.
##' @param name If the \code{object} argument is a path to a
##'     repository or a \code{git_repository}, the name of the tag to
##'     delete.
##' @return \code{invisible(NULL)}
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create two tags
##' tag(repo, "Tag1", "Tag message 1")
##' t2 <- tag(repo, "Tag2", "Tag message 2")
##'
##' ## List the two tags in the repository
##' tags(repo)
##'
##' ## Delete the two tags in the repository
##' tag_delete(repo, "Tag1")
##' tag_delete(t2)
##'
##' ## Show the empty list with tags in the repository
##' tags(repo)
##' }
tag_delete <- function(object = ".", name = NULL) {
    if (is_tag(object)) {
        name <- object$name
        object <- object$repo
    } else {
        object <- lookup_repository(object)
    }

    .Call(git2r_tag_delete, object, name)
    invisible(NULL)
}

##' Tags
##'
##' @template repo-param
##' @return list of tags in repository
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create tag
##' tag(repo, "Tagname", "Tag message")
##'
##' ## List tags
##' tags(repo)
##' }
tags <- function(repo = ".") {
    .Call(git2r_tag_list, lookup_repository(repo))
}

##' @export
format.git_tag <- function(x, ...) {
    sprintf("[%s] %s", substr(x$target, 1, 6), x$name)
}

##' @export
print.git_tag <- function(x, ...) {
    cat(format(x, ...), "\n", sep = "")
    invisible(x)
}

##' @export
summary.git_tag <- function(object, ...) {
    cat(sprintf(paste0("name:    %s\n",
                       "target:  %s\n",
                       "tagger:  %s <%s>\n",
                       "when:    %s\n",
                       "message: %s\n"),
                object$name,
                object$target,
                object$tagger$name,
                object$tagger$email,
                as.character(object$tagger$when),
                object$message))
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Get blame for file
##'
##' @rdname blame-methods
##' @docType methods
##' @param repo The repository
##' @param path Path to the file to consider
##' @return S4 class git_blame object
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a first user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create a second user and change the file
##' config(repo, user.name="Bob", user.email="bob@@example.org")
##' writeLines(c("Hello world!", "HELLO WORLD!", "HOLA"),
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Second commit message")
##'
##' ## Check blame
##' blame(repo, "example.txt")
##' }
setGeneric("blame",
           signature = c("repo", "path"),
           function(repo,
                    path)
           standardGeneric("blame"))

##' @rdname blame-methods
##' @export
setMethod("blame",
          signature(repo = "git_repository",
                    path = "character"),
          function(repo,
                   path)
          {
              .Call(git2r_blame_file, repo, path)
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

#' Number of files in git_diff object
#'
#' @docType methods
#' @param x The git_diff \code{object}
#' @return a non-negative integer
#' @keywords methods
#' @export
setMethod("length",
          signature(x = "git_diff"),
          function(x)
          {
              length(x@files)
          }
)

##' Show a diff
##'
##' @aliases show,git_diff-methods
##' @docType methods
##' @param object The diff \code{object}.
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file, add, commit
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit(repo, "Commit message")
##'
##' ## Change the file
##' writeLines(c("Hello again!", "Here is a second line", "And a third"),
##'            file.path(path, "test.txt"))
##'
##' ## Brief summary of diff between index and workdir
##' diff(repo)
##' }
setMethod("show",
          signature(object = "git_diff"),
          function(object)
          {
              cat("Old:  ")
              if (is.character(object@old)) {
                  cat(object@old, "\n")
              } else if (is(object@old, "git_tree")) {
                  show(object@old)
              } else {
                  cat("\n")
                  show(object@old)
              }
              cat("New:  ")
              if (is.character(object@new)) {
                  cat(object@new, "\n")
              } else if (is(object@new, "git_tree")) {
                  show(object@new)
              } else {
                  cat("\n")
                  show(object@new)
              }
          }
)

lines_per_file <- function(diff) {
    lapply(diff@files, function(x) {
        del <- add <- 0
        for (h in x@hunks) {
            for (l in h@lines) {
                if (l@origin == 45) {
                    del <- del + l@num_lines
                } else if (l@origin == 43) {
                    add <- add + l@num_lines
                }
            }
        }
        list(file=x@new_file, del=del, add=add)
    })
}

print_lines_per_file <- function(diff) {
  lpf <- lines_per_file(diff)
  files <- vapply(lpf, function(x) x$file, character(1))
  del <- vapply(lpf, function(x) x$del, numeric(1))
  add <- vapply(lpf, function(x) x$add, numeric(1))
  paste0(format(files), " | ", "-", format(del), " +", format(add))
}

hunks_per_file <- function(diff) {
    vapply(diff@files, function(x) length(x@hunks), numeric(1))
}

##' Show the summary of a diff
##'
##' @aliases summary,git_diff-methods
##' @docType methods
##' @param object The diff \code{object}.
##' @param ... Additional arguments affecting the summary produced.
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file, add, commit
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit(repo, "Commit message")
##'
##' ## Change the file
##' writeLines(c("Hello again!", "Here is a second line", "And a third"),
##'            file.path(path, "test.txt"))
##'
##' ## diff between index and workdir
##' diff_1 <- diff(repo)
##' summary(diff_1)
##'
##' ## Diff between index and HEAD is empty
##' diff_2 <- diff(repo, index=TRUE)
##' summary(diff_2)
##'
##' ## Diff between tree and working dir, same as diff_1
##' diff_3 <- diff(tree(commits(repo)[[1]]))
##' summary(diff_3)
##'
##' ## Add changes, diff between index and HEAD is the same as diff_1
##' add(repo, "test.txt")
##' diff_4 <- diff(repo, index=TRUE)
##' summary(diff_4)
##'
##' ## Diff between tree and index
##' diff_5 <- diff(tree(commits(repo)[[1]]), index=TRUE)
##' summary(diff_5)
##'
##' ## Diff between two trees
##' commit(repo, "Second commit")
##' tree_1 <- tree(commits(repo)[[2]])
##' tree_2 <- tree(commits(repo)[[1]])
##' diff_6 <- diff(tree_1, tree_2)
##' summary(diff_6)
##'
##' ## Binary files
##' set.seed(42)
##' writeBin(as.raw((sample(0:255, 1000, replace=TRUE))),
##'          con=file.path(path, "test.bin"))
##' add(repo, "test.bin")
##' diff_7 <- diff(repo, index=TRUE)
##' summary(diff_7)
##' }
setMethod("summary",
          signature(object = "git_diff"),
          function(object, ...)
          {
              show(object)
              if (length(object) > 0) {
                  plpf <- print_lines_per_file(object)
                  hpf <- hunks_per_file(object)
                  hunk_txt <- ifelse(hpf > 1, " hunks",
                                     ifelse(hpf > 0, " hunk",
                                            " hunk (binary file)"))
                  phpf <- paste0("  in ", format(hpf), hunk_txt)
                  cat("Summary:", paste0(plpf, phpf), sep="\n")
              } else {
                  cat("No changes.\n")
              }
          }
)

#' Changes between commits, trees, working tree, etc.
#'
#' @rdname diff-methods
#' @export
#' @docType methods
#' @param x A \code{git_repository} object or the old
#' \code{git_tree} object to compare to.
#' @param index \describe{
#'   \item{\emph{When object equals a git_repository}}{
#'     Whether to compare the index to HEAD. If FALSE (the default),
#'     then the working tree is compared to the index.
#'   }
#'   \item{\emph{When object equals a git_tree}}{
#'     Whether to use the working directory (by default), or the index
#'     (if set to TRUE) in the comparison to \code{object}.
#'   }
#' }
#' @param as_char logical: should the result be converted to a
#' character string?. Default is FALSE.
#' @param filename If as_char is TRUE, then the diff can be written to
#' a file with name filename (the file is overwritten if it
#' exists). Default is NULL.
#' @param ... Additional arguments affecting the diff produced
#' @return A \code{\linkS4class{git_diff}} object if as_char is
#' FALSE. If as_char is TRUE and filename is NULL, a character string,
#' else NULL.
#' @keywords methods
#' @examples
#' \dontrun{
#' ## Initialize a repository
#' path <- tempfile(pattern="git2r-")
#' dir.create(path)
#' repo <- init(path)
#'
#' ## Config user
#' config(repo, user.name="Alice", user.email="alice@@example.org")
#'
#' ## Create a file, add, commit
#' writeLines("Hello world!", file.path(path, "test.txt"))
#' add(repo, "test.txt")
#' commit(repo, "Commit message")
#'
#' ## Change the file
#' writeLines(c("Hello again!", "Here is a second line", "And a third"),
#'            file.path(path, "test.txt"))
#'
#' ## diff between index and workdir
#' diff_1 <- diff(repo)
#' summary(diff_1)
#' cat(diff(repo, as_char=TRUE))
#'
#' ## Diff between index and HEAD is empty
#' diff_2 <- diff(repo, index=TRUE)
#' summary(diff_2)
#' cat(diff(repo, index=TRUE, as_char=TRUE))
#'
#' ## Diff between tree and working dir, same as diff_1
#' diff_3 <- diff(tree(commits(repo)[[1]]))
#' summary(diff_3)
#' cat(diff(tree(commits(repo)[[1]]), as_char=TRUE))
#'
#' ## Add changes, diff between index and HEAD is the same as diff_1
#' add(repo, "test.txt")
#' diff_4 <- diff(repo, index=TRUE)
#' summary(diff_4)
#' cat(diff(repo, index=TRUE, as_char=TRUE))
#'
#' ## Diff between tree and index
#' diff_5 <- diff(tree(commits(repo)[[1]]), index=TRUE)
#' summary(diff_5)
#' cat(diff(tree(commits(repo)[[1]]), index=TRUE, as_char=TRUE))
#'
#' ## Diff between two trees
#' commit(repo, "Second commit")
#' tree_1 <- tree(commits(repo)[[2]])
#' tree_2 <- tree(commits(repo)[[1]])
#' diff_6 <- diff(tree_1, tree_2)
#' summary(diff_6)
#' cat(diff(tree_1, tree_2, as_char=TRUE))
#'
#' ## Binary files
#' set.seed(42)
#' writeBin(as.raw((sample(0:255, 1000, replace=TRUE))),
#'          con=file.path(path, "test.bin"))
#' add(repo, "test.bin")
#' diff_7 <- diff(repo, index=TRUE)
#' summary(diff_7)
#' cat(diff(repo, index=TRUE, as_char=TRUE))
#' }
setMethod("diff",
          signature(x = "git_repository"),
          function(x,
                   index    = FALSE,
                   as_char  = FALSE,
                   filename = NULL,
                   ...)
          {
              if (as_char) {
                  ## Make sure filename is character(0) to write to a
                  ## character vector or a character vector with path in
                  ## order to write to a file.
                  filename <- as.character(filename)
                  if (any(identical(filename, NA_character_),
                          identical(nchar(filename), 0L))) {
                      filename <- character(0)
                  } else if (length(filename)) {
                      filename <- normalizePath(filename, mustWork = FALSE)
                  }
              } else {
                  ## Make sure filename is NULL
                  filename <- NULL
              }

              .Call(git2r_diff, x, NULL, NULL, index, filename)
          }
)

#' @rdname diff-methods
#' @param new_tree The new git_tree object to compare, or NULL.
#'        If NULL, then we use the working directory or the index (see
#'        the \code{index} argument).
#' @export
setMethod("diff",
          signature(x = "git_tree"),
          function(x,
                   new_tree = NULL,
                   index    = FALSE,
                   as_char  = FALSE,
                   filename = NULL,
                   ...)
          {
              if (as_char) {
                  ## Make sure filename is character(0) to write to a
                  ## character vector or a character vector with path in
                  ## order to write to a file.
                  filename <- as.character(filename)
                  if (any(identical(filename, NA_character_),
                          identical(nchar(filename), 0L))) {
                      filename <- character(0)
                  } else if (length(filename)) {
                      filename <- normalizePath(filename, mustWork = FALSE)
                  }
              } else {
                  ## Make sure filename is NULL
                  filename <- NULL
              }

              if (!is.null(new_tree)) {
                  if (! is(new_tree, "git_tree")) {
                      stop("Not a git tree")
                  }
                  if (x@repo@path != new_tree@repo@path) {
                      stop("Cannot compare trees in different repositories")
                  }
              }

              .Call(git2r_diff, NULL, x, new_tree, index, filename)
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2018 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name="Alice", user.email="alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test.txt"))

## Add and commit
add(repo, "test.txt")
commit_1 <- commit(repo, "First commit message")

tools::assertError(revparse_single(repo, "HEAD^"))

## Add another commit
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Second commit message")

stopifnot(identical(revparse_single(repo, "HEAD^"), commit_1))
stopifnot(is_blob(revparse_single(repo, "HEAD:test.txt")))

## Cleanup
unlink(path, recursive=TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Get the configured remotes for a repo
##'
##' @rdname remotes-methods
##' @docType methods
##' @param repo The repository \code{object}
##' \code{\linkS4class{git_repository}}. If the \code{repo} argument
##' is missing, the repository is searched for with
##' \code{\link{discover_repository}} in the current working
##' directory.
##' @return Character vector with remotes
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Add a remote
##' remote_add(repo, "playground", "https://example.org/git2r/playground")
##' remotes(repo)
##' remote_url(repo, "playground")
##'
##' ## Rename a remote
##' remote_rename(repo, "playground", "foobar")
##' remotes(repo)
##' remote_url(repo, "foobar")
##'
##' ## Remove a remote
##' remote_remove(repo, "foobar")
##' remotes(repo)
##' }
setGeneric("remotes",
           signature = "repo",
           function(repo)
           standardGeneric("remotes"))

##' @rdname remotes-methods
##' @export
setMethod("remotes",
          signature(repo = "missing"),
          function()
          {
              callGeneric(repo = lookup_repository())
          }
)

##' @rdname remotes-methods
##' @export
setMethod("remotes",
          signature(repo = "git_repository"),
          function(repo)
          {
              .Call(git2r_remote_list, repo)
          }
)

##' Add a remote to a repo
##'
##' @rdname remote_add-methods
##' @docType methods
##' @param repo The repository to add the remote to
##' @param name Short name of the remote repository
##' @param url URL of the remote repository
##' @return NULL, invisibly
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Add a remote
##' remote_add(repo, "playground", "https://example.org/git2r/playground")
##' remotes(repo)
##' remote_url(repo, "playground")
##'
##' ## Rename a remote
##' remote_rename(repo, "playground", "foobar")
##' remotes(repo)
##' remote_url(repo, "foobar")
##'
##' ## Remove a remote
##' remote_remove(repo, "foobar")
##' remotes(repo)
##' }
setGeneric("remote_add",
           signature = c("repo", "name", "url"),
           function(repo, name, url)
           standardGeneric("remote_add"))

##' @rdname remote_add-methods
##' @export
setMethod("remote_add",
          signature(repo = "git_repository",
                    name = "character",
                    url  = "character"),
          function(repo, name, url)
          {
              ret <- .Call(git2r_remote_add, repo, name, url)
              invisible(ret)
          }
)

##' Rename a remote
##'
##' @rdname remote_rename-methods
##' @docType methods
##' @param repo The repository in which the remote should be renamed.
##' @param oldname Old name of the remote
##' @param newname New name of the remote
##' @return NULL, invisibly
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Add a remote
##' remote_add(repo, "playground", "https://example.org/git2r/playground")
##' remotes(repo)
##' remote_url(repo, "playground")
##'
##' ## Rename a remote
##' remote_rename(repo, "playground", "foobar")
##' remotes(repo)
##' remote_url(repo, "foobar")
##'
##' ## Remove a remote
##' remote_remove(repo, "foobar")
##' remotes(repo)
##' }
setGeneric("remote_rename",
           signature = c("repo", "oldname", "newname"),
           function(repo, oldname, newname)
           standardGeneric("remote_rename"))

##' @rdname remote_rename-methods
##' @export
setMethod("remote_rename",
          signature(repo    = "git_repository",
                    oldname = "character",
                    newname = "character"),
          function(repo, oldname, newname)
          {
              ret <- .Call(git2r_remote_rename, repo, oldname, newname)
              invisible(ret)
          }
)

##' Remove a remote
##'
##' All remote-tracking branches and configuration settings for the
##' remote will be removed.
##' @rdname remote_remove-methods
##' @docType methods
##' @param repo The repository to work on
##' @param name The name of the remote to remove
##' @return NULL, invisibly
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Add a remote
##' remote_add(repo, "playground", "https://example.org/git2r/playground")
##' remotes(repo)
##' remote_url(repo, "playground")
##'
##' ## Rename a remote
##' remote_rename(repo, "playground", "foobar")
##' remotes(repo)
##' remote_url(repo, "foobar")
##'
##' ## Remove a remote
##' remote_remove(repo, "foobar")
##' remotes(repo)
##' }
setGeneric("remote_remove",
           signature = c("repo", "name"),
           function(repo, name)
           standardGeneric("remote_remove"))

##' @rdname remote_remove-methods
##' @export
setMethod("remote_remove",
          signature(repo = "git_repository",
                    name = "character"),
          function(repo, name)
          {
              ret <- .Call(git2r_remote_remove, repo, name)
              invisible(ret)
          }
)

##' Set the remote's url in the configuration
##'
##' This assumes the common case of a single-url remote and will
##' otherwise raise an error.
##' @rdname remote_set_url-methods
##' @docType methods
##' @param repo The repository in which to perform the change
##' @param name The name of the remote
##' @param url The \code{url} to set
##' @return NULL, invisibly
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Add a remote
##' remote_add(repo, "playground", "https://example.org/git2r/playground")
##' remotes(repo)
##' remote_url(repo, "playground")
##'
##' ## Rename a remote
##' remote_rename(repo, "playground", "foobar")
##' remotes(repo)
##' remote_url(repo, "foobar")
##'
##' ## Set remote url
##' remote_set_url(repo, "foobar", "https://example.org/git2r/foobar")
##' remotes(repo)
##' remote_url(repo, "foobar")
##'
##' ## Remove a remote
##' remote_remove(repo, "foobar")
##' remotes(repo)
##' }
setGeneric("remote_set_url",
           signature = c("repo", "name", "url"),
           function(repo, name, url)
           standardGeneric("remote_set_url"))

##' @rdname remote_set_url-methods
##' @export
setMethod("remote_set_url",
          signature(repo = "git_repository",
                    name = "character",
                    url  = "character"),
          function(repo, name, url)
          {
              ret <- .Call(git2r_remote_set_url, repo, name, url)
              invisible(ret)
          }
)

##' Get the remote url for remotes in a repo
##'
##' @rdname remote_url-methods
##' @docType methods
##' @param repo The repository to get remote urls from
##' @param remote Character vector with the remotes to get the url
##' from. Default is the remotes of the repository.
##' @return Character vector with remote_url for each of the remote
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Add a remote
##' remote_add(repo, "playground", "https://example.org/git2r/playground")
##' remotes(repo)
##' remote_url(repo, "playground")
##'
##' ## Rename a remote
##' remote_rename(repo, "playground", "foobar")
##' remotes(repo)
##' remote_url(repo, "foobar")
##'
##' ## Remove a remote
##' remote_remove(repo, "foobar")
##' remotes(repo)
##' }
setGeneric("remote_url",
           signature = "repo",
           function(repo, remote = remotes(repo))
           standardGeneric("remote_url"))

##' @rdname remote_url-methods
##' @export
setMethod("remote_url",
          signature(repo = "git_repository"),
          function(repo, remote)
          {
              .Call(git2r_remote_url, repo, remote)
          }
)


##' List references in a remote repository
##'
##' Displays references available in a remote repository along with the
##' associated commit IDs.  Akin to the 'git ls-remote' command.
##' @rdname remote_ls-methods
##' @docType methods
##' @param name Character vector with the "remote" repository URL to query or
##' the name of the remote if a \code{repo} argument is given.
##' @param repo an optional repository object used if remotes are
##' specified by name.
##' @param credentials The credentials for remote repository
##' access. Default is NULL. To use and query an ssh-agent for the ssh
##' key credentials, let this parameter be NULL (the default).
##' @keywords methods
##' @return Character vector for each reference with the associated commit IDs.
##' @examples
##' \dontrun{
##' remote_ls("https://github.com/ropensci/git2r")
##' }
##' @export
setGeneric("remote_ls",
           signature = c("name"),
           function(name,
                    repo = NULL,
                    credentials = NULL)
           standardGeneric("remote_ls"))

##' @rdname remote_ls-methods
##' @export
setMethod("remote_ls",
          signature(name = "character"),
          function(name, repo, credentials)
          {
              ## FIXME: When updating to libgit 0.26 + 1, remove this
              ## and allow repo to be NULL, see 'git2r_remote_ls'.
              if (is.null(repo)) {
                  path <- tempdir()
                  repo <- git2r::init(path)
                  on.exit(unlink(file.path(path, ".git"), recursive = TRUE))
              }

              .Call(git2r_remote_ls, name, repo, credentials)
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@@example.org")

## Create a file, add and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message 1")

## Change file and commit
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
             con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message 2")

## Commit same content under different name in a sub-directory
dir.create(file.path(path, "sub-directory"))
file.copy(file.path(path, "test.txt"),
          file.path(path, "sub-directory", "copy.txt"))
add(repo, "sub-directory/copy.txt")
commit(repo, "Commit message 3")

## List blobs
b <- odb_blobs(repo)

## Order the data.frame before checking
b <- b[order(b$name), ]

## Check blobs
stopifnot(identical(nrow(b), 3L))
stopifnot(identical(
    colnames(b),
    c("sha", "path", "name", "len", "commit", "author", "when")))
stopifnot(identical(b$path, c("sub-directory", "", "")))
stopifnot(identical(b$name, c("copy.txt", "test.txt", "test.txt")))
stopifnot(identical(b$author, c("Alice", "Alice", "Alice")))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Create tag targeting HEAD commit in repository
##'
##' @rdname tag-methods
##' @docType methods
##' @param object The repository \code{object}.
##' @param name Name for the tag.
##' @param message The tag message.
##' @param session Add sessionInfo to tag message. Default is FALSE.
##' @param tagger The tagger (author) of the tag
##' @return invisible(\code{git_tag}) object
##' @keywords methods
##' @include commit.r
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create tag
##' tag(repo, "Tagname", "Tag message")
##'
##' ## List tags
##' tags(repo)
##' }
setGeneric("tag",
           signature = "object",
           function(object,
                    name,
                    message,
                    session = FALSE,
                    tagger  = default_signature(object))
           standardGeneric("tag"))

##' @rdname tag-methods
##' @export
setMethod("tag",
          signature(object = "git_repository"),
          function(object,
                   name,
                   message,
                   session,
                   tagger)
          {
              ## Argument checking
              stopifnot(is.character(name),
                        identical(length(name), 1L),
                        nchar(name[1]) > 0,
                        is.character(message),
                        identical(length(message), 1L),
                        nchar(message[1]) > 0,
                        is.logical(session),
                        identical(length(session), 1L),
                        is(tagger, "git_signature"))

              if (session)
                  message <- add_session_info(message)

              invisible(.Call(git2r_tag_create, object, name, message, tagger))
          }
)

##' Delete an existing tag reference
##'
##' @rdname tag_delete-methods
##' @docType methods
##' @param object Can be either a
##'     (\code{\linkS4class{git_repository}}) object, a
##'     \code{\linkS4class{git_tag}} object or the tag name. If the
##'     \code{object} argument is the tag name, the repository is
##'     searched for with \code{\link{discover_repository}} in the
##'     current working directory.
##' @param ... Additional arguments
##' @param name If the \code{object} argument is a
##'     \code{git_repository}, the name of the tag to delete.
##' @return \code{invisible(NULL)}
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create two tags
##' tag(repo, "Tag1", "Tag message 1")
##' t2 <- tag(repo, "Tag2", "Tag message 2")
##'
##' ## List the two tags in the repository
##' tags(repo)
##'
##' ## Delete the two tags in the repository
##' tag_delete(repo, "Tag1")
##' tag_delete(t2)
##'
##' ## Show the empty list with tags in the repository
##' tags(repo)
##' }
setGeneric("tag_delete",
           signature = "object",
           function(object, ...)
           standardGeneric("tag_delete"))

##' @rdname tag_delete-methods
##' @export
setMethod("tag_delete",
          signature(object = "git_repository"),
          function(object, name)
          {
              invisible(.Call(git2r_tag_delete, object, name))
          }
)

##' @rdname tag_delete-methods
##' @export
setMethod("tag_delete",
          signature(object = "character"),
          function(object)
          {
              callGeneric(object = lookup_repository(), name = object)
          }
)

##' @rdname tag_delete-methods
##' @export
setMethod("tag_delete",
          signature(object = "git_tag"),
          function(object)
          {
              callGeneric(object = object@repo, name = object@name)
          }
)

##' Tags
##'
##' @rdname tags-methods
##' @docType methods
##' @param repo The repository \code{object}
##' \code{\linkS4class{git_repository}}. If the \code{repo} argument
##' is missing, the repository is searched for with
##' \code{\link{discover_repository}} in the current working
##' directory.
##' @return list of tags in repository
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create tag
##' tag(repo, "Tagname", "Tag message")
##'
##' ## List tags
##' tags(repo)
##' }
setGeneric("tags",
           signature = "repo",
           function(repo)
           standardGeneric("tags"))

##' @rdname tags-methods
##' @export
setMethod("tags",
          signature(repo = "missing"),
          function()
          {
              callGeneric(repo = lookup_repository())
          }
)

##' @rdname tags-methods
##' @export
setMethod("tags",
          signature(repo = "git_repository"),
          function(repo)
          {
              .Call(git2r_tag_list, repo)
          }
)

##' Brief summary of a tag
##'
##' @aliases show,git_tag-methods
##' @docType methods
##' @param object The tag \code{object}
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create tag
##' tag(repo, "Tagname", "Tag message")
##'
##' ## View brief summary of tag
##' tags(repo)[[1]]
##' }
setMethod("show",
          signature(object = "git_tag"),
          function(object)
          {
              cat(sprintf("[%s] %s\n",
                          substr(object@target, 1 , 6),
                          object@name))
          }
)

##' Summary of a tag
##'
##' @aliases summary,git_tag-methods
##' @docType methods
##' @param object The tag \code{object}
##' @param ... Additional arguments affecting the summary produced.
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create tag
##' tag(repo, "Tagname", "Tag message")
##'
##' ## Summary of tag
##' summary(tags(repo)[[1]])
##' }
setMethod("summary",
          signature(object = "git_tag"),
          function(object, ...)
          {
              cat(sprintf(paste0("name:    %s\n",
                                 "target:  %s\n",
                                 "tagger:  %s <%s>\n",
                                 "when:    %s\n",
                                 "message: %s\n"),
                          object@name,
                          object@target,
                          object@tagger@name,
                          object@tagger@email,
                          as(object@tagger@when, "character"),
                          object@message))
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)

## Config repository
cfg <- config(repo, user.name="Alice", user.email="alice@example.org")

## Check configuration
stopifnot("local" %in% names(cfg))
stopifnot("user.name" %in% names(cfg$local))
stopifnot(identical(cfg$local$user.name, "Alice"))
stopifnot(identical(cfg$local$user.email, "alice@example.org"))

## Delete entries
cfg <- config(repo, user.name=NULL, user.email=NULL)

## Check configuration
stopifnot(is.null(cfg$local$user.name))
stopifnot(is.null(cfg$local$user.email))

## Supply values as objects
user.name <- "Alice"
user.email <- "alice@example.org"
cfg <- config(repo, user.name=user.name, user.email="alice@example.org")
stopifnot(identical(cfg$local$user.name, user.name))
stopifnot(identical(cfg$local$user.email, "alice@example.org"))
cfg <- config(repo, user.name="Alice", user.email=user.email)
stopifnot(identical(cfg$local$user.name, "Alice"))
stopifnot(identical(cfg$local$user.email, user.email))

## Cleanup
unlink(path, recursive=TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' When
##'
##' Help method to extract the time as a character string from the S4
##' classes \code{git_commit}, \code{git_signature}, \code{git_tag}
##' and \code{git_time}.
##' @rdname when-methods
##' @docType methods
##' @param object the \code{object} to extract the time slot from.
##' @return A \code{character} vector of length one.
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a first user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create tag
##' tag(repo, "Tagname", "Tag message")
##'
##' when(commits(repo)[[1]])
##' when(tags(repo)[[1]])
##' }
setGeneric("when",
           signature = "object",
           function(object)
           standardGeneric("when"))

##' @rdname when-methods
##' @export
setMethod("when",
          signature(object = "git_commit"),
          function(object)
          {
              as(object@author@when, "character")
          }
)

##' @rdname when-methods
##' @export
setMethod("when",
          signature(object = "git_signature"),
          function(object)
          {
              as(object@when, "character")
          }
)

##' @rdname when-methods
##' @export
setMethod("when",
          signature(object = "git_stash"),
          function(object)
          {
              as(object@stasher@when, "character")
          }
)

##' @rdname when-methods
##' @export
setMethod("when",
          signature(object = "git_tag"),
          function(object)
          {
              as(object@tagger@when, "character")
          }
)

##' @rdname when-methods
##' @export
setMethod("when",
          signature(object = "git_time"),
          function(object)
          {
              as(object, "character")
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2018 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Get the signature
##'
##' Get the signature according to the repository's configuration
##' @template repo-param
##' @return A \code{git_signature} object with entries:
## \describe{
##   \item{name}{
##     The full name of the author.
##   }
##   \item{email}{
##     Email of the author.
##   }
##   \item{when}{
##     Time when the action happened.
##   }
## }
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Get the default signature
##' default_signature(repo)
##'
##' ## Change user
##' config(repo, user.name = "Bob", user.email = "bob@@example.org")
##'
##' ## Get the default signature
##' default_signature(repo)
##' }
default_signature <- function(repo = ".") {
    .Call(git2r_signature_default, lookup_repository(repo))
}

##' @export
format.git_signature <- function(x, ...) {
    sprintf("name:  %s\nemail: %s\nwhen:  %s",
            x$name, x$email, as.character(x$when))
}

##' @export
print.git_signature <- function(x, ...) {
    cat(format(x, ...), "\n", sep = "")
    invisible(x)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2020 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Create a branch
##'
##' @param commit Commit to which the branch should point. The default
##'     is to use the \code{last_commit()} function to determine the
##'     commit to which the branch should point.
##' @param name Name for the branch
##' @param force Overwrite existing branch. Default = FALSE
##' @return invisible git_branch object
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' lines <- "Hello world!"
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##'
##' ## Create a branch
##' branch_1 <- branch_create(commit_1, name = "test-branch")
##'
##' ## Add one more commit
##' lines <- c("Hello world!", "HELLO WORLD!")
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_2 <- commit(repo, "Another commit message")
##'
##' ## Create a branch with the same name should fail
##' try(branch_create(commit_2, name = "test-branch"), TRUE)
##'
##' ## Force it
##' branch_2 <- branch_create(commit_2, name = "test-branch", force = TRUE)
##' }
branch_create <- function(commit = last_commit(), name = NULL,
                          force = FALSE) {
    invisible(.Call(git2r_branch_create, name, commit, force))
}

##' Delete a branch
##'
##' @param branch The branch
##' @return invisible NULL
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##'
##' ## Create a 'dev' branch
##' dev <- branch_create(commit_1, name = "dev")
##' branches(repo)
##'
##' ## Delete 'dev' branch
##' branch_delete(dev)
##' branches(repo)
##' }
branch_delete <- function(branch = NULL) {
    .Call(git2r_branch_delete, branch)
    invisible(NULL)
}

##' Remote name of a branch
##'
##' The name of remote that the remote tracking branch belongs to
##' @param branch The branch
##' @return character string with remote name
##' @export
##' @examples
##' \dontrun{
##' ## Initialize two temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## Get remote name
##' branch_remote_name(branches(repo)[[2]])
##' }
branch_remote_name <- function(branch = NULL) {
    .Call(git2r_branch_remote_name, branch)
}

##' Remote url of a branch
##'
##' @param branch The branch
##' @return character string with remote url
##' @export
##' @examples
##' \dontrun{
##' ## Initialize two temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## Get remote url of tracking branch to branch 'master'
##' branch_remote_url(branch_get_upstream(repository_head(repo)))
##' }
branch_remote_url <- function(branch = NULL) {
    .Call(git2r_branch_remote_url, branch)
}

##' Rename a branch
##'
##' @param branch Branch to rename
##' @param name The new name for the branch
##' @param force Overwrite existing branch. Default is FALSE
##' @return invisible renamed \code{git_branch} object
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Rename 'master' branch to 'dev'
##' branches(repo)
##' branch_rename(repository_head(repo), "dev")
##' branches(repo)
##' }
branch_rename <- function(branch = NULL, name = NULL, force = FALSE) {
    invisible(.Call(git2r_branch_rename, branch, name, force))
}

##' Get target (sha) pointed to by a branch
##'
##' @param branch The branch
##' @return sha or NA if not a direct reference
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Get target (sha) pointed to by 'master' branch
##' branch_target(repository_head(repo))
##' }
branch_target <- function(branch = NULL) {
    .Call(git2r_branch_target, branch)
}

##' Get remote tracking branch
##'
##' Get remote tracking branch, given a local branch.
##' @param branch The branch
##' @return \code{git_branch} object or NULL if no remote tracking
##'     branch.
##' @export
##' @examples
##' \dontrun{
##' ## Initialize two temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## Get remote tracking branch
##' branch_get_upstream(repository_head(repo))
##' }
branch_get_upstream <- function(branch = NULL) {
    .Call(git2r_branch_get_upstream, branch)
}

##' Set remote tracking branch
##'
##' Set the upstream configuration for a given local branch
##' @param branch The branch to configure
##' @param name remote-tracking or local branch to set as
##' upstream. Pass NULL to unset.
##' @return invisible NULL
##' @export
##' @examples
##' \dontrun{
##' ## Initialize two temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## Unset remote remote tracking branch
##' branch_get_upstream(repository_head(repo))
##' branch_set_upstream(repository_head(repo), NULL)
##' branch_get_upstream(repository_head(repo))
##'
##' ## Set remote tracking branch
##' branch_set_upstream(repository_head(repo), "origin/master")
##' branch_get_upstream(repository_head(repo))
##' }
branch_set_upstream <- function(branch = NULL, name) {
    if (missing(name))
        stop("Missing argument name")
    .Call(git2r_branch_set_upstream, branch, name)
    invisible(NULL)
}

##' Branches
##'
##' List branches in repository
##' @template repo-param
##' @param flags Filtering flags for the branch listing. Valid values
##'     are 'all', 'local' or 'remote'
##' @return list of branches in repository
##' @export
##' @examples
##' \dontrun{
##' ## Initialize repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config first user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## List branches
##' branches(repo)
##' }
branches <- function(repo = ".", flags=c("all", "local", "remote")) {
    flags <- switch(match.arg(flags),
                    local  = 1L,
                    remote = 2L,
                    all    = 3L)

    .Call(git2r_branch_list, lookup_repository(repo), flags)
}

##' Check if branch is head
##'
##' @param branch The branch \code{object} to check if it's head.
##' @return \code{TRUE} if branch is head, else \code{FALSE}.
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## List branches
##' branches(repo)
##'
##' ## Check that 'master' is_head
##' master <- branches(repo)[[1]]
##' is_head(master)
##'
##' ## Create and checkout 'dev' branch
##' checkout(repo, "dev", create = TRUE)
##'
##' ## List branches
##' branches(repo)
##'
##' ## Check that 'master' is no longer head
##' is_head(master)
##' }
is_head <- function(branch = NULL) {
    .Call(git2r_branch_is_head, branch)
}

##' Check if branch is local
##'
##' @param branch The branch \code{object} to check if it's local
##' @return \code{TRUE} if branch is local, else \code{FALSE}.
##' @export
##' @examples
##' \dontrun{
##' ## Initialize repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config first user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## List branches
##' branches(repo)
##'
##' ## Check if first branch is_local
##' is_local(branches(repo)[[1]])
##'
##' ## Check if second branch is_local
##' is_local(branches(repo)[[2]])
##' }
is_local <- function(branch) {
    if (!is_branch(branch))
        stop("argument 'branch' must be a 'git_branch' object")
    identical(branch$type, 1L)
}

##' @export
print.git_branch <- function(x, ...) {
    sha <- branch_target(x)
    if (!is.na(sha)) {
        cat(sprintf("[%s] ", substr(sha, 1, 6)))
    }

    if (is_local(x)) {
        cat("(Local) ")
    } else {
        cat(sprintf("(%s @ %s) ",
                    branch_remote_name(x),
                    branch_remote_url(x)))
    }

    if (is_head(x)) {
        cat("(HEAD) ")
    }

    if (is_local(x)) {
        cat(sprintf("%s\n", x$name))
    } else {
        cat(sprintf("%s\n",
                    substr(x$name,
                           start = nchar(branch_remote_name(x)) + 2,
                           stop = nchar(x$name))))
    }

    invisible(x)
}

##' Check if object is \code{git_branch}
##'
##' @param object Check if object is of class \code{git_branch}
##' @return TRUE if object is class \code{git_branch}, else FALSE
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' branch <- branches(repo)[[1]]
##'
##' ## Check if branch
##' is_branch(branch)
##' }
is_branch <- function(object) {
    inherits(object, "git_branch")
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a '.gitignore' file
writeLines("test.txt", file.path(path, ".gitignore"))
add(repo, ".gitignore")
commit(repo, "First commit message")

## Create a file
writeLines("Hello world!", file.path(path, "test.txt"))

## Check status
s_1 <- structure(list(staged = empty_named_list(),
                      unstaged = empty_named_list(),
                      untracked = empty_named_list(),
                      ignored = list(ignored = "test.txt")),
                 class = "git_status")
stopifnot(identical(status(repo, ignored = TRUE), s_1))

## The file is ignored and should not be added
add(repo, "test.txt")
stopifnot(identical(status(repo, ignored = TRUE), s_1))

## The file is ignored but should be added with force
s_2 <- structure(list(staged = list(new = "test.txt"),
                      unstaged = empty_named_list(),
                      untracked = empty_named_list(),
                      ignored = empty_named_list()),
                 class = "git_status")

add(repo, "test.txt", force = TRUE)
stopifnot(identical(status(repo, ignored = TRUE), s_2))

## Commit and check status
s_3 <- structure(list(staged = empty_named_list(),
                      unstaged = empty_named_list(),
                      untracked = empty_named_list(),
                      ignored = empty_named_list()),
                 class = "git_status")

commit(repo, "Second commit message")
stopifnot(identical(status(repo, ignored = TRUE), s_3))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create first commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "First commit message")

## Edit file and checkout
writeLines(c("Hello world!", "Hello world!"), file.path(path, "test.txt"))
status_exp_1 <- structure(list(staged = structure(list(),
                                   .Names = character(0)),
                               unstaged = structure(list(modified = "test.txt"),
                                   .Names = "modified"),
                               untracked = structure(list(),
                                   .Names = character(0))),
                          .Names = c("staged", "unstaged", "untracked"),
                          class = "git_status")
status_obs_1 <- status(repo)
str(status_exp_1)
str(status_obs_1)
stopifnot(identical(status_obs_1, status_exp_1))
checkout(repo, path = "test.txt")
status_exp_2 <- structure(list(staged = structure(list(),
                                   .Names = character(0)),
                               unstaged = structure(list(),
                                   .Names = character(0)),
                               untracked = structure(list(),
                                   .Names = character(0))),
                          .Names = c("staged", "unstaged", "untracked"),
                          class = "git_status")
status_obs_2 <- status(repo)
str(status_exp_2)
str(status_obs_2)
stopifnot(identical(status_obs_2, status_exp_2))

## Create second commit
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Second commit message")
tag(repo, "commit_2", "Tag message")

## Create third commit
writeLines(c("Hello world!", "HELLO WORLD!", "HeLlO wOrLd!"),
           file.path(path, "test.txt"))
add(repo, "test.txt")
commit_3 <- commit(repo, "Third commit message")

## Check HEAD
stopifnot(identical(is_detached(repo), FALSE))
stopifnot(identical(repository_head(repo)$name, "master"))

## Check show and summary
repo
summary(repo)

## Checkout first commit
checkout(commit_1, TRUE)
stopifnot(identical(is_detached(repo), TRUE))
stopifnot(identical(repository_head(repo), commit_1))
stopifnot(identical(readLines(file.path(path, "test.txt")), "Hello world!"))

## Check show and summary
repo
summary(repo)

## Checkout tag
checkout(tags(repo)[[1]], TRUE)
stopifnot(identical(is_detached(repo), TRUE))
stopifnot(identical(readLines(file.path(path, "test.txt")),
                    c("Hello world!", "HELLO WORLD!")))

## Check is_detached with missing repo argument
wd <- setwd(path)
stopifnot(identical(is_detached(), TRUE))
if (!is.null(wd))
    setwd(wd)

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Fetch new data and update tips
##'
##' @template repo-param
##' @param name the remote's name
##' @param credentials The credentials for remote repository
##'     access. Default is NULL. To use and query an ssh-agent for the
##'     ssh key credentials, let this parameter be NULL (the default).
##' @param verbose Print information each time a reference is updated
##'     locally. Default is \code{TRUE}.
##' @param refspec The refs to fetch and which local refs to update,
##'     see examples. Pass NULL to use the
##'     \code{remote.<repository>.fetch} variable. Default is
##'     \code{NULL}.
##' @return invisible list of class \code{git_transfer_progress}
##'     with statistics from the fetch operation:
##' \describe{
##'   \item{total_objects}{
##'     Number of objects in the packfile being downloaded
##'   }
##'   \item{indexed_objects}{
##'     Received objects that have been hashed
##'   }
##'   \item{received_objects}{
##'     Objects which have been downloaded
##'   }
##'   \item{total_deltas}{
##'     Total number of deltas in the pack
##'   }
##'   \item{indexed_deltas}{
##'     Deltas which have been indexed
##'   }
##'   \item{local_objects}{
##'     Locally-available objects that have been injected in order to
##'     fix a thin pack
##'   }
##'   \item{received_bytes}{
##'     Size of the packfile received up to now
##'   }
##' }
##' @export
##' @examples
##' \dontrun{
##' ## Initialize three temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo_1 <- tempfile(pattern="git2r-")
##' path_repo_2 <- tempfile(pattern="git2r-")
##'
##' dir.create(path_bare)
##' dir.create(path_repo_1)
##' dir.create(path_repo_2)
##'
##' bare_repo <- init(path_bare, bare = TRUE)
##' repo_1 <- clone(path_bare, path_repo_1)
##' repo_2 <- clone(path_bare, path_repo_2)
##'
##' config(repo_1, user.name = "Alice", user.email = "alice@@example.org")
##' config(repo_2, user.name = "Bob", user.email = "bob@@example.org")
##'
##' ## Add changes to repo 1
##' writeLines("Lorem ipsum dolor sit amet",
##'            con = file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "Commit message")
##'
##' ## Push changes from repo 1 to origin (bare_repo)
##' push(repo_1, "origin", "refs/heads/master")
##'
##' ## Fetch changes from origin (bare_repo) to repo 2
##' fetch(repo_2, "origin")
##'
##' ## List updated heads
##' fetch_heads(repo_2)
##'
##' ## Checking out GitHub pull requests locally
##' path <- tempfile(pattern="ghit-")
##' repo <- clone("https://github.com/leeper/ghit", path)
##' fetch(repo, "origin", refspec = "pull/13/head:refs/heads/BRANCHNAME")
##' checkout(repo, "BRANCHNAME")
##' summary(repo)
##' }
fetch <- function(repo = ".", name = NULL, credentials = NULL,
                  verbose = TRUE, refspec = NULL) {
    invisible(.Call(git2r_remote_fetch, lookup_repository(repo),
                    name, credentials, "fetch", verbose, refspec))
}

##' Get updated heads during the last fetch.
##'
##' @template repo-param
##' @return list with \code{git_fetch_head} entries. NULL if there is
##'     no FETCH_HEAD file.
##' @export
##' @examples
##' \dontrun{
##' ## Initialize three temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo_1 <- tempfile(pattern="git2r-")
##' path_repo_2 <- tempfile(pattern="git2r-")
##'
##' dir.create(path_bare)
##' dir.create(path_repo_1)
##' dir.create(path_repo_2)
##'
##' bare_repo <- init(path_bare, bare = TRUE)
##' repo_1 <- clone(path_bare, path_repo_1)
##' repo_2 <- clone(path_bare, path_repo_2)
##'
##' config(repo_1, user.name = "Alice", user.email = "alice@@example.org")
##' config(repo_2, user.name = "Bob", user.email = "bob@@example.org")
##'
##' ## Add changes to repo 1
##' writeLines("Lorem ipsum dolor sit amet",
##'            con = file.path(path_repo_1, "example.txt"))
##' add(repo_1, "example.txt")
##' commit(repo_1, "Commit message")
##'
##' ## Push changes from repo 1 to origin (bare_repo)
##' push(repo_1, "origin", "refs/heads/master")
##'
##' ## Fetch changes from origin (bare_repo) to repo 2
##' fetch(repo_2, "origin")
##'
##' ## List updated heads
##' fetch_heads(repo_2)
##' }
fetch_heads <- function(repo = ".")  {
    .Call(git2r_repository_fetch_heads, lookup_repository(repo))
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Status case 1
status_exp_1 <- structure(list(staged = empty_named_list(),
                               unstaged = empty_named_list(),
                               untracked = empty_named_list()),
                          class = "git_status")
status_obs_1 <- status(repo)
stopifnot(identical(print(status_obs_1), status_obs_1))
str(status_exp_1)
str(status_obs_1)
stopifnot(identical(status_obs_1, status_exp_1))
stopifnot(identical(capture.output(status(repo)),
                    "working directory clean"))

## Status case 2, include ignored files
status_exp_2 <- structure(list(staged = empty_named_list(),
                               unstaged = empty_named_list(),
                               untracked = empty_named_list(),
                               ignored = empty_named_list()),
                          class = "git_status")
status_obs_2 <- status(repo, ignored = TRUE)
status_obs_2
str(status_exp_2)
str(status_obs_2)
stopifnot(identical(status_obs_2, status_exp_2))
stopifnot(identical(capture.output(status(repo, ignored = TRUE)),
                    "working directory clean"))

## Create 4 files
writeLines("File-1", file.path(path, "test-1.txt"))
writeLines("File-2", file.path(path, "test-2.txt"))
writeLines("File-3", file.path(path, "test-3.txt"))
writeLines("File-4", file.path(path, "test-4.txt"))

## Status case 3: 4 untracked files
status_exp_3 <- structure(list(staged = empty_named_list(),
                               unstaged = empty_named_list(),
                               untracked = list(untracked = "test-1.txt",
                                                untracked = "test-2.txt",
                                                untracked = "test-3.txt",
                                                untracked = "test-4.txt")),
                          class = "git_status")
status_obs_3 <- status(repo)
status_obs_3
str(status_exp_3)
str(status_obs_3)
stopifnot(identical(status_obs_3, status_exp_3))

## Add file 1 and 2 to the repository and commit
add(repo, c("test-1.txt", "test-2.txt"))
commit(repo, "Commit message")

## Status case 4: 2 untracked files
status_exp_4 <- structure(list(staged = empty_named_list(),
                               unstaged = empty_named_list(),
                               untracked = list(untracked = "test-3.txt",
                                                untracked = "test-4.txt")),
                          class = "git_status")
status_obs_4 <- status(repo)
status_obs_4
str(status_exp_4)
str(status_obs_4)
stopifnot(identical(status_obs_4, status_exp_4))

## Update file 1 & 2
writeLines(c("File-1", "Hello world"), file.path(path, "test-1.txt"))
writeLines(c("File-2", "Hello world"), file.path(path, "test-2.txt"))

## Add file 1
add(repo, "test-1.txt")

## Status case 5: 1 staged file, 1 unstaged file and 2 untracked files
status_exp_5 <- structure(list(staged = list(modified = "test-1.txt"),
                               unstaged = list(modified = "test-2.txt"),
                               untracked = list(untracked = "test-3.txt",
                                                untracked = "test-4.txt")),
                          class = "git_status")
status_obs_5 <- status(repo)
status_obs_5
str(status_exp_5)
str(status_obs_5)
stopifnot(identical(status_obs_5, status_exp_5))

## Add .gitignore file with file test-4.txt
writeLines("test-4.txt", file.path(path, ".gitignore"))

## Status case 6: 1 staged file, 1 unstaged file, 2 untracked files
## and 1 ignored file
status_exp_6 <- structure(list(staged = list(modified = "test-1.txt"),
                               unstaged = list(modified = "test-2.txt"),
                               untracked = list(untracked = ".gitignore",
                                                untracked = "test-3.txt"),
                               ignored = list(ignored = "test-4.txt")),
                          class = "git_status")
status_obs_6 <- status(repo, ignore = TRUE)
status_obs_6
str(status_exp_6)
str(status_obs_6)
stopifnot(identical(status_obs_6, status_exp_6))

## Cleanup
unlink(path, recursive = TRUE)

# We need https
if(getRversion() < "3.3.0") setInternet2()
VERSION <- commandArgs(TRUE)

# Downloads libssh2 + dependencies
if(!file.exists(sprintf("../windows/libgit2-%s/include/git2.h", VERSION))){
  download.file(sprintf("https://github.com/rwinlib/libgit2/archive/v%s.zip", VERSION), "lib.zip", quiet = TRUE)
  dir.create("../windows", showWarnings = FALSE)
  unzip("lib.zip", exdir = "../windows")
  unlink("lib.zip")
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2018 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Check when method
w1 <- structure(list(time = 1395567947, offset = 60),
               class = "git_time")
stopifnot(identical(when(w1), "2014-03-23 09:45:47 GMT"))
stopifnot(identical(when(w1, usetz = FALSE), "2014-03-23 09:45:47"))
stopifnot(identical(when(w1, tz = "Europe/Stockholm", origin = "1980-02-02"),
                    "2024-04-23 11:45:47 CEST"))

s1 <- structure(list(name = "Alice", email = "alice@example.org", when = w1),
                class = "git_signature")
stopifnot(identical(when(s1), "2014-03-23 09:45:47 GMT"))
stopifnot(identical(when(s1, usetz = FALSE), "2014-03-23 09:45:47"))
stopifnot(identical(when(s1, tz = "Europe/Stockholm", origin = "1980-02-02"),
                    "2024-04-23 11:45:47 CEST"))

w2 <- structure(list(time = 1395567950, offset = 60),
               class = "git_time")
s2 <- structure(list(name = "Alice", email = "alice@example.org", when = w2),
                class = "git_signature")
c1 <- structure(list(sha = "166f3f779fd7e4165aaa43f2828050ce040052b0",
                     author = s1,
                     committer = s2,
                     summary = "A commit summary",
                     message = "A commit message"),
                class = "git_commit")
stopifnot(identical(when(c1), "2014-03-23 09:45:47 GMT"))
stopifnot(identical(when(c1, usetz = FALSE), "2014-03-23 09:45:47"))
stopifnot(identical(when(c1, tz = "Europe/Stockholm", origin = "1980-02-02"),
                    "2024-04-23 11:45:47 CEST"))

t1 <- structure(list(sha = "166f3f779fd7e4165aaa43f2828050ce040052b0",
                     message = "A tag message",
                     name = "A tage name",
                     tagger = s1,
                     target = "166f3f779fd7e4165aaa43f2828050ce040052b0"),
                class = "git_tag")
stopifnot(identical(when(t1), "2014-03-23 09:45:47 GMT"))
stopifnot(identical(when(t1, usetz = FALSE), "2014-03-23 09:45:47"))
stopifnot(identical(when(t1, tz = "Europe/Stockholm", origin = "1980-02-02"),
                    "2024-04-23 11:45:47 CEST"))

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Brief summary of signature
##'
##' @aliases show,git_signature-methods
##' @docType methods
##' @param object The repository \code{object}
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Brief summary of default signature
##' default_signature(repo)
##' }
setMethod("show",
          signature(object = "git_signature"),
          function(object)
          {
              cat(sprintf(paste0("name:  %s\n",
                                 "email: %s\n",
                                 "when:  %s\n"),
                          object@name,
                          object@email,
                          as(object@when, "character")))
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Create a branch
##'
##' @rdname branch_create-methods
##' @docType methods
##' @param commit Commit to which branch should point.
##' @param name Name for the branch
##' @param force Overwrite existing branch. Default = FALSE
##' @return invisible S4 class git_branch object
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##'
##' ## Create a branch
##' branch_1 <- branch_create(commit_1, name = "test-branch")
##'
##' ## Add one more commit
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_2 <- commit(repo, "Another commit message")
##'
##' ## Create a branch with the same name should fail
##' try(branch_create(commit_2, name = "test-branch"), TRUE)
##'
##' ## Force it
##' branch_2 <- branch_create(commit_2, name = "test-branch", force = TRUE)
##' }
setGeneric("branch_create",
           signature = "commit",
           function(commit, name, force = FALSE)
           standardGeneric("branch_create"))

##' @rdname branch_create-methods
##' @export
setMethod("branch_create",
          signature = "git_commit",
          function(commit, name, force)
          {
              invisible(.Call(git2r_branch_create, name, commit, force))
          }
)

##' Delete a branch
##'
##' @rdname branch_delete-methods
##' @docType methods
##' @param branch The branch
##' @return invisible NULL
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##'
##' ## Create a 'dev' branch
##' dev <- branch_create(commit_1, name = "dev")
##' branches(repo)
##'
##' ## Delete 'dev' branch
##' branch_delete(dev)
##' branches(repo)
##' }
setGeneric("branch_delete",
           signature = "branch",
           function(branch)
           standardGeneric("branch_delete"))

##' @rdname branch_delete-methods
##' @export
setMethod("branch_delete",
          signature = "git_branch",
          function(branch)
          {
              invisible(.Call(git2r_branch_delete, branch))
          }
)

##' Remote name of a branch
##'
##' The name of remote that the remote tracking branch belongs to
##' @rdname branch_remote_name-methods
##' @docType methods
##' @param branch The branch
##' @return character string with remote name
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize two temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## Get remote name
##' branch_remote_name(branches(repo)[[2]])
##' }
setGeneric("branch_remote_name",
           signature = "branch",
           function(branch)
           standardGeneric("branch_remote_name"))

##' @rdname branch_remote_name-methods
##' @export
setMethod("branch_remote_name",
          signature = "git_branch",
          function(branch)
          {
              .Call(git2r_branch_remote_name, branch)
          }
)

##' Remote url of a branch
##'
##' @rdname branch_remote_url-methods
##' @docType methods
##' @param branch The branch
##' @return character string with remote url
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize two temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## Get remote url of tracking branch to branch 'master'
##' branch_remote_url(branch_get_upstream(head(repo)))
##' }
setGeneric("branch_remote_url",
           signature = "branch",
           function(branch)
           standardGeneric("branch_remote_url"))

##' @rdname branch_remote_url-methods
##' @export
setMethod("branch_remote_url",
          signature = "git_branch",
          function(branch)
          {
              .Call(git2r_branch_remote_url, branch)
          }
)

##' Rename a branch
##'
##' @rdname branch_rename-methods
##' @docType methods
##' @param branch Branch to rename
##' @param name The new name for the branch
##' @param force Overwrite existing branch. Default is FALSE
##' @return invisible renamed S4 class git_branch
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Rename 'master' branch to 'dev'
##' branches(repo)
##' branch_rename(head(repo), "dev")
##' branches(repo)
##' }
setGeneric("branch_rename",
           signature = "branch",
           function(branch, name, force = FALSE)
           standardGeneric("branch_rename"))

##' @rdname branch_rename-methods
##' @export
setMethod("branch_rename",
          signature = "git_branch",
          function(branch, name, force)
          {
              invisible(.Call(git2r_branch_rename, branch, name, force))
          }
)

##' Get target (sha) pointed to by a branch
##'
##' @rdname branch_target-methods
##' @docType methods
##' @param branch The branch
##' @return sha or NA if not a direct reference
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Get target (sha) pointed to by 'master' branch
##' branch_target(head(repo))
##' }
setGeneric("branch_target",
           signature = "branch",
           function(branch)
           standardGeneric("branch_target"))

##' @rdname branch_target-methods
##' @export
setMethod("branch_target",
          signature = "git_branch",
          function(branch)
          {
              .Call(git2r_branch_target, branch)
          }
)

##' Get remote tracking branch
##'
##' Get remote tracking branch, given a local branch.
##' @rdname branch_get_upstream-methods
##' @docType methods
##' @param branch The branch
##' @return S4 class git_branch or NULL if no remote tracking branch.
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize two temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## Get remote tracking branch
##' branch_get_upstream(head(repo))
##' }
setGeneric("branch_get_upstream",
           signature = "branch",
           function(branch)
           standardGeneric("branch_get_upstream"))

##' @rdname branch_get_upstream-methods
##' @export
setMethod("branch_get_upstream",
          signature = "git_branch",
          function(branch)
          {
              .Call(git2r_branch_get_upstream, branch)
          }
)

##' Set remote tracking branch
##'
##' Set the upstream configuration for a given local branch
##' @rdname branch_set_upstream-methods
##' @docType methods
##' @param branch The branch to configure
##' @param name remote-tracking or local branch to set as
##' upstream. Pass NULL to unset.
##' @return invisible NULL
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize two temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## Unset remote remote tracking branch
##' branch_get_upstream(head(repo))
##' branch_set_upstream(head(repo), NULL)
##' branch_get_upstream(head(repo))
##'
##' ## Set remote tracking branch
##' branch_set_upstream(head(repo), "origin/master")
##' branch_get_upstream(head(repo))
##' }
setGeneric("branch_set_upstream",
           signature = "branch",
           function(branch, name)
           standardGeneric("branch_set_upstream"))

##' @rdname branch_set_upstream-methods
##' @export
setMethod("branch_set_upstream",
          signature(branch = "git_branch"),
          function(branch, name)
          {
              if (missing(name)) {
                  stop("Missing argument name")
              }
              invisible(.Call(git2r_branch_set_upstream, branch, name))
          }
)

##' Branches
##'
##' List branches in repository
##' @rdname branches-methods
##' @docType methods
##' @param repo The repository \code{object}
##' \code{\linkS4class{git_repository}}. If the \code{repo} argument
##' is missing, the repository is searched for with
##' \code{\link{discover_repository}} in the current working
##' directory.
##' @param flags Filtering flags for the branch listing. Valid values
##' are 'all', 'local' or 'remote'
##' @return list of branches in repository
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config first user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## List branches
##' branches(repo)
##' }
setGeneric("branches",
           signature = "repo",
           function(repo, flags=c("all", "local", "remote"))
           standardGeneric("branches"))

##' @rdname branches-methods
##' @export
setMethod("branches",
          signature(repo = "missing"),
          function(flags)
          {
              callGeneric(repo = lookup_repository(), flags = flags)
          }
)

##' @rdname branches-methods
##' @export
setMethod("branches",
          signature(repo = "git_repository"),
          function(repo, flags)
          {
              flags <- switch(match.arg(flags),
                              local  = 1L,
                              remote = 2L,
                              all    = 3L)

              .Call(git2r_branch_list, repo, flags)
          }
)

##' Check if branch is head
##'
##' @rdname is_head-methods
##' @docType methods
##' @param branch The branch \code{object} to check if it's head
##' @return TRUE if branch is head, else FALSE
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## List branches
##' branches(repo)
##'
##' ## Check that 'master' is_head
##' master <- branches(repo)[[1]]
##' is_head(master)
##'
##' ## Create and checkout 'dev' branch
##' checkout(repo, "dev", create = TRUE)
##'
##' ## List branches
##' branches(repo)
##'
##' ## Check that 'master' is no longer head
##' is_head(master)
##' }
setGeneric("is_head",
           signature = "branch",
           function(branch)
           standardGeneric("is_head"))

##' @rdname is_head-methods
##' @export
setMethod("is_head",
          signature(branch = "git_branch"),
          function(branch)
          {
              .Call(git2r_branch_is_head, branch)
          }
)

##' Check if branch is local
##'
##' @rdname is_local-methods
##' @docType methods
##' @param branch The branch \code{object} to check if it's local
##' @return TRUE if branch is local, else FALSE
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Initialize repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config first user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## List branches
##' branches(repo)
##'
##' ## Check if first branch is_local
##' is_local(branches(repo)[[1]])
##'
##' ## Check if second branch is_local
##' is_local(branches(repo)[[2]])
##' }
setGeneric("is_local",
           signature = "branch",
           function(branch)
           standardGeneric("is_local"))

##' @rdname is_local-methods
##' @export
setMethod("is_local",
          signature(branch = "git_branch"),
          function(branch)
          {
              identical(branch@type, 1L)
          }
)

##' Brief summary of branch
##'
##' @aliases show,git_branch-methods
##' @docType methods
##' @param object The branch \code{object}
##' @return None (invisible 'NULL').
##' @keywords methods
##' @include S4_classes.r
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Brief summary of the branch in the repository
##' branches(repo)[[1]]
##' }
setMethod("show",
          signature(object = "git_branch"),
          function(object)
          {
              sha <- branch_target(object)
              if (!is.na(sha)) {
                  cat(sprintf("[%s] ", substr(sha, 1 , 6)))
              }

              if (is_local(object)) {
                  cat("(Local) ")
              } else {
                  cat(sprintf("(%s @ %s) ",
                              branch_remote_name(object),
                              branch_remote_url(object)))
              }

              if (is_head(object)) {
                  cat("(HEAD) ")
              }

              if (is_local(object)) {
                  cat(sprintf("%s\n", object@name))
              } else {
                  cat(sprintf("%s\n",
                              substr(object@name,
                                     start = nchar(branch_remote_name(object)) + 2,
                                     stop = nchar(object@name))))
              }
          }
)

##' Check if object is S4 class git_branch
##'
##' @param object Check if object is S4 class git_branch
##' @return TRUE if object is S4 class git_branch, else FALSE
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' branch <- branches(repo)[[1]]
##'
##' ## Check if branch
##' is_branch(branch)
##' }
is_branch <- function(object) {
    methods::is(object = object, class2 = "git_branch")
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Default notes reference
##'
##' Get the default notes reference for a repository
##' @template repo-param
##' @return Character vector of length one with name of default notes
##'     reference
##' @export
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## View default notes reference
##' note_default_ref(repo)
##' }
note_default_ref <- function(repo = ".") {
    .Call(git2r_note_default_ref, lookup_repository(repo))
}

##' Add note for a object
##'
##' @param object The object to annotate (git_blob, git_commit or
##'     git_tree).
##' @param message Content of the note to add
##' @param ref Canonical name of the reference to use. Default is
##'     \code{note_default_ref}.
##' @param author Signature of the notes note author
##' @param committer Signature of the notes note committer
##' @param force Overwrite existing note. Default is FALSE
##' @return git_note
##' @export
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##'
##' ## Create another commit
##' writeLines(c("Hello world!",
##'              "HELLO WORLD!"),
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_2 <- commit(repo, "Commit message 2")
##'
##' ## Check that notes is an empty list
##' notes(repo)
##'
##' ## Create note in default namespace
##' note_create(commit_1, "Note-1")
##'
##' ## Create note in named (review) namespace
##' note_create(commit_1, "Note-2", ref="refs/notes/review")
##' note_create(commit_2, "Note-3", ref="review")
##'
##' ## Create note on blob and tree
##' note_create(tree(commit_1), "Note-4")
##' note_create(tree(commit_1)["example.txt"], "Note-5")
##' }
note_create <- function(object    = NULL,
                        message   = NULL,
                        ref       = NULL,
                        author    = NULL,
                        committer = NULL,
                        force     = FALSE) {
    if (is.null(object))
        stop("'object' is missing")
    if (!any(is_blob(object), is_commit(object), is_tree(object)))
        stop("'object' must be a 'git_blob', 'git_commit' or 'git_tree' object")

    repo <- object$repo
    sha <- object$sha
    if (is.null(ref))
        ref <- note_default_ref(repo)
    stopifnot(is.character(ref), identical(length(ref), 1L))
    if (!length(grep("^refs/notes/", ref)))
        ref <- paste0("refs/notes/", ref)
    if (is.null(author))
        author <- default_signature(repo)
    if (is.null(committer))
        committer <- default_signature(repo)
    .Call(git2r_note_create, repo, sha, message, ref, author, committer, force)
}

##' List notes
##'
##' List all the notes within a specified namespace.
##' @template repo-param
##' @param ref Reference to read from. Default (ref = NULL) is to call
##'     \code{note_default_ref}.
##' @return list with git_note objects
##' @export
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##'
##' ## Create another commit
##' writeLines(c("Hello world!",
##'              "HELLO WORLD!"),
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_2 <- commit(repo, "Commit message 2")
##'
##' ## Create note in default namespace
##' note_create(commit_1, "Note-1")
##' note_create(commit_1, "Note-2", force = TRUE)
##'
##' ## Create note in named (review) namespace
##' note_create(commit_1, "Note-3", ref="refs/notes/review")
##' note_create(commit_2, "Note-4", ref="review")
##'
##' ## Create note on blob and tree
##' note_create(tree(commit_1), "Note-5")
##' note_create(tree(commit_1)["example.txt"], "Note-6")
##'
##' ## List notes in default namespace
##' notes(repo)
##'
##' ## List notes in 'review' namespace
##' notes(repo, "review")
##' }
notes <- function(repo = ".", ref = NULL) {
    repo <- lookup_repository(repo)
    if (is.null(ref))
        ref <- note_default_ref(repo)
    stopifnot(is.character(ref), identical(length(ref), 1L))
    if (!length(grep("^refs/notes/", ref)))
        ref <- paste0("refs/notes/", ref)
    .Call(git2r_notes, repo, ref)
}

##' Remove the note for an object
##'
##' @param note The note to remove
##' @param author Signature of the notes commit author.
##' @param committer Signature of the notes commit committer.
##' @return invisible NULL
##' @export
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##'
##'
##' ## Create note in default namespace
##' note_1 <- note_create(commit_1, "Note-1")
##'
##' ## Create note in named (review) namespace
##' note_2 <- note_create(commit_1, "Note-2", ref="refs/notes/review")
##'
##' ## List notes in default namespace
##' notes(repo)
##'
##' ## List notes in 'review' namespace
##' notes(repo, "review")
##'
##' ## Remove notes
##' note_remove(note_1)
##' note_remove(note_2)
##'
##' ## List notes in default namespace
##' notes(repo)
##'
##' ## List notes in 'review' namespace
##' notes(repo, "review")
##' }
note_remove <- function(note      = NULL,
                        author    = NULL,
                        committer = NULL) {
    if (!inherits(note, "git_note"))
        stop("'note' is not a git_note")
    if (is.null(author))
        author <- default_signature(note$repo)
    if (is.null(committer))
        committer <- default_signature(note$repo)
    .Call(git2r_note_remove, note, author, committer)
    invisible(NULL)
}

##' @export
format.git_note <- function(x, ...) {
    sprintf("note:  %s", x$sha)
}

##' @export
print.git_note <- function(x, ...) {
    cat(format(x, ...), "\n", sep = "")
    invisible(x)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test.txt"))

## add and commit
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message")

## Check branch
stopifnot(identical(length(branches(repo)), 1L))
stopifnot(identical(is_head(branches(repo)[[1]]), TRUE))
stopifnot(identical(is_local(branches(repo)[[1]]), TRUE))
stopifnot(identical(branches(repo)[[1]]$name, "master"))
stopifnot(identical(branches(repo)[[1]], repository_head(repo)))
stopifnot(identical(branches(repo)$master, repository_head(repo)))

## Check branch argument
res <- tools::assertError(is_local(1))
stopifnot(length(grep("argument 'branch' must be a 'git_branch' object",
                      res[[1]]$message)) > 0)

## Check branch name argument
res <- tools::assertError(branch_set_upstream())
stopifnot(length(grep("Missing argument name",
                      res[[1]]$message)) > 0)

## Print branch
b <- repository_head(repo)
stopifnot(identical(print(b), b))

## Create a branch
b <- branch_create(commit_1, name = "test")
stopifnot(identical(b$name, "test"))
stopifnot(identical(b$type, 1L))
stopifnot(identical(sha(b), branch_target(b)))
stopifnot(identical(length(branches(repo)), 2L))
stopifnot(identical(branch_target(branches(repo)[[1]]),
                    branch_target(branches(repo)[[2]])))

## Check is_branch
stopifnot(identical(is_branch(b), TRUE))
stopifnot(identical(is_branch(5), FALSE))

## Add one more commit
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Another commit message")

## Now the first branch should have moved on
stopifnot(!identical(branch_target(branches(repo)[[1]]),
                     branch_target(branches(repo)[[2]])))

## Create a branch with the same name should fail
tools::assertError(branch_create(commit_2, name = "test"))

## Force it and check the branches are identical again
b <- branch_create(commit_2, name = "test", force = TRUE)
stopifnot(identical(branch_target(branches(repo)[[1]]),
                    branch_target(branches(repo)[[2]])))

## Test arguments
res <- tools::assertError(.Call(git2r:::git2r_branch_delete, NULL))
stopifnot(length(grep("'branch' must be an S3 class git_branch",
                      res[[1]]$message)) > 0)
res <- tools::assertError(.Call(git2r:::git2r_branch_delete, 3))
stopifnot(length(grep("'branch' must be an S3 class git_branch",
                      res[[1]]$message)) > 0)
res <- tools::assertError(.Call(git2r:::git2r_branch_delete, repo))
stopifnot(length(grep("'branch' must be an S3 class git_branch",
                      res[[1]]$message)) > 0)
b_tmp <- b
b_tmp$name <- NA_character_
res <- tools::assertError(.Call(git2r:::git2r_branch_delete, b_tmp))
stopifnot(length(grep("'branch' must be an S3 class git_branch",
                      res[[1]]$message)) > 0)
b_tmp <- b
b_tmp$type <- NA_integer_
res <- tools::assertError(.Call(git2r:::git2r_branch_delete, b_tmp))
stopifnot(length(grep("'branch' must be an S3 class git_branch",
                      res[[1]]$message)) > 0)
b_tmp$type <- 3L
res <- tools::assertError(.Call(git2r:::git2r_branch_delete, b_tmp))
stopifnot(length(grep("'branch' must be an S3 class git_branch",
                      res[[1]]$message)) > 0)

## Delete branch
branch_delete(b)
stopifnot(identical(length(branches(repo)), 1L))

## Add one more commit
writeLines(c("Hello world!", "HELLO WORLD!", "hello world"),
           file.path(path, "test.txt"))
add(repo, "test.txt")
commit_3 <- commit(repo, "Another third commit message")

## Create and test renaming of branches
b_1 <- branch_create(commit_1, name = "test-1")
b_2 <- branch_create(commit_2, name = "test-2")
b_3 <- branch_create(commit_3, name = "test-3")
stopifnot(identical(length(branches(repo)), 4L))
b_1 <- branch_rename(b_1, name = "test-1-new-name")
stopifnot(identical(length(branches(repo)), 4L))
stopifnot(identical(b_1$name, "test-1-new-name"))
tools::assertError(branch_rename(b_1, name = "test-2"))
branch_rename(b_1, name = "test-2", force = TRUE)
stopifnot(identical(length(branches(repo)), 3L))

## Check branches method with missing repo argument
wd <- setwd(path)
stopifnot(identical(length(branches()), 3L))
if (!is.null(wd))
    setwd(wd)

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library(git2r)
source("util/check.R")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message 1")

## Create first branch, checkout, add file and commit
b_1 <- branch_create(commit_1, "branch1")
checkout(b_1)
writeLines("Branch 1", file.path(path, "branch-1.txt"))
add(repo, "branch-1.txt")
commit_2 <- commit(repo, "Commit message branch 1")

## Create second branch, checkout, add file and commit
b_2 <- branch_create(commit_1, "branch2")
checkout(b_2)
writeLines("Branch 2", file.path(path, "branch-2.txt"))
add(repo, "branch-2.txt")
commit_3 <- commit(repo, "Commit message branch 2")
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit_4 <- commit(repo, "Second commit message branch 2")

## Check that merge base equals commit_1
stopifnot(identical(merge_base(commit_2, commit_3), commit_1))

## Checkout master
b <- branches(repo)
checkout(b[sapply(b, "[", "name") == "master"][[1]], force = TRUE)

## Merge branch 1
m_1 <- merge(b[sapply(b, "[", "name") == "branch1"][[1]])
stopifnot(identical(m_1$fast_forward, TRUE))
stopifnot(identical(m_1$conflicts, FALSE))
stopifnot(identical(sha(m_1), NA_character_))
stopifnot(identical(print(m_1), m_1))

## Merge branch 1 again
m_1_again <- merge(b[sapply(b, "[", "name") == "branch1"][[1]])
stopifnot(identical(m_1_again$up_to_date, TRUE))
stopifnot(identical(m_1_again$fast_forward, FALSE))
stopifnot(identical(m_1_again$conflicts, FALSE))
stopifnot(identical(sha(m_1_again), NA_character_))

## Merge branch 2
m_2 <- merge(b[sapply(b, "[", "name") == "branch2"][[1]])
stopifnot(identical(m_2$fast_forward, FALSE))
stopifnot(identical(m_2$conflicts, FALSE))
stopifnot(identical(sha(m_2), sha(commits(repo)[[1]])))

## Create third branch, checkout, change file and commit
b_3 <- branch_create(lookup(repo, sha(m_2)), "branch3")
checkout(b_3)
writeLines(c("Lorem ipsum dolor amet sit, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message branch 3")

## Checkout master and create a change that creates a conflict on
## merge
b <- branches(repo)
checkout(b[sapply(b, "[", "name") == "master"][[1]], force = TRUE)
writeLines(c("Lorem ipsum dolor sit amet, adipisicing consectetur elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Some commit message branch 1")

## Merge branch 3 with fail = TRUE
m_3 <- merge(b[sapply(b, "[", "name") == "branch3"][[1]], fail = TRUE)
stopifnot(identical(m_3$up_to_date, FALSE))
stopifnot(identical(m_3$fast_forward, FALSE))
stopifnot(identical(m_3$conflicts, TRUE))
stopifnot(identical(sha(m_3), NA_character_))
m_3

## Check status; Expect to have a clean working directory
wd <- structure(list(staged = empty_named_list(),
                     unstaged = empty_named_list(),
                     untracked = empty_named_list()),
                class = "git_status")
stopifnot(identical(status(repo), wd))

## Merge branch 3
m_3 <- merge(b[sapply(b, "[", "name") == "branch3"][[1]])
stopifnot(identical(m_3$up_to_date, FALSE))
stopifnot(identical(m_3$fast_forward, FALSE))
stopifnot(identical(m_3$conflicts, TRUE))
stopifnot(identical(sha(m_3), NA_character_))
m_3

## Check status; Expect to have one unstaged unmerged conflict.
stopifnot(identical(status(repo),
                    structure(list(staged = empty_named_list(),
                                   unstaged = list(conflicted = "test.txt"),
                                   untracked = empty_named_list()),
                              class = "git_status")))

## Cleanup
unlink(path, recursive = TRUE)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Plot commits over time
##'
##' @rdname plot-methods
##' @aliases plot
##' @aliases plot-methods
##' @docType methods
##' @param x The repository to plot
##' @param breaks Default is \code{month}. Change to year, quarter,
##' week or day as necessary.
##' @param main Default title for the plot is "Commits on repo:" and
##' repository workdir basename. Supply a new title if you desire one.
##' @param ... Additional arguments affecting the plot
##' @keywords methods
##' @include S4_classes.r
##' @importFrom graphics axis
##' @importFrom graphics barplot
##' @export
##' @examples
##' \dontrun{
##' ## Initialize repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- clone("https://github.com/ropensci/git2r.git", path)
##'
##' ## Plot commits
##' plot(repo)
##' }
setMethod("plot",
          signature(x = "git_repository"),
          function(x,
                   breaks = c("month", "year", "quarter", "week", "day"),
                   main = NULL,
                   ...)
          {
              breaks = match.arg(breaks)

              df <- contributions(x, breaks = breaks, by = "commits")
              tmp <- data.frame(when = seq(from = min(df$when),
                                           to   = max(df$when),
                                           by   = breaks),
                                n = 0)
              i <- match(df$when, tmp$when)
              tmp$n[i] <- df$n
              df <- tmp

              xlab <- sprintf("Time [%s]", breaks)
              ylab <- "Number of commits"
              if (is.null(main)) {
                  if (is_bare(x)) {
                      main <- "Commits"
                  } else {
                      main <- sprintf("Commits on repository: %s",
                                      basename(workdir(x)))
                  }
              }

              mp <- graphics::barplot(df$n, xlab = xlab, ylab = ylab,
                                      main = main, ...)
              graphics::axis(1, at = mp, labels = seq(min(df$when),
                                             max(df$when), breaks))
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

stash_index <- function(object, index) {
    if (inherits(object, "git_stash")) {
        ## Determine the index of the stash in the stash list
        index <- match(object$sha, vapply(stash_list(object$repo),
                                          "[[", character(1), "sha"))
    }

    ## The stash list is zero-based
    if (abs(index - round(index)) >= .Machine$double.eps^0.5)
        stop("'index' must be an integer")
    as.integer(index) - 1L
}

stash_object <- function(object) {
    if (inherits(object, "git_stash"))
        return(object$repo)
    lookup_repository(object)
}

##' Apply stash
##'
##' Apply a single stashed state from the stash list.
##'
##' If local changes in the working directory conflict with changes in
##' the stash then an error will be raised. In this case, the index
##' will always remain unmodified and all files in the working
##' directory will remain unmodified. However, if you are restoring
##' untracked files or ignored files and there is a conflict when
##' applying the modified files, then those files will remain in the
##' working directory.
##' @param object path to a repository, or a \code{git_repository}
##'     object, or the stash \code{object} to pop. Default is a
##'     \code{path = '.'} to a reposiory.
##' @param index The index to the stash to apply. Only used when
##'     \code{object} is a path to a repository or a
##'     \code{git_repository} object. Default is \code{index = 1}.
##' @return invisible NULL
##' @export
##' @examples \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' # Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' add(repo, 'test.txt')
##' commit(repo, "Commit message")
##'
##' # Change file
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
##'
##' # Create stash in repository
##' stash(repo)
##'
##' # Change file
##' writeLines(c("Hello world!", "HeLlO wOrLd!"), file.path(path, "test.txt"))
##'
##' # Create stash in repository
##' stash(repo)
##'
##' # View stashes
##' stash_list(repo)
##'
##' # Read file
##' readLines(file.path(path, "test.txt"))
##'
##' # Apply latest git_stash object in repository
##' stash_apply(stash_list(repo)[[1]])
##'
##' # Read file
##' readLines(file.path(path, "test.txt"))
##'
##' # View stashes
##' stash_list(repo)
##' }
stash_apply <- function(object = ".", index = 1) {
    .Call(git2r_stash_apply, stash_object(object), stash_index(object, index))
    invisible(NULL)
}

##' Drop stash
##'
##' @param object path to a repository, or a \code{git_repository}
##'     object, or the stash \code{object} to drop. Default is a
##'     \code{path = '.'} to a reposiory.
##' @param index The index to the stash to drop. Only used when
##'     \code{object} is a path to a repository or a
##'     \code{git_repository} object. Default is \code{index = 1}.
##' @return invisible NULL
##' @export
##' @examples \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' # Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' add(repo, 'test.txt')
##' commit(repo, "Commit message")
##'
##' # Change file
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
##'
##' # Create stash in repository
##' stash(repo)
##'
##' # Change file
##' writeLines(c("Hello world!", "HeLlO wOrLd!"), file.path(path, "test.txt"))
##'
##' # Create stash in repository
##' stash(repo)
##'
##' # View stashes
##' stash_list(repo)
##'
##' # Drop git_stash object in repository
##' stash_drop(stash_list(repo)[[1]])
##'
##' ## Drop stash using an index to stash
##' stash_drop(repo, 1)
##'
##' # View stashes
##' stash_list(repo)
##' }
stash_drop <- function(object = ".", index = 1) {
    .Call(git2r_stash_drop, stash_object(object), stash_index(object, index))
    invisible(NULL)
}

##' Stash
##'
##' @template repo-param
##' @param message Optional description. Defaults to current time.
##' @param index All changes already added to the index are left
##'     intact in the working directory. Default is FALSE
##' @param untracked All untracked files are also stashed and then
##'     cleaned up from the working directory. Default is FALSE
##' @param ignored All ignored files are also stashed and then cleaned
##'     up from the working directory. Default is FALSE
##' @param stasher Signature with stasher and time of stash
##' @return invisible \code{git_stash} object if anything to stash
##'     else NULL
##' @export
##' @examples \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' # Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' add(repo, 'test.txt')
##' commit(repo, "Commit message")
##'
##' # Change file
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
##'
##' # Check status of repository
##' status(repo)
##'
##' # Create stash in repository
##' stash(repo)
##'
##' # Check status of repository
##' status(repo)
##'
##' # View stash
##' stash_list(repo)
##' }
stash <- function(repo = ".",
                  message   = as.character(Sys.time()),
                  index     = FALSE,
                  untracked = FALSE,
                  ignored   = FALSE,
                  stasher   = NULL) {
    repo <- lookup_repository(repo)
    if (is.null(stasher))
        stasher <- default_signature(repo)
    invisible(.Call(git2r_stash_save, repo, message, index,
                    untracked, ignored, stasher))
}

##' List stashes in repository
##'
##' @template repo-param
##' @return list of stashes in repository
##' @export
##' @examples \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' # Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test-1.txt"))
##' add(repo, 'test-1.txt')
##' commit(repo, "Commit message")
##'
##' # Make one more commit
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-1.txt"))
##' add(repo, 'test-1.txt')
##' commit(repo, "Next commit message")
##'
##' # Create one more file
##' writeLines("Hello world!", file.path(path, "test-2.txt"))
##'
##' # Check that there are no stashes
##' stash_list(repo)
##'
##' # Stash
##' stash(repo)
##'
##' # Only untracked changes, therefore no stashes
##' stash_list(repo)
##'
##' # Stash and include untracked changes
##' stash(repo, "Stash message", untracked=TRUE)
##'
##' # View stash
##' stash_list(repo)
##' }
stash_list <- function(repo = ".") {
    .Call(git2r_stash_list, lookup_repository(repo))
}

##' Pop stash
##'
##' Apply a single stashed state from the stash list and remove it
##' from the list if successful.
##' @param object path to a repository, or a \code{git_repository}
##'     object, or the stash \code{object} to pop. Default is a
##'     \code{path = '.'} to a reposiory.
##' @param index The index to the stash to pop. Only used when
##'     \code{object} is a path to a repository or a
##'     \code{git_repository} object. Default is \code{index = 1}.
##' @return invisible NULL
##' @export
##' @examples \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' # Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' add(repo, 'test.txt')
##' commit(repo, "Commit message")
##'
##' # Change file
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
##'
##' # Create stash in repository
##' stash(repo)
##'
##' # Change file
##' writeLines(c("Hello world!", "HeLlO wOrLd!"), file.path(path, "test.txt"))
##'
##' # Create stash in repository
##' stash(repo)
##'
##' # View stashes
##' stash_list(repo)
##'
##' # Read file
##' readLines(file.path(path, "test.txt"))
##'
##' # Pop latest git_stash object in repository
##' stash_pop(stash_list(repo)[[1]])
##'
##' # Read file
##' readLines(file.path(path, "test.txt"))
##'
##' # View stashes
##' stash_list(repo)
##' }
stash_pop <- function(object = ".", index = 1) {
    .Call(git2r_stash_pop, stash_object(object), stash_index(object, index))
    invisible(NULL)
}

##' @export
format.git_stash <- function(x, ...) {
    x$message
}

##' @export
print.git_stash <- function(x, ...) {
    cat(format(x, ...), "\n", sep = "")
    invisible(x)
}

##' Summary of a stash
##'
##' @param object The stash \code{object}
##' @param ... Additional arguments affecting the summary produced.
##' @return None (invisible 'NULL').
##' @export
##' @examples \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' # Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' add(repo, 'test.txt')
##' commit(repo, "Commit message")
##'
##' # Change file
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
##'
##' # Create stash in repository
##' stash(repo, "Stash message")
##'
##' # View summary of stash
##' summary(stash_list(repo)[[1]])
##' }
summary.git_stash <- function(object, ...) {
    cat(sprintf(paste0("message: %s\n",
                       "stasher: %s <%s>\n",
                       "when:    %s\n",
                       "sha:     %s\n\n"),
                object$summary,
                object$author$name,
                object$author$email,
                as.character(object$author$when),
                object$sha))
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2016 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Determine previous branch name
##'
##' @param repo The repository.
##' @noRd
previous_branch_name <- function(repo)
{
    branch <- revparse_single(repo, "@{-1}")@sha

    branch <- sapply(references(repo), function(x) {
        ifelse(x@sha == branch, x@shorthand, NA_character_)
    })
    branch <- branch[vapply(branch, Negate(is.na), logical(1))]

    branch <- sapply(branches(repo, "local"), function(x) {
        ifelse(x@name %in% branch, x@name, NA_character_)
    })
    branch <- branch[vapply(branch, Negate(is.na), logical(1))]

    if (any(!is.character(branch), !identical(length(branch), 1L))) {
        stop("'branch' must be a character vector of length one")
    }

    branch
}

##' Checkout
##'
##' Update files in the index and working tree to match the content of
##' the tree pointed at by the treeish object (commit, tag or tree).
##' Checkout using the default GIT_CHECKOUT_SAFE_CREATE strategy
##' (force = FALSE) or GIT_CHECKOUT_FORCE (force = TRUE).
##' @rdname checkout-methods
##' @docType methods
##' @param object A repository, commit, tag or tree which content will
##' be used to update the working directory.
##' @param ... Additional arguments affecting the checkout
##' @param branch If object is a repository, the name of the branch to
##' check out. Default is NULL.
##' @param create If object is a repository, then branch is created if
##' doesn't exist.
##' @param force If TRUE, then make working directory match
##' target. This will throw away local changes. Default is FALSE.
##' @param path Limit the checkout operation to only certain
##' paths. This argument is only used if branch is NULL. Default is
##' NULL.
##' @return invisible NULL
##' @keywords methods
##' @include S4_classes.r
##' @examples
##' \dontrun{
##' ## Create directories and initialize repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo_1 <- tempfile(pattern="git2r-")
##' path_repo_2 <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo_1)
##' dir.create(path_repo_2)
##' repo_bare <- init(path_bare, bare = TRUE)
##'
##' ## Clone to repo 1 and config user
##' repo_1 <- clone(path_bare, path_repo_1)
##' config(repo_1, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Add changes to repo 1 and push to bare
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            con = file.path(path_repo_1, "test.txt"))
##' add(repo_1, "test.txt")
##' commit(repo_1, "First commit message")
##' push(repo_1, "origin", "refs/heads/master")
##'
##' ## Create and checkout 'dev' branch in repo 1
##' checkout(repo_1, "dev", create = TRUE)
##'
##' ## Add changes to 'dev' branch in repo 1 and push to bare
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
##'            con = file.path(path_repo_1, "test.txt"))
##' add(repo_1, "test.txt")
##' commit(repo_1, "Second commit message")
##' push(repo_1, "origin", "refs/heads/dev")
##'
##' ## Clone to repo 2
##' repo_2 <- clone(path_bare, path_repo_2)
##' config(repo_2, user.name="Bob", user.email="bob@@example.org")
##'
##' ## Read content of 'test.txt'
##' readLines(file.path(path_repo_2, "test.txt"))
##'
##' ## Checkout dev branch
##' checkout(repo_2, "dev")
##'
##' ## Read content of 'test.txt'
##' readLines(file.path(path_repo_2, "test.txt"))
##'
##' ## Edit "test.txt" in repo_2
##' writeLines("Hello world!", con = file.path(path_repo_2, "test.txt"))
##'
##' ## Check status
##' status(repo_2)
##'
##' ## Checkout "test.txt"
##' checkout(repo_2, path = "test.txt")
##'
##' ## Check status
##' status(repo_2)
##' }
setGeneric("checkout",
           signature = "object",
           function(object, ...)
           standardGeneric("checkout")
)

##' @rdname checkout-methods
##' @export
setMethod("checkout",
          signature(object = "git_repository"),
          function(object,
                   branch = NULL,
                   create = FALSE,
                   force = FALSE,
                   path = NULL,
                   ...)
          {
              if (!is.null(branch)) {
                  if (any(!is.character(branch), !identical(length(branch), 1L)))
                      stop("'branch' must be a character vector of length one")

                  if (is_empty(object)) {
                      if (!isTRUE(create))
                          stop(sprintf("'%s' did not match any branch", branch))
                      ref_name <- paste0("refs/heads/", branch)
                      .Call(git2r_repository_set_head, object, ref_name)
                  } else {
                      if (identical(branch, "-"))
                          branch <- previous_branch_name(object)

                      ## Check if branch exists in a local branch
                      lb <- branches(object, "local")
                      lb <- lb[vapply(lb, slot, character(1), "name") == branch]
                      if (length(lb)) {
                          checkout(lb[[1]], force = force)
                      } else {
                          ## Check if there exists exactly one remote
                          ## branch with a matching name.
                          rb <- branches(object, "remote")

                          ## Split remote/name to check for a unique name
                          name <- vapply(rb, function(x) {
                                      remote <- strsplit(x@name, "/")[[1]][1]
                                      sub(paste0("^", remote, "/"), "", x@name)
                                  },
                                  character(1))
                          i <- which(name == branch)
                          if (identical(length(i), 1L)) {
                              ## Create branch and track remote
                              commit <- lookup(object, branch_target(rb[[i]]))
                              branch <- branch_create(commit, branch)
                              branch_set_upstream(branch, rb[[i]]@name)
                              checkout(branch, force = force)
                          } else {
                              if (!isTRUE(create))
                                  stop(sprintf("'%s' did not match any branch", branch))

                              ## Create branch
                              commit <- lookup(object, branch_target(head(object)))
                              checkout(branch_create(commit, branch), force = force)
                          }
                      }
                  }
              } else if (!is.null(path)) {
                  .Call(git2r_checkout_path, object, path)
              } else {
                  stop("missing 'branch' or 'path' argument")
              }

              invisible(NULL)
          }
)

##' @rdname checkout-methods
##' @export
setMethod("checkout",
          signature(object = "git_branch"),
          function(object, force = FALSE)
          {
              ref_name <- paste0("refs/heads/", object@name)
              .Call(git2r_checkout_tree, object@repo, ref_name, force)
              .Call(git2r_repository_set_head, object@repo, ref_name)
              invisible(NULL)
          }
)

##' @rdname checkout-methods
##' @export
setMethod("checkout",
          signature(object = "git_commit"),
          function(object, force = FALSE)
          {
              .Call(git2r_checkout_tree, object@repo, object@sha, force)
              .Call(git2r_repository_set_head_detached, object)
              invisible(NULL)
          }
)

##' @rdname checkout-methods
##' @export
setMethod("checkout",
          signature(object = "git_tag"),
          function(object, force = FALSE)
          {
              .Call(git2r_checkout_tree, object@repo, object@target, force)
              .Call(git2r_repository_set_head_detached,
                    lookup(object@repo, object@target))
              invisible(NULL)
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Default notes reference
##'
##' Get the default notes reference for a repository
##' @rdname note_default_ref-methods
##' @docType methods
##' @param repo The repository
##' @return Character vector of length one with name of default notes
##' reference
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## View default notes reference
##' note_default_ref(repo)
##' }
setGeneric("note_default_ref",
           signature = "repo",
           function(repo)
           standardGeneric("note_default_ref"))

##' @rdname note_default_ref-methods
##' @export
setMethod("note_default_ref",
          signature = "git_repository",
          function(repo)
          {
              .Call(git2r_note_default_ref, repo)
          }
)

##' Add note for a object
##'
##' @rdname note_create-methods
##' @docType methods
##' @param object The object to annotate
##' @param message Content of the note to add
##' @param ref Canonical name of the reference to use. Default is
##' \code{note_default_ref}.
##' @param author Signature of the notes note author
##' @param committer Signature of the notes note committer
##' @param force Overwrite existing note. Default is FALSE
##' @return S4 class git_note
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##'
##' ## Create another commit
##' writeLines(c("Hello world!",
##'              "HELLO WORLD!"),
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_2 <- commit(repo, "Commit message 2")
##'
##' ## Check that notes is an empty list
##' notes(repo)
##'
##' ## Create note in default namespace
##' note_create(commit_1, "Note-1")
##'
##' ## Create note in named (review) namespace
##' note_create(commit_1, "Note-2", ref="refs/notes/review")
##' note_create(commit_2, "Note-3", ref="review")
##'
##' ## Create note on blob and tree
##' note_create(tree(commit_1), "Note-4")
##' note_create(tree(commit_1)["example.txt"], "Note-5")
##' }
setGeneric("note_create",
           signature = "object",
           function(object,
                    message,
                    ref       = note_default_ref(object@repo),
                    author    = default_signature(object@repo),
                    committer = default_signature(object@repo),
                    force     = FALSE)
           standardGeneric("note_create"))

##' @rdname note_create-methods
##' @export
setMethod("note_create",
          signature = "git_blob",
          function(object,
                   message,
                   ref,
                   author,
                   committer,
                   force)
          {
              stopifnot(is.character(ref))
              stopifnot(identical(length(ref), 1L))
              if (!length(grep("^refs/notes/", ref)))
                  ref <- paste0("refs/notes/", ref)
              .Call(git2r_note_create,
                    object@repo,
                    object@sha,
                    message,
                    ref,
                    author,
                    committer,
                    force)
          }
)

##' @rdname note_create-methods
##' @export
setMethod("note_create",
          signature = "git_commit",
          function(object,
                   message,
                   ref,
                   author,
                   committer,
                   force)
          {
              stopifnot(is.character(ref))
              stopifnot(identical(length(ref), 1L))
              if (!length(grep("^refs/notes/", ref)))
                  ref <- paste0("refs/notes/", ref)
              .Call(git2r_note_create,
                    object@repo,
                    object@sha,
                    message,
                    ref,
                    author,
                    committer,
                    force)
          }
)

##' @rdname note_create-methods
##' @export
setMethod("note_create",
          signature = "git_tree",
          function(object,
                   message,
                   ref,
                   author,
                   committer,
                   force)
          {
              stopifnot(is.character(ref))
              stopifnot(identical(length(ref), 1L))
              if (!length(grep("^refs/notes/", ref)))
                  ref <- paste0("refs/notes/", ref)
              .Call(git2r_note_create,
                    object@repo,
                    object@sha,
                    message,
                    ref,
                    author,
                    committer,
                    force)
          }
)

##' List notes
##'
##' List all the notes within a specified namespace.
##' @rdname notes-methods
##' @docType methods
##' @param repo The repository
##' @param ref Reference to read from. Default is
##' "refs/notes/commits".
##' @return list with S4 class git_note objects
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##'
##' ## Create another commit
##' writeLines(c("Hello world!",
##'              "HELLO WORLD!"),
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_2 <- commit(repo, "Commit message 2")
##'
##' ## Create note in default namespace
##' note_create(commit_1, "Note-1")
##' note_create(commit_1, "Note-2", force = TRUE)
##'
##' ## Create note in named (review) namespace
##' note_create(commit_1, "Note-3", ref="refs/notes/review")
##' note_create(commit_2, "Note-4", ref="review")
##'
##' ## Create note on blob and tree
##' note_create(tree(commit_1), "Note-5")
##' note_create(tree(commit_1)["example.txt"], "Note-6")
##'
##' ## List notes in default namespace
##' notes(repo)
##'
##' ## List notes in 'review' namespace
##' notes(repo, "review")
##' }
setGeneric("notes",
           signature = c("repo", "ref"),
           function(repo,
                    ref = note_default_ref(repo))
           standardGeneric("notes"))

##' @rdname notes-methods
##' @export
setMethod("notes",
          signature = "git_repository",
          function(repo, ref)
          {
              stopifnot(is.character(ref), identical(length(ref), 1L))
              if (!length(grep("^refs/notes/", ref)))
                  ref <- paste0("refs/notes/", ref)
              .Call(git2r_notes, repo, ref)
          }
)

##' Remove the note for an object
##'
##' @rdname note_remove-methods
##' @docType methods
##' @param note The note to remove
##' @param author Signature of the notes commit author.
##' @param committer Signature of the notes commit committer.
##' @return invisible NULL
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##'
##'
##' ## Create note in default namespace
##' note_1 <- note_create(commit_1, "Note-1")
##'
##' ## Create note in named (review) namespace
##' note_2 <- note_create(commit_1, "Note-2", ref="refs/notes/review")
##'
##' ## List notes in default namespace
##' notes(repo)
##'
##' ## List notes in 'review' namespace
##' notes(repo, "review")
##'
##' ## Remove notes
##' note_remove(note_1)
##' note_remove(note_2)
##'
##' ## List notes in default namespace
##' notes(repo)
##'
##' ## List notes in 'review' namespace
##' notes(repo, "review")
##' }
setGeneric("note_remove",
           signature = "note",
           function(note,
                    author    = default_signature(note@repo),
                    committer = default_signature(note@repo))
           standardGeneric("note_remove"))

##' @rdname note_remove-methods
##' @export
setMethod("note_remove",
          signature = "git_note",
          function(note,
                   author,
                   committer)
          {
              invisible(.Call(git2r_note_remove, note, author, committer))
          }
)

##' Brief summary of note
##'
##' @aliases show,git_note-methods
##' @docType methods
##' @param object The \linkS4class{git_note} \code{object}
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Create note
##' note_create(commits(repo)[[1]], "My note")
##' }
setMethod("show",
          signature(object = "git_note"),
          function(object)
          {
              cat(sprintf("note:  %s\n", object@sha))
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' List and view reflog information
##'
##' @template repo-param
##' @param refname The name of the reference to list. 'HEAD' by
##'     default.
##' @return S3 class \code{git_reflog} with git_reflog_entry objects.
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Change file and commit
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Second commit message")
##'
##' ## Change file again and commit
##' lines <- c(
##'   "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'   "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
##'   "minim veniam, quis nostrud exercitation ullamco laboris nisi ut")
##' writeLines(lines, file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Third commit message")
##'
##' ## View reflog
##' reflog(repo)
##' }
reflog <- function(repo = ".", refname = "HEAD") {
    structure(.Call(git2r_reflog_list, lookup_repository(repo), refname),
              class = "git_reflog")
}

##' @export
print.git_reflog <- function(x, ...) {
    lapply(x, print)
    invisible(x)
}

##' Print a reflog entry
##'
##' @param x The reflog entry
##' @param ... Unused
##' @return None (invisible 'NULL').
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## View repository HEAD reflog
##' reflog(repo)
##' }
print.git_reflog_entry <- function(x, ...) {
    cat(sprintf("[%s] %s@{%i}: %s\n",
                substring(x$sha, 1, 7),
                x$refname,
                x$index,
                x$message))
    invisible(x)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2018 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Get all references that can be found in a repository.
##' @template repo-param
##' @return Character vector with references
##' @export
##' @examples
##' \dontrun{
##' ## Initialize two temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config user and commit a file
##' config(repo, user.name = "Alice", user.email = "alice@@example.org")
##'
##' ## Write to a file and commit
##' lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
##' writeLines(lines, file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## Add tag to HEAD
##' tag(repo, "v1.0", "First version")
##'
##' ## Create a note
##' note_create(commits(repo)[[1]], "My note")
##'
##' ## List all references in repository
##' references(repo)
##' }
##'
references <- function(repo = ".") {
    .Call(git2r_reference_list, lookup_repository(repo))
}

##' @export
format.git_reference <- function(x, ...) {
    if (identical(x$type, 1L))
        return(sprintf("[%s] %s", substr(x$sha, 1, 6), x$shorthand))
    sprintf("%s => %s", x$name, x$target)
}

##' @export
print.git_reference <- function(x, ...) {
    cat(format(x, ...), "\n", sep = "")
    invisible(x)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Create push refspec from arguments
##'
##' @param repo a \code{git_repository} object.
##' @param name The remote's name. Default is NULL.
##' @param refspec The refspec to be pushed. Default is NULL.
##' @param opts List with push options. Default is NULL.
##' @return List with remote (character vector) and refspec (character
##' vector).
##' @noRd
get_refspec <- function(repo = NULL, remote = NULL, spec = NULL, opts = NULL) {
    stopifnot(inherits(repo, "git_repository"))

    if (is_detached(repo))
        stop("You are not currently on a branch.")

    ## Options:
    if (!is.null(opts)) {
        stopifnot(is.list(opts))
    } else {
        opts <- list()
    }

    ## Remote:
    ## From: http://git-scm.com/docs/git-push
    ## When the command line does not specify where to push with the
    ## <repository> argument, branch.*.remote configuration for the
    ## current branch is consulted to determine where to push. If the
    ## configuration is missing, it defaults to origin.
    if (!is.null(remote)) {
        stopifnot(is.character(remote), identical(length(remote), 1L))
        remote <- sub("^[[:space:]]*", "", sub("[[:space:]]*$", "", remote))
        if (identical(nchar(remote), 0L))
            remote <- NULL
    }
    if (is.null(remote)) {
        remote <- .Call(git2r_config_get_string,
                        repo,
                        paste0("branch.",
                               repository_head(repo)$name,
                               ".remote"))
        if (is.null(remote))
            remote <- "origin"
    }

    ## Refspec:
    stopifnot(is.character(spec))

    if (isTRUE(opts$force))
        spec <- paste0("+", spec)

    list(remote = remote, refspec = spec)
}

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Drop stash
##'
##' @rdname stash_drop-methods
##' @docType methods
##' @param object The stash \code{object} to drop or a zero-based
##' integer to the stash to drop. The last stash has index 0.
##' @param ... Additional arguments affecting the stash_drop
##' @param index Zero based index to the stash to drop. Only used when
##' \code{object} is a \code{git_repository}.
##' @return invisible NULL
##' @keywords methods
##' @examples \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' # Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' add(repo, 'test.txt')
##' commit(repo, "Commit message")
##'
##' # Change file
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
##'
##' # Create stash in repository
##' stash(repo)
##'
##' # Change file
##' writeLines(c("Hello world!", "HeLlO wOrLd!"), file.path(path, "test.txt"))
##'
##' # Create stash in repository
##' stash(repo)
##'
##' # View stashes
##' stash_list(repo)
##'
##' # Drop git_stash object in repository
##' stash_drop(stash_list(repo)[[1]])
##'
##' ## Drop stash using an index to stash
##' stash_drop(repo, 0)
##'
##' # View stashes
##' stash_list(repo)
##' }
setGeneric("stash_drop",
           signature = "object",
           function(object, ...)
           standardGeneric("stash_drop"))

##' @rdname stash_drop-methods
##' @include S4_classes.r
##' @export
setMethod("stash_drop",
          signature(object = "git_repository"),
          function(object, index)
          {
              if (missing(index))
                  stop("missing argument 'index'")
              if (abs(index - round(index)) >= .Machine$double.eps^0.5)
                  stop("'index' must be an integer")
              index <- as.integer(index)
              .Call(git2r_stash_drop, object, index)
              invisible(NULL)
          }
)

##' @rdname stash_drop-methods
##' @export
setMethod("stash_drop",
          signature(object = "git_stash"),
          function(object)
          {
              ## Determine the index of the stash in the stash list
              i <- match(object@sha, vapply(stash_list(object@repo), slot, character(1), "sha"))

              ## The stash list is zero-based
              .Call(git2r_stash_drop, object@repo, i - 1L)
              invisible(NULL)
          }
)

##' Stash
##'
##' @rdname stash-methods
##' @docType methods
##' @param object The repository \code{object}.
##' @param message Optional description. Defaults to current time.
##' @param index All changes already added to the index are left
##' intact in the working directory. Default is FALSE
##' @param untracked All untracked files are also stashed and then
##' cleaned up from the working directory. Default is FALSE
##' @param ignored All ignored files are also stashed and then cleaned
##' up from the working directory. Default is FALSE
##' @param stasher Signature with stasher and time of stash
##' @return invisible S4 class git_stash if anything to stash else NULL
##' @keywords methods
##' @examples \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' # Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' add(repo, 'test.txt')
##' commit(repo, "Commit message")
##'
##' # Change file
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
##'
##' # Check status of repository
##' status(repo)
##'
##' # Create stash in repository
##' stash(repo)
##'
##' # Check status of repository
##' status(repo)
##'
##' # View stash
##' stash_list(repo)
##' }
setGeneric("stash",
           signature = "object",
           function(object,
                    message   = as.character(Sys.time()),
                    index     = FALSE,
                    untracked = FALSE,
                    ignored   = FALSE,
                    stasher   = default_signature(object))
           standardGeneric("stash"))

##' @rdname stash-methods
##' @include S4_classes.r
##' @export
setMethod("stash",
          signature(object = "git_repository"),
          function(object,
                   message,
                   index,
                   untracked,
                   ignored,
                   stasher)
          {
              invisible(.Call(git2r_stash_save,
                              object,
                              message,
                              index,
                              untracked,
                              ignored,
                              stasher))
          }
)

##' List stashes in repository
##'
##' @rdname stash_list-methods
##' @docType methods
##' @param repo The repository \code{object}
##' \code{\linkS4class{git_repository}}. If the \code{repo} argument
##' is missing, the repository is searched for with
##' \code{\link{discover_repository}} in the current working
##' directory.
##' @return list of stashes in repository
##' @keywords methods
##' @include S4_classes.r
##' @examples \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' # Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test-1.txt"))
##' add(repo, 'test-1.txt')
##' commit(repo, "Commit message")
##'
##' # Make one more commit
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-1.txt"))
##' add(repo, 'test-1.txt')
##' commit(repo, "Next commit message")
##'
##' # Create one more file
##' writeLines("Hello world!", file.path(path, "test-2.txt"))
##'
##' # Check that there are no stashes
##' stash_list(repo)
##'
##' # Stash
##' stash(repo)
##'
##' # Only untracked changes, therefore no stashes
##' stash_list(repo)
##'
##' # Stash and include untracked changes
##' stash(repo, "Stash message", untracked=TRUE)
##'
##' # View stash
##' stash_list(repo)
##' }
setGeneric("stash_list",
           signature = "repo",
           function(repo)
           standardGeneric("stash_list"))

##' @rdname stash_list-methods
##' @export
setMethod("stash_list",
          signature(repo = "missing"),
          function()
          {
              callGeneric(repo = lookup_repository())
          }
)

##' @rdname stash_list-methods
##' @export
setMethod("stash_list",
          signature(repo = "git_repository"),
          function(repo)
          {
              .Call(git2r_stash_list, repo)
          }
)

##' Brief summary of a stash
##'
##' @aliases show,git_stash-methods
##' @docType methods
##' @param object The stash \code{object}
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' # Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' add(repo, 'test.txt')
##' commit(repo, "Commit message")
##'
##' # Change file
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
##'
##' # Create stash in repository
##' stash(repo, "Stash message")
##'
##' # View brief summary of stash
##' stash_list(repo)[[1]]
##' }
setMethod("show",
          signature(object = "git_stash"),
          function(object)
          {
              cat(sprintf("%s\n", object@message))
          }
)

##' Summary of a stash
##'
##' @aliases summary,git_stash-methods
##' @docType methods
##' @param object The stash \code{object}
##' @param ... Additional arguments affecting the summary produced.
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' # Configure a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' # Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "test.txt"))
##' add(repo, 'test.txt')
##' commit(repo, "Commit message")
##'
##' # Change file
##' writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))
##'
##' # Create stash in repository
##' stash(repo, "Stash message")
##'
##' # View summary of stash
##' summary(stash_list(repo)[[1]])
##' }
setMethod("summary",
          signature(object = "git_stash"),
          function(object, ...)
          {
              cat(sprintf(paste0("message: %s\n",
                                 "stasher: %s <%s>\n",
                                 "when:    %s\n",
                                 "sha:     %s\n\n"),
                          object@summary,
                          object@author@name,
                          object@author@email,
                          as(object@author@when, "character"),
                          object@sha))
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Ahead Behind
##'
##' Count the number of unique commits between two commit objects.
##' @rdname ahead_behind-methods
##' @docType methods
##' @param local a S4 class git_commit \code{object}.
##' @param upstream a S4 class git_commit \code{object}.
##' @return An integer vector of length 2 with number of commits that
##' the upstream commit is ahead and behind the local commit
##' @keywords methods
##' @examples \dontrun{
##' ## Create a directory in tempdir
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##'
##' ## Initialize a repository
##' repo <- init(path)
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            con = file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##'
##' # Change file and commit
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
##'              con = file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit_2 <- commit(repo, "Commit message 2")
##'
##' ahead_behind(commit_1, commit_2)
##' }
setGeneric("ahead_behind",
           signature = c("local", "upstream"),
           function(local, upstream)
           standardGeneric("ahead_behind"))

##' @rdname ahead_behind-methods
##' @export
setMethod("ahead_behind",
          signature(local = "git_commit", upstream = "git_commit"),
          function(local, upstream)
          {
              stopifnot(identical(local@repo, upstream@repo))
              .Call(git2r_graph_ahead_behind, local, upstream)
          }
)

##' Add sessionInfo to message
##'
##' @param message The message.
##' @return message with appended sessionInfo
##' @importFrom utils capture.output
##' @importFrom utils sessionInfo
##' @keywords internal
add_session_info <- function(message)
{
    paste0(message, "\n\nsessionInfo:\n",
           paste0(utils::capture.output(utils::sessionInfo()),
                  collapse="\n"))
}

##' Commit
##'
##' @rdname commit-methods
##' @docType methods
##' @param repo The repository \code{object}.
##' @param message The commit message.
##' @param all Stage modified and deleted files. Files not added to
##' Git are not affected.
##' @param session Add sessionInfo to commit message. Default is
##' FALSE.
##' @param reference Name of the reference that will be updated to
##' point to this commit.
##' @param author Signature with author and author time of commit.
##' @param committer Signature with committer and commit time of commit.
##' @return \code{\linkS4class{git_commit}} object
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##' }
setGeneric("commit",
           signature = "repo",
           function(repo,
                    message   = NULL,
                    all       = FALSE,
                    session   = FALSE,
                    reference = "HEAD",
                    author    = default_signature(repo),
                    committer = default_signature(repo))
           standardGeneric("commit"))

##' @rdname commit-methods
##' @export
setMethod("commit",
          signature(repo = "git_repository"),
          function(repo,
                   message,
                   all,
                   session,
                   reference,
                   author,
                   committer)
          {
              ## Argument checking
              stopifnot(is.character(message),
                        identical(length(message), 1L),
                        is.logical(all),
                        identical(length(all), 1L),
                        is.logical(session),
                        identical(length(session), 1L))

              if (!nchar(message[1]))
                  stop("Aborting commit due to empty commit message.")

              if (all) {
                  s <- status(repo,
                              unstaged  = TRUE,
                              staged    = FALSE,
                              untracked = FALSE,
                              ignored   = FALSE)

                  # Convert list of lists to character vector
                  unstaged <- unlist(s$unstaged)
                  for (i in seq_along(unstaged)) {
                    if (names(unstaged)[i] == "modified") {
                      ## Stage modified files
                      add(repo, unstaged[i])
                    } else if (names(unstaged)[i] == "deleted") {
                      ## Stage deleted files
                      .Call(git2r_index_remove_bypath, repo, unstaged[i])
                    }
                  }

              }

              if (session)
                  message <- add_session_info(message)

              .Call(git2r_commit, repo, message, author, committer)
          }
)

##' Commits
##'
##' @rdname commits-methods
##' @docType methods
##' @param repo The repository \code{object}
##' \code{\linkS4class{git_repository}}. If the \code{repo} argument
##' is missing, the repository is searched for with
##' \code{\link{discover_repository}} in the current working
##' directory.
##' @param ... Additional arguments to commits.
##' @param topological Sort the commits in topological order (parents
##' before children); can be combined with time sorting. Default is
##' TRUE.
##' @param time Sort the commits by commit time; Can be combined with
##' topological sorting. Default is TRUE.
##' @param reverse Sort the commits in reverse order; can be combined
##' with topological and/or time sorting. Default is FALSE.
##' @param n The upper limit of the number of commits to output. The
##' defualt is NULL for unlimited number of commits.
##' @return list of commits in repository
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Change file and commit
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Second commit message")
##'
##' ## Change file again and commit
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
##'              "minim veniam, quis nostrud exercitation ullamco laboris nisi ut"),
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Third commit message")
##'
##' ## List commits in repository
##' commits(repo)
##' }
setGeneric("commits",
           signature = "repo",
           function(repo, ...)
           standardGeneric("commits"))

##' @rdname commits-methods
##' @export
setMethod("commits",
          signature(repo = "missing"),
          function(topological = TRUE,
                   time        = TRUE,
                   reverse     = FALSE,
                   n           = NULL,
                   ...)
          {
              callGeneric(repo        = lookup_repository(),
                          topological = topological,
                          time        = time,
                          reverse     = reverse,
                          n           = n,
                          ...)
          }
)

##' @rdname commits-methods
##' @include S4_classes.r
##' @export
setMethod("commits",
          signature(repo = "git_repository"),
          function(repo,
                   topological = TRUE,
                   time        = TRUE,
                   reverse     = FALSE,
                   n           = NULL,
                   ...)
          {
              ## Check limit in number of commits
              if (is.null(n)) {
                  n <- -1L
              } else if (is.numeric(n)) {
                  if (!identical(length(n), 1L))
                      stop("'n' must be integer")
                  if (abs(n - round(n)) >= .Machine$double.eps^0.5)
                      stop("'n' must be integer")
                  n <- as.integer(n)
              } else {
                  stop("'n' must be integer")
              }

              if (is_shallow(repo)) {
                  ## FIXME: Remove this if-statement when libgit2
                  ## supports shallow clones, see #219.  Note: This
                  ## workaround does not use the 'topological', 'time'
                  ## and 'reverse' flags.

                  ## List to hold result
                  result <- list()

                  ## Get latest commit
                  x <- lookup(repo, branch_target(head(repo)))

                  ## Repeat until no more parent commits
                  repeat {
                      if (n == 0) {
                          break
                      } else if (n > 0) {
                          n <- n - 1
                      }

                      if (is.null(x))
                          break
                      result <- append(result, x)

                      ## Get parent to commit
                      x <- tryCatch(parents(x)[[1]], error = function(e) NULL)
                  }

                  return(result)
              }

              .Call(git2r_revwalk_list, repo, topological, time, reverse, n)
          }
)

##' Descendant
##'
##' Determine if a commit is the descendant of another commit
##' @rdname descendant_of-methods
##' @docType methods
##' @param commit a S4 class git_commit \code{object}.
##' @param ancestor a S4 class git_commit \code{object} to check if
##' ancestor to \code{commit}.
##' @return TRUE if commit is descendant of \code{ancestor}, else
##' FALSE
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Create a directory in tempdir
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##'
##' ## Initialize a repository
##' repo <- init(path)
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            con = file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit_1 <- commit(repo, "Commit message 1")
##'
##' # Change file and commit
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
##'              con = file.path(path, "test.txt"))
##' add(repo, "test.txt")
##' commit_2 <- commit(repo, "Commit message 2")
##'
##' descendant_of(commit_1, commit_2)
##' descendant_of(commit_2, commit_1)
##' }
setGeneric("descendant_of",
           signature = c("commit", "ancestor"),
           function(commit, ancestor)
           standardGeneric("descendant_of"))

##' @rdname descendant_of-methods
##' @export
setMethod("descendant_of",
          signature(commit = "git_commit", ancestor = "git_commit"),
          function(commit, ancestor)
          {
              stopifnot(identical(commit@repo, ancestor@repo))
              .Call(git2r_graph_descendant_of, commit, ancestor)
          }
)

##' Check if object is S4 class git_commit
##'
##' @param object Check if object is S4 class git_commit
##' @return TRUE if object is S4 class git_commit, else FALSE
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Commit a text file
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##'
##' ## Check if commit
##' is_commit(commit_1)
##' }
is_commit <- function(object) {
    is(object = object, class2 = "git_commit")
}

##' Is merge
##'
##' Determine if a commit is a merge commit, i.e. has more than one
##' parent.
##' @rdname is_merge-methods
##' @docType methods
##' @param commit a S4 class git_commit \code{object}.
##' @return TRUE if commit has more than one parent, else FALSE
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines(c("First line in file 1.", "Second line in file 1."),
##'            file.path(path, "example-1.txt"))
##' add(repo, "example-1.txt")
##' commit(repo, "First commit message")
##'
##' ## Create and add one more file
##' writeLines(c("First line in file 2.", "Second line in file 2."),
##'            file.path(path, "example-2.txt"))
##' add(repo, "example-2.txt")
##' commit(repo, "Second commit message")
##'
##' ## Create a new branch 'fix'
##' checkout(repo, "fix", create = TRUE)
##'
##' ## Update 'example-1.txt' (swap words in first line) and commit
##' writeLines(c("line First in file 1.", "Second line in file 1."),
##'            file.path(path, "example-1.txt"))
##' add(repo, "example-1.txt")
##' commit(repo, "Third commit message")
##'
##' checkout(repo, "master")
##'
##' ## Update 'example-2.txt' (swap words in second line) and commit
##' writeLines(c("First line in file 2.", "line Second in file 2."),
##'            file.path(path, "example-2.txt"))
##' add(repo, "example-2.txt")
##' commit(repo, "Fourth commit message")
##'
##' ## Merge 'fix'
##' merge(repo, "fix")
##'
##' ## Display parents of last commit
##' parents(lookup(repo, branch_target(head(repo))))
##'
##' ## Check that last commit is a merge
##' is_merge(lookup(repo, branch_target(head(repo))))
##' }
setGeneric("is_merge",
           signature = c("commit"),
           function(commit)
           standardGeneric("is_merge"))

##' @rdname is_merge-methods
##' @export
setMethod("is_merge",
          signature(commit = "git_commit"),
          function(commit)
          {
              length(parents(commit)) > 1
          }
)

##' Parents
##'
##' Get parents of a commit.
##' @rdname parents-methods
##' @docType methods
##' @param object a S4 class git_commit \code{object}.
##' @return list of S4 git_commit objects
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("First line.",
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_1 <- commit(repo, "First commit message")
##'
##' ## commit_1 has no parents
##' parents(commit_1)
##'
##' ## Update 'example.txt' and commit
##' writeLines(c("First line.", "Second line."),
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit_2 <- commit(repo, "Second commit message")
##'
##' ## commit_2 has commit_1 as parent
##' parents(commit_2)
##' }
setGeneric("parents",
           signature = "object",
           function(object)
           standardGeneric("parents"))

##' @rdname parents-methods
##' @export
setMethod("parents",
          signature(object = "git_commit"),
          function(object)
          {
              .Call(git2r_commit_parent_list, object)
          }
)

##' Brief summary of commit
##'
##' Displays the first seven characters of the sha, the date and the
##' summary of the commit message:
##' \code{[shortened sha] yyyy-mm-dd: summary}
##' @aliases show,git_commit-methods
##' @docType methods
##' @param object The commit \code{object}
##' @return None (invisible 'NULL').
##' @keywords methods
##' @include S4_classes.r
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Brief summary of commit in repository
##' show(commits(repo)[[1]])
##' }
setMethod("show",
          signature(object = "git_commit"),
          function(object)
          {
              cat(sprintf("[%s] %s: %s\n",
                          substring(object@sha, 1, 7),
                          substring(as(object@author@when, "character"), 1, 10),
                          object@summary))
          }
)

##' Summary of commit
##'
##' @aliases summary,git_commit-methods
##' @docType methods
##' @param object The commit \code{object}
##' @param ... Additional arguments affecting the summary produced.
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Summary of commit in repository
##' summary(commits(repo)[[1]])
##' }
setMethod("summary",
          signature(object = "git_commit"),
          function(object, ...)
          {
              is_merge_commit <- is_merge(object)
              po <- parents(object)

              cat(sprintf("Commit:  %s\n", object@sha))

              if (is_merge_commit) {
                  sha <- vapply(po, slot, character(1), "sha")
                  cat(sprintf("Merge:   %s\n", sha[1]))
                  cat(paste0("         ", sha[-1]), sep="\n")
              }

              cat(sprintf(paste0("Author:  %s <%s>\n",
                                 "When:    %s\n\n"),
                          object@author@name,
                          object@author@email,
                          as(object@author@when, "character")))

              msg <- paste0("    ", readLines(textConnection(object@message)))
              cat("", sprintf("%s\n", msg))

              if (is_merge_commit) {
                  cat("\n")
                  lapply(po, function(parent) {
                      cat("Commit message: ", parent@sha, "\n")
                      msg <- paste0("    ",
                                    readLines(textConnection(parent@message)))
                      cat("", sprintf("%s\n", msg), "\n")
                  })
              }

              if (identical(length(po), 1L)) {
                  df <- diff(tree(po[[1]]), tree(object))
                  if (length(df) > 0) {
                      if (length(df) > 1) {
                          cat(sprintf("%i files changed, ", length(df)))
                      } else {
                          cat("1 file changed, ")
                      }

                      cat(sprintf("%i insertions, %i deletions\n",
                                  sum(vapply(lines_per_file(df), "[[", numeric(1), "add")),
                                  sum(vapply(lines_per_file(df), "[[", numeric(1), "del"))))

                      plpf <- print_lines_per_file(df)
                      hpf <- hunks_per_file(df)
                      hunk_txt <- ifelse(hpf > 1, " hunks",
                                         ifelse(hpf > 0, " hunk",
                                                " hunk (binary file)"))
                      phpf <- paste0("  in ", format(hpf), hunk_txt)
                      cat(paste0(plpf, phpf), sep="\n")
                  }

                  cat("\n")
              }

              invisible(NULL)
          }
)

##' Coerce a commit to a \code{data.frame}
##'
##' The commit is coerced to a \code{data.frame}
##'
##'
##' The \code{data.frame} have the following columns:
##' \describe{
##'
##'   \item{sha}{
##'     The 40 character hexadecimal string of the SHA-1
##'   }
##'
##'   \item{summary}{
##'     the short "summary" of the git commit message.
##'   }
##'
##'   \item{message}{
##'     the full message of a commit
##'   }
##'
##'   \item{author}{
##'     full name of the author
##'   }
##'
##'   \item{email}{
##'     email of the author
##'   }
##'
##'   \item{when}{
##'     time when the commit happened
##'   }
##'
##' }
##' @name coerce-git_commit-method
##' @aliases coerce,git_commit,data.frame-method
##' @docType methods
##' @param from The commit \code{object}
##' @return \code{data.frame}
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file and commit
##' writeLines("Example file",  file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' c1 <- commit(repo, "Commit message")
##'
##' ## Coerce the commit to a data.frame
##' df <- as(c1, "data.frame")
##' df
##' }
setAs(from = "git_commit",
      to   = "data.frame",
      def  = function(from)
      {
          data.frame(sha              = from@sha,
                     summary          = from@summary,
                     message          = from@message,
                     author           = from@author@name,
                     email            = from@author@email,
                     when             = as(from@author@when, "POSIXct"),
                     stringsAsFactors = FALSE)
      }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2016 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' List and view reflog information
##'
##' @rdname reflog-methods
##' @docType methods
##' @param repo S4 class git_repository \code{object}.
##' @param refname The name of the reference to list. 'HEAD' by
##'     default.
##' @return S3 class \code{git_reflog} with git_reflog_entry objects.
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize a repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Config user
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Change file and commit
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Second commit message")
##'
##' ## Change file again and commit
##' writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'              "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
##'              "minim veniam, quis nostrud exercitation ullamco laboris nisi ut"),
##'            file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "Third commit message")
##'
##' ## View reflog
##' reflog(repo)
##' }
setGeneric("reflog",
           signature = c("repo", "refname"),
           function(repo, refname)
           standardGeneric("reflog"))

##' @rdname reflog-methods
##' @export
setMethod("reflog",
          signature(repo    = "git_repository",
                    refname = "missing"),
          function(repo)
          {
              callGeneric(repo = repo, refname = "HEAD")
          }
)

##' @rdname reflog-methods
##' @export
setMethod("reflog",
          signature(repo    = "git_repository",
                    refname = "character"),
          function(repo, refname)
          {
              structure(.Call(git2r_reflog_list, repo, refname),
                        class = "git_reflog")
          }
)

##' @export
print.git_reflog <- function(x, ...)
{
    lapply(x, show)
    invisible(NULL)
}

##' Brief summary of a reflog entry
##'
##' @aliases show,git_reflog_entry-methods
##' @docType methods
##' @param object The reflog entry \code{object}
##' @return None (invisible 'NULL').
##' @keywords methods
##' @export
##' @examples
##' \dontrun{
##' ## Initialize a temporary repository
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##'
##' ## Create a user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## View repository HEAD reflog
##' reflog(repo)
##' }
setMethod("show",
          signature(object = "git_reflog_entry"),
          function(object)
          {
              cat(sprintf("[%s] %s@{%i}: %s\n",
                          substring(object@sha, 1, 7),
                          object@refname,
                          object@index,
                          object@message))
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2015 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

##' Get all references that can be found in a repository.
##'
##' @rdname references-methods
##' @docType methods
##' @param repo The repository \code{object}
##' \code{\linkS4class{git_repository}}. If the \code{repo} argument
##' is missing, the repository is searched for with
##' \code{\link{discover_repository}} in the current working
##' directory.
##' @return Character vector with references
##' @keywords methods
##' @examples
##' \dontrun{
##' ## Initialize two temporary repositories
##' path_bare <- tempfile(pattern="git2r-")
##' path_repo <- tempfile(pattern="git2r-")
##' dir.create(path_bare)
##' dir.create(path_repo)
##' repo_bare <- init(path_bare, bare = TRUE)
##' repo <- clone(path_bare, path_repo)
##'
##' ## Config user and commit a file
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Write to a file and commit
##' writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
##'            file.path(path_repo, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Push commits from repository to bare repository
##' ## Adds an upstream tracking branch to branch 'master'
##' push(repo, "origin", "refs/heads/master")
##'
##' ## Add tag to HEAD
##' tag(repo, "v1.0", "First version")
##'
##' ## Create a note
##' note_create(commits(repo)[[1]], "My note")
##'
##' ## List all references in repository
##' references(repo)
##' }
##'
setGeneric("references",
           signature = "repo",
           function(repo) standardGeneric("references"))

##' @rdname references-methods
##' @include S4_classes.r
##' @export
setMethod("references",
          signature(repo = "missing"),
          function()
          {
              callGeneric(repo = lookup_repository())
          }
)

##' @rdname references-methods
##' @include S4_classes.r
##' @export
setMethod("references",
          signature(repo = "git_repository"),
          function(repo)
          {
              .Call(git2r_reference_list, repo)
          }
)

##' Brief summary of reference
##'
##' @aliases show,git_reference-methods
##' @docType methods
##' @param object The reference \code{object}
##' @return None (invisible 'NULL').
##' @keywords methods
##' @include S4_classes.r
##' @export
##' @examples
##' \dontrun{
##' ## Create and initialize a repository in a temporary directory
##' path <- tempfile(pattern="git2r-")
##' dir.create(path)
##' repo <- init(path)
##' config(repo, user.name="Alice", user.email="alice@@example.org")
##'
##' ## Create a file, add and commit
##' writeLines("Hello world!", file.path(path, "example.txt"))
##' add(repo, "example.txt")
##' commit(repo, "First commit message")
##'
##' ## Brief summary of reference
##' references(repo)[[1]]
##' }
setMethod("show",
          signature(object = "git_reference"),
          function(object)
          {
              if (identical(object@type, 1L)) {
                  cat(sprintf("[%s] %s\n",
                              substr(object@sha, 1 , 6),
                              object@shorthand))
              } else if (identical(object@type, 2L)) {
                  cat(sprintf("%s => %s\n",
                              object@name,
                              object@target))
              } else {
                  stop("Unexpected reference type")
              }
          }
)

## git2r, R bindings to the libgit2 library.
## Copyright (C) 2013-2019 The git2r contributors
##
## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License, version 2,
## as published by the Free Software Foundation.
##
## git2r is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.
##
## You should have received a copy of the GNU General Public License along
## with this program; if not, write to the Free Software Foundation, Inc.,
## 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

library("git2r")

## For debugging
sessionInfo()

## Create a directory in tempdir
path <- tempfile(pattern = "git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Add files
invisible(lapply(file.path(path, letters[1:4]), writeLines, text = ""))
add(repo, letters)
commit(repo, "init")

## Remove one file
rm_file(repo, letters[1])
commit(repo, "remove")

## Remove two files. Don't raise warnings
withCallingHandlers(rm_file(repo, letters[2:3]), warning = function(w) stop(w))

## Remove one file using the absolute path to the file.
rm_file(repo, file.path(path, letters[4]))

## Cleanup
unlink(path, recursive = TRUE)

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.R
\name{is_head}
\alias{is_head}
\title{Check if branch is head}
\usage{
is_head(branch = NULL)
}
\arguments{
\item{branch}{The branch \code{object} to check if it's head.}
}
\value{
\code{TRUE} if branch is head, else \code{FALSE}.
}
\description{
Check if branch is head
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## List branches
branches(repo)

## Check that 'master' is_head
master <- branches(repo)[[1]]
is_head(master)

## Create and checkout 'dev' branch
checkout(repo, "dev", create = TRUE)

## List branches
branches(repo)

## Check that 'master' is no longer head
is_head(master)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_stash-class}
\alias{git_stash-class}
\title{S4 class to handle a git stash}
\description{
Class \code{"git_stash"}
}
\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{head.git_repository}
\alias{head.git_repository}
\title{Get HEAD for a repository}
\usage{
\method{head}{git_repository}(x, ...)
}
\arguments{
\item{x}{The repository \code{x} to check head}

\item{...}{Additional arguments. Unused.}
}
\value{
NULL if unborn branch or not found. A git_branch if not a
    detached head. A git_commit if detached head
}
\description{
Get HEAD for a repository
}
\examples{
\dontrun{
## Create and initialize a repository in a temporary directory
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Commit message")

## Get HEAD of repository
repository_head(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge.r
\docType{methods}
\name{show,git_merge_result-method}
\alias{show,git_merge_result-method}
\alias{show,git_merge_result-methods}
\title{Brief summary of merge result}
\usage{
\S4method{show}{git_merge_result}(object)
}
\arguments{
\item{object}{The \code{\linkS4class{git_merge_result}} \code{object}}
}
\value{
None (invisible 'NULL').
}
\description{
Brief summary of merge result
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")
writeLines("First line.",
           file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Create and checkout a new branch. Update 'example.txt' and commit
checkout(repo, "new_branch", create=TRUE)
writeLines(c("First line.", "Second line."),
           file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Second commit message")

## Checkout 'master' branch
checkout(repo, "master", force = TRUE)

## Display 'example.txt'
readLines(file.path(path, "example.txt"))

## Merge and display brief summary of the fast-forward merge
merge(repo, "new_branch")

## Display 'example.txt'
readLines(file.path(path, "example.txt"))
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/status.r
\docType{methods}
\name{status}
\alias{status}
\alias{status,missing-method}
\alias{status,git_repository-method}
\title{Status}
\usage{
status(repo, staged = TRUE, unstaged = TRUE, untracked = TRUE,
  ignored = FALSE, all_untracked = FALSE, ...)

\S4method{status}{missing}(repo, staged = TRUE, unstaged = TRUE,
  untracked = TRUE, ignored = FALSE, all_untracked = FALSE, ...)

\S4method{status}{git_repository}(repo, staged = TRUE, unstaged = TRUE,
  untracked = TRUE, ignored = FALSE, all_untracked = FALSE, ...)
}
\arguments{
\item{repo}{The repository \code{object}
\code{\linkS4class{git_repository}} to get status from. If the
\code{repo} argument is missing, the repository is searched
for with \code{\link{discover_repository}} in the current
working directory.}

\item{staged}{Include staged files. Default TRUE.}

\item{unstaged}{Include unstaged files. Default TRUE.}

\item{untracked}{Include untracked files and directories. Default
TRUE.}

\item{ignored}{Include ignored files. Default FALSE.}

\item{all_untracked}{Shows individual files in untracked
directories if \code{untracked} is \code{TRUE}.}

\item{...}{Additional arguments to status.}
}
\value{
S3 class \code{git_status} with repository status
}
\description{
Display state of the repository working directory and the staging
area.
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user
config(repo, user.name="Alice", user.email="alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test.txt"))

## Check status; untracked file
status(repo)

## Add file
add(repo, "test.txt")

## Check status; staged file
status(repo)

## Commit
commit(repo, "First commit message")

## Check status; clean
status(repo)

## Change the file
writeLines(c("Hello again!", "Here is a second line", "And a third"),
           file.path(path, "test.txt"))

## Check status; unstaged file
status(repo)

## Add file and commit
add(repo, "test.txt")
commit(repo, "Second commit message")

## Check status; clean
status(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/push.R
\name{push}
\alias{push}
\title{Push}
\usage{
push(
  object = ".",
  name = NULL,
  refspec = NULL,
  force = FALSE,
  credentials = NULL,
  set_upstream = FALSE
)
}
\arguments{
\item{object}{path to repository, or a \code{git_repository} or
\code{git_branch}.}

\item{name}{The remote's name. Default is NULL.}

\item{refspec}{The refspec to be pushed. Default is NULL.}

\item{force}{Force your local revision to the remote repo. Use it
with care. Default is FALSE.}

\item{credentials}{The credentials for remote repository
access. Default is NULL. To use and query an ssh-agent for the
ssh key credentials, let this parameter be NULL (the default).}

\item{set_upstream}{Set the current local branch to track the
remote branch. Default is FALSE.}
}
\value{
invisible(NULL)
}
\description{
Push
}
\examples{
\dontrun{
## Initialize two temporary repositories
path_bare <- tempfile(pattern="git2r-")
path_repo <- tempfile(pattern="git2r-")
dir.create(path_bare)
dir.create(path_repo)
repo_bare <- init(path_bare, bare = TRUE)

## Clone the bare repository. This creates remote-tracking
## branches for each branch in the cloned repository.
repo <- clone(path_bare, path_repo)

## Config user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Write to a file and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path_repo, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Push commits from repository to bare repository
push(repo, "origin", "refs/heads/master")

## Now, unset the remote-tracking branch to NULL to demonstrate
## the 'set_upstream' argument. Then push with 'set_upstream = TRUE'
## to add the upstream tracking branch to branch 'master' again.
branch_get_upstream(repository_head(repo))
branch_set_upstream(repository_head(repo), NULL)
branch_get_upstream(repository_head(repo))
push(repo, "origin", "refs/heads/master", set_upstream = TRUE)
branch_get_upstream(repository_head(repo))

## Change file and commit
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path_repo, "example.txt"))
add(repo, "example.txt")
commit(repo, "Second commit message")

## Push commits from repository to bare repository
push(repo)

## List commits in repository and bare repository
commits(repo)
commits(repo_bare)
}
}
\seealso{
\code{\link{cred_user_pass}}, \code{\link{cred_ssh_key}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/index.R
\name{rm_file}
\alias{rm_file}
\title{Remove files from the working tree and from the index}
\usage{
rm_file(repo = ".", path = NULL)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{path}{character vector with filenames to remove. Only files
known to Git are removed.}
}
\value{
invisible(NULL)
}
\description{
Remove files from the working tree and from the index
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "file-to-remove.txt"))

## Add file to repository
add(repo, "file-to-remove.txt")
commit(repo, "First commit message")

## Remove file
rm_file(repo, "file-to-remove.txt")

## View status of repository
status(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.r
\docType{methods}
\name{coerce-git_commit-method}
\alias{coerce-git_commit-method}
\alias{coerce,git_commit,data.frame-method}
\title{Coerce a commit to a \code{data.frame}}
\arguments{
\item{from}{The commit \code{object}}
}
\value{
\code{data.frame}
}
\description{
The commit is coerced to a \code{data.frame}
}
\details{
The \code{data.frame} have the following columns:
\describe{

\item{sha}{
    The 40 character hexadecimal string of the SHA-1
  }

\item{summary}{
    the short "summary" of the git commit message.
  }

\item{message}{
    the full message of a commit
  }

\item{author}{
    full name of the author
  }

\item{email}{
    email of the author
  }

\item{when}{
    time when the commit happened
  }

}
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name="Alice", user.email="alice@example.org")

## Create a file and commit
writeLines("Example file",  file.path(path, "example.txt"))
add(repo, "example.txt")
c1 <- commit(repo, "Commit message")

## Coerce the commit to a data.frame
df <- as(c1, "data.frame")
df
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/contributions.R
\name{contributions}
\alias{contributions}
\title{Contributions}
\usage{
contributions(
  repo = ".",
  breaks = c("month", "year", "quarter", "week", "day"),
  by = c("commits", "author")
)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{breaks}{Default is \code{month}. Change to year, quarter,
week or day as necessary.}

\item{by}{Contributions by "commits" or "author". Default is "commits".}
}
\value{
A \code{data.frame} with contributions.
}
\description{
See contributions to a Git repo
}
\examples{
\dontrun{
## Create directories and initialize repositories
path_bare <- tempfile(pattern="git2r-")
path_repo_1 <- tempfile(pattern="git2r-")
path_repo_2 <- tempfile(pattern="git2r-")
dir.create(path_bare)
dir.create(path_repo_1)
dir.create(path_repo_2)
repo_bare <- init(path_bare, bare = TRUE)

## Clone to repo 1 and config user
repo_1 <- clone(path_bare, path_repo_1)
config(repo_1, user.name = "Alice", user.email = "alice@example.org")

## Add changes to repo 1 and push to bare
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path_repo_1, "test.txt"))
add(repo_1, "test.txt")
commit(repo_1, "First commit message")

## Add more changes to repo 1
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path_repo_1, "test.txt"))
add(repo_1, "test.txt")
commit(repo_1, "Second commit message")

## Push to bare
push(repo_1, "origin", "refs/heads/master")

## Clone to repo 2
repo_2 <- clone(path_bare, path_repo_2)
config(repo_2, user.name = "Bob", user.email = "bob@example.org")

## Add changes to repo 2
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
  "minim veniam, quis nostrud exercitation ullamco laboris nisi ut")
writeLines(lines, file.path(path_repo_2, "test.txt"))
add(repo_2, "test.txt")
commit(repo_2, "Third commit message")

## Push to bare
push(repo_2, "origin", "refs/heads/master")

## Pull changes to repo 1
pull(repo_1)

## View contributions by day
contributions(repo_1)

## View contributions by author and day
contributions(repo_1, by = "author")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/note.R
\name{note_default_ref}
\alias{note_default_ref}
\title{Default notes reference}
\usage{
note_default_ref(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
Character vector of length one with name of default notes
    reference
}
\description{
Get the default notes reference for a repository
}
\examples{
\dontrun{
## Create and initialize a repository in a temporary directory
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## View default notes reference
note_default_ref(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{in_repository}
\alias{in_repository}
\title{Determine if a directory is in a git repository}
\usage{
in_repository(path = ".")
}
\arguments{
\item{path}{The path to the directory.}
}
\value{
TRUE if directory is in a git repository else FALSE
}
\description{
The lookup start from path and walk across parent directories if
nothing has been found.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Check if path is in a git repository
in_repository(path)

## Check if working directory is in a git repository
setwd(path)
in_repository()
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.r
\docType{methods}
\name{coerce-git_tree-data.frame-method}
\alias{coerce-git_tree-data.frame-method}
\alias{coerce,git_tree,data.frame-method}
\title{Coerce entries in a git_tree to a \code{data.frame}}
\arguments{
\item{from}{The tree \code{object}}
}
\value{
\code{data.frame}
}
\description{
The entries in a tree are coerced to a \code{data.frame}
}
\details{
The \code{data.frame} have the following columns:
\describe{

\item{filemode}{
    The UNIX file attributes of a tree entry
  }

\item{type}{
    String representation of the tree entry type
  }

\item{sha}{
    The sha of a tree entry
  }

\item{name}{
    The filename of a tree entry
  }

}
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
dir.create(file.path(path, "subfolder"))
repo <- init(path)

## Create a user
config(repo, user.name="Alice", user.email="alice@example.org")

## Create three files and commit
writeLines("First file",  file.path(path, "example-1.txt"))
writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
writeLines("Third file",  file.path(path, "example-3.txt"))
add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
new_commit <- commit(repo, "Commit message")

## Coerce tree to a data.frame
df <- as(tree(new_commit), "data.frame")
df
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remote.R
\name{remotes}
\alias{remotes}
\title{Get the configured remotes for a repo}
\usage{
remotes(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
Character vector with remotes
}
\description{
Get the configured remotes for a repo
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Add a remote
remote_add(repo, "playground", "https://example.org/git2r/playground")
remotes(repo)
remote_url(repo, "playground")

## Rename a remote
remote_rename(repo, "playground", "foobar")
remotes(repo)
remote_url(repo, "foobar")

## Set remote url
remote_set_url(repo, "foobar", "https://example.org/git2r/foobar")
remotes(repo)
remote_url(repo, "foobar")

## Remove a remote
remote_remove(repo, "foobar")
remotes(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/libgit2.r
\name{libgit2_sha}
\alias{libgit2_sha}
\title{SHA of the libgit2 library}
\usage{
libgit2_sha()
}
\value{
The 40 character hexadecimal string of the SHA-1
}
\description{
SHA of the libgit2 library that the bundled source code is based
on
}
\examples{
libgit2_sha()
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/revparse.R
\name{revparse_single}
\alias{revparse_single}
\title{Revparse}
\usage{
revparse_single(repo = ".", revision = NULL)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{revision}{The revision string, see
http://git-scm.com/docs/git-rev-parse.html#_specifying_revisions}
}
\value{
a \code{git_commit} or \code{git_tag} or \code{git_tree}
object
}
\description{
Find object specified by revision.
}
\examples{
\dontrun{
## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "First commit message")

# Change file and commit
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Second commit message")

revparse_single(repo, "HEAD^")
revparse_single(repo, "HEAD:test.txt")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{summary.git_repository}
\alias{summary.git_repository}
\title{Summary of repository}
\usage{
\method{summary}{git_repository}(object, ...)
}
\arguments{
\item{object}{The repository \code{object}}

\item{...}{Additional arguments affecting the summary produced.}
}
\value{
None (invisible 'NULL').
}
\description{
Summary of repository
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test.txt"))
summary(repo)

## Add file
add(repo, "test.txt")
summary(repo)

## Commit
commit(repo, "First commit message")
summary(repo)

## Change the file
writeLines(c("Hello again!", "Here is a second line", "And a third"),
           file.path(path, "test.txt"))
summary(repo)

## Add file and commit
add(repo, "test.txt")
commit(repo, "Second commit message")
summary(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/index.r
\docType{methods}
\name{index_remove_bypath}
\alias{index_remove_bypath}
\alias{index_remove_bypath,git_repository,character-method}
\title{Remove an index entry corresponding to a file on disk}
\usage{
index_remove_bypath(repo, path)

\S4method{index_remove_bypath}{git_repository,character}(repo, path)
}
\arguments{
\item{repo}{The repository \code{object}.}

\item{path}{character vector with filenames to remove. The path
must be relative to the repository's working folder. It may
exist. If this file currently is the result of a merge
conflict, this file will no longer be marked as
conflicting. The data about the conflict will be moved to the
"resolve undo" (REUC) section.}
}
\value{
invisible(NULL)
}
\description{
Remove an index entry corresponding to a file on disk
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name="Alice", user.email="alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "file-to-remove.txt"))

## Add file to repository
add(repo, "file-to-remove.txt")

## View status of repository
status(repo)

## Remove file
index_remove_bypath(repo, "file-to-remove.txt")

## View status of repository
status(repo)
}

}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.r
\docType{methods}
\name{branch_remote_url}
\alias{branch_remote_url}
\alias{branch_remote_url,git_branch-method}
\title{Remote url of a branch}
\usage{
branch_remote_url(branch)

\S4method{branch_remote_url}{git_branch}(branch)
}
\arguments{
\item{branch}{The branch}
}
\value{
character string with remote url
}
\description{
Remote url of a branch
}
\examples{
\dontrun{
## Initialize two temporary repositories
path_bare <- tempfile(pattern="git2r-")
path_repo <- tempfile(pattern="git2r-")
dir.create(path_bare)
dir.create(path_repo)
repo_bare <- init(path_bare, bare = TRUE)
repo <- clone(path_bare, path_repo)

## Config user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")

## Write to a file and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           file.path(path_repo, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Push commits from repository to bare repository
## Adds an upstream tracking branch to branch 'master'
push(repo, "origin", "refs/heads/master")

## Get remote url of tracking branch to branch 'master'
branch_remote_url(branch_get_upstream(head(repo)))
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/note.R
\name{note_create}
\alias{note_create}
\title{Add note for a object}
\usage{
note_create(
  object = NULL,
  message = NULL,
  ref = NULL,
  author = NULL,
  committer = NULL,
  force = FALSE
)
}
\arguments{
\item{object}{The object to annotate (git_blob, git_commit or
git_tree).}

\item{message}{Content of the note to add}

\item{ref}{Canonical name of the reference to use. Default is
\code{note_default_ref}.}

\item{author}{Signature of the notes note author}

\item{committer}{Signature of the notes note committer}

\item{force}{Overwrite existing note. Default is FALSE}
}
\value{
git_note
}
\description{
Add note for a object
}
\examples{
\dontrun{
## Create and initialize a repository in a temporary directory
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit_1 <- commit(repo, "Commit message 1")

## Create another commit
writeLines(c("Hello world!",
             "HELLO WORLD!"),
           file.path(path, "example.txt"))
add(repo, "example.txt")
commit_2 <- commit(repo, "Commit message 2")

## Check that notes is an empty list
notes(repo)

## Create note in default namespace
note_create(commit_1, "Note-1")

## Create note in named (review) namespace
note_create(commit_1, "Note-2", ref="refs/notes/review")
note_create(commit_2, "Note-3", ref="review")

## Create note on blob and tree
note_create(tree(commit_1), "Note-4")
note_create(tree(commit_1)["example.txt"], "Note-5")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reflog.R
\name{reflog}
\alias{reflog}
\title{List and view reflog information}
\usage{
reflog(repo = ".", refname = "HEAD")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{refname}{The name of the reference to list. 'HEAD' by
default.}
}
\value{
S3 class \code{git_reflog} with git_reflog_entry objects.
}
\description{
List and view reflog information
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Write to a file and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Change file and commit
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Second commit message")

## Change file again and commit
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
  "minim veniam, quis nostrud exercitation ullamco laboris nisi ut")
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Third commit message")

## View reflog
reflog(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reference.r
\docType{methods}
\name{references}
\alias{references}
\alias{references,missing-method}
\alias{references,git_repository-method}
\title{Get all references that can be found in a repository.}
\usage{
references(repo)

\S4method{references}{missing}()

\S4method{references}{git_repository}(repo)
}
\arguments{
\item{repo}{The repository \code{object}
\code{\linkS4class{git_repository}}. If the \code{repo} argument
is missing, the repository is searched for with
\code{\link{discover_repository}} in the current working
directory.}
}
\value{
Character vector with references
}
\description{
Get all references that can be found in a repository.
}
\examples{
\dontrun{
## Initialize two temporary repositories
path_bare <- tempfile(pattern="git2r-")
path_repo <- tempfile(pattern="git2r-")
dir.create(path_bare)
dir.create(path_repo)
repo_bare <- init(path_bare, bare = TRUE)
repo <- clone(path_bare, path_repo)

## Config user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")

## Write to a file and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           file.path(path_repo, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Push commits from repository to bare repository
## Adds an upstream tracking branch to branch 'master'
push(repo, "origin", "refs/heads/master")

## Add tag to HEAD
tag(repo, "v1.0", "First version")

## Create a note
note_create(commits(repo)[[1]], "My note")

## List all references in repository
references(repo)
}

}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blob.R
\name{blob_create}
\alias{blob_create}
\title{Create blob from file on disk}
\usage{
blob_create(repo = ".", path = NULL, relative = TRUE)
}
\arguments{
\item{repo}{The repository where the blob(s) will be written. Can
be a bare repository. A \code{git_repository} object, or a
path to a repository, or \code{NULL}.  If the \code{repo}
argument is \code{NULL}, the repository is searched for with
\code{\link{discover_repository}} in the current working
directory.}

\item{path}{The file(s) from which the blob will be created.}

\item{relative}{TRUE if the file(s) from which the blob will be
created is relative to the repository's working dir. Default
is TRUE.}
}
\value{
list of S3 class git_blob \code{objects}
}
\description{
Read a file from the filesystem and write its content to the
Object Database as a loose blob. The method is vectorized and
accepts a vector of files to create blobs from.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create blobs from files relative to workdir
writeLines("Hello, world!", file.path(path, "example-1.txt"))
writeLines("test content", file.path(path, "example-2.txt"))
blob_list_1 <- blob_create(repo, c("example-1.txt",
                                   "example-2.txt"))

## Create blobs from files not relative to workdir
temp_file_1 <- tempfile()
temp_file_2 <- tempfile()
writeLines("Hello, world!", temp_file_1)
writeLines("test content", temp_file_2)
blob_list_2 <- blob_create(repo, c(temp_file_1, temp_file_2),
                           relative = FALSE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.r
\docType{methods}
\name{init}
\alias{init}
\alias{init,character-method}
\title{Init a repository}
\usage{
init(path, bare = FALSE)

\S4method{init}{character}(path, bare = FALSE)
}
\arguments{
\item{path}{A path to where to init a git repository}

\item{bare}{If TRUE, a Git repository without a working directory
is created at the pointed path. If FALSE, provided path will be
considered as the working directory into which the .git directory
will be created.}
}
\value{
A S4 \code{\linkS4class{git_repository}} object
}
\description{
Init a repository
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
is_bare(repo)

## Initialize a bare repository
path_bare <- tempfile(pattern="git2r-")
dir.create(path_bare)
repo_bare <- init(path_bare, bare = TRUE)
is_bare(repo_bare)
}
}
\seealso{
\link{repository}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reset.R
\name{reset}
\alias{reset}
\title{Reset current HEAD to the specified state}
\usage{
reset(object, reset_type = c("soft", "mixed", "hard"), path = NULL)
}
\arguments{
\item{object}{Either a \code{git_commit}, a \code{git_repository}
or a character vector. If \code{object} is a
\code{git_commit}, HEAD is moved to the \code{git_commit}. If
\code{object} is a \code{git_repository}, resets the index
entries in the \code{path} argument to their state at HEAD. If
\code{object} is a character vector with paths, resets the
index entries in \code{object} to their state at HEAD if the
current working directory is in a repository.}

\item{reset_type}{If object is a 'git_commit', the kind of reset
operation to perform. 'soft' means the HEAD will be moved to
the commit. 'mixed' reset will trigger a 'soft' reset, plus
the index will be replaced with the content of the commit
tree. 'hard' reset will trigger a 'mixed' reset and the
working directory will be replaced with the content of the
index.}

\item{path}{If object is a 'git_repository', resets the index
entries for all paths to their state at HEAD.}
}
\value{
invisible NULL
}
\description{
Reset current HEAD to the specified state
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

# Configure a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
writeLines("Hello world!", file.path(path, "test-1.txt"))
add(repo, "test-1.txt")
commit_1 <- commit(repo, "Commit message")

## Change and stage the file
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-1.txt"))
add(repo, "test-1.txt")
status(repo)

## Unstage file
reset(repo, path = "test-1.txt")
status(repo)

## Make one more commit
add(repo, "test-1.txt")
commit(repo, "Next commit message")

## Create one more file
writeLines("Hello world!", file.path(path, "test-2.txt"))

## 'soft' reset to first commit and check status
reset(commit_1)
status(repo)

## 'mixed' reset to first commit and check status
commit(repo, "Next commit message")
reset(commit_1, "mixed")
status(repo)

## 'hard' reset to first commit and check status
add(repo, "test-1.txt")
commit(repo, "Next commit message")
reset(commit_1, "hard")
status(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blame.R
\name{blame}
\alias{blame}
\title{Get blame for file}
\usage{
blame(repo = ".", path = NULL)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{path}{Path to the file to consider}
}
\value{
git_blame object with the following entries:
\describe{
  \item{path}{
    The path to the file of the blame
  }
  \item{hunks}{
    List of blame hunks
  }
  \item{repo}{
    The git_repository that contains the file
  }
}
\describe{
  \item{lines_in_hunk}{
    The number of lines in this hunk
  }
  \item{final_commit_id}{
    The sha of the commit where this line was last changed
  }
  \item{final_start_line_number}{
    The 1-based line number where this hunk begins, in the final
    version of the file
  }
  \item{final_signature}{
    Final committer
  }
  \item{orig_commit_id}{
    The sha of the commit where this hunk was found. This will usually
    be the same as 'final_commit_id'.
  }
  \item{orig_start_line_number}{
     The 1-based line number where this hunk begins in the file
     named by 'orig_path' in the commit specified by 'orig_commit_id'.
  }
  \item{orig_signature}{
    Origin committer
  }
  \item{orig_path}{
    The path to the file where this hunk originated, as of the commit
    specified by 'orig_commit_id'
  }
  \item{boundary}{
    TRUE iff the hunk has been tracked to a boundary commit.
  }
  \item{repo}{
    The \code{git_repository} object that contains the blame hunk
  }
}
}
\description{
Get blame for file
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a first user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Create a second user and change the file
config(repo, user.name = "Bob", user.email = "bob@example.org")
writeLines(c("Hello world!", "HELLO WORLD!", "HOLA"),
           file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Second commit message")

## Check blame
blame(repo, "example.txt")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remote.r
\docType{methods}
\name{remote_set_url}
\alias{remote_set_url}
\alias{remote_set_url,git_repository,character,character-method}
\title{Set the remote's url in the configuration}
\usage{
remote_set_url(repo, name, url)

\S4method{remote_set_url}{git_repository,character,character}(repo, name, url)
}
\arguments{
\item{repo}{The repository in which to perform the change}

\item{name}{The name of the remote}

\item{url}{The \code{url} to set}
}
\value{
NULL, invisibly
}
\description{
This assumes the common case of a single-url remote and will
otherwise raise an error.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Add a remote
remote_add(repo, "playground", "https://example.org/git2r/playground")
remotes(repo)
remote_url(repo, "playground")

## Rename a remote
remote_rename(repo, "playground", "foobar")
remotes(repo)
remote_url(repo, "foobar")

## Set remote url
remote_set_url(repo, "foobar", "https://example.org/git2r/foobar")
remotes(repo)
remote_url(repo, "foobar")

## Remove a remote
remote_remove(repo, "foobar")
remotes(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tag.R
\name{is_tag}
\alias{is_tag}
\title{Check if object is a git_tag object}
\usage{
is_tag(object)
}
\arguments{
\item{object}{Check if object is a git_tag object}
}
\value{
TRUE if object is a git_tag, else FALSE
}
\description{
Check if object is a git_tag object
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Commit a text file
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Create tag
tag(repo, "Tagname", "Tag message")

is_tag(tags(repo)[[1]])
is_tag(last_commit(repo))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{config}
\alias{config}
\title{Config}
\usage{
config(repo = NULL, global = FALSE, user.name, user.email, ...)
}
\arguments{
\item{repo}{The \code{repository}. Default is NULL.}

\item{global}{Write option(s) to global configuration
file. Default is FALSE.}

\item{user.name}{The user name. Use NULL to delete the entry}

\item{user.email}{The e-mail address. Use NULL to delete the entry}

\item{...}{Additional options to write or delete from the
configuration.}
}
\value{
S3 class \code{git_config}. When writing options, the
configuration is returned invisible.
}
\description{
Config file management. To display the configuration variables,
call method \code{config} without the \code{user.name},
\code{user.email} or \code{...} options.
}
\details{
There are two ways git2r can find the local repository when
writing local options (1) Use the \code{repo} argument. (2) If the
\code{repo} argument is \code{NULL} but the current working
directory is inside the local repository, then \code{git2r} uses
that repository.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern = "git2r-")
dir.create(path)
repo <- init(path)

## Set user name and email.
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Display configuration
config(repo)

## Delete user email.
config(repo, user.email = NULL)

## Display configuration
config(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_signature-class}
\alias{git_signature-class}
\title{S4 class to handle a git signature}
\description{
Class \code{"git_signature"}
}
\section{Slots}{

\describe{
  \item{name}{
    The full name of the author.
  }
  \item{email}{
    Email of the author.
  }
  \item{when}{
    Time when the action happened.
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{is_bare}
\alias{is_bare}
\title{Check if repository is bare}
\usage{
is_bare(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
\code{TRUE} if bare repository, else \code{FALSE}
}
\description{
Check if repository is bare
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
is_bare(repo)

## Initialize a bare repository
path_bare <- tempfile(pattern="git2r-")
dir.create(path_bare)
repo_bare <- init(path_bare, bare = TRUE)
is_bare(repo_bare)
}
}
\seealso{
\link{init}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/signature.R
\name{default_signature}
\alias{default_signature}
\title{Get the signature}
\usage{
default_signature(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
A \code{git_signature} object with entries:
}
\description{
Get the signature according to the repository's configuration
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Get the default signature
default_signature(repo)

## Change user
config(repo, user.name = "Bob", user.email = "bob@example.org")

## Get the default signature
default_signature(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_blame-class}
\alias{git_blame-class}
\title{S4 class to handle a git blame for a single file}
\description{
Class \code{"git_blame"}
}
\section{Slots}{

\describe{
  \item{path}{
    The path to the file of the blame
  }
  \item{hunks}{
    List of blame hunks
  }
  \item{repo}{
    The S4 class git_repository that contains the file
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.R
\name{diff.git_repository}
\alias{diff.git_repository}
\alias{diff.git_tree}
\title{Changes between commits, trees, working tree, etc.}
\usage{
\method{diff}{git_repository}(
  x,
  index = FALSE,
  as_char = FALSE,
  filename = NULL,
  context_lines = 3,
  interhunk_lines = 0,
  old_prefix = "a",
  new_prefix = "b",
  id_abbrev = NULL,
  path = NULL,
  max_size = NULL,
  ...
)

\method{diff}{git_tree}(
  x,
  new_tree = NULL,
  index = FALSE,
  as_char = FALSE,
  filename = NULL,
  context_lines = 3,
  interhunk_lines = 0,
  old_prefix = "a",
  new_prefix = "b",
  id_abbrev = NULL,
  path = NULL,
  max_size = NULL,
  ...
)
}
\arguments{
\item{x}{A \code{git_repository} object or the old \code{git_tree}
object to compare to.}

\item{index}{\describe{
  \item{\emph{When object equals a git_repository}}{
    Whether to compare the index to HEAD. If FALSE (the default),
    then the working tree is compared to the index.
  }
  \item{\emph{When object equals a git_tree}}{
    Whether to use the working directory (by default), or the index
    (if set to TRUE) in the comparison to \code{object}.
  }
}}

\item{as_char}{logical: should the result be converted to a
character string?. Default is FALSE.}

\item{filename}{If as_char is TRUE, then the diff can be written
to a file with name filename (the file is overwritten if it
exists). Default is NULL.}

\item{context_lines}{The number of unchanged lines that define the
boundary of a hunk (and to display before and after). Defaults
to 3.}

\item{interhunk_lines}{The maximum number of unchanged lines
between hunk boundaries before the hunks will be merged into
one. Defaults to 0.}

\item{old_prefix}{The virtual "directory" prefix for old file
names in hunk headers. Default is "a".}

\item{new_prefix}{The virtual "directory" prefix for new file
names in hunk headers. Defaults to "b".}

\item{id_abbrev}{The abbreviation length to use when formatting
object ids. Defaults to the value of 'core.abbrev' from the
config, or 7 if NULL.}

\item{path}{A character vector of paths / fnmatch patterns to
constrain diff. Default is NULL which include all paths.}

\item{max_size}{A size (in bytes) above which a blob will be
marked as binary automatically; pass a negative value to
disable. Defaults to 512MB when max_size is NULL.}

\item{...}{Not used.}

\item{new_tree}{The new git_tree object to compare, or NULL.  If
NULL, then we use the working directory or the index (see the
\code{index} argument).}
}
\value{
A \code{git_diff} object if as_char is FALSE. If as_char
    is TRUE and filename is NULL, a character string, else NULL.
}
\description{
Changes between commits, trees, working tree, etc.
}
\section{Line endings}{


Different operating systems handle line endings
differently. Windows uses both a carriage-return character and a
linefeed character to represent a newline in a file. While Linux
and macOS use only the linefeed character for a newline in a
file. To avoid problems in your diffs, you can configure Git to
properly handle line endings using the \verb{core.autocrlf}
setting in the Git config file, see the Git documentation
(\url{https://git-scm.com/}).
}

\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add, commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message")

## Change the file
writeLines(c("Hello again!", "Here is a second line", "And a third"),
           file.path(path, "test.txt"))

## diff between index and workdir
diff_1 <- diff(repo)
summary(diff_1)
cat(diff(repo, as_char=TRUE))

## Diff between index and HEAD is empty
diff_2 <- diff(repo, index=TRUE)
summary(diff_2)
cat(diff(repo, index=TRUE, as_char=TRUE))

## Diff between tree and working dir, same as diff_1
diff_3 <- diff(tree(commits(repo)[[1]]))
summary(diff_3)
cat(diff(tree(commits(repo)[[1]]), as_char=TRUE))

## Add changes, diff between index and HEAD is the same as diff_1
add(repo, "test.txt")
diff_4 <- diff(repo, index=TRUE)
summary(diff_4)
cat(diff(repo, index=TRUE, as_char=TRUE))

## Diff between tree and index
diff_5 <- diff(tree(commits(repo)[[1]]), index=TRUE)
summary(diff_5)
cat(diff(tree(commits(repo)[[1]]), index=TRUE, as_char=TRUE))

## Diff between two trees
commit(repo, "Second commit")
tree_1 <- tree(commits(repo)[[2]])
tree_2 <- tree(commits(repo)[[1]])
diff_6 <- diff(tree_1, tree_2)
summary(diff_6)
cat(diff(tree_1, tree_2, as_char=TRUE))

## Binary files
set.seed(42)
writeBin(as.raw((sample(0:255, 1000, replace=TRUE))),
         con=file.path(path, "test.bin"))
add(repo, "test.bin")
diff_7 <- diff(repo, index=TRUE)
summary(diff_7)
cat(diff(repo, index=TRUE, as_char=TRUE))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{discover_repository}
\alias{discover_repository}
\title{Find path to repository for any file}
\usage{
discover_repository(path = ".", ceiling = NULL)
}
\arguments{
\item{path}{A character vector specifying the path to a file or
folder}

\item{ceiling}{The default is to not use the ceiling argument and
start the lookup from path and walk across parent
directories. When ceiling is 0, the lookup is only in
path. When ceiling is 1, the lookup is in both the path and
the parent to path.}
}
\value{
Character vector with path (terminated by a file
    separator) to repository or NULL if this cannot be
    established.
}
\description{
Find path to repository for any file
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "example-1.txt"))
add(repo, "example-1.txt")
commit(repo, "First commit message")

## Create a second file. The file is not added for version control
## in the repository.
dir.create(file.path(path, "example"))
file_2 <- file.path(path, "example/example-2.txt")
writeLines("Not under version control", file_2)

## Find the path to the repository using the path to the second file
discover_repository(file_2)

## Demonstrate the 'ceiling' argument
wd <- workdir(repo)
dir.create(file.path(wd, "temp"))

## Lookup repository in 'file.path(wd, "temp")'. Should return NULL
discover_repository(file.path(wd, "temp"), ceiling = 0)

## Lookup repository in parent to 'file.path(wd, "temp")'.
## Should not return NULL
discover_repository(file.path(wd, "temp"), ceiling = 1)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{repository}
\alias{repository}
\title{Open a repository}
\usage{
repository(path = ".", discover = TRUE)
}
\arguments{
\item{path}{A path to an existing local git repository.}

\item{discover}{Discover repository from path. Default is TRUE.}
}
\value{
A \code{git_repository} object with entries:
\describe{
  \item{path}{
    Path to a git repository
  }
}
}
\description{
Open a repository
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

# Configure a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
writeLines("Hello world!", file.path(path, "test-1.txt"))
add(repo, 'test-1.txt')
commit_1 <- commit(repo, "Commit message")

## Make one more commit
writeLines(c("Hello world!", "HELLO WORLD!"),
           file.path(path, "test-1.txt"))
add(repo, 'test-1.txt')
commit(repo, "Next commit message")

## Create one more file
writeLines("Hello world!",
           file.path(path, "test-2.txt"))

## Brief summary of repository
repo

## Summary of repository
summary(repo)

## Workdir of repository
workdir(repo)

## Check if repository is bare
is_bare(repo)

## Check if repository is empty
is_empty(repo)

## Check if repository is a shallow clone
is_shallow(repo)

## List all references in repository
references(repo)

## List all branches in repository
branches(repo)

## Get HEAD of repository
repository_head(repo)

## Check if HEAD is head
is_head(repository_head(repo))

## Check if HEAD is local
is_local(repository_head(repo))

## List all tags in repository
tags(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.R
\name{branch_create}
\alias{branch_create}
\title{Create a branch}
\usage{
branch_create(commit = last_commit(), name = NULL, force = FALSE)
}
\arguments{
\item{commit}{Commit to which the branch should point. The default
is to use the \code{last_commit()} function to determine the
commit to which the branch should point.}

\item{name}{Name for the branch}

\item{force}{Overwrite existing branch. Default = FALSE}
}
\value{
invisible git_branch object
}
\description{
Create a branch
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
lines <- "Hello world!"
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit_1 <- commit(repo, "First commit message")

## Create a branch
branch_1 <- branch_create(commit_1, name = "test-branch")

## Add one more commit
lines <- c("Hello world!", "HELLO WORLD!")
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit_2 <- commit(repo, "Another commit message")

## Create a branch with the same name should fail
try(branch_create(commit_2, name = "test-branch"), TRUE)

## Force it
branch_2 <- branch_create(commit_2, name = "test-branch", force = TRUE)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{is_empty}
\alias{is_empty}
\title{Check if repository is empty}
\usage{
is_empty(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
\code{TRUE} if repository is empty else \code{FALSE}.
}
\description{
Check if repository is empty
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Check if it's an empty repository
is_empty(repo)

## Commit a file
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Check if it's an empty repository
is_empty(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge.r
\docType{methods}
\name{merge,git_repository,character-method}
\alias{merge,git_repository,character-method}
\alias{merge,git_branch,missing-method}
\title{Merge a branch into HEAD}
\usage{
\S4method{merge}{git_repository,character}(x, y, ...,
  commit_on_success = TRUE, merger = default_signature(x))

\S4method{merge}{git_branch,missing}(x, ..., commit_on_success = TRUE,
  merger = default_signature(x@repo))
}
\arguments{
\item{x}{A \code{\linkS4class{git_branch}} or
\code{\linkS4class{git_repository}} object.}

\item{y}{If \code{x} is a \code{\linkS4class{git_repository}}, the
name of the branch to merge into HEAD. Not used if \code{x} is a
\code{\linkS4class{git_branch}}.}

\item{...}{Additional arguments affecting the merge}

\item{commit_on_success}{If there are no conflicts written to the
index, the merge commit will be committed. Default is TRUE.}

\item{merger}{Who made the merge.}
}
\value{
A \code{\linkS4class{git_merge_result}} object.
}
\description{
Merge a branch into HEAD
}
\examples{
\dontrun{
## Create a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
config(repo, user.name="Alice", user.email = "alice@example.org")

## Create a file, add and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message 1")

## Create first branch, checkout, add file and commit
checkout(repo, "branch1", create = TRUE)
writeLines("Branch 1", file.path(path, "branch-1.txt"))
add(repo, "branch-1.txt")
commit(repo, "Commit message branch 1")

## Create second branch, checkout, add file and commit
b_2 <- branch_create(commit_1, "branch2")
checkout(b_2)
writeLines("Branch 2", file.path(path, "branch-2.txt"))
add(repo, "branch-2.txt")
commit(repo, "Commit message branch 2")

## Make a change to 'test.txt'
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Second commit message branch 2")

## Checkout master
checkout(repo, "master", force = TRUE)

## Merge branch 1
merge(repo, "branch1")

## Merge branch 2
merge(repo, "branch2")

## Create third branch, checkout, change file and commit
checkout(repo, "branch3", create=TRUE)
writeLines(c("Lorem ipsum dolor amet sit, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message branch 3")

## Checkout master and create a change that creates a merge conflict
checkout(repo, "master", force=TRUE)
writeLines(c("Lorem ipsum dolor sit amet, adipisicing consectetur elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Some commit message branch 1")

## Merge branch 3
merge(repo, "branch3")

## Check status; Expect to have one unstaged unmerged conflict.
status(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{fetch_heads}
\alias{fetch_heads}
\title{Get updated heads during the last fetch.}
\usage{
fetch_heads(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
list with \code{git_fetch_head} entries. NULL if there is
    no FETCH_HEAD file.
}
\description{
Get updated heads during the last fetch.
}
\examples{
\dontrun{
## Initialize three temporary repositories
path_bare <- tempfile(pattern="git2r-")
path_repo_1 <- tempfile(pattern="git2r-")
path_repo_2 <- tempfile(pattern="git2r-")

dir.create(path_bare)
dir.create(path_repo_1)
dir.create(path_repo_2)

bare_repo <- init(path_bare, bare = TRUE)
repo_1 <- clone(path_bare, path_repo_1)
repo_2 <- clone(path_bare, path_repo_2)

config(repo_1, user.name = "Alice", user.email = "alice@example.org")
config(repo_2, user.name = "Bob", user.email = "bob@example.org")

## Add changes to repo 1
writeLines("Lorem ipsum dolor sit amet",
           con = file.path(path_repo_1, "example.txt"))
add(repo_1, "example.txt")
commit(repo_1, "Commit message")

## Push changes from repo 1 to origin (bare_repo)
push(repo_1, "origin", "refs/heads/master")

## Fetch changes from origin (bare_repo) to repo 2
fetch(repo_2, "origin")

## List updated heads
fetch_heads(repo_2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_diff_hunk-class}
\alias{git_diff_hunk-class}
\title{Git diff hunk}
\description{
Git diff hunk
}
\section{Slots}{

\describe{
  \item{old_start}{
    Starting line number in old_file.
  }
  \item{old_lines}{
    Number of lines in old_file.
  }
  \item{new_start}{
    Starting line number in new_file.
  }
  \item{new_lines}{
    Number of lines in new_file.
  }
  \item{header}{
    Header text.
  }
  \item{lines}{
    List with \code{\linkS4class{git_diff_line}} objects.
  }
}
}

\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge.R
\name{merge.git_branch}
\alias{merge.git_branch}
\alias{merge.git_repository}
\alias{merge.character}
\title{Merge a branch into HEAD}
\usage{
\method{merge}{git_branch}(x, y = NULL, commit_on_success = TRUE, merger = NULL, fail = FALSE, ...)

\method{merge}{git_repository}(x, y = NULL, commit_on_success = TRUE, merger = NULL, fail = FALSE, ...)

\method{merge}{character}(
  x = ".",
  y = NULL,
  commit_on_success = TRUE,
  merger = NULL,
  fail = FALSE,
  ...
)
}
\arguments{
\item{x}{A path (default '.') to a repository, or a
\code{git_repository} object, or a \code{git_branch}.}

\item{y}{If \code{x} is a \code{git_repository}, the name of the
branch to merge into HEAD. Not used if \code{x} is a
\code{git_branch}.}

\item{commit_on_success}{If there are no conflicts written to the
index, the merge commit will be committed. Default is TRUE.}

\item{merger}{Who made the merge. The default (\code{NULL}) is to
use \code{default_signature} for the repository.}

\item{fail}{If a conflict occurs, exit immediately instead of
attempting to continue resolving conflicts. Default is
\code{FALSE}.}

\item{...}{Additional arguments (unused).}
}
\value{
A list of class \code{git_merge_result} with entries:
\describe{
  \item{up_to_date}{
    TRUE if the merge is already up-to-date, else FALSE.
  }
  \item{fast_forward}{
    TRUE if a fast-forward merge, else FALSE.
  }
  \item{conflicts}{
    TRUE if the index contain entries representing file conflicts,
    else FALSE.
  }
  \item{sha}{
    If the merge created a merge commit, the sha of the merge
    commit. NA if no merge commit created.
  }
}
}
\description{
Merge a branch into HEAD
}
\examples{
\dontrun{
## Create a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
config(repo, user.name="Alice", user.email = "alice@example.org")

## Create a file, add and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message 1")

## Create first branch, checkout, add file and commit
checkout(repo, "branch1", create = TRUE)
writeLines("Branch 1", file.path(path, "branch-1.txt"))
add(repo, "branch-1.txt")
commit(repo, "Commit message branch 1")

## Create second branch, checkout, add file and commit
b_2 <- branch_create(commit_1, "branch2")
checkout(b_2)
writeLines("Branch 2", file.path(path, "branch-2.txt"))
add(repo, "branch-2.txt")
commit(repo, "Commit message branch 2")

## Make a change to 'test.txt'
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Second commit message branch 2")

## Checkout master
checkout(repo, "master", force = TRUE)

## Merge branch 1
merge(repo, "branch1")

## Merge branch 2
merge(repo, "branch2")

## Create third branch, checkout, change file and commit
checkout(repo, "branch3", create=TRUE)
writeLines(c("Lorem ipsum dolor amet sit, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message branch 3")

## Checkout master and create a change that creates a merge conflict
checkout(repo, "master", force=TRUE)
writeLines(c("Lorem ipsum dolor sit amet, adipisicing consectetur elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Some commit message branch 1")

## Merge branch 3
merge(repo, "branch3")

## Check status; Expect to have one unstaged unmerged conflict.
status(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.R
\name{branch_target}
\alias{branch_target}
\title{Get target (sha) pointed to by a branch}
\usage{
branch_target(branch = NULL)
}
\arguments{
\item{branch}{The branch}
}
\value{
sha or NA if not a direct reference
}
\description{
Get target (sha) pointed to by a branch
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Get target (sha) pointed to by 'master' branch
branch_target(repository_head(repo))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{summary.git_tree}
\alias{summary.git_tree}
\title{Summary of tree}
\usage{
\method{summary}{git_tree}(object, ...)
}
\arguments{
\item{object}{The tree \code{object}}

\item{...}{Additional arguments affecting the summary produced.}
}
\value{
None (invisible 'NULL').
}
\description{
Summary of tree
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

summary(tree(last_commit(repo)))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.r
\docType{methods}
\name{show,git_diff-method}
\alias{show,git_diff-method}
\alias{show,git_diff-methods}
\title{Show a diff}
\usage{
\S4method{show}{git_diff}(object)
}
\arguments{
\item{object}{The diff \code{object}.}
}
\description{
Show a diff
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user
config(repo, user.name="Alice", user.email="alice@example.org")

## Create a file, add, commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message")

## Change the file
writeLines(c("Hello again!", "Here is a second line", "And a third"),
           file.path(path, "test.txt"))

## Brief summary of diff between index and workdir
diff(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stash.R
\name{stash}
\alias{stash}
\title{Stash}
\usage{
stash(
  repo = ".",
  message = as.character(Sys.time()),
  index = FALSE,
  untracked = FALSE,
  ignored = FALSE,
  stasher = NULL
)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{message}{Optional description. Defaults to current time.}

\item{index}{All changes already added to the index are left
intact in the working directory. Default is FALSE}

\item{untracked}{All untracked files are also stashed and then
cleaned up from the working directory. Default is FALSE}

\item{ignored}{All ignored files are also stashed and then cleaned
up from the working directory. Default is FALSE}

\item{stasher}{Signature with stasher and time of stash}
}
\value{
invisible \code{git_stash} object if anything to stash
    else NULL
}
\description{
Stash
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

# Configure a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

# Create a file, add and commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, 'test.txt')
commit(repo, "Commit message")

# Change file
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))

# Check status of repository
status(repo)

# Create stash in repository
stash(repo)

# Check status of repository
status(repo)

# View stash
stash_list(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{as.data.frame.git_tree}
\alias{as.data.frame.git_tree}
\title{Coerce entries in a git_tree to a \code{data.frame}}
\usage{
\method{as.data.frame}{git_tree}(x, ...)
}
\arguments{
\item{x}{The tree \code{object}}

\item{...}{Additional arguments. Not used.}
}
\value{
\code{data.frame}
}
\description{
The entries in a tree are coerced to a \code{data.frame}
}
\details{
The \code{data.frame} have the following columns:
\describe{

  \item{filemode}{
    The UNIX file attributes of a tree entry
  }

  \item{type}{
    String representation of the tree entry type
  }

  \item{sha}{
    The sha of a tree entry
  }

  \item{name}{
    The filename of a tree entry
  }

}
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
dir.create(file.path(path, "subfolder"))
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create three files and commit
writeLines("First file",  file.path(path, "example-1.txt"))
writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
writeLines("Third file",  file.path(path, "example-3.txt"))
add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
commit(repo, "Commit message")

## Display tree
tree(last_commit(repo))

## Coerce tree to a data.frame
df <- as.data.frame(tree(last_commit(repo)))
df
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.R
\name{branch_set_upstream}
\alias{branch_set_upstream}
\title{Set remote tracking branch}
\usage{
branch_set_upstream(branch = NULL, name)
}
\arguments{
\item{branch}{The branch to configure}

\item{name}{remote-tracking or local branch to set as
upstream. Pass NULL to unset.}
}
\value{
invisible NULL
}
\description{
Set the upstream configuration for a given local branch
}
\examples{
\dontrun{
## Initialize two temporary repositories
path_bare <- tempfile(pattern="git2r-")
path_repo <- tempfile(pattern="git2r-")
dir.create(path_bare)
dir.create(path_repo)
repo_bare <- init(path_bare, bare = TRUE)
repo <- clone(path_bare, path_repo)

## Config user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")

## Write to a file and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           file.path(path_repo, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Push commits from repository to bare repository
## Adds an upstream tracking branch to branch 'master'
push(repo, "origin", "refs/heads/master")

## Unset remote remote tracking branch
branch_get_upstream(repository_head(repo))
branch_set_upstream(repository_head(repo), NULL)
branch_get_upstream(repository_head(repo))

## Set remote tracking branch
branch_set_upstream(repository_head(repo), "origin/master")
branch_get_upstream(repository_head(repo))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blob.R
\name{content}
\alias{content}
\title{Content of blob}
\usage{
content(blob = NULL, split = TRUE, raw = FALSE)
}
\arguments{
\item{blob}{The blob object.}

\item{split}{Split blob content to text lines. Default TRUE.}

\item{raw}{When \code{TRUE}, get the content of the blob as a raw
vector, else as a character vector. Default is \code{FALSE}.}
}
\value{
The content of the blob. NA_character_ if the blob is
    binary and \code{raw} is \code{FALSE}.
}
\description{
Content of blob
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Display content of blob.
content(tree(commits(repo)[[1]])["example.txt"])
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/pull.R
\name{pull}
\alias{pull}
\title{Pull}
\usage{
pull(repo = ".", credentials = NULL, merger = NULL)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{credentials}{The credentials for remote repository
access. Default is NULL. To use and query an ssh-agent for the
ssh key credentials, let this parameter be NULL (the default).}

\item{merger}{Who made the merge, if the merge is non-fast forward
merge that creates a merge commit. The
\code{default_signature} for \code{repo} is used if this
parameter is \code{NULL}.}
}
\value{
A list of class \code{git_merge_result} with entries:
\describe{
  \item{up_to_date}{
    TRUE if the merge is already up-to-date, else FALSE.
  }
  \item{fast_forward}{
    TRUE if a fast-forward merge, else FALSE.
  }
  \item{conflicts}{
    TRUE if the index contain entries representing file conflicts,
    else FALSE.
  }
  \item{sha}{
    If the merge created a merge commit, the sha of the merge
    commit. NA if no merge commit created.
  }
}
}
\description{
Pull
}
\examples{
\dontrun{
## Initialize repositories
path_bare <- tempfile(pattern="git2r-")
path_repo_1 <- tempfile(pattern="git2r-")
path_repo_2 <- tempfile(pattern="git2r-")
dir.create(path_bare)
dir.create(path_repo_1)
dir.create(path_repo_2)
repo_bare <- init(path_bare, bare = TRUE)
repo_1 <- clone(path_bare, path_repo_1)

## Config first user and commit a file
config(repo_1, user.name = "Alice", user.email = "alice@example.org")

## Write to a file and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path_repo_1, "example.txt"))
add(repo_1, "example.txt")
commit(repo_1, "First commit message")

## Push commits from first repository to bare repository
## Adds an upstream tracking branch to branch 'master'
push(repo_1, "origin", "refs/heads/master")

## Clone to second repository
repo_2 <- clone(path_bare, path_repo_2)
config(repo_2, user.name = "Bob", user.email = "bob@example.org")

## Change file and commit
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path_repo_1, "example.txt"))
add(repo_1, "example.txt")
commit(repo_1, "Second commit message")

## Push commits from first repository to bare repository
push(repo_1)

## Pull changes to repo_2
pull(repo_2)

## Change file again and commit. This time in repository 2
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
  "minim veniam, quis nostrud exercitation ullamco laboris nisi ut")
writeLines(lines, file.path(path_repo_2, "example.txt"))
add(repo_2, "example.txt")
commit(repo_2, "Third commit message")

## Push commits from second repository to bare repository
push(repo_2)

## Pull changes to repo_1
pull(repo_1)

## List commits in repositories
commits(repo_1)
commits(repo_2)
commits(repo_bare)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkout.r
\docType{methods}
\name{checkout}
\alias{checkout}
\alias{checkout,git_repository-method}
\alias{checkout,git_branch-method}
\alias{checkout,git_commit-method}
\alias{checkout,git_tag-method}
\title{Checkout}
\usage{
checkout(object, ...)

\S4method{checkout}{git_repository}(object, branch = NULL, create = FALSE,
  force = FALSE, path = NULL, ...)

\S4method{checkout}{git_branch}(object, force = FALSE)

\S4method{checkout}{git_commit}(object, force = FALSE)

\S4method{checkout}{git_tag}(object, force = FALSE)
}
\arguments{
\item{object}{A repository, commit, tag or tree which content will
be used to update the working directory.}

\item{...}{Additional arguments affecting the checkout}

\item{branch}{If object is a repository, the name of the branch to
check out. Default is NULL.}

\item{create}{If object is a repository, then branch is created if
doesn't exist.}

\item{force}{If TRUE, then make working directory match
target. This will throw away local changes. Default is FALSE.}

\item{path}{Limit the checkout operation to only certain
paths. This argument is only used if branch is NULL. Default is
NULL.}
}
\value{
invisible NULL
}
\description{
Update files in the index and working tree to match the content of
the tree pointed at by the treeish object (commit, tag or tree).
Checkout using the default GIT_CHECKOUT_SAFE_CREATE strategy
(force = FALSE) or GIT_CHECKOUT_FORCE (force = TRUE).
}
\examples{
\dontrun{
## Create directories and initialize repositories
path_bare <- tempfile(pattern="git2r-")
path_repo_1 <- tempfile(pattern="git2r-")
path_repo_2 <- tempfile(pattern="git2r-")
dir.create(path_bare)
dir.create(path_repo_1)
dir.create(path_repo_2)
repo_bare <- init(path_bare, bare = TRUE)

## Clone to repo 1 and config user
repo_1 <- clone(path_bare, path_repo_1)
config(repo_1, user.name="Alice", user.email="alice@example.org")

## Add changes to repo 1 and push to bare
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path_repo_1, "test.txt"))
add(repo_1, "test.txt")
commit(repo_1, "First commit message")
push(repo_1, "origin", "refs/heads/master")

## Create and checkout 'dev' branch in repo 1
checkout(repo_1, "dev", create = TRUE)

## Add changes to 'dev' branch in repo 1 and push to bare
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           con = file.path(path_repo_1, "test.txt"))
add(repo_1, "test.txt")
commit(repo_1, "Second commit message")
push(repo_1, "origin", "refs/heads/dev")

## Clone to repo 2
repo_2 <- clone(path_bare, path_repo_2)
config(repo_2, user.name="Bob", user.email="bob@example.org")

## Read content of 'test.txt'
readLines(file.path(path_repo_2, "test.txt"))

## Checkout dev branch
checkout(repo_2, "dev")

## Read content of 'test.txt'
readLines(file.path(path_repo_2, "test.txt"))

## Edit "test.txt" in repo_2
writeLines("Hello world!", con = file.path(path_repo_2, "test.txt"))

## Check status
status(repo_2)

## Checkout "test.txt"
checkout(repo_2, path = "test.txt")

## Check status
status(repo_2)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.R
\name{descendant_of}
\alias{descendant_of}
\title{Descendant}
\usage{
descendant_of(commit = NULL, ancestor = NULL)
}
\arguments{
\item{commit}{a git_commit object. Can also be a tag or a branch,
and in that case the commit will be the target of the tag or
branch.}

\item{ancestor}{a git_commit object to check if ancestor to
\code{commit}. Can also be a tag or a branch, and in that case
the commit will be the target of the tag or branch.}
}
\value{
TRUE if \code{commit} is descendant of \code{ancestor},
    else FALSE
}
\description{
Determine if a commit is the descendant of another commit
}
\examples{
\dontrun{
## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message 1")
tag_1 <- tag(repo, "Tagname1", "Tag message 1")

# Change file and commit
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Commit message 2")
tag_2 <- tag(repo, "Tagname2", "Tag message 2")

descendant_of(commit_1, commit_2)
descendant_of(commit_2, commit_1)
descendant_of(tag_1, tag_2)
descendant_of(tag_2, tag_1)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.R
\name{length.git_diff}
\alias{length.git_diff}
\title{Number of files in git_diff object}
\usage{
\method{length}{git_diff}(x)
}
\arguments{
\item{x}{The git_diff \code{object}}
}
\value{
a non-negative integer
}
\description{
Number of files in git_diff object
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_tree-class}
\alias{git_tree-class}
\title{S4 class to handle a git tree}
\description{
Class \code{"git_tree"}
}
\section{Slots}{

\describe{
  \item{sha}{
    The 40 character hexadecimal string of the SHA-1
  }
  \item{filemode}{
    The UNIX file attributes of a tree entry
  }
  \item{type}{
    String representation of the tree entry type
  }
  \item{id}{
    The sha id of a tree entry
  }
  \item{name}{
    The filename of a tree entry
  }
  \item{repo}{
    The S4 class git_repository that contains the commit
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tag.R
\name{tag_delete}
\alias{tag_delete}
\title{Delete an existing tag reference}
\usage{
tag_delete(object = ".", name = NULL)
}
\arguments{
\item{object}{Can be either the path (default is ".") to a
repository, or a \code{git_repository} object, or a
\code{git_tag} object. or the tag name.}

\item{name}{If the \code{object} argument is a path to a
repository or a \code{git_repository}, the name of the tag to
delete.}
}
\value{
\code{invisible(NULL)}
}
\description{
Delete an existing tag reference
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Commit a text file
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Create two tags
tag(repo, "Tag1", "Tag message 1")
t2 <- tag(repo, "Tag2", "Tag message 2")

## List the two tags in the repository
tags(repo)

## Delete the two tags in the repository
tag_delete(repo, "Tag1")
tag_delete(t2)

## Show the empty list with tags in the repository
tags(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.r
\docType{methods}
\name{summary,git_diff-method}
\alias{summary,git_diff-method}
\alias{summary,git_diff-methods}
\title{Show the summary of a diff}
\usage{
\S4method{summary}{git_diff}(object, ...)
}
\arguments{
\item{object}{The diff \code{object}.}

\item{...}{Additional arguments affecting the summary produced.}
}
\value{
None (invisible 'NULL').
}
\description{
Show the summary of a diff
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user
config(repo, user.name="Alice", user.email="alice@example.org")

## Create a file, add, commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message")

## Change the file
writeLines(c("Hello again!", "Here is a second line", "And a third"),
           file.path(path, "test.txt"))

## diff between index and workdir
diff_1 <- diff(repo)
summary(diff_1)

## Diff between index and HEAD is empty
diff_2 <- diff(repo, index=TRUE)
summary(diff_2)

## Diff between tree and working dir, same as diff_1
diff_3 <- diff(tree(commits(repo)[[1]]))
summary(diff_3)

## Add changes, diff between index and HEAD is the same as diff_1
add(repo, "test.txt")
diff_4 <- diff(repo, index=TRUE)
summary(diff_4)

## Diff between tree and index
diff_5 <- diff(tree(commits(repo)[[1]]), index=TRUE)
summary(diff_5)

## Diff between two trees
commit(repo, "Second commit")
tree_1 <- tree(commits(repo)[[2]])
tree_2 <- tree(commits(repo)[[1]])
diff_6 <- diff(tree_1, tree_2)
summary(diff_6)

## Binary files
set.seed(42)
writeBin(as.raw((sample(0:255, 1000, replace=TRUE))),
         con=file.path(path, "test.bin"))
add(repo, "test.bin")
diff_7 <- diff(repo, index=TRUE)
summary(diff_7)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential.R
\name{cred_user_pass}
\alias{cred_user_pass}
\title{Create a new plain-text username and password credential object}
\usage{
cred_user_pass(username = NULL, password = NULL)
}
\arguments{
\item{username}{The username of the credential}

\item{password}{The password of the credential. If getPass is installed
and the only input is username, \code{getPass::getPass()} will be
called to allow for interactive and obfuscated interactive
input of the password.}
}
\value{
A list of class \code{cred_user_pass} with entries:
\describe{
  \item{username}{
    The username of the credential
  }
  \item{password}{
    The password of the credential
  }
}
}
\description{
Create a new plain-text username and password credential object
}
\examples{
\dontrun{
## Create a plain-text username and password credential object
cred_user_pass("Random Developer", "SecretPassword")
}
}
\seealso{
Other git credential functions: 
\code{\link{cred_env}()},
\code{\link{cred_ssh_key}()},
\code{\link{cred_token}()}
}
\concept{git credential functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blob.R
\name{is_binary}
\alias{is_binary}
\title{Is blob binary}
\usage{
is_binary(blob = NULL)
}
\arguments{
\item{blob}{The blob \code{object}.}
}
\value{
TRUE if binary data, FALSE if not.
}
\description{
Is blob binary
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Commit a text file
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit_1 <- commit(repo, "First commit message")

## Check if binary
b_text <- tree(commit_1)["example.txt"]
is_binary(b_text)

## Commit plot file (binary)
x <- 1:100
y <- x^2
png(file.path(path, "plot.png"))
plot(y ~ x, type = "l")
dev.off()
add(repo, "plot.png")
commit_2 <- commit(repo, "Second commit message")

## Check if binary
b_png <- tree(commit_2)["plot.png"]
is_binary(b_png)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.R
\name{branch_delete}
\alias{branch_delete}
\title{Delete a branch}
\usage{
branch_delete(branch = NULL)
}
\arguments{
\item{branch}{The branch}
}
\value{
invisible NULL
}
\description{
Delete a branch
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit_1 <- commit(repo, "First commit message")

## Create a 'dev' branch
dev <- branch_create(commit_1, name = "dev")
branches(repo)

## Delete 'dev' branch
branch_delete(dev)
branches(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fetch.R
\name{fetch}
\alias{fetch}
\title{Fetch new data and update tips}
\usage{
fetch(
  repo = ".",
  name = NULL,
  credentials = NULL,
  verbose = TRUE,
  refspec = NULL
)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{name}{the remote's name}

\item{credentials}{The credentials for remote repository
access. Default is NULL. To use and query an ssh-agent for the
ssh key credentials, let this parameter be NULL (the default).}

\item{verbose}{Print information each time a reference is updated
locally. Default is \code{TRUE}.}

\item{refspec}{The refs to fetch and which local refs to update,
see examples. Pass NULL to use the
\code{remote.<repository>.fetch} variable. Default is
\code{NULL}.}
}
\value{
invisible list of class \code{git_transfer_progress}
    with statistics from the fetch operation:
\describe{
  \item{total_objects}{
    Number of objects in the packfile being downloaded
  }
  \item{indexed_objects}{
    Received objects that have been hashed
  }
  \item{received_objects}{
    Objects which have been downloaded
  }
  \item{total_deltas}{
    Total number of deltas in the pack
  }
  \item{indexed_deltas}{
    Deltas which have been indexed
  }
  \item{local_objects}{
    Locally-available objects that have been injected in order to
    fix a thin pack
  }
  \item{received_bytes}{
    Size of the packfile received up to now
  }
}
}
\description{
Fetch new data and update tips
}
\examples{
\dontrun{
## Initialize three temporary repositories
path_bare <- tempfile(pattern="git2r-")
path_repo_1 <- tempfile(pattern="git2r-")
path_repo_2 <- tempfile(pattern="git2r-")

dir.create(path_bare)
dir.create(path_repo_1)
dir.create(path_repo_2)

bare_repo <- init(path_bare, bare = TRUE)
repo_1 <- clone(path_bare, path_repo_1)
repo_2 <- clone(path_bare, path_repo_2)

config(repo_1, user.name = "Alice", user.email = "alice@example.org")
config(repo_2, user.name = "Bob", user.email = "bob@example.org")

## Add changes to repo 1
writeLines("Lorem ipsum dolor sit amet",
           con = file.path(path_repo_1, "example.txt"))
add(repo_1, "example.txt")
commit(repo_1, "Commit message")

## Push changes from repo 1 to origin (bare_repo)
push(repo_1, "origin", "refs/heads/master")

## Fetch changes from origin (bare_repo) to repo 2
fetch(repo_2, "origin")

## List updated heads
fetch_heads(repo_2)

## Checking out GitHub pull requests locally
path <- tempfile(pattern="ghit-")
repo <- clone("https://github.com/leeper/ghit", path)
fetch(repo, "origin", refspec = "pull/13/head:refs/heads/BRANCHNAME")
checkout(repo, "BRANCHNAME")
summary(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/note.R
\name{notes}
\alias{notes}
\title{List notes}
\usage{
notes(repo = ".", ref = NULL)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{ref}{Reference to read from. Default (ref = NULL) is to call
\code{note_default_ref}.}
}
\value{
list with git_note objects
}
\description{
List all the notes within a specified namespace.
}
\examples{
\dontrun{
## Create and initialize a repository in a temporary directory
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit_1 <- commit(repo, "Commit message 1")

## Create another commit
writeLines(c("Hello world!",
             "HELLO WORLD!"),
           file.path(path, "example.txt"))
add(repo, "example.txt")
commit_2 <- commit(repo, "Commit message 2")

## Create note in default namespace
note_create(commit_1, "Note-1")
note_create(commit_1, "Note-2", force = TRUE)

## Create note in named (review) namespace
note_create(commit_1, "Note-3", ref="refs/notes/review")
note_create(commit_2, "Note-4", ref="review")

## Create note on blob and tree
note_create(tree(commit_1), "Note-5")
note_create(tree(commit_1)["example.txt"], "Note-6")

## List notes in default namespace
notes(repo)

## List notes in 'review' namespace
notes(repo, "review")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{is_tree}
\alias{is_tree}
\title{Check if object is S3 class git_tree}
\usage{
is_tree(object)
}
\arguments{
\item{object}{Check if object is S3 class git_tree}
}
\value{
TRUE if object is S3 class git_tree, else FALSE
}
\description{
Check if object is S3 class git_tree
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Commit a text file
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit_1 <- commit(repo, "First commit message")
tree_1 <- tree(commit_1)

## Check if tree
is_tree(commit_1)
is_tree(tree_1)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_repository-class}
\alias{git_repository-class}
\title{S4 class to handle a git repository}
\description{
Class \code{"git_repository"}
}
\section{Slots}{

\describe{
  \item{path}{
    Path to a git repository
  }
}
}

\section{Methods}{

\describe{
  \item{is_bare}{\code{signature(object = "git_repository")}}
  \item{is_empty}{\code{signature(object = "git_repository")}}
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.r
\docType{methods}
\name{show,git_branch-method}
\alias{show,git_branch-method}
\alias{show,git_branch-methods}
\title{Brief summary of branch}
\usage{
\S4method{show}{git_branch}(object)
}
\arguments{
\item{object}{The branch \code{object}}
}
\value{
None (invisible 'NULL').
}
\description{
Brief summary of branch
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Brief summary of the branch in the repository
branches(repo)[[1]]
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stash.R
\name{stash_apply}
\alias{stash_apply}
\title{Apply stash}
\usage{
stash_apply(object = ".", index = 1)
}
\arguments{
\item{object}{path to a repository, or a \code{git_repository}
object, or the stash \code{object} to pop. Default is a
\code{path = '.'} to a reposiory.}

\item{index}{The index to the stash to apply. Only used when
\code{object} is a path to a repository or a
\code{git_repository} object. Default is \code{index = 1}.}
}
\value{
invisible NULL
}
\description{
Apply a single stashed state from the stash list.
}
\details{
If local changes in the working directory conflict with changes in
the stash then an error will be raised. In this case, the index
will always remain unmodified and all files in the working
directory will remain unmodified. However, if you are restoring
untracked files or ignored files and there is a conflict when
applying the modified files, then those files will remain in the
working directory.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

# Configure a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

# Create a file, add and commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, 'test.txt')
commit(repo, "Commit message")

# Change file
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))

# Create stash in repository
stash(repo)

# Change file
writeLines(c("Hello world!", "HeLlO wOrLd!"), file.path(path, "test.txt"))

# Create stash in repository
stash(repo)

# View stashes
stash_list(repo)

# Read file
readLines(file.path(path, "test.txt"))

# Apply latest git_stash object in repository
stash_apply(stash_list(repo)[[1]])

# Read file
readLines(file.path(path, "test.txt"))

# View stashes
stash_list(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential.r
\docType{methods}
\name{cred_user_pass}
\alias{cred_user_pass}
\alias{cred_user_pass,character,character-method}
\title{Create a new plain-text username and password credential object}
\usage{
cred_user_pass(username, password)

\S4method{cred_user_pass}{character,character}(username, password)
}
\arguments{
\item{username}{The username of the credential}

\item{password}{The password of the credential}
}
\value{
A S4 \code{cred_user_pass} object
}
\description{
Create a new plain-text username and password credential object
}
\examples{
\dontrun{
## Create a plain-text username and password credential object
cred_user_pass("Random Developer", "SecretPassword")
}
}
\seealso{
Other git credential functions: \code{\link{cred_env}},
  \code{\link{cred_ssh_key}}, \code{\link{cred_token}}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time.R
\name{git_time}
\alias{git_time}
\alias{as.character.git_time}
\alias{format.git_time}
\alias{as.POSIXct.git_time}
\alias{print.git_time}
\title{Time}
\usage{
\method{as.character}{git_time}(x, tz = "GMT", origin = "1970-01-01", usetz = TRUE, ...)

\method{format}{git_time}(x, tz = "GMT", origin = "1970-01-01", usetz = TRUE, ...)

\method{as.POSIXct}{git_time}(x, tz = "GMT", origin = "1970-01-01", ...)

\method{print}{git_time}(x, tz = "GMT", origin = "1970-01-01", usetz = TRUE, ...)
}
\arguments{
\item{x}{\R  object to be converted.}

\item{tz}{time zone specification to be used for the conversion,
    \emph{if one is required}.  System-specific (see \link[base]{time zones}),
    but \code{""} is the current time zone, and \code{"GMT"} is UTC
    (Universal Time, Coordinated).  Invalid values are most commonly
    treated as UTC, on some platforms with a warning.}

\item{origin}{a date-time object, or something which can be coerced by
    \code{as.POSIXct(tz = "GMT")} to such an object.}

\item{usetz}{logical.  Should the time zone abbreviation be appended
    to the output?  This is used in printing times, and more reliable
    than using \code{"\%Z"}.}

\item{...}{further arguments to be passed to or from other methods.}
}
\description{
The class \code{git_time} stores the time a Git object was created.
}
\details{
The default is to use \code{tz = "GMT"} and \code{origin =
"1970-01-01"}. To use your local timezone, set \code{tz =
Sys.timezone()}.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a first user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Create tag
tag(repo, "Tagname", "Tag message")

as.POSIXct(commits(repo)[[1]]$author$when)
as.POSIXct(tags(repo)[[1]]$tagger$when)
as.POSIXct(tags(repo)[[1]]$tagger$when, tz = Sys.timezone())
}
}
\seealso{
\code{\link{when}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{[.git_tree}
\alias{[.git_tree}
\title{Extract object from tree}
\usage{
\method{[}{git_tree}(x, i)
}
\arguments{
\item{x}{The tree \code{object}}

\item{i}{The index (integer or logical) of the tree object to
extract. If negative values, all elements except those indicated
are selected. A character vector to match against the names of
objects to extract.}
}
\value{
Git object
}
\description{
Lookup a tree entry by its position in the tree
}
\examples{
\dontrun{
##' Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
dir.create(file.path(path, "subfolder"))
repo <- init(path)

##' Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

##' Create three files and commit
writeLines("First file",  file.path(path, "example-1.txt"))
writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
writeLines("Third file",  file.path(path, "example-3.txt"))
add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
new_commit <- commit(repo, "Commit message")

##' Pick a tree in the repository
tree_object <- tree(new_commit)

##' Display tree
tree_object

##' Select item by name
tree_object["example-1.txt"]

##' Select first item in tree
tree_object[1]

##' Select first three items in tree
tree_object[1:3]

##' Select all blobs in tree
tree_object[vapply(as(tree_object, 'list'), is_blob, logical(1))]
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{clone}
\alias{clone}
\title{Clone a remote repository}
\usage{
clone(
  url,
  local_path,
  bare = FALSE,
  branch = NULL,
  checkout = TRUE,
  credentials = NULL,
  progress = TRUE,
  proxy = NULL
)
}
\arguments{
\item{url}{The remote repository to clone, or a local repository path.}

\item{local_path}{Local directory to clone to.}

\item{bare}{Create a bare repository. Default is FALSE.}

\item{branch}{The name of the branch to checkout. Default is NULL
which means to use the remote's default branch.}

\item{checkout}{Checkout HEAD after the clone is complete. Default
is TRUE.}

\item{credentials}{The credentials for remote repository
access. Default is NULL. To use and query an ssh-agent for the
ssh key credentials, let this parameter be NULL (the default).}

\item{progress}{Show progress. Default is TRUE.}

\item{proxy}{Either \code{NULL} (the default) to disable proxy usage,
\code{TRUE} for automatic proxy detection, or a character string
with a proxy URL (for example, \code{"http://proxy.example.org:3128"}).}
}
\value{
A \code{git_repository} object.
}
\description{
Clone a remote repository
}
\examples{
\dontrun{
## Initialize repository
path_repo_1 <- tempfile(pattern="git2r-")
path_repo_2 <- tempfile(pattern="git2r-")
dir.create(path_repo_1)
dir.create(path_repo_2)
repo_1 <- init(path_repo_1)

## Config user and commit a file
config(repo_1, user.name = "Alice", user.email = "alice@example.org")

## Write to a file and commit
writeLines(
    "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
    file.path(path_repo_1, "example.txt"))
add(repo_1, "example.txt")
commit(repo_1, "First commit message")

## Change file and commit
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path_repo_1, "example.txt"))
add(repo_1, "example.txt")
commit(repo_1, "Second commit message")

## Change file again and commit.
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
  "minim veniam, quis nostrud exercitation ullamco laboris nisi ut")
writeLines(lines, file.path(path_repo_1, "example.txt"))
add(repo_1, "example.txt")
commit(repo_1, "Third commit message")

## Clone to second repository
repo_2 <- clone(path_repo_1, path_repo_2)

## List commits in repositories
commits(repo_1)
commits(repo_2)
}
}
\seealso{
\link{repository}, \code{\link{cred_user_pass}},
    \code{\link{cred_ssh_key}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odb.R
\name{odb_objects}
\alias{odb_objects}
\title{List all objects available in the database}
\usage{
odb_objects(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
A data.frame with the following columns:
\describe{
  \item{sha}{The sha of the object}
  \item{type}{The type of the object}
  \item{len}{The length of the object}
}
}
\description{
List all objects available in the database
}
\examples{
\dontrun{
## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message 1")

## Create tag
tag(repo, "Tagname", "Tag message")

## List objects in repository
odb_objects(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/config.R
\name{git_config_files}
\alias{git_config_files}
\title{Locate the path to configuration files}
\usage{
git_config_files(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
a \code{data.frame} with one row per potential
    configuration file where \code{NA} means not found.
}
\description{
Potential configuration files:
\describe{
  \item{system}{
    Locate the path to the system configuration file. If
    '/etc/gitconfig' doesn't exist, it will look for
    '\%PROGRAMFILES\%'.
  }
  \item{xdg}{
    Locate the path to the global xdg compatible configuration
    file. The xdg compatible configuration file is usually located
    in '$HOME/.config/git/config'. This method will try to guess
    the full path to that file, if the file exists.
  }
  \item{global}{
    The user or global configuration file is usually located in
    '$HOME/.gitconfig'. This method will try to guess the full
    path to that file, if the file exists.
  }
  \item{local}{
    Locate the path to the repository specific configuration file,
    if the file exists.
  }
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/checkout.R
\name{checkout}
\alias{checkout}
\title{Checkout}
\usage{
checkout(
  object = NULL,
  branch = NULL,
  create = FALSE,
  force = FALSE,
  path = NULL,
  ...
)
}
\arguments{
\item{object}{A path to a repository, or a \code{git_repository}
object, or a \code{git_commit} object, or a \code{git_tag}
object, or a \code{git_tree} object.}

\item{branch}{name of the branch to check out. Only used if object
is a path to a repository or a \code{git_repository} object.}

\item{create}{create branch if it doesn't exist. Only used if
object is a path to a repository or a \code{git_repository}
object.}

\item{force}{If \code{TRUE}, then make working directory match
target. This will throw away local changes. Default is
\code{FALSE}.}

\item{path}{Limit the checkout operation to only certain
paths. This argument is only used if branch is NULL. Default
is \code{NULL}.}

\item{...}{Additional arguments. Not used.}
}
\value{
invisible NULL
}
\description{
Update files in the index and working tree to match the content of
the tree pointed at by the treeish object (commit, tag or tree).
The default checkout strategy (\code{force = FALSE}) will only
make modifications that will not lose changes. Use \code{force =
TRUE} to force working directory to look like index.
}
\examples{
\dontrun{
## Create directories and initialize repositories
path_bare <- tempfile(pattern="git2r-")
path_repo_1 <- tempfile(pattern="git2r-")
path_repo_2 <- tempfile(pattern="git2r-")
dir.create(path_bare)
dir.create(path_repo_1)
dir.create(path_repo_2)
repo_bare <- init(path_bare, bare = TRUE)

## Clone to repo 1 and config user
repo_1 <- clone(path_bare, path_repo_1)
config(repo_1, user.name = "Alice", user.email = "alice@example.org")

## Add changes to repo 1 and push to bare
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path_repo_1, "test.txt"))
add(repo_1, "test.txt")
commit(repo_1, "First commit message")
push(repo_1, "origin", "refs/heads/master")

## Create and checkout 'dev' branch in repo 1
checkout(repo_1, "dev", create = TRUE)

## Add changes to 'dev' branch in repo 1 and push to bare
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path_repo_1, "test.txt"))
add(repo_1, "test.txt")
commit(repo_1, "Second commit message")
push(repo_1, "origin", "refs/heads/dev")

## Clone to repo 2
repo_2 <- clone(path_bare, path_repo_2)
config(repo_2, user.name = "Bob", user.email = "bob@example.org")

## Read content of 'test.txt'
readLines(file.path(path_repo_2, "test.txt"))

## Checkout dev branch
checkout(repo_2, "dev")

## Read content of 'test.txt'
readLines(file.path(path_repo_2, "test.txt"))

## Edit "test.txt" in repo_2
writeLines("Hello world!", con = file.path(path_repo_2, "test.txt"))

## Check status
status(repo_2)

## Checkout "test.txt"
checkout(repo_2, path = "test.txt")

## Check status
status(repo_2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_merge_result-class}
\alias{git_merge_result-class}
\title{S4 class to handle the merge result}
\description{
Class \code{"git_merge_result"}
}
\section{Slots}{

\describe{
  \item{up_to_date}{
    TRUE if the merge is already up-to-date, else FALSE.
  }
  \item{fast_forward}{
    TRUE if a fast-forward merge, else FALSE.
  }
  \item{conflicts}{
    TRUE if the index contain entries representing file conflicts,
    else FALSE.
  }
  \item{sha}{
    If the merge created a merge commit, the sha of the merge
    commit. NA if no merge commit created.
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_reflog_entry-class}
\alias{git_reflog_entry-class}
\title{S4 class to handle a git reflog entry.}
\description{
Class \code{"git_reflog_entry"}
}
\section{Slots}{

\describe{
  \item{sha}{
    The 40 character hexadecimal string of the SHA-1
  }
  \item{message}{
    The log message of the entry
  }
  \item{index}{
    The index (zero-based) of the entry in the reflog
  }
  \item{committer}{
    The committer signature
  }
  \item{refname}{
    Name of the reference
  }
  \item{repo}{
    The S4 class git_repository that contains the reflog entry
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{is_shallow}
\alias{is_shallow}
\title{Determine if the repository is a shallow clone}
\usage{
is_shallow(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
\code{TRUE} if shallow clone, else \code{FALSE}
}
\description{
Determine if the repository is a shallow clone
}
\examples{
\dontrun{
## Initialize repository
path_repo_1 <- tempfile(pattern="git2r-")
path_repo_2 <- tempfile(pattern="git2r-")
dir.create(path_repo_1)
dir.create(path_repo_2)
repo_1 <- init(path_repo_1)

## Config user and commit a file
config(repo_1, user.name = "Alice", user.email = "alice@example.org")

## Write to a file and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path_repo_1, "example.txt"))
add(repo_1, "example.txt")
commit(repo_1, "First commit message")

## Change file and commit
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path_repo_1, "example.txt"))
add(repo_1, "example.txt")
commit(repo_1, "Second commit message")

## Change file again and commit.
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
  "minim veniam, quis nostrud exercitation ullamco laboris nisi ut")
writeLines(lines, file.path(path_repo_1, "example.txt"))
add(repo_1, "example.txt")
commit(repo_1, "Third commit message")

## Clone to second repository
repo_2 <- clone(path_repo_1, path_repo_2)

## Check if it's a shallow clone
is_shallow(repo_2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tag.R
\name{tags}
\alias{tags}
\title{Tags}
\usage{
tags(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
list of tags in repository
}
\description{
Tags
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Commit a text file
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Create tag
tag(repo, "Tagname", "Tag message")

## List tags
tags(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_fetch_head-class}
\alias{git_fetch_head-class}
\title{S4 class to handle a fetch head}
\description{
Class \code{"git_fetch_head"}
}
\section{Slots}{

\describe{
  \item{ref_name}{
    The name of the ref.
  }
  \item{remote_url}{
    The url of the remote.
  }
  \item{sha}{
    The SHA-1 of the remote head that were updated during the last
    fetch.
  }
  \item{is_merge}{
    Is head for merge.
  }
  \item{repo}{
    The S4 class git_repository that contains the fetch head.
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.R
\name{commit}
\alias{commit}
\title{Commit}
\usage{
commit(
  repo = ".",
  message = NULL,
  all = FALSE,
  session = FALSE,
  author = NULL,
  committer = NULL
)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{message}{The commit message.}

\item{all}{Stage modified and deleted files. Files not added to
Git are not affected.}

\item{session}{Add sessionInfo to commit message. Default is
FALSE.}

\item{author}{Signature with author and author time of commit.}

\item{committer}{Signature with committer and commit time of
commit.}
}
\value{
A list of class \code{git_commit} with entries:
\describe{
  \item{sha}{
    The 40 character hexadecimal string of the SHA-1
  }
  \item{author}{
    An author signature
  }
  \item{committer}{
    The committer signature
  }
  \item{summary}{
    The short "summary" of a git commit message, comprising the first
    paragraph of the message with whitespace trimmed and squashed.
  }
  \item{message}{
    The message of a commit
  }
  \item{repo}{
    The \code{git_repository} object that contains the commit
  }
}
}
\description{
Commit
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Write to a file and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{as.list.git_tree}
\alias{as.list.git_tree}
\title{Coerce entries in a git_tree to a list of entry objects}
\usage{
\method{as.list}{git_tree}(x, ...)
}
\arguments{
\item{x}{The tree \code{object}}

\item{...}{Unused}
}
\value{
list of entry objects
}
\description{
Coerce entries in a git_tree to a list of entry objects
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
dir.create(file.path(path, "subfolder"))
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create three files and commit
writeLines("First file",  file.path(path, "example-1.txt"))
writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
writeLines("Third file",  file.path(path, "example-3.txt"))
add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
commit(repo, "Commit message")

## Inspect size of each blob in tree
invisible(lapply(as(tree(last_commit(repo)), "list"),
  function(obj) {
    if (is_blob(obj))
      summary(obj)
    NULL
  }))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bundle_r_package.R
\name{bundle_r_package}
\alias{bundle_r_package}
\title{Bundle bare repo of package}
\usage{
bundle_r_package(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
Invisible bundled \code{git_repository} object
}
\description{
Clone the package git repository as a bare repository to
\code{pkg/inst/pkg.git}
}
\examples{
\dontrun{
## Initialize repository
path <- tempfile()
dir.create(path)
path <- file.path(path, "git2r")
repo <- clone("https://github.com/ropensci/git2r.git", path)

## Bundle bare repository in package
bundle_r_package(repo)

## Build and install bundled package
wd <- setwd(dirname(path))
system(sprintf("R CMD build \%s", path))
pkg <- list.files(".", pattern = "[.]tar[.]gz$")
system(sprintf("R CMD INSTALL \%s", pkg))
setwd(wd)

## Reload package
detach("package:git2r", unload = TRUE)
library(git2r)

## Summarize last five commits of bundled repo
repo <- repository(system.file("git2r.git", package = "git2r"))
invisible(lapply(commits(repo, n = 5), summary))

## Plot content of bundled repo
plot(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blob.r
\docType{methods}
\name{content}
\alias{content}
\alias{content,git_blob-method}
\title{Content of blob}
\usage{
content(blob, split = TRUE)

\S4method{content}{git_blob}(blob, split = TRUE)
}
\arguments{
\item{blob}{The blob \code{object}.}

\item{split}{Split blob content to text lines. Default TRUE.}
}
\value{
The content of the blob. NA_character_ if the blob is binary.
}
\description{
Content of blob
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Display content of blob.
content(tree(commits(repo)[[1]])["example.txt"])
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge.R
\name{merge_base}
\alias{merge_base}
\title{Find a merge base between two commits}
\usage{
merge_base(one = NULL, two = NULL)
}
\arguments{
\item{one}{One of the commits}

\item{two}{The other commit}
}
\value{
git_commit
}
\description{
Find a merge base between two commits
}
\examples{
\dontrun{
## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
writeLines("Master branch", file.path(path, "master_branch.txt"))
add(repo, "master_branch.txt")
commit_1 <- commit(repo, "Commit message 1")

## Create first branch, checkout, add file and commit
branch_1 <- branch_create(commit_1, "branch_1")
checkout(branch_1)
writeLines("Branch 1", file.path(path, "branch_1.txt"))
add(repo, "branch_1.txt")
commit_2 <- commit(repo, "Commit message branch_1")

## Create second branch, checkout, add file and commit
branch_2 <- branch_create(commit_1, "branch_2")
checkout(branch_2)
writeLines("Branch 2", file.path(path, "branch_2.txt"))
add(repo, "branch_2.txt")
commit_3 <- commit(repo, "Commit message branch_2")

## Check that merge base equals commit_1
stopifnot(identical(merge_base(commit_2, commit_3), commit_1))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{cred_ssh_key-class}
\alias{cred_ssh_key-class}
\title{S4 class to handle a passphrase-protected ssh key
credential object}
\description{
Class \code{"cred_ssh_key"}
}
\section{Slots}{

\describe{
  \item{publickey}{
    The path to the public key of the credential
  }
  \item{privatekey}{
    The path to the private key of the credential
  }
  \item{passphrase}{
    The passphrase of the credential
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/note.R
\name{note_remove}
\alias{note_remove}
\title{Remove the note for an object}
\usage{
note_remove(note = NULL, author = NULL, committer = NULL)
}
\arguments{
\item{note}{The note to remove}

\item{author}{Signature of the notes commit author.}

\item{committer}{Signature of the notes commit committer.}
}
\value{
invisible NULL
}
\description{
Remove the note for an object
}
\examples{
\dontrun{
## Create and initialize a repository in a temporary directory
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit_1 <- commit(repo, "Commit message 1")


## Create note in default namespace
note_1 <- note_create(commit_1, "Note-1")

## Create note in named (review) namespace
note_2 <- note_create(commit_1, "Note-2", ref="refs/notes/review")

## List notes in default namespace
notes(repo)

## List notes in 'review' namespace
notes(repo, "review")

## Remove notes
note_remove(note_1)
note_remove(note_2)

## List notes in default namespace
notes(repo)

## List notes in 'review' namespace
notes(repo, "review")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_branch-class}
\alias{git_branch-class}
\title{S4 class to handle a git branch}
\description{
Class \code{git_branch}
}
\section{Slots}{

\describe{
  \item{name}{
    Name of the branch.
  }
  \item{type}{
    Type of the branch, either 1 (local) or 2 (remote).
  }
  \item{repo}{
    The S4 class git_repository that contains the branch
  }
}
}

\section{Methods}{

\describe{
  \item{is_head}{\code{signature(object = "git_branch")}}
  \item{is_local}{\code{signature(object = "git_branch")}}
  \item{show}{\code{signature(object = "git_branch")}}
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{workdir}
\alias{workdir}
\title{Workdir of repository}
\usage{
workdir(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
Character vector with the path of the workdir. If the
repository is bare, \code{NULL} will be returned.
}
\description{
Workdir of repository
}
\examples{
\dontrun{
## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)

## Get the path of the workdir for repository
workdir(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.R
\name{commits}
\alias{commits}
\title{Commits}
\usage{
commits(
  repo = ".",
  topological = TRUE,
  time = TRUE,
  reverse = FALSE,
  n = NULL,
  ref = NULL,
  path = NULL
)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{topological}{Sort the commits in topological order (parents
before children); can be combined with time sorting. Default
is TRUE.}

\item{time}{Sort the commits by commit time; Can be combined with
topological sorting. Default is TRUE.}

\item{reverse}{Sort the commits in reverse order; can be combined
with topological and/or time sorting. Default is FALSE.}

\item{n}{The upper limit of the number of commits to output. The
default is NULL for unlimited number of commits.}

\item{ref}{The name of a reference to list commits from e.g. a tag
or a branch. The default is NULL for the current branch.}

\item{path}{The path to a file. If not NULL, only commits modifying
this file will be returned. Note that modifying commits that
occurred before the file was given its present name are not
returned; that is, the output of \code{git log} with
\code{--no-follow} is reproduced.}
}
\value{
list of commits in repository
}
\description{
Commits
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Write to a file and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Change file and commit
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Second commit message")

## Create a tag
tag(repo, "Tagname", "Tag message")

## Change file again and commit
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
  "minim veniam, quis nostrud exercitation ullamco laboris nisi ut")
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Third commit message")

## Create a new file containing R code, and commit.
writeLines(c("x <- seq(1,100)",
             "print(mean(x))"),
           file.path(path, "mean.R"))
add(repo, "mean.R")
commit(repo, "Fourth commit message")

## List the commits in the repository
commits(repo)

## List the commits starting from the tag
commits(repo, ref = "Tagname")

## List the commits modifying example.txt and mean.R.
commits(repo, path = "example.txt")
commits(repo, path = "mean.R")

## Create and checkout 'dev' branch in the repo
checkout(repo, "dev", create = TRUE)

## Add changes to the 'dev' branch
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Commit message in dev branch")

## Checkout the 'master' branch again and list the commits
## starting from the 'dev' branch.
checkout(repo, "master")
commits(repo, ref = "dev")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{cred_token-class}
\alias{cred_token-class}
\title{S4 class to handle a personal access token credential
object}
\description{
Class \code{"cred_token"}
}
\section{Slots}{

\describe{
  \item{token}{
    The name of the environmental variable that holds
    the personal access token for the authentication.
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_diff-class}
\alias{git_diff-class}
\title{Git diff}
\description{
Git diff
}
\section{Slots}{

\describe{
  \item{old}{
    Represents the 'from' side of the diff.
  }
  \item{new}{
    Represents the 'to' side of the diff.
  }
  \item{files}{
    List of \code{\linkS4class{git_diff_file}} objects.
  }
}
}

\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.R
\name{parents}
\alias{parents}
\title{Parents}
\usage{
parents(object = NULL)
}
\arguments{
\item{object}{a git_commit object.}
}
\value{
list of git_commit objects
}
\description{
Get parents of a commit.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
writeLines("First line.",
           file.path(path, "example.txt"))
add(repo, "example.txt")
commit_1 <- commit(repo, "First commit message")

## commit_1 has no parents
parents(commit_1)

## Update 'example.txt' and commit
writeLines(c("First line.", "Second line."),
           file.path(path, "example.txt"))
add(repo, "example.txt")
commit_2 <- commit(repo, "Second commit message")

## commit_2 has commit_1 as parent
parents(commit_2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{lookup}
\alias{lookup}
\title{Lookup}
\usage{
lookup(repo = ".", sha = NULL)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{sha}{The identity of the object to lookup. Must be 4 to 40
characters long.}
}
\value{
a \code{git_blob} or \code{git_commit} or \code{git_tag}
or \code{git_tree} object
}
\description{
Lookup one object in a repository.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit_1 <- commit(repo, "First commit message")

## Create tag
tag(repo, "Tagname", "Tag message")

## First, get SHAs to lookup in the repository
sha_commit <- sha(commit_1)
sha_tree <- sha(tree(commit_1))
sha_blob <- sha(tree(commit_1)["example.txt"])
sha_tag <- sha(tags(repo)[[1]])

## SHAs
sha_commit
sha_tree
sha_blob
sha_tag

## Lookup objects
lookup(repo, sha_commit)
lookup(repo, sha_tree)
lookup(repo, sha_blob)
lookup(repo, sha_tag)

## Lookup objects, using only the first seven characters
lookup(repo, substr(sha_commit, 1, 7))
lookup(repo, substr(sha_tree, 1, 7))
lookup(repo, substr(sha_blob, 1, 7))
lookup(repo, substr(sha_tag, 1, 7))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tag.r
\docType{methods}
\name{tags}
\alias{tags}
\alias{tags,missing-method}
\alias{tags,git_repository-method}
\title{Tags}
\usage{
tags(repo)

\S4method{tags}{missing}()

\S4method{tags}{git_repository}(repo)
}
\arguments{
\item{repo}{The repository \code{object}
\code{\linkS4class{git_repository}}. If the \code{repo} argument
is missing, the repository is searched for with
\code{\link{discover_repository}} in the current working
directory.}
}
\value{
list of tags in repository
}
\description{
Tags
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name="Alice", user.email="alice@example.org")

## Commit a text file
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Create tag
tag(repo, "Tagname", "Tag message")

## List tags
tags(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_diff_line-class}
\alias{git_diff_line-class}
\title{Git diff line}
\description{
Git diff line
}
\section{Slots}{

\describe{
  \item{origin}{
    Line origin constants.
  }
  \item{old_lineno}{
    Line number in old file or -1 for added line.
  }
  \item{new_lineno}{
    Line number in new file or -1 for deleted line.
  }
  \item{num_lines}{
    Number of newline characters in content.
  }
  \item{content}{
    The diff text.
  }
}
}

\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.r
\docType{methods}
\name{is_detached}
\alias{is_detached}
\alias{is_detached,missing-method}
\alias{is_detached,git_repository-method}
\title{Check if HEAD of repository is detached}
\usage{
is_detached(repo)

\S4method{is_detached}{missing}()

\S4method{is_detached}{git_repository}(repo)
}
\arguments{
\item{repo}{The repository \code{object}
\code{\linkS4class{git_repository}}. If the \code{repo} argument
is missing, the repository is searched for with
\code{\link{discover_repository}} in the current working
directory.}
}
\value{
TRUE if repository HEAD is detached, else FALSE
}
\description{
Check if HEAD of repository is detached
}
\examples{
\dontrun{
## Create and initialize a repository in a temporary directory
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
config(repo, user.name="Alice", user.email="alice@example.org")

## Create a file, add and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           file.path(path, "example.txt"))
add(repo, "example.txt")
commit_1 <- commit(repo, "Commit message 1")

## Change file, add and commit
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
             file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Commit message 2")

## HEAD of repository is not detached
is_detached(repo)

## Checkout first commit
checkout(commit_1)

## HEAD of repository is detached
is_detached(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.R
\name{is_branch}
\alias{is_branch}
\title{Check if object is \code{git_branch}}
\usage{
is_branch(object)
}
\arguments{
\item{object}{Check if object is of class \code{git_branch}}
}
\value{
TRUE if object is class \code{git_branch}, else FALSE
}
\description{
Check if object is \code{git_branch}
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Commit a text file
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

branch <- branches(repo)[[1]]

## Check if branch
is_branch(branch)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{length.git_tree}
\alias{length.git_tree}
\title{Number of entries in tree}
\usage{
\method{length}{git_tree}(x)
}
\arguments{
\item{x}{The tree \code{object}}
}
\value{
a non-negative integer or double (which will be rounded
down)
}
\description{
Number of entries in tree
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential.R
\name{ssh_path}
\alias{ssh_path}
\title{Compose usual path to ssh keys}
\usage{
ssh_path(file = "")
}
\arguments{
\item{file}{basename of file for which path is requested}
}
\value{
Full path to the file
}
\description{
This function provides a consistent means across OS-types to access the
\code{.ssh} directory.
}
\details{
On Windows-based systems,
\code{path.expand("~")} returns \code{"C:/Users/username/Documents"},
whereas the usual path to the \code{.ssh} directory is
\code{"C:/Users/username"}.

On other operating systems, \code{path.expand("~")} returns the usual path
to the \code{.ssh} directory.

Calling \code{ssh_path()} with no arguments will return the usual path to
the \code{.ssh} directory.
}
\examples{
ssh_path()
ssh_path("is_rsa.pub")
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/note.r
\docType{methods}
\name{note_default_ref}
\alias{note_default_ref}
\alias{note_default_ref,git_repository-method}
\title{Default notes reference}
\usage{
note_default_ref(repo)

\S4method{note_default_ref}{git_repository}(repo)
}
\arguments{
\item{repo}{The repository}
}
\value{
Character vector of length one with name of default notes
reference
}
\description{
Get the default notes reference for a repository
}
\examples{
\dontrun{
## Create and initialize a repository in a temporary directory
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
config(repo, user.name="Alice", user.email="alice@example.org")

## View default notes reference
note_default_ref(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/push.r
\docType{methods}
\name{push}
\alias{push}
\alias{push,git_branch-method}
\alias{push,git_repository-method}
\title{Push}
\usage{
push(object, ...)

\S4method{push}{git_branch}(object, force = FALSE, credentials = NULL)

\S4method{push}{git_repository}(object, name = NULL, refspec = NULL,
  force = FALSE, credentials = NULL)
}
\arguments{
\item{object}{S4 class \code{git_repository} or \code{git_branch}.}

\item{...}{Additional arguments affecting the push.}

\item{force}{Force your local revision to the remote repo. Use it
with care. Default is FALSE.}

\item{credentials}{The credentials for remote repository
access. Default is NULL. To use and query an ssh-agent for the ssh
key credentials, let this parameter be NULL (the default).}

\item{name}{The remote's name. Default is NULL.}

\item{refspec}{The refspec to be pushed. Default is NULL.}
}
\value{
invisible(NULL)
}
\description{
Push
}
\examples{
\dontrun{
## Initialize two temporary repositories
path_bare <- tempfile(pattern="git2r-")
path_repo <- tempfile(pattern="git2r-")
dir.create(path_bare)
dir.create(path_repo)
repo_bare <- init(path_bare, bare = TRUE)
repo <- clone(path_bare, path_repo)

## Config user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")

## Write to a file and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           file.path(path_repo, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Push commits from repository to bare repository
## Adds an upstream tracking branch to branch 'master'
push(repo, "origin", "refs/heads/master")

## Change file and commit
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           file.path(path_repo, "example.txt"))
add(repo, "example.txt")
commit(repo, "Second commit message")

## Push commits from repository to bare repository
push(repo)

## List commits in repository and bare repository
commits(repo)
commits(repo_bare)
}
}
\seealso{
\code{\linkS4class{cred_user_pass}},
\code{\linkS4class{cred_ssh_key}}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.R
\name{branch_rename}
\alias{branch_rename}
\title{Rename a branch}
\usage{
branch_rename(branch = NULL, name = NULL, force = FALSE)
}
\arguments{
\item{branch}{Branch to rename}

\item{name}{The new name for the branch}

\item{force}{Overwrite existing branch. Default is FALSE}
}
\value{
invisible renamed \code{git_branch} object
}
\description{
Rename a branch
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Rename 'master' branch to 'dev'
branches(repo)
branch_rename(repository_head(repo), "dev")
branches(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_commit-class}
\alias{git_commit-class}
\title{S4 class to handle a git commit.}
\description{
Class \code{"git_commit"}
}
\section{Slots}{

\describe{
  \item{sha}{
    The 40 character hexadecimal string of the SHA-1
  }
  \item{author}{
    An author signature
  }
  \item{committer}{
    The committer signature
  }
  \item{summary}{
    The short "summary" of a git commit message, comprising the first
    paragraph of the message with whitespace trimmed and squashed.
  }
  \item{message}{
    The message of a commit
  }
  \item{repo}{
    The S4 class git_repository that contains the commit
  }
}
}

\section{Methods}{

\describe{
  \item{show}{\code{signature(object = "git_commit")}}
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odb.r
\docType{methods}
\name{odb_objects}
\alias{odb_objects}
\alias{odb_objects,missing-method}
\alias{odb_objects,git_repository-method}
\title{List all objects available in the database}
\usage{
odb_objects(repo)

\S4method{odb_objects}{missing}()

\S4method{odb_objects}{git_repository}(repo)
}
\arguments{
\item{repo}{The repository}
}
\value{
A data.frame with the following columns:
\describe{
  \item{sha}{The sha of the object}
  \item{type}{The type of the object}
  \item{len}{The length of the object}
}
}
\description{
List all objects available in the database
}
\examples{
\dontrun{
## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name="Alice", user.email="alice@example.org")

## Create a file, add and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message 1")

## Create tag
tag(repo, "Tagname", "Tag message")

## List objects in repository
odb_objects(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_time-class}
\alias{git_time-class}
\alias{coerce,git_time,character-method}
\title{S4 class to handle a git time in a signature}
\description{
Class \code{"git_time"}
}
\section{Slots}{

\describe{
  \item{time}{
    Time in seconds from epoch
  }
  \item{offset}{
    Timezone offset, in minutes
  }
}
}

\section{Methods}{

\describe{
  \item{show}{\code{signature(object = "git_time")}}
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{as.data.frame.git_repository}
\alias{as.data.frame.git_repository}
\title{Coerce Git repository to a \code{data.frame}}
\usage{
\method{as.data.frame}{git_repository}(x, ...)
}
\arguments{
\item{x}{The repository \code{object}}

\item{...}{Additional arguments. Not used.}
}
\value{
\code{data.frame}
}
\description{
The commits in the repository are coerced to a \code{data.frame}
}
\details{
The \code{data.frame} have the following columns:
\describe{

  \item{sha}{
    The 40 character hexadecimal string of the SHA-1
  }

  \item{summary}{
    the short "summary" of the git commit message.
  }

  \item{message}{
    the full message of a commit
  }

  \item{author}{
    full name of the author
  }

  \item{email}{
    email of the author
  }

  \item{when}{
    time when the commit happened
  }

}
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create three files and commit
writeLines("First file",  file.path(path, "example-1.txt"))
writeLines("Second file", file.path(path, "example-2.txt"))
writeLines("Third file",  file.path(path, "example-3.txt"))
add(repo, "example-1.txt")
commit(repo, "Commit first file")
add(repo, "example-2.txt")
commit(repo, "Commit second file")
add(repo, "example-3.txt")
commit(repo, "Commit third file")

## Coerce commits to a data.frame
df <- as.data.frame(repo)
df
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/odb.R
\name{odb_blobs}
\alias{odb_blobs}
\title{Blobs in the object database}
\usage{
odb_blobs(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
A data.frame with the following columns:
\describe{
  \item{sha}{The sha of the blob}
  \item{path}{The path to the blob from the tree and sub-trees}
  \item{name}{The name of the blob from the tree that contains the blob}
  \item{len}{The length of the blob}
  \item{commit}{The sha of the commit}
  \item{author}{The author of the commit}
  \item{when}{The timestamp of the author signature in the commit}
}
}
\description{
List all blobs reachable from the commits in the object
database. For each commit, list blob's in the commit tree and
sub-trees.
}
\note{
A blob sha can have several entries
}
\examples{
\dontrun{
## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message 1")

## Change file and commit
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message 2")

## Commit same content under different name in a sub-directory
dir.create(file.path(path, "sub-directory"))
file.copy(file.path(path, "test.txt"),
          file.path(path, "sub-directory", "copy.txt"))
add(repo, "sub-directory/copy.txt")
commit(repo, "Commit message 3")

## List blobs
odb_blobs(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.r
\docType{methods}
\name{commits}
\alias{commits}
\alias{commits,missing-method}
\alias{commits,git_repository-method}
\title{Commits}
\usage{
commits(repo, ...)

\S4method{commits}{missing}(topological = TRUE, time = TRUE,
  reverse = FALSE, n = NULL, ...)

\S4method{commits}{git_repository}(repo, topological = TRUE, time = TRUE,
  reverse = FALSE, n = NULL, ...)
}
\arguments{
\item{repo}{The repository \code{object}
\code{\linkS4class{git_repository}}. If the \code{repo} argument
is missing, the repository is searched for with
\code{\link{discover_repository}} in the current working
directory.}

\item{...}{Additional arguments to commits.}

\item{topological}{Sort the commits in topological order (parents
before children); can be combined with time sorting. Default is
TRUE.}

\item{time}{Sort the commits by commit time; Can be combined with
topological sorting. Default is TRUE.}

\item{reverse}{Sort the commits in reverse order; can be combined
with topological and/or time sorting. Default is FALSE.}

\item{n}{The upper limit of the number of commits to output. The
defualt is NULL for unlimited number of commits.}
}
\value{
list of commits in repository
}
\description{
Commits
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user
config(repo, user.name="Alice", user.email="alice@example.org")

## Write to a file and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Change file and commit
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Second commit message")

## Change file again and commit
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
             "minim veniam, quis nostrud exercitation ullamco laboris nisi ut"),
           file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Third commit message")

## List commits in repository
commits(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/git2r.R
\docType{package}
\name{git2r}
\alias{git2r}
\alias{git2r-package}
\title{git2r: R bindings to the libgit2 library}
\description{
git2r: R bindings to the libgit2 library.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://docs.ropensci.org/git2r/}
  \item \url{https://github.com/ropensci/git2r}
  \item Report bugs at \url{https://github.com/ropensci/git2r/issues}
}

}
\author{
\strong{Maintainer}: Stefan Widgren \email{stefan.widgren@gmail.com} (\href{https://orcid.org/0000-0001-5745-2284}{ORCID})

Other contributors:
\itemize{
  \item Gabor Csardi [contributor]
  \item Gregory Jefferis [contributor]
  \item Jennifer Bryan [contributor]
  \item Jeroen Ooms [contributor]
  \item Jim Hester [contributor]
  \item John Blischak [contributor]
  \item Karthik Ram [contributor]
  \item Peter Carbonetto [contributor]
  \item Scott Chamberlain [contributor]
  \item Thomas Rosendal [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{ls_tree}
\alias{ls_tree}
\title{List the contents of a tree object}
\usage{
ls_tree(tree = NULL, repo = ".", recursive = TRUE)
}
\arguments{
\item{tree}{default (\code{NULL}) is the tree of the last commit
in \code{repo}. Can also be a \code{git_tree} object or a
character that identifies a tree in the repository (see
\sQuote{Examples}).}

\item{repo}{never used if \code{tree} is a \code{git_tree}
object. A \code{git_repository} object, or a path (default =
'.') to a repository.}

\item{recursive}{default is to recurse into sub-trees.}
}
\value{
A data.frame with the following columns: \describe{
    \item{mode}{UNIX file attribute of the tree entry}
    \item{type}{type of object} \item{sha}{sha of the object}
    \item{path}{path relative to the root tree}
    \item{name}{filename of the tree entry} \item{len}{object size
    of blob (file) entries. NA for other objects.}  }
}
\description{
Traverse the entries in a tree and its subtrees.  Akin to the 'git
ls-tree' command.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
dir.create(file.path(path, "subfolder"))
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create three files and commit
writeLines("First file",  file.path(path, "example-1.txt"))
writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
writeLines("Third file",  file.path(path, "example-3.txt"))
add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
commit(repo, "Commit message")

## Traverse tree entries and its subtrees.
## Various approaches that give identical result.
ls_tree(tree = tree(last_commit(path)))
ls_tree(tree = tree(last_commit(repo)))
ls_tree(repo = path)
ls_tree(repo = repo)

## Skip content in subfolder
ls_tree(repo = repo, recursive = FALSE)

## Start in subfolder
ls_tree(tree = "HEAD:subfolder", repo = repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.r
\docType{methods}
\name{workdir}
\alias{workdir}
\alias{workdir,missing-method}
\alias{workdir,git_repository-method}
\title{Workdir of repository}
\usage{
workdir(repo)

\S4method{workdir}{missing}()

\S4method{workdir}{git_repository}(repo)
}
\arguments{
\item{repo}{The repository \code{object}
\code{\linkS4class{git_repository}}. If the \code{repo} argument
is missing, the repository is searched for with
\code{\link{discover_repository}} in the current working
directory.}
}
\value{
Character vector with the path of the workdir. If the
repository is bare, \code{NULL} will be returned.
}
\description{
Workdir of repository
}
\examples{
\dontrun{
## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)

## Get the path of the workdir for repository
workdir(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stash.R
\name{stash_list}
\alias{stash_list}
\title{List stashes in repository}
\usage{
stash_list(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
list of stashes in repository
}
\description{
List stashes in repository
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

# Configure a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

# Create a file, add and commit
writeLines("Hello world!", file.path(path, "test-1.txt"))
add(repo, 'test-1.txt')
commit(repo, "Commit message")

# Make one more commit
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-1.txt"))
add(repo, 'test-1.txt')
commit(repo, "Next commit message")

# Create one more file
writeLines("Hello world!", file.path(path, "test-2.txt"))

# Check that there are no stashes
stash_list(repo)

# Stash
stash(repo)

# Only untracked changes, therefore no stashes
stash_list(repo)

# Stash and include untracked changes
stash(repo, "Stash message", untracked=TRUE)

# View stash
stash_list(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.r
\docType{methods}
\name{is_local}
\alias{is_local}
\alias{is_local,git_branch-method}
\title{Check if branch is local}
\usage{
is_local(branch)

\S4method{is_local}{git_branch}(branch)
}
\arguments{
\item{branch}{The branch \code{object} to check if it's local}
}
\value{
TRUE if branch is local, else FALSE
}
\description{
Check if branch is local
}
\examples{
\dontrun{
## Initialize repositories
path_bare <- tempfile(pattern="git2r-")
path_repo <- tempfile(pattern="git2r-")
dir.create(path_bare)
dir.create(path_repo)
repo_bare <- init(path_bare, bare = TRUE)
repo <- clone(path_bare, path_repo)

## Config first user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")

## Write to a file and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           file.path(path_repo, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Push commits from repository to bare repository
## Adds an upstream tracking branch to branch 'master'
push(repo, "origin", "refs/heads/master")

## List branches
branches(repo)

## Check if first branch is_local
is_local(branches(repo)[[1]])

## Check if second branch is_local
is_local(branches(repo)[[2]])
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.R
\name{ahead_behind}
\alias{ahead_behind}
\title{Ahead Behind}
\usage{
ahead_behind(local = NULL, upstream = NULL)
}
\arguments{
\item{local}{a git_commit object. Can also be a tag or a branch,
and in that case the commit will be the target of the tag or
branch.}

\item{upstream}{a git_commit object. Can also be a tag or a
branch, and in that case the commit will be the target of the
tag or branch.}
}
\value{
An integer vector of length 2 with number of commits that
    the upstream commit is ahead and behind the local commit
}
\description{
Count the number of unique commits between two commit objects.
}
\examples{
\dontrun{
## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message 1")
tag_1 <- tag(repo, "Tagname1", "Tag message 1")

# Change file and commit
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Commit message 2")
tag_2 <- tag(repo, "Tagname2", "Tag message 2")

ahead_behind(commit_1, commit_2)
ahead_behind(tag_1, tag_2)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.r
\docType{methods}
\name{head,git_repository-method}
\alias{head,git_repository-method}
\title{Get HEAD for a repository}
\usage{
\S4method{head}{git_repository}(x)
}
\arguments{
\item{x}{The repository \code{x} to check head}
}
\value{
NULL if unborn branch or not found. S4 class git_branch if
not a detached head. S4 class git_commit if detached head
}
\description{
Get HEAD for a repository
}
\examples{
\dontrun{
## Create and initialize a repository in a temporary directory
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
config(repo, user.name="Alice", user.email="alice@example.org")

## Create a file, add and commit
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Commit message")

## Get HEAD of repository
head(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/plot.R
\name{plot.git_repository}
\alias{plot.git_repository}
\title{Plot commits over time}
\usage{
\method{plot}{git_repository}(
  x,
  breaks = c("month", "year", "quarter", "week", "day"),
  main = NULL,
  ...
)
}
\arguments{
\item{x}{The repository to plot}

\item{breaks}{Default is \code{month}. Change to year, quarter,
week or day as necessary.}

\item{main}{Default title for the plot is "Commits on repo:" and
repository workdir basename. Supply a new title if you desire one.}

\item{...}{Additional arguments affecting the plot}
}
\description{
Plot commits over time
}
\examples{
\dontrun{
## Initialize repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- clone("https://github.com/ropensci/git2r.git", path)

## Plot commits
plot(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/when.R
\name{when}
\alias{when}
\title{When}
\usage{
when(object, tz = "GMT", origin = "1970-01-01", usetz = TRUE)
}
\arguments{
\item{object}{the \code{object} to extract the time slot from.}

\item{tz}{a character string.  The time zone specification to be used
    for the conversion, \emph{if one is required}.  System-specific (see
    \link[base]{time zones}), but \code{""} is the current time zone, and
    \code{"GMT"} is UTC (Universal Time, Coordinated).  Invalid values
    are most commonly treated as UTC, on some platforms with a warning.}

\item{origin}{a date-time object, or something which can be coerced by
    \code{as.POSIXct(tz = "GMT")} to such an object.  Optional since \R
    4.3.0, where the equivalent of \code{"1970-01-01"} is used.}

\item{usetz}{logical.  Should the time zone abbreviation be appended
    to the output?  This is used in printing times, and more reliable
    than using \code{"\%Z"}.}
}
\value{
A \code{character} vector of length one.
}
\description{
Help method to extract the time as a character string from a
git_commit, git_signature, git_tag and git_time object.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a first user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Create tag
tag(repo, "Tagname", "Tag message")

when(commits(repo)[[1]])
when(tags(repo)[[1]])
when(tags(repo)[[1]], tz = Sys.timezone())
}
}
\seealso{
\code{\link{git_time}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.r
\docType{methods}
\name{is_empty}
\alias{is_empty}
\alias{is_empty,missing-method}
\alias{is_empty,git_repository-method}
\title{Check if repository is empty}
\usage{
is_empty(repo)

\S4method{is_empty}{missing}()

\S4method{is_empty}{git_repository}(repo)
}
\arguments{
\item{repo}{The repository \code{object}
\code{\linkS4class{git_repository}} to check if it's empty. If the
\code{repo} argument is missing, the repository is searched for
with \code{\link{discover_repository}} in the current working
directory.}
}
\value{
TRUE if empty else FALSE
}
\description{
Check if repository is empty
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name="Alice", user.email="alice@example.org")

## Check if it's an empty repository
is_empty(repo)

## Commit a file
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Check if it's an empty repository
is_empty(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/reflog.R
\name{print.git_reflog_entry}
\alias{print.git_reflog_entry}
\title{Print a reflog entry}
\usage{
\method{print}{git_reflog_entry}(x, ...)
}
\arguments{
\item{x}{The reflog entry}

\item{...}{Unused}
}
\value{
None (invisible 'NULL').
}
\description{
Print a reflog entry
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## View repository HEAD reflog
reflog(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/refspec.r
\name{get_refspec}
\alias{get_refspec}
\title{Create push refspec from arguments}
\usage{
get_refspec(repo = NULL, remote = NULL, spec = NULL, opts = NULL)
}
\arguments{
\item{repo}{S4 class \code{git_repository}}

\item{opts}{List with push options. Default is NULL.}

\item{name}{The remote's name. Default is NULL.}

\item{refspec}{The refspec to be pushed. Default is NULL.}
}
\value{
List with remote (character vector) and refspec (character
vector).
}
\description{
Create push refspec from arguments
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stash.R
\name{summary.git_stash}
\alias{summary.git_stash}
\title{Summary of a stash}
\usage{
\method{summary}{git_stash}(object, ...)
}
\arguments{
\item{object}{The stash \code{object}}

\item{...}{Additional arguments affecting the summary produced.}
}
\value{
None (invisible 'NULL').
}
\description{
Summary of a stash
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

# Configure a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

# Create a file, add and commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, 'test.txt')
commit(repo, "Commit message")

# Change file
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))

# Create stash in repository
stash(repo, "Stash message")

# View summary of stash
summary(stash_list(repo)[[1]])
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/merge.r
\name{merge_branch}
\alias{merge_branch}
\title{Merge branch into HEAD}
\usage{
merge_branch(branch, commit_on_success, merger)
}
\arguments{
\item{branch}{The branch}

\item{commit_on_success}{If there are no conflicts written to the
index, the merge commit will be committed.}

\item{merger}{Who made the merge.}
}
\value{
A \code{\linkS4class{git_merge_result}} object.
}
\description{
Merge branch into HEAD
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remote.R
\name{remote_ls}
\alias{remote_ls}
\title{List references in a remote repository}
\usage{
remote_ls(name = NULL, repo = NULL, credentials = NULL)
}
\arguments{
\item{name}{Character vector with the "remote" repository URL to
query or the name of the remote if a \code{repo} argument is
given.}

\item{repo}{an optional repository object used if remotes are
specified by name.}

\item{credentials}{The credentials for remote repository
access. Default is NULL. To use and query an ssh-agent for the
ssh key credentials, let this parameter be NULL (the default).}
}
\value{
Character vector for each reference with the associated
    commit IDs.
}
\description{
Displays references available in a remote repository along with
the associated commit IDs.  Akin to the 'git ls-remote' command.
}
\examples{
\dontrun{
remote_ls("https://github.com/ropensci/git2r")
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential.R
\name{cred_token}
\alias{cred_token}
\title{Create a new personal access token credential object}
\usage{
cred_token(token = "GITHUB_PAT")
}
\arguments{
\item{token}{The name of the environmental variable that holds the
personal access token for the authentication. Default is
\code{GITHUB_PAT}.}
}
\value{
A list of class \code{cred_token} with entry:
\describe{
  \item{token}{
    The name of the environmental variable that holds
    the personal access token for the authentication.
  }
}
}
\description{
The personal access token is stored in an envrionmental variable.
Environmental variables can be written to the file
\code{.Renviron}. This file is read by \emph{R} during startup,
see \code{\link[base]{Startup}}. On GitHub, personal access tokens
function like ordinary OAuth access tokens. They can be used
instead of a password for Git over HTTPS, see
\url{https://help.github.com/articles/creating-an-access-token-for-command-line-use}
}
\examples{
\dontrun{
## Create a personal access token credential object.
## This example assumes that the token is stored in
## the 'GITHUB_PAT' environmental variable.
repo <- repository("git2r")
cred <- cred_token()
push(repo, credentials = cred)
}
}
\seealso{
Other git credential functions: 
\code{\link{cred_env}()},
\code{\link{cred_ssh_key}()},
\code{\link{cred_user_pass}()}
}
\concept{git credential functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.R
\name{branch_get_upstream}
\alias{branch_get_upstream}
\title{Get remote tracking branch}
\usage{
branch_get_upstream(branch = NULL)
}
\arguments{
\item{branch}{The branch}
}
\value{
\code{git_branch} object or NULL if no remote tracking
    branch.
}
\description{
Get remote tracking branch, given a local branch.
}
\examples{
\dontrun{
## Initialize two temporary repositories
path_bare <- tempfile(pattern="git2r-")
path_repo <- tempfile(pattern="git2r-")
dir.create(path_bare)
dir.create(path_repo)
repo_bare <- init(path_bare, bare = TRUE)
repo <- clone(path_bare, path_repo)

## Config user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Write to a file and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path_repo, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Push commits from repository to bare repository
## Adds an upstream tracking branch to branch 'master'
push(repo, "origin", "refs/heads/master")

## Get remote tracking branch
branch_get_upstream(repository_head(repo))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{is_detached}
\alias{is_detached}
\title{Check if HEAD of repository is detached}
\usage{
is_detached(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\value{
\code{TRUE} if repository HEAD is detached, else
    \code{FALSE}.
}
\description{
Check if HEAD of repository is detached
}
\examples{
\dontrun{
## Create and initialize a repository in a temporary directory
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit_1 <- commit(repo, "Commit message 1")

## Change file, add and commit
lines <- c(
  "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
  "eiusmod tempor incididunt ut labore et dolore magna aliqua.")
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Commit message 2")

## HEAD of repository is not detached
is_detached(repo)

## Checkout first commit
checkout(commit_1)

## HEAD of repository is detached
is_detached(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/punch_card.r
\docType{methods}
\name{punch_card}
\alias{punch_card}
\alias{punch_card,git_repository-method}
\title{Punch card}
\usage{
punch_card(repo, main = NULL, ...)

\S4method{punch_card}{git_repository}(repo, main = NULL, ...)
}
\arguments{
\item{repo}{The repository}

\item{main}{Default title for the plot is "Punch card on repo:"
and repository workdir basename. Supply a new title if you desire
one.}

\item{...}{Additional arguments affecting the plot}
}
\value{
invisible NULL
}
\description{
Punch card
}
\examples{
\dontrun{
## Initialize repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- clone("https://github.com/ropensci/git2r.git", path)

## Plot
punch_card(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.R
\name{last_commit}
\alias{last_commit}
\title{Last commit}
\usage{
last_commit(repo = ".")
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}
}
\description{
Get last commit in the current branch.
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Write to a file and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Get last commit
last_commit(repo)
last_commit(path)

## Coerce the last commit to a data.frame
as.data.frame(last_commit(path), "data.frame")

## Summary of last commit in repository
summary(last_commit(repo))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tag.R
\name{tag}
\alias{tag}
\title{Create tag targeting HEAD commit in repository}
\usage{
tag(
  object = ".",
  name = NULL,
  message = NULL,
  session = FALSE,
  tagger = NULL,
  force = FALSE
)
}
\arguments{
\item{object}{The repository \code{object}.}

\item{name}{Name for the tag.}

\item{message}{The tag message. Specify a tag message to create an
annotated tag. A lightweight tag is created if the message
parameter is \code{NULL}.}

\item{session}{Add sessionInfo to tag message. Default is FALSE.}

\item{tagger}{The tagger (author) of the tag}

\item{force}{Overwrite existing tag. Default = FALSE}
}
\value{
invisible(\code{git_tag}) object
}
\description{
Create tag targeting HEAD commit in repository
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Commit a text file
filename <- file.path(path, "example.txt")
writeLines("Hello world!", filename)
add(repo, "example.txt")
commit(repo, "First commit message")

## Create an annotated tag
tag(repo, "v1.0", "Tag message")

## List tags
tags(repo)

## Make a change to the text file and commit.
writeLines(c("Hello world!", "HELLO WORLD!"), filename)
add(repo, "example.txt")
commit(repo, "Second commit message")

## Create a lightweight tag
tag(repo, "v2.0")

## List tags
tags(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/status.R
\name{status}
\alias{status}
\title{Status}
\usage{
status(
  repo = ".",
  staged = TRUE,
  unstaged = TRUE,
  untracked = TRUE,
  ignored = FALSE,
  all_untracked = FALSE
)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{staged}{Include staged files. Default TRUE.}

\item{unstaged}{Include unstaged files. Default TRUE.}

\item{untracked}{Include untracked files and directories. Default
TRUE.}

\item{ignored}{Include ignored files. Default FALSE.}

\item{all_untracked}{Shows individual files in untracked
directories if \code{untracked} is \code{TRUE}.}
}
\value{
\code{git_status} with repository status
}
\description{
Display state of the repository working directory and the staging
area.
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("Hello world!", file.path(path, "test.txt"))

## Check status; untracked file
status(repo)

## Add file
add(repo, "test.txt")

## Check status; staged file
status(repo)

## Commit
commit(repo, "First commit message")

## Check status; clean
status(repo)

## Change the file
writeLines(c("Hello again!", "Here is a second line", "And a third"),
           file.path(path, "test.txt"))

## Check status; unstaged file
status(repo)

## Add file and commit
add(repo, "test.txt")
commit(repo, "Second commit message")

## Check status; clean
status(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{init}
\alias{init}
\title{Init a repository}
\usage{
init(path = ".", bare = FALSE)
}
\arguments{
\item{path}{A path to where to init a git repository}

\item{bare}{If TRUE, a Git repository without a working directory
is created at the pointed path. If FALSE, provided path will
be considered as the working directory into which the .git
directory will be created.}
}
\value{
A \code{git_repository} object
}
\description{
Init a repository
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
is_bare(repo)

## Initialize a bare repository
path_bare <- tempfile(pattern="git2r-")
dir.create(path_bare)
repo_bare <- init(path_bare, bare = TRUE)
is_bare(repo_bare)
}
}
\seealso{
\link{repository}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.r
\docType{methods}
\name{is_shallow}
\alias{is_shallow}
\alias{is_shallow,missing-method}
\alias{is_shallow,git_repository-method}
\title{Determine if the repository was a shallow clone}
\usage{
is_shallow(repo)

\S4method{is_shallow}{missing}()

\S4method{is_shallow}{git_repository}(repo)
}
\arguments{
\item{repo}{The repository \code{object}
\code{\linkS4class{git_repository}}. If the \code{repo} argument
is missing, the repository is searched for with
\code{\link{discover_repository}} in the current working
directory.}
}
\value{
TRUE if shallow clone, else FALSE
}
\description{
Determine if the repository was a shallow clone
}
\examples{
\dontrun{
## Initialize repository
path_repo_1 <- tempfile(pattern="git2r-")
path_repo_2 <- tempfile(pattern="git2r-")
dir.create(path_repo_1)
dir.create(path_repo_2)
repo_1 <- init(path_repo_1)

## Config user and commit a file
config(repo_1, user.name="Alice", user.email="alice@example.org")

## Write to a file and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           file.path(path_repo_1, "example.txt"))
add(repo_1, "example.txt")
commit(repo_1, "First commit message")

## Change file and commit
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
           file.path(path_repo_1, "example.txt"))
add(repo_1, "example.txt")
commit(repo_1, "Second commit message")

## Change file again and commit.
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad",
             "minim veniam, quis nostrud exercitation ullamco laboris nisi ut"),
           file.path(path_repo_1, "example.txt"))
add(repo_1, "example.txt")
commit(repo_1, "Third commit message")

## Clone to second repository
repo_2 <- clone(path_repo_1, path_repo_2)

## Check if it's a shallow clone
is_shallow(repo_2)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/time.r
\docType{methods}
\name{show,git_time-method}
\alias{show,git_time-method}
\alias{show,git_time-methods}
\title{Brief summary of \code{git_time}}
\usage{
\S4method{show}{git_time}(object)
}
\arguments{
\item{object}{The time \code{object}}
}
\value{
None (invisible 'NULL').
}
\description{
Brief summary of \code{git_time}
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name="Alice", user.email="alice@example.org")

## Brief summary of git_time from the default signature
default_signature(repo)@when
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/sha.R
\name{sha}
\alias{sha}
\alias{sha.git_blob}
\alias{sha.git_branch}
\alias{sha.git_commit}
\alias{sha.git_note}
\alias{sha.git_reference}
\alias{sha.git_reflog_entry}
\alias{sha.git_tag}
\alias{sha.git_tree}
\alias{sha.git_fetch_head}
\alias{sha.git_merge_result}
\title{Get the SHA-1 of a git object}
\usage{
sha(object)

\method{sha}{git_blob}(object)

\method{sha}{git_branch}(object)

\method{sha}{git_commit}(object)

\method{sha}{git_note}(object)

\method{sha}{git_reference}(object)

\method{sha}{git_reflog_entry}(object)

\method{sha}{git_tag}(object)

\method{sha}{git_tree}(object)

\method{sha}{git_fetch_head}(object)

\method{sha}{git_merge_result}(object)
}
\arguments{
\item{object}{a git object to get the SHA-1 from.}
}
\value{
The 40 character hexadecimal string of the SHA-1.
}
\description{
Get the 40 character hexadecimal string of the SHA-1.
}
\examples{
\dontrun{
## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message 1")

## Get the SHA-1 of the last commit
sha(last_commit(repo))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.R
\name{is_merge}
\alias{is_merge}
\title{Is merge}
\usage{
is_merge(commit = NULL)
}
\arguments{
\item{commit}{a git_commit object.}
}
\value{
TRUE if commit has more than one parent, else FALSE
}
\description{
Determine if a commit is a merge commit, i.e. has more than one
parent.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
writeLines(c("First line in file 1.", "Second line in file 1."),
           file.path(path, "example-1.txt"))
add(repo, "example-1.txt")
commit(repo, "First commit message")

## Create and add one more file
writeLines(c("First line in file 2.", "Second line in file 2."),
           file.path(path, "example-2.txt"))
add(repo, "example-2.txt")
commit(repo, "Second commit message")

## Create a new branch 'fix'
checkout(repo, "fix", create = TRUE)

## Update 'example-1.txt' (swap words in first line) and commit
writeLines(c("line First in file 1.", "Second line in file 1."),
           file.path(path, "example-1.txt"))
add(repo, "example-1.txt")
commit(repo, "Third commit message")

checkout(repo, "master")

## Update 'example-2.txt' (swap words in second line) and commit
writeLines(c("First line in file 2.", "line Second in file 2."),
           file.path(path, "example-2.txt"))
add(repo, "example-2.txt")
commit(repo, "Fourth commit message")

## Merge 'fix'
merge(repo, "fix")

## Display parents of last commit
parents(lookup(repo, branch_target(repository_head(repo))))

## Check that last commit is a merge
is_merge(lookup(repo, branch_target(repository_head(repo))))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/punch_card.R
\name{punch_card}
\alias{punch_card}
\title{Punch card}
\usage{
punch_card(repo = ".", main = NULL, ...)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{main}{Default title for the plot is "Punch card on repo:"
and repository workdir basename. Supply a new title if you
desire one.}

\item{...}{Additional arguments affecting the plot}
}
\value{
invisible NULL
}
\description{
Punch card
}
\examples{
\dontrun{
## Initialize repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- clone("https://github.com/ropensci/git2r.git", path)

## Plot
punch_card(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stash.r
\docType{methods}
\name{show,git_stash-method}
\alias{show,git_stash-method}
\alias{show,git_stash-methods}
\title{Brief summary of a stash}
\usage{
\S4method{show}{git_stash}(object)
}
\arguments{
\item{object}{The stash \code{object}}
}
\value{
None (invisible 'NULL').
}
\description{
Brief summary of a stash
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

# Configure a user
config(repo, user.name="Alice", user.email="alice@example.org")

# Create a file, add and commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, 'test.txt')
commit(repo, "Commit message")

# Change file
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))

# Create stash in repository
stash(repo, "Stash message")

# View brief summary of stash
stash_list(repo)[[1]]
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/libgit2.R
\name{libgit2_features}
\alias{libgit2_features}
\title{Compile time options for libgit2.}
\usage{
libgit2_features()
}
\value{
A list with threads, https and ssh set to TRUE/FALSE.
}
\description{
Compile time options for libgit2.
}
\examples{
libgit2_features()
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_diff_file-class}
\alias{git_diff_file-class}
\title{Git diff file}
\description{
Git diff file
}
\section{Slots}{

\describe{
  \item{old_file}{
    Filename for old side of diff.
  }
  \item{new_file}{
    Filename for new side of diff.
  }
  \item{hunks}{
    List of \code{\linkS4class{git_diff_hunk}} objects.
  }
}
}

\keyword{classes}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential.R
\name{cred_env}
\alias{cred_env}
\title{Create a new environmental credential object}
\usage{
cred_env(username = NULL, password = NULL)
}
\arguments{
\item{username}{The name of the environmental variable that holds
the username for the authentication.}

\item{password}{The name of the environmental variable that holds
the password for the authentication.}
}
\value{
A list of class \code{cred_env} with entries:
\describe{
  \item{username}{
    The name of the environmental variable that holds
    the username for the authentication.
  }
  \item{password}{
    The name of the environmental variable that holds
    the password for the authentication.
  }
}
}
\description{
Environmental variables can be written to the file
\code{.Renviron}. This file is read by \emph{R} during startup,
see \code{\link[base]{Startup}}.
}
\examples{
\dontrun{
## Create an environmental credential object for the username and
## password.
cred <- cred_env("NAME_OF_ENV_VARIABLE_WITH_USERNAME",
                 "NAME_OF_ENV_VARIABLE_WITH_PASSWORD")
repo <- repository("git2r")
push(repo, credentials = cred)
}
}
\seealso{
Other git credential functions: 
\code{\link{cred_ssh_key}()},
\code{\link{cred_token}()},
\code{\link{cred_user_pass}()}
}
\concept{git credential functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blame.r
\docType{methods}
\name{blame}
\alias{blame}
\alias{blame,git_repository,character-method}
\title{Get blame for file}
\usage{
blame(repo, path)

\S4method{blame}{git_repository,character}(repo, path)
}
\arguments{
\item{repo}{The repository}

\item{path}{Path to the file to consider}
}
\value{
S4 class git_blame object
}
\description{
Get blame for file
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a first user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Create a second user and change the file
config(repo, user.name="Bob", user.email="bob@example.org")
writeLines(c("Hello world!", "HELLO WORLD!", "HOLA"),
           file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "Second commit message")

## Check blame
blame(repo, "example.txt")
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_reference-class}
\alias{git_reference-class}
\title{S4 class to handle a git reference}
\description{
Class \code{"git_reference"}
}
\section{Slots}{

\describe{
  \item{name}{
    The full name of the reference.
  }
  \item{type}{
    Type of the reference, either direct (GIT_REF_OID == 1) or
    symbolic (GIT_REF_SYMBOLIC == 2)
  }
  \item{sha}{
    The 40 character hexadecimal string of the SHA-1
  }
  \item{target}{
    The full name to the reference pointed to by a
    symbolic reference.
  }
  \item{shorthand}{
    The reference's short name
  }
}
}

\section{Methods}{

\describe{
  \item{show}{\code{signature(object = "git_reference")}}
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.R
\name{lookup_commit}
\alias{lookup_commit}
\alias{lookup_commit.git_branch}
\alias{lookup_commit.git_commit}
\alias{lookup_commit.git_tag}
\alias{lookup_commit.git_reference}
\title{Lookup the commit related to a git object}
\usage{
lookup_commit(object)

\method{lookup_commit}{git_branch}(object)

\method{lookup_commit}{git_commit}(object)

\method{lookup_commit}{git_tag}(object)

\method{lookup_commit}{git_reference}(object)
}
\arguments{
\item{object}{a git object to get the related commit from.}
}
\value{
A git commit object.
}
\description{
Lookup the commit related to a git_reference, git_tag or
git_branch object.
}
\examples{
\dontrun{
## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file, add and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit(repo, "Commit message 1")

## Get the commit pointed to by the 'master' branch
lookup_commit(repository_head(repo))

## Create a tag
a_tag <- tag(repo, "Tagname", "Tag message")

## Get the commit pointed to by 'a_tag'
lookup_commit(a_tag)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_tag-class}
\alias{git_tag-class}
\title{S4 class to handle a git tag}
\description{
Class \code{"git_tag"}
}
\section{Slots}{

\describe{
  \item{sha}{
    The 40 character hexadecimal string of the SHA-1
  }
  \item{message}{
    The message of the tag
  }
  \item{name}{
    The name of the tag
  }
  \item{tagger}{
    The tagger (author) of the tag
  }
  \item{target}{
    The target of the tag
  }
  \item{repo}{
    The S4 class git_repository that contains the tag
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/diff.R, R/merge.R, R/repository.R, R/tree.R
\docType{import}
\name{reexports}
\alias{reexports}
\alias{diff}
\alias{merge}
\alias{head}
\alias{as.data.frame}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{base}{\code{\link[base]{as.data.frame}}, \code{\link[base]{diff}}, \code{\link[base]{merge}}}

  \item{utils}{\code{\link[utils]{head}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/index.R
\name{add}
\alias{add}
\title{Add file(s) to index}
\usage{
add(repo = ".", path = NULL, force = FALSE)
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{path}{Character vector with file names or shell glob
patterns that will matched against files in the repository's
working directory. Each file that matches will be added to the
index (either updating an existing entry or adding a new
entry).}

\item{force}{Add ignored files. Default is FALSE.}
}
\value{
invisible(NULL)
}
\description{
Add file(s) to index
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Create a file
writeLines("a", file.path(path, "a.txt"))

## Add file to repository and view status
add(repo, "a.txt")
status(repo)

## Add file with a leading './' when the repository working
## directory is the current working directory
setwd(path)
writeLines("b", file.path(path, "b.txt"))
add(repo, "./b.txt")
status(repo)

## Add a file in a sub-folder with sub-folder as the working
## directory. Create a file in the root of the repository
## working directory that will remain untracked.
dir.create(file.path(path, "sub_dir"))
setwd("./sub_dir")
writeLines("c", file.path(path, "c.txt"))
writeLines("c", file.path(path, "sub_dir/c.txt"))
add(repo, "c.txt")
status(repo)

## Add files with glob expansion when the current working
## directory is outside the repository's working directory.
setwd(tempdir())
dir.create(file.path(path, "glob_dir"))
writeLines("d", file.path(path, "glob_dir/d.txt"))
writeLines("e", file.path(path, "glob_dir/e.txt"))
writeLines("f", file.path(path, "glob_dir/f.txt"))
writeLines("g", file.path(path, "glob_dir/g.md"))
add(repo, "glob_dir/*txt")
status(repo)

## Add file with glob expansion with a relative path when
## the current working directory is inside the repository's
## working directory.
setwd(path)
add(repo, "./glob_dir/*md")
status(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/when.r
\docType{methods}
\name{when}
\alias{when}
\alias{when,git_commit-method}
\alias{when,git_signature-method}
\alias{when,git_stash-method}
\alias{when,git_tag-method}
\alias{when,git_time-method}
\title{When}
\usage{
when(object)

\S4method{when}{git_commit}(object)

\S4method{when}{git_signature}(object)

\S4method{when}{git_stash}(object)

\S4method{when}{git_tag}(object)

\S4method{when}{git_time}(object)
}
\arguments{
\item{object}{the \code{object} to extract the time slot from.}
}
\value{
A \code{character} vector of length one.
}
\description{
Help method to extract the time as a character string from the S4
classes \code{git_commit}, \code{git_signature}, \code{git_tag}
and \code{git_time}.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a first user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Create tag
tag(repo, "Tagname", "Tag message")

when(commits(repo)[[1]])
when(tags(repo)[[1]])
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stash.R
\name{stash_drop}
\alias{stash_drop}
\title{Drop stash}
\usage{
stash_drop(object = ".", index = 1)
}
\arguments{
\item{object}{path to a repository, or a \code{git_repository}
object, or the stash \code{object} to drop. Default is a
\code{path = '.'} to a reposiory.}

\item{index}{The index to the stash to drop. Only used when
\code{object} is a path to a repository or a
\code{git_repository} object. Default is \code{index = 1}.}
}
\value{
invisible NULL
}
\description{
Drop stash
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

# Configure a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

# Create a file, add and commit
writeLines("Hello world!", file.path(path, "test.txt"))
add(repo, 'test.txt')
commit(repo, "Commit message")

# Change file
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test.txt"))

# Create stash in repository
stash(repo)

# Change file
writeLines(c("Hello world!", "HeLlO wOrLd!"), file.path(path, "test.txt"))

# Create stash in repository
stash(repo)

# View stashes
stash_list(repo)

# Drop git_stash object in repository
stash_drop(stash_list(repo)[[1]])

## Drop stash using an index to stash
stash_drop(repo, 1)

# View stashes
stash_list(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.r
\docType{methods}
\name{is_bare}
\alias{is_bare}
\alias{is_bare,git_repository-method}
\alias{is_bare,missing-method}
\title{Check if repository is bare}
\usage{
is_bare(repo)

\S4method{is_bare}{git_repository}(repo)

\S4method{is_bare}{missing}()
}
\arguments{
\item{repo}{The repository \code{object}
\code{\linkS4class{git_repository}} to check if it's bare. If the
\code{repo} argument is missing, the repository is searched for
with \code{\link{discover_repository}} in the current working
directory.}
}
\value{
TRUE if bare repository, else FALSE
}
\description{
Check if repository is bare
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)
is_bare(repo)

## Initialize a bare repository
path_bare <- tempfile(pattern="git2r-")
dir.create(path_bare)
repo_bare <- init(path_bare, bare = TRUE)
is_bare(repo_bare)
}
}
\seealso{
\link{init}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blob.R
\name{length.git_blob}
\alias{length.git_blob}
\title{Size in bytes of the contents of a blob}
\usage{
\method{length}{git_blob}(x)
}
\arguments{
\item{x}{The blob \code{object}}
}
\value{
a non-negative integer
}
\description{
Size in bytes of the contents of a blob
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Commit a text file
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit_1 <- commit(repo, "First commit message")
blob_1 <- tree(commit_1)["example.txt"]

## Get length in size of bytes of the content of the blob
length(blob_1)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential.R
\name{cred_ssh_key}
\alias{cred_ssh_key}
\title{Create a new passphrase-protected ssh key credential object}
\usage{
cred_ssh_key(
  publickey = ssh_path("id_rsa.pub"),
  privatekey = ssh_path("id_rsa"),
  passphrase = character(0)
)
}
\arguments{
\item{publickey}{The path to the public key of the
credential. Default is \code{ssh_path("id_rsa.pub")}}

\item{privatekey}{The path to the private key of the
credential. Default is \code{ssh_path("id_rsa")}}

\item{passphrase}{The passphrase of the credential. Default is
\code{character(0)}. If getPass is installed and private key
is passphrase protected \code{getPass::getPass()} will be
called to allow for interactive and obfuscated interactive
input of the passphrase.}
}
\value{
A list of class \code{cred_ssh_key} with entries:
\describe{
  \item{publickey}{
    The path to the public key of the credential
  }
  \item{privatekey}{
    The path to the private key of the credential
  }
  \item{passphrase}{
    The passphrase of the credential
  }
}
}
\description{
Create a new passphrase-protected ssh key credential object
}
\examples{
\dontrun{
## Create a ssh key credential object. It can optionally be
## passphrase-protected
cred <- cred_ssh_key(ssh_path("id_rsa.pub"), ssh_path("id_rsa"))
repo <- repository("git2r")
push(repo, credentials = cred)
}
}
\seealso{
Other git credential functions: 
\code{\link{cred_env}()},
\code{\link{cred_token}()},
\code{\link{cred_user_pass}()}
}
\concept{git credential functions}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/libgit2.R
\name{ssl_cert_locations}
\alias{ssl_cert_locations}
\title{Set the SSL certificate-authority locations}
\usage{
ssl_cert_locations(filename = NULL, path = NULL)
}
\arguments{
\item{filename}{Location of a file containing several certificates
concatenated together. Default NULL.}

\item{path}{Location of a directory holding several certificates,
one per file. Default NULL.}
}
\value{
invisible(NULL)
}
\description{
Set the SSL certificate-authority locations
}
\note{
Either parameter may be 'NULL', but not both.
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.R
\name{tree}
\alias{tree}
\title{Tree}
\usage{
tree(object = NULL)
}
\arguments{
\item{object}{the \code{commit} or \code{stash} object}
}
\value{
A S3 class git_tree object
}
\description{
Get the tree pointed to by a commit or stash.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a first user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

tree(last_commit(repo))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/branch.R
\name{branches}
\alias{branches}
\title{Branches}
\usage{
branches(repo = ".", flags = c("all", "local", "remote"))
}
\arguments{
\item{repo}{a path to a repository or a \code{git_repository}
object. Default is '.'}

\item{flags}{Filtering flags for the branch listing. Valid values
are 'all', 'local' or 'remote'}
}
\value{
list of branches in repository
}
\description{
List branches in repository
}
\examples{
\dontrun{
## Initialize repositories
path_bare <- tempfile(pattern="git2r-")
path_repo <- tempfile(pattern="git2r-")
dir.create(path_bare)
dir.create(path_repo)
repo_bare <- init(path_bare, bare = TRUE)
repo <- clone(path_bare, path_repo)

## Config first user and commit a file
config(repo, user.name = "Alice", user.email = "alice@example.org")

## Write to a file and commit
lines <- "Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do"
writeLines(lines, file.path(path_repo, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Push commits from repository to bare repository
## Adds an upstream tracking branch to branch 'master'
push(repo, "origin", "refs/heads/master")

## List branches
branches(repo)
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/libgit2.R
\name{libgit2_version}
\alias{libgit2_version}
\title{Version of the libgit2 library}
\usage{
libgit2_version()
}
\value{
A list with major, minor and rev
}
\description{
Version of the libgit2 library that the bundled source code is
based on
}
\examples{
libgit2_version()
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blob.R
\name{hashfile}
\alias{hashfile}
\title{Determine the sha from a blob in a file}
\usage{
hashfile(path = NULL)
}
\arguments{
\item{path}{The path vector with files to hash.}
}
\value{
A vector with the sha for each file in path.
}
\description{
The blob is not written to the object database.
}
\examples{
\dontrun{
## Create a file. NOTE: The line endings from writeLines gives
## LF (line feed) on Unix/Linux and CRLF (carriage return, line feed)
## on Windows. The example use writeChar to have more control.
path <- tempfile()
f <- file(path, "wb")
writeChar("Hello, world!\n", f, eos = NULL)
close(f)

## Generate hash
hashfile(path)
identical(hashfile(path), hash("Hello, world!\n"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/blob.R
\name{hash}
\alias{hash}
\title{Determine the sha from a blob string}
\usage{
hash(data = NULL)
}
\arguments{
\item{data}{The string vector to hash.}
}
\value{
A string vector with the sha for each string in data.
}
\description{
The blob is not written to the object database.
}
\examples{
\dontrun{
identical(hash(c("Hello, world!\n",
                 "test content\n")),
               c("af5626b4a114abcb82d63db7c8082c3c4756e51b",
                 "d670460b4b4aece5915caf5c68d12f560a9fe3e4"))
}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_note-class}
\alias{git_note-class}
\title{S4 class to handle a git note}
\description{
Class \code{git_note}
}
\section{Slots}{

\describe{
  \item{sha}{
    The 40 character hexadecimal string of the SHA-1 of the blob
    containing the message
  }
  \item{annotated}{
    The 40 character hexadecimal string of the SHA-1 of the git
    object being annotated
  }
  \item{message}{
    The note message
  }
  \item{refname}{
    Name of the reference
  }
  \item{repo}{
    The S4 class git_repository that contains the note
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.r
\name{add_session_info}
\alias{add_session_info}
\title{Add sessionInfo to message}
\usage{
add_session_info(message)
}
\arguments{
\item{message}{The message.}
}
\value{
message with appended sessionInfo
}
\description{
Add sessionInfo to message
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_blob-class}
\alias{git_blob-class}
\title{S4 class to handle a git blob}
\description{
Class \code{"git_blob"}
}
\section{Slots}{

\describe{
  \item{sha}{
    The 40 character hexadecimal string of the SHA-1
  }
  \item{repo}{
    The S4 class git_repository that contains the blob
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tree.r
\docType{methods}
\name{[,git_tree,integer,missing-method}
\alias{[,git_tree,integer,missing-method}
\alias{[,git_tree,numeric,missing-method}
\alias{[,git_tree,logical,missing-method}
\alias{[,git_tree,character,missing-method}
\title{Extract object from tree}
\usage{
\S4method{[}{git_tree,integer,missing}(x, i)

\S4method{[}{git_tree,numeric,missing}(x, i)

\S4method{[}{git_tree,logical,missing}(x, i)

\S4method{[}{git_tree,character,missing}(x, i)
}
\arguments{
\item{x}{The tree \code{object}}

\item{i}{The index (integer or logical) of the tree object to
extract. If negative values, all elements except those indicated
are selected. A character vector to match against the names of
objects to extract.}
}
\value{
Git object
}
\description{
Lookup a tree entry by its position in the tree
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
dir.create(file.path(path, "subfolder"))
repo <- init(path)

## Create a user
config(repo, user.name="Alice", user.email="alice@example.org")

## Create three files and commit
writeLines("First file",  file.path(path, "example-1.txt"))
writeLines("Second file", file.path(path, "subfolder/example-2.txt"))
writeLines("Third file",  file.path(path, "example-3.txt"))
add(repo, c("example-1.txt", "subfolder/example-2.txt", "example-3.txt"))
new_commit <- commit(repo, "Commit message")

## Pick a tree in the repository
tree_object <- tree(new_commit)

## Summarize tree
summary(tree_object)

## Select item by name
tree_object["example-1.txt"]

## Select first item in tree
tree_object[1]

## Select first three items in tree
tree_object[1:3]

## Select all blobs in tree
tree_object[vapply(as(tree_object, 'list'), is_blob, logical(1))]
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_blame_hunk-class}
\alias{git_blame_hunk-class}
\title{S4 class to represent a blame hunk}
\description{
Class \code{"git_blame_hunk"}
}
\section{Slots}{

\describe{
  \item{lines_in_hunk}{
    The number of lines in this hunk
  }
  \item{final_commit_id}{
    The sha of the commit where this line was last changed
  }
  \item{final_start_line_number}{
    The 1-based line number where this hunk begins, in the final
    version of the file
  }
  \item{final_signature}{
    Final committer
  }
  \item{orig_commit_id}{
    The sha of the commit where this hunk was found. This will usually
    be the same as 'final_commit_id'.
  }
  \item{orig_start_line_number}{
     The 1-based line number where this hunk begins in the file
     named by 'orig_path' in the commit specified by 'orig_commit_id'.
  }
  \item{orig_signature}{
    Origin committer
  }
  \item{orig_path}{
    The path to the file where this hunk originated, as of the commit
    specified by 'orig_commit_id'
  }
  \item{boundary}{
    TRUE iff the hunk has been tracked to a boundary commit.
  }
  \item{repo}{
    The S4 class git_repository that contains the blame hunk
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.r
\docType{methods}
\name{descendant_of}
\alias{descendant_of}
\alias{descendant_of,git_commit,git_commit-method}
\title{Descendant}
\usage{
descendant_of(commit, ancestor)

\S4method{descendant_of}{git_commit,git_commit}(commit, ancestor)
}
\arguments{
\item{commit}{a S4 class git_commit \code{object}.}

\item{ancestor}{a S4 class git_commit \code{object} to check if
ancestor to \code{commit}.}
}
\value{
TRUE if commit is descendant of \code{ancestor}, else
FALSE
}
\description{
Determine if a commit is the descendant of another commit
}
\examples{
\dontrun{
## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name="Alice", user.email="alice@example.org")

## Create a file, add and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message 1")

# Change file and commit
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
             con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Commit message 2")

descendant_of(commit_1, commit_2)
descendant_of(commit_2, commit_1)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.r
\docType{methods}
\name{default_signature}
\alias{default_signature}
\alias{default_signature,git_repository-method}
\title{Get the signature}
\usage{
default_signature(repo)

\S4method{default_signature}{git_repository}(repo)
}
\arguments{
\item{repo}{The repository \code{object} to check signature}
}
\value{
S4 class git_signature
}
\description{
Get the signature according to the repository's configuration
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name="Alice", user.email="alice@example.org")

## Get the default signature
default_signature(repo)

## Change user
config(repo, user.name="Bob", user.email="bob@example.org")

## Get the default signature
default_signature(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.r
\docType{methods}
\name{commit}
\alias{commit}
\alias{commit,git_repository-method}
\title{Commit}
\usage{
commit(repo, message = NULL, all = FALSE, session = FALSE,
  reference = "HEAD", author = default_signature(repo),
  committer = default_signature(repo))

\S4method{commit}{git_repository}(repo, message = NULL, all = FALSE,
  session = FALSE, reference = "HEAD", author = default_signature(repo),
  committer = default_signature(repo))
}
\arguments{
\item{repo}{The repository \code{object}.}

\item{message}{The commit message.}

\item{all}{Stage modified and deleted files. Files not added to
Git are not affected.}

\item{session}{Add sessionInfo to commit message. Default is
FALSE.}

\item{reference}{Name of the reference that will be updated to
point to this commit.}

\item{author}{Signature with author and author time of commit.}

\item{committer}{Signature with committer and commit time of commit.}
}
\value{
\code{\linkS4class{git_commit}} object
}
\description{
Commit
}
\examples{
\dontrun{
## Initialize a repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Config user
config(repo, user.name="Alice", user.email="alice@example.org")

## Write to a file and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/stash.r
\docType{methods}
\name{stash_list}
\alias{stash_list}
\alias{stash_list,missing-method}
\alias{stash_list,git_repository-method}
\title{List stashes in repository}
\usage{
stash_list(repo)

\S4method{stash_list}{missing}()

\S4method{stash_list}{git_repository}(repo)
}
\arguments{
\item{repo}{The repository \code{object}
\code{\linkS4class{git_repository}}. If the \code{repo} argument
is missing, the repository is searched for with
\code{\link{discover_repository}} in the current working
directory.}
}
\value{
list of stashes in repository
}
\description{
List stashes in repository
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

# Configure a user
config(repo, user.name="Alice", user.email="alice@example.org")

# Create a file, add and commit
writeLines("Hello world!", file.path(path, "test-1.txt"))
add(repo, 'test-1.txt')
commit(repo, "Commit message")

# Make one more commit
writeLines(c("Hello world!", "HELLO WORLD!"), file.path(path, "test-1.txt"))
add(repo, 'test-1.txt')
commit(repo, "Next commit message")

# Create one more file
writeLines("Hello world!", file.path(path, "test-2.txt"))

# Check that there are no stashes
stash_list(repo)

# Stash
stash(repo)

# Only untracked changes, therefore no stashes
stash_list(repo)

# Stash and include untracked changes
stash(repo, "Stash message", untracked=TRUE)

# View stash
stash_list(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/credential.r
\name{ssh_key_needs_passphrase}
\alias{ssh_key_needs_passphrase}
\title{Check if private key is passphrase protected}
\usage{
ssh_key_needs_passphrase(privatekey = "~/.ssh/id_rsa")
}
\arguments{
\item{privatekey}{The path to the private key of the
credential. Default is \code{'~/.ssh/id_rsa'}}
}
\description{
Check if private key is passphrase protected
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{git_transfer_progress-class}
\alias{git_transfer_progress-class}
\title{Class \code{"git_transfer_progress"}}
\description{
Statistics from the fetch operation.
}
\section{Slots}{

\describe{
  \item{total_objects}{
    Number of objects in the packfile being downloaded
  }
  \item{indexed_objects}{
    Received objects that have been hashed
  }
  \item{received_objects}{
    Objects which have been downloaded
  }
  \item{total_deltas}{
    Total number of deltas in the pack
  }
  \item{indexed_deltas}{
    Deltas which have been indexed
  }
  \item{local_objects}{
    Locally-available objects that have been injected in order to
    fix a thin pack
  }
  \item{received_bytes}{
    Size of the packfile received up to now
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remote.r
\docType{methods}
\name{remote_ls}
\alias{remote_ls}
\alias{remote_ls,character-method}
\title{List references in a remote repository}
\usage{
remote_ls(name, repo = NULL, credentials = NULL)

\S4method{remote_ls}{character}(name, repo = NULL, credentials = NULL)
}
\arguments{
\item{name}{Character vector with the "remote" repository URL to query or
the name of the remote if a \code{repo} argument is given.}

\item{repo}{an optional repository object used if remotes are
specified by name.}

\item{credentials}{The credentials for remote repository
access. Default is NULL. To use and query an ssh-agent for the ssh
key credentials, let this parameter be NULL (the default).}
}
\value{
Character vector for each reference with the associated commit IDs.
}
\description{
Displays references available in a remote repository along with the
associated commit IDs.  Akin to the 'git ls-remote' command.
}
\examples{
\dontrun{
remote_ls("https://github.com/ropensci/git2r")
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.r
\docType{methods}
\name{discover_repository}
\alias{discover_repository}
\alias{discover_repository,character,missing-method}
\alias{discover_repository,character,numeric-method}
\title{Find path to repository for any file}
\usage{
discover_repository(path, ceiling)

\S4method{discover_repository}{character,missing}(path)

\S4method{discover_repository}{character,numeric}(path, ceiling)
}
\arguments{
\item{path}{A character vector specifying the path to a file or folder}

\item{ceiling}{The defult is to not use the ceiling argument and
start the lookup from path and walk across parent
directories. When ceiling is 0, the lookup is only in path. When
ceiling is 1, the lookup is in both the path and the parent to
path.}
}
\value{
Character vector with path to repository or NULL if this
cannot be established.
}
\description{
libgit's git_discover_repository is used to identify the location
of the repository. The path will therefore be terminated by a file
separator.
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           file.path(path, "example-1.txt"))
add(repo, "example-1.txt")
commit(repo, "First commit message")

## Create a second file. The file is not added for version control
## in the repository.
dir.create(file.path(path, "example"))
file_2 <- file.path(path, "example/example-2.txt")
writeLines("Not under version control", file_2)

## Find the path to the repository using the path to the second file
discover_repository(file_2)

## Demonstrate the 'ceiling' argument
wd <- workdir(repo)
dir.create(file.path(wd, "temp"))

## Lookup repository in 'file.path(wd, "temp")'. Should return NULL
discover_repository(file.path(wd, "temp"), ceiling = 0)

## Lookup repository in parent to 'file.path(wd, "temp")'.
## Should not return NULL
discover_repository(file.path(wd, "temp"), ceiling = 1)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tag.r
\docType{methods}
\name{summary,git_tag-method}
\alias{summary,git_tag-method}
\alias{summary,git_tag-methods}
\title{Summary of a tag}
\usage{
\S4method{summary}{git_tag}(object, ...)
}
\arguments{
\item{object}{The tag \code{object}}

\item{...}{Additional arguments affecting the summary produced.}
}
\value{
None (invisible 'NULL').
}
\description{
Summary of a tag
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name="Alice", user.email="alice@example.org")

## Commit a text file
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Create tag
tag(repo, "Tagname", "Tag message")

## Summary of tag
summary(tags(repo)[[1]])
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/S4_classes.r
\docType{class}
\name{cred_env-class}
\alias{cred_env-class}
\title{S4 class to handle environmental variables credential
object}
\description{
Class \code{"cred_env"}
}
\section{Slots}{

\describe{
  \item{username}{
    The name of the environmental variable that holds
    the username for the authentication.
  }
  \item{password}{
    The name of the environmental variable that holds
    the password for the authentication.
  }
}
}

\keyword{classes}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tag.r
\docType{methods}
\name{tag}
\alias{tag}
\alias{tag,git_repository-method}
\title{Create tag targeting HEAD commit in repository}
\usage{
tag(object, name, message, session = FALSE,
  tagger = default_signature(object))

\S4method{tag}{git_repository}(object, name, message, session = FALSE,
  tagger = default_signature(object))
}
\arguments{
\item{object}{The repository \code{object}.}

\item{name}{Name for the tag.}

\item{message}{The tag message.}

\item{session}{Add sessionInfo to tag message. Default is FALSE.}

\item{tagger}{The tagger (author) of the tag}
}
\value{
invisible(\code{git_tag}) object
}
\description{
Create tag targeting HEAD commit in repository
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user
config(repo, user.name="Alice", user.email="alice@example.org")

## Commit a text file
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Create tag
tag(repo, "Tagname", "Tag message")

## List tags
tags(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repository.r
\name{lookup_repository}
\alias{lookup_repository}
\title{Lookup repository for methods with missing repo argument}
\usage{
lookup_repository()
}
\value{
S4 class git_repository
}
\description{
Lookup repository for methods with missing repo argument
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/remote.r
\docType{methods}
\name{remotes}
\alias{remotes}
\alias{remotes,missing-method}
\alias{remotes,git_repository-method}
\title{Get the configured remotes for a repo}
\usage{
remotes(repo)

\S4method{remotes}{missing}()

\S4method{remotes}{git_repository}(repo)
}
\arguments{
\item{repo}{The repository \code{object}
\code{\linkS4class{git_repository}}. If the \code{repo} argument
is missing, the repository is searched for with
\code{\link{discover_repository}} in the current working
directory.}
}
\value{
Character vector with remotes
}
\description{
Get the configured remotes for a repo
}
\examples{
\dontrun{
## Initialize a temporary repository
path <- tempfile(pattern="git2r-")
dir.create(path)
repo <- init(path)

## Create a user and commit a file
config(repo, user.name="Alice", user.email="alice@example.org")
writeLines("Hello world!", file.path(path, "example.txt"))
add(repo, "example.txt")
commit(repo, "First commit message")

## Add a remote
remote_add(repo, "playground", "https://example.org/git2r/playground")
remotes(repo)
remote_url(repo, "playground")

## Rename a remote
remote_rename(repo, "playground", "foobar")
remotes(repo)
remote_url(repo, "foobar")

## Remove a remote
remote_remove(repo, "foobar")
remotes(repo)
}
}
\keyword{methods}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/commit.r
\docType{methods}
\name{ahead_behind}
\alias{ahead_behind}
\alias{ahead_behind,git_commit,git_commit-method}
\title{Ahead Behind}
\usage{
ahead_behind(local, upstream)

\S4method{ahead_behind}{git_commit,git_commit}(local, upstream)
}
\arguments{
\item{local}{a S4 class git_commit \code{object}.}

\item{upstream}{a S4 class git_commit \code{object}.}
}
\value{
An integer vector of length 2 with number of commits that
the upstream commit is ahead and behind the local commit
}
\description{
Count the number of unique commits between two commit objects.
}
\examples{
\dontrun{
## Create a directory in tempdir
path <- tempfile(pattern="git2r-")
dir.create(path)

## Initialize a repository
repo <- init(path)
config(repo, user.name="Alice", user.email="alice@example.org")

## Create a file, add and commit
writeLines("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
           con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit_1 <- commit(repo, "Commit message 1")

# Change file and commit
writeLines(c("Lorem ipsum dolor sit amet, consectetur adipisicing elit, sed do",
             "eiusmod tempor incididunt ut labore et dolore magna aliqua."),
             con = file.path(path, "test.txt"))
add(repo, "test.txt")
commit_2 <- commit(repo, "Commit message 2")

ahead_behind(commit_1, commit_2)
}
}
\keyword{methods}
